@@ -1,6 +1,7 @@
 import py
 
 import os, sys, subprocess
+from os.path import join, dirname
 
 import pypy
 from pypy.interpreter import gateway
@@ -18,9 +19,9 @@ from pypy.module.sys.version import CPYTHON_VERSION
 thisdir = py.path.local(__file__).dirpath()
 
 try:
-    this_dir = os.path.dirname(__file__)
+    this_dir = dirname(__file__)
 except NameError:
-    this_dir = os.path.dirname(sys.argv[0])
+    this_dir = dirname(sys.argv[0])
 
 def debug(msg):
     try:
@@ -75,12 +76,10 @@ def create_entry_point(space, w_dict):
                 w_executable = space.newfilename(argv[0])
                 w_argv = space.newlist([space.newfilename(s)
                                         for s in argv[1:]])
-                w_exitcode = space.call_function(w_entry_point, w_executable, w_argv)
+                w_bargv = space.newlist([space.newbytes(s)
+                                        for s in argv[1:]])
+                w_exitcode = space.call_function(w_entry_point, w_executable, w_bargv, w_argv)
                 exitcode = space.int_w(w_exitcode)
-                # try to pull it all in
-            ##    from pypy.interpreter import main, interactive, error
-            ##    con = interactive.PyPyConsole(space)
-            ##    con.interact()
             except OperationError as e:
                 debug("OperationError:")
                 debug(" operror-type: " + e.w_type.getname(space))
@@ -119,7 +118,7 @@ def get_additional_entrypoints(space, w_initstdio):
         verbose = rffi.cast(lltype.Signed, verbose)
         if ll_home and ord(ll_home[0]):
             home1 = rffi.charp2str(ll_home)
-            home = os.path.join(home1, 'x') # <- so that 'll_home' can be
+            home = join(home1, 'x') # <- so that 'll_home' can be
                                             # directly the root directory
         else:
             home1 = "pypy's shared library location"
@@ -268,7 +267,7 @@ class PyPyTarget(object):
         return pypy_optiondescription
 
     def target(self, driver, args):
-        driver.exe_name = 'pypy3-%(backend)s'
+        driver.exe_name = 'pypy%d.%d-%%(backend)s' % CPYTHON_VERSION[:2]
 
         config = driver.config
         parser = self.opt_parser(config)
@@ -287,7 +286,8 @@ class PyPyTarget(object):
 
         config.translation.suggest(check_str_without_nul=True)
         config.translation.suggest(shared=True)
-        config.translation.suggest(icon=os.path.join(this_dir, 'pypy.ico'))
+        config.translation.suggest(icon=join(this_dir, 'pypy.ico'))
+        config.translation.suggest(manifest=join(this_dir, 'python.manifest'))
         if config.translation.shared:
             if config.translation.output is not None:
                 raise Exception("Cannot use the --output option with PyPy "
@@ -345,10 +345,15 @@ class PyPyTarget(object):
                        "pypy/goal/targetpypystandalone.py.")
 
         if config.objspace.usemodules.cpyext:
-            if config.translation.gc not in ('incminimark', 'boehm'):
-                raise Exception("The 'cpyext' module requires the 'incminimark'"
-                    " or 'boehm' GC.  You need either 'targetpypystandalone.py"
-                    " --withoutmod-cpyext', or use one of these two GCs.")
+            if config.translation.reverse_debugger:
+                assert config.translation.gc == 'boehm'
+                # RevDB uses Boehm but overwrites some operations, which seems
+                # to be sufficient for cpyext.
+            elif config.translation.gc != 'incminimark':
+                raise Exception("The 'cpyext' module requires the default"
+                    " 'incminimark' GC.  You need 'targetpypystandalone.py"
+                    " --withoutmod-cpyext' with other GCs.  (It almost"
+                    " works with Boehm, the fix should be quick)")
 
         config.translating = True
 
@@ -360,9 +365,13 @@ class PyPyTarget(object):
         options = make_dict(config)
         wrapstr = 'space.wrap(%r)' % (options)  # import time
         SysModule.interpleveldefs['pypy_translation_info'] = wrapstr
-        if config.objspace.usemodules._cffi_backend:
+        
+        if 'compile' in driver._disabled:
+            driver.default_goal = 'source'
+        elif config.objspace.usemodules._cffi_backend:
             self.hack_for_cffi_modules(driver)
-
+        else:
+            driver.default_goal = 'compile' 
         return self.get_entry_point(config)
 
     def hack_for_cffi_modules(self, driver):
@@ -378,14 +387,11 @@ class PyPyTarget(object):
         @taskdef([compile_goal], "Create cffi bindings for modules")
         def task_build_cffi_imports(self):
             ''' Use cffi to compile cffi interfaces to modules'''
-            filename = os.path.join(pypydir, '..', 'lib_pypy', 'pypy_tools',
+            filename = join(pypydir, '..', 'lib_pypy', 'pypy_tools',
                                    'build_cffi_imports.py')
-            if sys.platform == 'darwin':
-                argv = [filename, '--embed-dependencies']
-            else:
-                argv = [filename,]
-            status, out, err = run_subprocess(str(driver.compute_exe_name()),
-                                              argv)
+            argv = [filename,]
+            exe_name = py.path.local(driver.c_entryp)
+            status, out, err = run_subprocess(str(exe_name), argv)
             sys.stdout.write(out)
             sys.stderr.write(err)
         driver.task_build_cffi_imports = types.MethodType(task_build_cffi_imports, driver)
@@ -408,13 +414,13 @@ class PyPyTarget(object):
         self.space = make_objspace(config)
 
         # manually imports app_main.py
-        filename = os.path.join(pypydir, 'interpreter', 'app_main.py')
+        filename = join(pypydir, 'interpreter', 'app_main.py')
         app = gateway.applevel(open(filename).read(), 'app_main.py', 'app_main')
         app.hidden_applevel = False
         w_dict = app.getwdict(self.space)
         entry_point, _ = create_entry_point(self.space, w_dict)
 
-        return entry_point, None, PyPyAnnotatorPolicy()
+        return entry_point, None, PyPyAnnotatorPolicy(self.space)
 
     def interface(self, ns):
         for name in ['take_options', 'handle_config', 'print_help', 'target',
