@@ -1,5 +1,5 @@
 from pypy.interpreter import gateway
-from rpython.rlib.objectmodel import we_are_translated
+from rpython.rlib.objectmodel import dict_to_switch
 from rpython.rlib.unroll import unrolling_iterable
 
 
@@ -26,7 +26,7 @@ def syntax_warning(space, msg, fn, lineno, offset):
     _emit_syntax_warning(space, w_msg, w_filename, w_lineno, w_offset)
 
 
-def parse_future(tree, feature_flags):
+def parse_future(space, tree, feature_flags):
     from pypy.interpreter.astcompiler import ast
     future_lineno = 0
     future_column = 0
@@ -40,11 +40,14 @@ def parse_future(tree, feature_flags):
     if body is None:
         return 0, 0, 0
     for stmt in body:
-        if isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Str):
-            if have_docstring:
-                break
-            else:
-                have_docstring = True
+        if isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Constant):
+            constant = stmt.value
+            assert isinstance(constant, ast.Constant)
+            if space.isinstance_w(constant.value, space.w_unicode):
+                if have_docstring:
+                    break
+                else:
+                    have_docstring = True
         elif isinstance(stmt, ast.ImportFrom):
             if stmt.module == "__future__":
                 future_lineno = stmt.lineno
@@ -68,31 +71,15 @@ class ForbiddenNameAssignment(Exception):
         self.node = node
 
 
-def check_forbidden_name(space, name, node=None):
+def check_forbidden_name(space, name):
     """Raise an error if the name cannot be assigned to."""
-    if name in ("None", "__debug__"):
-        raise ForbiddenNameAssignment(name, node)
+    if name == "__debug__":
+        return True
     if name in ("async", "await"):
         space.warn(space.newtext(
             "'async' and 'await' will become reserved keywords"
             " in Python 3.7"), space.w_DeprecationWarning)
-    # XXX Warn about using True and False
-
-
-def dict_to_switch(d):
-    """Convert of dictionary with integer keys to a switch statement."""
-    def lookup(query):
-        if we_are_translated():
-            for key, value in unrolling_iteritems:
-                if key == query:
-                    return value
-            else:
-                raise KeyError
-        else:
-            return d[query]
-    lookup._always_inline_ = True
-    unrolling_iteritems = unrolling_iterable(d.iteritems())
-    return lookup
+    return False
 
 
 def mangle(name, klass):
