@@ -20,14 +20,14 @@ CONST_TRUE = 1
 
 class __extend__(ast.AST):
 
-    def as_constant_truth(self, space):
+    def as_constant_truth(self, space, compile_info):
         """Return the truth of this node if known."""
-        const = self.as_constant()
+        const = self.as_constant(space, compile_info)
         if const is None:
             return CONST_NOT_CONST
         return int(space.is_true(const))
 
-    def as_constant(self):
+    def as_constant(self, space, compile_info):
         """Return the value of this node as a wrapped constant if possible."""
         return None
 
@@ -38,55 +38,36 @@ class __extend__(ast.AST):
 class __extend__(ast.expr):
 
     def accept_jump_if(self, gen, condition, target):
-        self.walkabout(gen)
-        if condition:
-            gen.emit_jump(ops.POP_JUMP_IF_TRUE, target, True)
-        else:
-            gen.emit_jump(ops.POP_JUMP_IF_FALSE, target, True)
-
-
-class __extend__(ast.Num):
-
-    def as_constant(self):
-        return self.n
-
-
-class __extend__(ast.Str):
-
-    def as_constant(self):
-        return self.s
-
-
-class __extend__(ast.Bytes):
-
-    def as_constant(self):
-        return self.s
-
-
-class __extend__(ast.Ellipsis):
-
-    def as_constant_truth(self, space):
-        return True
+        test_constant = self.as_constant_truth(
+            gen.space, gen.compile_info)
+        if test_constant == CONST_NOT_CONST:
+            self.walkabout(gen)
+            if condition:
+                gen.emit_jump(ops.POP_JUMP_IF_TRUE, target)
+            else:
+                gen.emit_jump(ops.POP_JUMP_IF_FALSE, target)
+            return
+        gen.emit_line_tracing_nop(self)
+        if test_constant == condition:
+            gen.emit_jump(ops.JUMP_FORWARD, target)
 
 
 class __extend__(ast.Constant):
 
-    def as_constant(self):
+    def as_constant(self, space, compile_info):
         return self.value
 
-class __extend__(ast.NameConstant):
-
-    def as_constant(self):
-        return self.value
-
-class __extend__(ast.Index):
-    def as_constant(self):
-        return self.value.as_constant()
+class __extend__(ast.Name):
+    def as_constant(self, space, compile_info):
+        if self.id == '__debug__':
+            return space.newbool(compile_info.optimize == 0)
+        else:
+            return None
 
 class __extend__(ast.Slice):
-    def as_constant(self):
+    def as_constant(self, space, compile_info):
         # XXX: this ought to return a slice object if all the indices are
-        # constants, but we don't have a space here.
+        # constants
         return None
 
 class __extend__(ast.UnaryOp):
@@ -136,6 +117,16 @@ def _fold_pow(space, w_left, w_right):
     """)
     return space.pow(w_left, w_right, space.w_None)
 
+def _fold_lshift(space, w_left, w_right):
+    # don't constant-fold if "w_left" and "w_right" are integers and
+    # the estimated bit length of the result is unreasonably large
+    space.appexec([w_left, w_right], """(left, right):
+        if isinstance(left, int) and isinstance(right, int):
+            if left.bit_length() + right > 1000:
+                raise OverflowError
+    """)
+    return space.lshift(w_left, w_right)
+
 def _fold_not(space, operand):
     return space.newbool(not space.is_true(operand))
 
@@ -148,7 +139,7 @@ binary_folders = {
     ast.FloorDiv : _binary_fold("floordiv"),
     ast.Mod : _binary_fold("mod"),
     ast.Pow : _fold_pow,
-    ast.LShift : _binary_fold("lshift"),
+    ast.LShift : _fold_lshift,
     ast.RShift : _binary_fold("rshift"),
     ast.BitOr : _binary_fold("or_"),
     ast.BitXor : _binary_fold("xor"),
@@ -189,9 +180,9 @@ class OptimizingVisitor(ast.ASTVisitor):
         return node
 
     def visit_BinOp(self, binop):
-        left = binop.left.as_constant()
+        left = binop.left.as_constant(self.space, self.compile_info)
         if left is not None:
-            right = binop.right.as_constant()
+            right = binop.right.as_constant(self.space, self.compile_info)
             if right is not None:
                 op = binop.op
                 try:
@@ -214,11 +205,11 @@ class OptimizingVisitor(ast.ASTVisitor):
                     else:
                         if self.space.int_w(w_len) > 20:
                             return binop
-                    return ast.Constant(w_const, binop.lineno, binop.col_offset)
+                    return self.new_constant(w_const, binop)
         return binop
 
     def visit_UnaryOp(self, unary):
-        w_operand = unary.operand.as_constant()
+        w_operand = unary.operand.as_constant(self.space, self.compile_info)
         op = unary.op
         if w_operand is not None:
             try:
@@ -235,7 +226,7 @@ class OptimizingVisitor(ast.ASTVisitor):
             except OperationError:
                 pass
             else:
-                return ast.Constant(w_const, unary.lineno, unary.col_offset)
+                return self.new_constant(w_const, unary)
         elif op == ast.Not:
             compare = unary.operand
             if isinstance(compare, ast.Compare) and len(compare.ops) == 1:
@@ -249,46 +240,16 @@ class OptimizingVisitor(ast.ASTVisitor):
                     return compare
         return unary
 
-    def visit_BoolOp(self, bop):
-        values = bop.values
-        we_are_and = bop.op == ast.And
-        i = 0
-        while i < len(values) - 1:
-            truth = values[i].as_constant_truth(self.space)
-            if truth != CONST_NOT_CONST:
-                if (truth != CONST_TRUE) == we_are_and:
-                    del values[i + 1:]
-                    break
-                else:
-                    del values[i]
-            else:
-                i += 1
-        if len(values) == 1:
-            return values[0]
-        return bop
-
-    def visit_Repr(self, rep):
-        w_const = rep.value.as_constant()
-        if w_const is not None:
-            w_repr = self.space.repr(w_const)
-            return ast.Constant(w_repr, rep.lineno, rep.col_offset)
-        return rep
-
     def visit_Name(self, name):
         """Turn loading None, True, and False into a constant lookup."""
-        if name.ctx == ast.Del:
+        if name.ctx == ast.Del or name.ctx == ast.Store:
             return name
         space = self.space
-        iden = name.id
         w_const = None
-        if iden == "None":
-            w_const = space.w_None
-        elif iden == "True":
-            w_const = space.w_True
-        elif iden == "False":
-            w_const = space.w_False
+        if name.id == '__debug__':
+            w_const = space.newbool(self.compile_info.optimize == 0)
         if w_const is not None:
-            return ast.NameConstant(w_const, name.lineno, name.col_offset)
+            return self.new_constant(w_const, name)
         return name
 
     def visit_Tuple(self, tup):
@@ -300,11 +261,11 @@ class OptimizingVisitor(ast.ASTVisitor):
             consts_w = [None]*len(tup.elts)
             for i in range(len(tup.elts)):
                 node = tup.elts[i]
-                w_const = node.as_constant()
+                w_const = node.as_constant(self.space, self.compile_info)
                 if w_const is None:
                     new_elts = self._optimize_constant_star_unpacks(tup.elts)
                     if new_elts is not None:
-                        return ast.Tuple(new_elts, ast.Load, tup.lineno, tup.col_offset)
+                        return ast.Tuple(new_elts, ast.Load, tup.lineno, tup.col_offset, tup.end_lineno, tup.end_col_offset)
                     return tup
                 consts_w[i] = w_const
             # intern the string constants packed into the tuple here,
@@ -315,13 +276,14 @@ class OptimizingVisitor(ast.ASTVisitor):
         else:
             consts_w = []
         w_consts = self.space.newtuple(consts_w)
-        return ast.Constant(w_consts, tup.lineno, tup.col_offset)
+        return self.new_constant(w_consts, tup)
 
     def _make_starred_tuple_const(self, consts_w, firstelt):
         w_consts = self.space.newtuple(consts_w[:])
-        return ast.Starred(ast.Constant(
-                    w_consts, firstelt.lineno, firstelt.col_offset),
-                ast.Load, firstelt.lineno, firstelt.col_offset)
+        return ast.Starred(self.new_constant(
+                    w_consts, firstelt),
+                ast.Load, firstelt.lineno, firstelt.col_offset,
+                firstelt.end_lineno, firstelt.end_col_offset)
 
     def _optimize_constant_star_unpacks(self, elts):
         # turn (1, 2, 3, *a) into (*(1, 2, 3), *a) with a constant (1, 2, 3)
@@ -341,7 +303,8 @@ class OptimizingVisitor(ast.ASTVisitor):
         for i in range(len(elts)):
             elt = elts[i]
             if isinstance(elt, ast.Starred):
-                if const_since_last_star_w is not None:
+                if (const_since_last_star_w is not None and
+                        len(const_since_last_star_w) > 0):
                     firstelt = elts[after_last_star_index]
                     new_elts.append(self._make_starred_tuple_const(
                         const_since_last_star_w, firstelt))
@@ -350,7 +313,7 @@ class OptimizingVisitor(ast.ASTVisitor):
                 after_last_star_index = i + 1
                 new_elts.append(elt)
             elif const_since_last_star_w is not None:
-                w_const = elt.as_constant()
+                w_const = elt.as_constant(self.space, self.compile_info)
                 if w_const is None:
                     new_elts.extend(elts[after_last_star_index:i + 1])
                     const_since_last_star_w = None
@@ -370,14 +333,14 @@ class OptimizingVisitor(ast.ASTVisitor):
         if l.ctx == ast.Load and l.elts:
             new_elts = self._optimize_constant_star_unpacks(l.elts)
             if new_elts:
-                return ast.List(new_elts, ast.Load, l.lineno, l.col_offset)
+                return ast.List(new_elts, ast.Load, l.lineno, l.col_offset, l.end_lineno, l.end_col_offset)
         return l
 
     def visit_Subscript(self, subs):
         if subs.ctx == ast.Load:
-            w_obj = subs.value.as_constant()
+            w_obj = subs.value.as_constant(self.space, self.compile_info)
             if w_obj is not None:
-                w_idx = subs.slice.as_constant()
+                w_idx = subs.slice.as_constant(self.space, self.compile_info)
                 if w_idx is not None:
                     try:
                         w_const = self.space.getitem(w_obj, w_idx)
@@ -408,6 +371,10 @@ class OptimizingVisitor(ast.ASTVisitor):
                         # See test_const_fold_unicode_subscr
                         return subs
 
-                    return ast.Constant(w_const, subs.lineno, subs.col_offset)
+                    return self.new_constant(w_const, subs)
 
         return subs
+
+    def new_constant(self, const, node):
+        return ast.Constant(const, self.space.w_None, node.lineno, node.col_offset,
+                node.end_lineno, node.end_col_offset)
