@@ -1,6 +1,6 @@
 import string
 import py
-from pypy.interpreter.astcompiler import ast, astbuilder, symtable, consts
+from pypy.interpreter.astcompiler import ast, symtable, consts
 from pypy.interpreter.pyparser import pyparse
 from pypy.interpreter.pyparser.error import SyntaxError
 
@@ -8,13 +8,12 @@ from pypy.interpreter.pyparser.error import SyntaxError
 class TestSymbolTable:
 
     def setup_class(cls):
-        cls.parser = pyparse.PythonParser(cls.space)
+        cls.parser = pyparse.PegParser(cls.space)
 
     def mod_scope(self, source, mode="exec"):
-        info = pyparse.CompileInfo("<test>", mode,
-                                   consts.CO_FUTURE_WITH_STATEMENT)
-        tree = self.parser.parse_source(source, info)
-        module = astbuilder.ast_from_node(self.space, tree, info)
+        ec = self.space.getexecutioncontext()
+        module = ec.compiler.compile_to_ast(source, "<test>", mode, 0)
+        info = pyparse.CompileInfo("<test>", mode, 0)
         builder = symtable.SymtableBuilder(self.space, module, info)
         scope = builder.find_scope(module)
         assert isinstance(scope, symtable.ModuleScope)
@@ -66,6 +65,10 @@ class TestSymbolTable:
         input = "def f(x, x): pass"
         exc = py.test.raises(SyntaxError, self.mod_scope, input).value
         assert exc.msg == "duplicate argument 'x' in function definition"
+        input = "def f(x,\nx): pass"
+        exc = py.test.raises(SyntaxError, self.mod_scope, input).value
+        assert exc.msg == "duplicate argument 'x' in function definition"
+        assert exc.lineno == 2
 
     def test_function_defaults(self):
         scp = self.mod_scope("y = w = 4\ndef f(x=y, *, z=w): return x")
@@ -297,8 +300,8 @@ class TestSymbolTable:
 
         src = "def f(x):\n   global x"
         exc = py.test.raises(SyntaxError, self.func_scope, src).value
-        assert exc.lineno == 2
         assert exc.msg == "name 'x' is parameter and global"
+        assert exc.lineno == 2
 
     def test_global_nested(self):
         src = """
@@ -333,6 +336,7 @@ def f():
     nonlocal x"""
         exc = py.test.raises(SyntaxError, self.func_scope, src).value
         assert exc.msg == "no binding for nonlocal 'x' found"
+        assert exc.lineno == 4
 
         src = str(py.code.Source("""
                      def f(x):
@@ -340,6 +344,7 @@ def f():
                  """))
         exc = py.test.raises(SyntaxError, self.func_scope, src).value
         assert exc.msg == "name 'x' is parameter and nonlocal"
+        assert exc.lineno == 3
         #
         src = str(py.code.Source("""
                      def f():
@@ -369,6 +374,8 @@ def f():
         exc = py.test.raises(SyntaxError, self.func_scope, src).value
         assert exc.msg == "name 'x' is nonlocal and global"
         assert exc.lineno == 4
+        assert exc.offset == 5
+        assert exc.end_offset == 13
 
         src = """
 def f():
@@ -477,11 +484,6 @@ def f(x):
             input = "def f():\n    yield y\n    %s\n    yield y" % (input,)
             assert not self.func_scope(input).has_yield_inside_try
 
-    def test_return(self):
-        for input in ("class x: return", "return"):
-            exc = py.test.raises(SyntaxError, self.func_scope, input).value
-            assert exc.msg == "return outside function"
-
     def test_tmpnames(self):
         scp = self.mod_scope("with x: pass")
         assert scp.lookup("_[1]") == symtable.SCOPE_LOCAL
@@ -502,6 +504,14 @@ def f(x):
         assert exc_global.msg == "annotated name 'x' can't be global"
         assert exc_global.lineno == 3
 
+    def test_annotation_global4(self):
+        src_global = ("class A:\n"
+                      "    global x\n"
+                      "    x: int\n")
+        exc_global = py.test.raises(SyntaxError, self.func_scope, src_global).value
+        assert exc_global.msg == "annotated name 'x' can't be global"
+        assert exc_global.lineno == 3
+
     def test_annotation_nonlocal(self):
         src_nonlocal = ("def f():\n"
                         "    x: int\n"
@@ -531,3 +541,7 @@ def f(x):
     def test_issue13343(self):
         scp = self.mod_scope("lambda *, k1=x, k2: None")
         assert scp.lookup("x") == symtable.SCOPE_GLOBAL_IMPLICIT
+
+    def test_named_expr_list_comprehension(self):
+        fscp = self.func_scope("def f(): [(y := x) for x in range(5)]")
+        assert fscp.lookup("y") == symtable.SCOPE_CELL
