@@ -3,8 +3,8 @@ import py
 
 from rpython.rlib.cache import Cache
 from rpython.tool.uid import HUGEVAL_BYTES
-from rpython.rlib import jit, types, rutf8
-from rpython.rlib.debug import make_sure_not_resized
+from rpython.rlib import jit, types, rutf8, rstring
+from rpython.rlib.debug import make_sure_not_resized, check_not_access_directly
 from rpython.rlib.objectmodel import (we_are_translated, newlist_hint,
      compute_unique_id, specialize, not_rpython)
 from rpython.rlib.signature import signature
@@ -18,6 +18,7 @@ from pypy.interpreter.executioncontext import (ExecutionContext, ActionFlag,
 from pypy.interpreter.error import OperationError, new_exception_class, oefmt
 from pypy.interpreter.argument import Arguments
 from pypy.interpreter.miscutils import ThreadLocals, make_weak_value_dictionary
+from pypy.tool.generate_stdlib_module_names import get_stdlib_names
 
 
 __all__ = ['ObjSpace', 'OperationError', 'W_Root']
@@ -72,7 +73,11 @@ class W_Root(object):
 
     # to be used directly only by space.type implementations
     def getclass(self, space):
-        return space.gettypeobject(self.typedef)
+        # make sure that we never annotate this with access_directly set to
+        # True. otherwise every call to getclass (and other methods) has an
+        # extra indirection due to a much more complicated function set
+        check_not_access_directly(self)
+        return space.gettypefor(self.__class__)
 
     def setclass(self, space, w_subtype):
         raise oefmt(space.w_TypeError,
@@ -225,6 +230,18 @@ class W_Root(object):
         from pypy.module.cpyext.pyobject import w_root_attach_pyobj
         return w_root_attach_pyobj(self, space, py_obj)
 
+    # -------------------------------------------------------------------
+
+    # hpy support
+
+    def _hpy_get_raw_storage(self, space):
+        raise oefmt(space.w_SystemError, "trying to read hpy storage from wrong object of type %T", self)
+
+    def _hpy_get_gc_storage(self, space):
+        raise oefmt(space.w_SystemError, "trying to read hpy storage from wrong object of type %T", self)
+
+    def _hpy_set_raw_storage(self, space, storage):
+        raise oefmt(space.w_SystemError, "trying to write hpy storage into wrong object of type %T", self)
 
     # -------------------------------------------------------------------
 
@@ -250,13 +267,13 @@ class W_Root(object):
         self._typed_unwrap_error(space, "bytes")
 
     def text_w(self, space):
-        self._typed_unwrap_error(space, "unicode")
+        self._typed_unwrap_error(space, "str")
 
     def utf8_w(self, space):
-        self._typed_unwrap_error(space, "unicode")
+        self._typed_unwrap_error(space, "str")
 
     def convert_to_w_unicode(self, space):
-        self._typed_unwrap_error(space, "unicode")
+        self._typed_unwrap_error(space, "str")
 
     def bytearray_list_of_chars_w(self, space):
         self._typed_unwrap_error(space, "bytearray")
@@ -303,7 +320,9 @@ class W_Root(object):
         from pypy.objspace.std.intobject import W_AbstractIntObject
         w_impl = space.lookup(self, '__int__')
         if w_impl is None:
-            self._typed_unwrap_error(space, "integer")
+            w_impl = space.lookup(self, '__index__')
+            if w_impl is None:
+                    self._typed_unwrap_error(space, "integer")
         w_result = space.get_and_call_function(w_impl, self)
 
         if space.is_w(space.type(w_result), space.w_int):
@@ -411,6 +430,8 @@ class ObjSpace(object):
     @not_rpython
     def __init__(self, config=None):
         "Basic initialization of objects."
+        self.w_default_importlib_import = None
+
         self.fromcache = InternalSpaceCache(self).getorbuild
         self.threadlocals = ThreadLocals()
         # set recursion limit
@@ -460,7 +481,16 @@ class ObjSpace(object):
         ret = 0
         self.wait_for_thread_shutdown()
         w_atexit = self.getbuiltinmodule('atexit')
-        self.call_method(w_atexit, '_run_exitfuncs')
+        try:
+            self.call_method(w_atexit, '_run_exitfuncs')
+        except OperationError as e:
+            try:
+                e.write_unraisable(self, '_run_exitfuncs')
+            except Exception:
+                pass
+            # discard exceptions, see call_py_exitfuncs in pylifecycle.c in
+            # CPython
+            pass
         self.sys.finalizing = True
         if self.sys.flush_std_files(self) < 0:
             ret = -1
@@ -561,10 +591,23 @@ class ObjSpace(object):
         except AttributeError:
             pass
 
-        modules = ['__pypy__']  # Install __pypy__ first for bootstrapping
+        # Install __pypy__ first for bootstrapping
+        modules = ['__pypy__']
+
+        # _frozen_importlib imports lib-python/3/importlib/_bootstrap_external,
+        # which imports many builtins. Make sure it is imported last
+        append__frozen_importlib = False
+        # and zipimport even laster
+        append_zipimport = False
 
         # You can enable more modules by specifying --usemodules=xxx,yyy
         for name, value in self.config.objspace.usemodules:
+            if name == '_frozen_importlib':
+                append__frozen_importlib = True
+                continue
+            if name == "zipimport":
+                append_zipimport = True
+                continue
             if value and name not in modules:
                 modules.append(name)
 
@@ -572,7 +615,10 @@ class ObjSpace(object):
             for name in self.config.objspace.extmodules.split(','):
                 if name not in modules:
                     modules.append(name)
-
+        if append__frozen_importlib:
+            modules.append('_frozen_importlib')
+        if append_zipimport:
+            modules.append('zipimport')
         self._builtinmodule_list = modules
         return self._builtinmodule_list
 
@@ -612,17 +658,23 @@ class ObjSpace(object):
             self.setitem(self.builtin.w_dict, self.newtext(name), w_type)
 
         # install mixed modules
-        bootstrap_modules = set(('sys', 'imp', 'builtins', 'exceptions'))
+        bootstrap_modules = set(('sys', 'imp', 'builtins', 'exceptions',
+                                 'zipimport', '_frozen_importlib'))
         for mixedname in self.get_builtinmodule_to_install():
             if mixedname not in bootstrap_modules:
                 self.install_mixedmodule(mixedname)
 
-        w_builtin_module_names = self.newtuple(
-            [self.newtext(name) for name in sorted(self.builtin_modules)])
+        w_builtin_module_names = self.newtuple([self.newtext(name) for name in sorted(self.builtin_modules)])
+        stdlib = [self.newtext(name) for name in sorted(list(self.builtin_modules.keys()) + get_stdlib_names())]
+        w_stdlib_module_names = self.newfrozenset(stdlib)
 
-        # force this value into the dict without unlazyfying everything
+
+        # force these value into the dict without unlazyfying everything
         self.setitem(self.sys.w_dict, self.newtext('builtin_module_names'),
                      w_builtin_module_names)
+        self.setitem(self.sys.w_dict, self.newtext('stdlib_module_names'),
+                     w_stdlib_module_names)
+
 
     def get_builtin_types(self):
         """Get a dictionary mapping the names of builtin types to the type
@@ -642,7 +694,7 @@ class ObjSpace(object):
                     raise
                 break
             name = self.text_w(w_name)
-            if not name.startswith('__'):
+            if not name.startswith('_'):
                 excname = name
                 w_exc = self.getitem(w_dic, w_name)
                 exc_types_w[name] = w_exc
@@ -661,12 +713,24 @@ class ObjSpace(object):
             # is required to initialise sys on Windows.
             from pypy.module.cpyext.state import State
             self.fromcache(State).build_api()
+        elif self.config.objspace.usemodules._cffi_backend:
+            from pypy.module._cffi_backend import copy_includes
+            copy_includes.main()
+
+        # now we can setup _frozen_importlib, after the dll is installed
+        self.install_mixedmodule('_frozen_importlib')
+
         self.getbuiltinmodule('sys')
         self.getbuiltinmodule('_imp')
-        self.getbuiltinmodule('_frozen_importlib')
+        frozen_importlib = self.getbuiltinmodule('_frozen_importlib')
         self.getbuiltinmodule('builtins')
         for mod in self.builtin_modules.values():
             mod.setup_after_space_initialization()
+        self.w_default_importlib_import = frozen_importlib.w_import
+        # special-casing zipimport: it needs a lot of things already in place,
+        # so we even install it here!
+        self.install_mixedmodule('zipimport')
+        self.getbuiltinmodule('zipimport')
 
     @not_rpython
     def initialize(self):
@@ -721,6 +785,16 @@ class ObjSpace(object):
             self.default_compiler = compiler
             return compiler
 
+    def createnewcompiler(self):
+        "Factory function creating a compiler object."
+        try:
+            return self.default_newcompiler
+        except AttributeError:
+            from pypy.interpreter.pycompiler import PythonAstCompiler
+            compiler = PythonAstCompiler(self)
+            self.default_newcompiler = compiler
+            return compiler
+
     def createframe(self, code, w_globals, outer_func=None):
         "Create an empty PyFrame suitable for this code object."
         return self.FrameClass(self, code, w_globals, outer_func)
@@ -772,17 +846,15 @@ class ObjSpace(object):
         w_result = w_obj.immutable_unique_id(self)
         if w_result is None:
             # in the common case, returns an unsigned value
-            w_result = self.newint(r_uint(compute_unique_id(w_obj)))
+            id = compute_unique_id(w_obj)
+            if id >= 0:
+                w_result = self.newint(id)
+            else:
+                # the following path returns W_LongObject, but it should happen
+                # only on 32-bit platforms
+                w_result = self.newint(r_uint(id))
         return w_result
 
-    def hash_w(self, w_obj):
-        """shortcut for space.int_w(space.hash(w_obj))"""
-        return self.int_w(self.hash(w_obj))
-
-    def len_w(self, w_obj):
-        """shortcut for space.int_w(space.len(w_obj))"""
-        return self.int_w(self.len(w_obj))
-
     def contains_w(self, w_container, w_item):
         """shortcut for space.is_true(space.contains(w_container, w_item))"""
         return self.is_true(self.contains(w_container, w_item))
@@ -905,28 +977,16 @@ class ObjSpace(object):
             return None
         if not isinstance(w_obj, RequiredClass):   # or obj is None
             raise oefmt(self.w_TypeError,
-                        "'%s' object expected, got '%N' instead",
+                        "'%s' object expected, got '%T' instead",
                         wrappable_class_name(RequiredClass),
-                        w_obj.getclass(self))
+                        w_obj)
         return w_obj
 
-    def _iter_unpackiterable(self, w_iterable):
-        try:
-            return self.iter(w_iterable)
-        except OperationError as e:
-            if e.got_any_traceback():
-                raise
-            if not e.match(self, self.w_TypeError):
-                raise
-            raise oefmt(self.w_TypeError,
-                    "cannot unpack non-iterable %T object",
-                    w_iterable)
-
     def unpackiterable(self, w_iterable, expected_length=-1):
         """Unpack an iterable into a real (interpreter-level) list.
 
         Raise an OperationError(w_ValueError) if the length is wrong."""
-        w_iterator = self._iter_unpackiterable(w_iterable)
+        w_iterator = self.iter(w_iterable)
         if expected_length == -1:
             if self.is_generator(w_iterator):
                 # special hack for speed
@@ -998,7 +1058,7 @@ class ObjSpace(object):
         # Like unpackiterable(), but for the cases where we have
         # an expected_length and want to unroll when JITted.
         # Returns a fixed-size list.
-        w_iterator = self._iter_unpackiterable(w_iterable)
+        w_iterator = self.iter(w_iterable)
         assert expected_length != -1
         return self._unpackiterable_known_length_jitlook(w_iterator,
                                                          expected_length)
@@ -1154,8 +1214,8 @@ class ObjSpace(object):
         return self.call_args(w_callable, args)
 
     def _try_fetch_pycode(self, w_func):
-        from pypy.interpreter.function import Function, Method
-        if isinstance(w_func, Method):
+        from pypy.interpreter.function import Function, _Method
+        if isinstance(w_func, _Method):
             w_func = w_func.w_function
         if isinstance(w_func, Function):
             return w_func.code
@@ -1165,8 +1225,8 @@ class ObjSpace(object):
         nargs = len(args_w) # used for pruning funccall versions
         if not self.config.objspace.disable_call_speedhacks and nargs < 5:
             # start of hack for performance
-            from pypy.interpreter.function import Function, Method
-            if isinstance(w_func, Method):
+            from pypy.interpreter.function import Function, _Method
+            if isinstance(w_func, _Method):
                 if nargs < 4:
                     func = w_func.w_function
                     if isinstance(func, Function):
@@ -1179,18 +1239,19 @@ class ObjSpace(object):
         args = Arguments(self, list(args_w))
         return self.call_args(w_func, args)
 
-    def call_valuestack(self, w_func, nargs, frame, methodcall=False):
+    def call_valuestack(self, w_func, nargs, frame, dropvalues, methodcall=False):
         # methodcall is only used for better error messages in argument.py
-        from pypy.interpreter.function import Function, Method, is_builtin_code
+        from pypy.interpreter.function import Function, _Method, is_builtin_code
         if frame.get_is_being_profiled() and is_builtin_code(w_func):
             # XXX: this code is copied&pasted :-( from the slow path below
             # call_valuestack().
             args = frame.make_arguments(nargs, w_function=w_func)
+            frame.dropvalues(dropvalues)
             return self.call_args_and_c_profile(frame, w_func, args)
 
         if not self.config.objspace.disable_call_speedhacks:
             # start of hack for performance
-            if isinstance(w_func, Method):
+            if isinstance(w_func, _Method):
                 # reuse callable stack place for w_inst
                 frame.settopvalue(w_func.w_instance, nargs)
                 nargs += 1
@@ -1199,10 +1260,12 @@ class ObjSpace(object):
 
             if isinstance(w_func, Function):
                 return w_func.funccall_valuestack(
-                        nargs, frame, methodcall=methodcall)
+                        nargs, frame, methodcall=methodcall, dropvalues=dropvalues)
+
             # end of hack for performance
 
         args = frame.make_arguments(nargs, w_function=w_func)
+        frame.dropvalues(dropvalues)
         return self.call_args(w_func, args)
 
     def call_args_and_c_profile(self, frame, w_func, args):
@@ -1335,9 +1398,7 @@ class ObjSpace(object):
             filename = '?'
         from pypy.interpreter.pycode import PyCode
         if isinstance(statement, str):
-            compiler = self.createcompiler()
-            statement = compiler.compile(statement, filename, 'exec', 0,
-                                         hidden_applevel=hidden_applevel)
+            statement = self._cached_compile(filename, statement, 'exec', 0, hidden_applevel)
         if not isinstance(statement, PyCode):
             raise TypeError('space.exec_(): expected a string, code or PyCode object')
         w_key = self.newtext('__builtins__')
@@ -1345,6 +1406,50 @@ class ObjSpace(object):
             self.setitem(w_globals, w_key, self.builtin)
         return statement.exec_code(self, w_globals, w_locals)
 
+
+    @not_rpython
+    def _cached_compile(self, filename, source, mode, flags, hidden_applevel, ast_transform=None):
+        import os
+        from hashlib import md5
+        from rpython.config.translationoption import CACHE_DIR
+        from rpython.tool.gcc_cache import try_atomic_write
+        from pypy.module.marshal import interp_marshal
+        from pypy.interpreter.pycode import default_magic
+        h = md5(str(default_magic))
+        h.update(filename)
+        h.update(source)
+        h.update(mode)
+        h.update(str(flags))
+        h.update(str(hidden_applevel))
+        addition = ''
+        if ast_transform:
+            addition = ast_transform.func_name
+
+        cachename = os.path.join(
+            CACHE_DIR, "applevel_exec_%s_%s" % (addition, h.hexdigest()))
+        try:
+            if self.config.translating:
+                raise IOError("don't use the cache when translating pypy")
+            with open(cachename, 'rb') as f:
+                w_bin = self.newbytes(f.read())
+                w_code = interp_marshal._loads(self, w_bin, hidden_applevel)
+        except IOError:
+            # must (re)compile the source
+            ec = self.getexecutioncontext()
+            if ast_transform:
+                c = self.createcompiler()
+                tree = c.compile_to_ast(source, filename, "exec", 0)
+                tree = ast_transform(self, tree)
+                w_code = c.compile_ast(tree, filename, 'exec', 0)
+            else:
+                w_code = ec.compiler.compile(
+                    source, filename, mode, flags, hidden_applevel)
+            w_bin = interp_marshal.dumps(self, w_code)
+            content = self.bytes_w(w_bin)
+            try_atomic_write(cachename, content)
+        return w_code
+
+
     @not_rpython
     def appdef(self, source):
         '''Create interp-level function object from app-level source.
@@ -1408,29 +1513,13 @@ class ObjSpace(object):
                 if op == 'ge': return self.ge(w_x1, w_x2)
                 assert False, "bad value for op"
 
-    def decode_index(self, w_index_or_slice, seqlength):
-        """Helper for custom sequence implementations
-             -> (index, 0, 0) or
-                (start, stop, step)
-        """
-        if self.isinstance_w(w_index_or_slice, self.w_slice):
-            from pypy.objspace.std.sliceobject import W_SliceObject
-            assert isinstance(w_index_or_slice, W_SliceObject)
-            start, stop, step = w_index_or_slice.indices3(self, seqlength)
-        else:
-            start = self.int_w(w_index_or_slice, allow_conversion=False)
-            if start < 0:
-                start += seqlength
-            if not (0 <= start < seqlength):
-                raise oefmt(self.w_IndexError, "index out of range")
-            stop = 0
-            step = 0
-        return start, stop, step
-
-    def decode_index4(self, w_index_or_slice, seqlength):
+    def decode_index4_unsafe(self, w_index_or_slice, seqlength):
         """Helper for custom sequence implementations
              -> (index, 0, 0, 1) or
                 (start, stop, step, slice_length)
+
+        UNSAFE: if the __index__ method of one of the slice's start/stop/step
+        fields changes the sequence, then the result can be incorrect
         """
         if self.isinstance_w(w_index_or_slice, self.w_slice):
             from pypy.objspace.std.sliceobject import W_SliceObject
@@ -1438,7 +1527,7 @@ class ObjSpace(object):
             start, stop, step, length = w_index_or_slice.indices4(self,
                                                                   seqlength)
         else:
-            start = self.int_w(w_index_or_slice, allow_conversion=False)
+            start = self.getindex_w(w_index_or_slice, self.w_OverflowError)
             if start < 0:
                 start += seqlength
             if not (0 <= start < seqlength):
@@ -1448,6 +1537,32 @@ class ObjSpace(object):
             length = 1
         return start, stop, step, length
 
+    def decode_index4(self, w_index_or_slice, w_seq):
+        """Helper for custom sequence implementations
+             -> (index, 0, 0, 1) or
+                (start, stop, step, slice_length)
+        """
+        if self.isinstance_w(w_index_or_slice, self.w_slice):
+            from pypy.objspace.std.sliceobject import W_SliceObject
+            assert isinstance(w_index_or_slice, W_SliceObject)
+            # it's important to first unpack the slice and then read the length
+            # of the sequence, because the former can change the latter
+            start, stop, step = w_index_or_slice.unpack(self)
+            seqlength = self.len_w(w_seq)
+            start, stop, step, length = w_index_or_slice.adjust_indices(start, stop, step, seqlength)
+        else:
+            start = self.getindex_w(w_index_or_slice, self.w_OverflowError)
+            seqlength = self.len_w(w_seq)
+            if start < 0:
+                start += seqlength
+            if not (0 <= start < seqlength):
+                raise oefmt(self.w_IndexError, "index out of range")
+            stop = 0
+            step = 0
+            length = 1
+        return start, stop, step, length
+
+
     def getindex_w(self, w_obj, w_exception, objdescr=None):
         """Return w_obj.__index__() as an RPython int.
         If w_exception is None, silently clamp in case of overflow;
@@ -1459,7 +1574,7 @@ class ObjSpace(object):
             if objdescr is None or not err.match(self, self.w_TypeError):
                 raise
             raise oefmt(self.w_TypeError,
-                        "%s indices must be integers or slices, not %T",
+                        "%s indices must be integers or slices, not '%T'",
                         objdescr, w_obj)
         try:
             # allow_conversion=False it's not really necessary because the
@@ -1546,7 +1661,7 @@ class ObjSpace(object):
             return self._try_buffer_w(w_obj, flags)
         except BufferInterfaceNotFound:
             raise oefmt(self.w_TypeError,
-                        "'%T' does not support the buffer interface", w_obj)
+                        "a bytes-like object is required, not '%T'", w_obj)
 
     def readbuf_w(self, w_obj):
         # Old buffer interface, returns a readonly buffer (PyObject_AsReadBuffer)
@@ -1573,7 +1688,7 @@ class ObjSpace(object):
         if self.is_none(w_obj):
             e = oefmt(self.w_TypeError, "a %s is required, not None", expected)
         else:
-            e = oefmt(self.w_TypeError, "a %s is required, not %T", expected, w_obj)
+            e = oefmt(self.w_TypeError, "a %s is required, not '%T'", expected, w_obj)
         raise e
 
     @specialize.arg(1)
@@ -1618,9 +1733,13 @@ class ObjSpace(object):
 
         if self.isinstance_w(w_obj, self.w_unicode):
             return w_obj.charbuf_w(self)
+
     def text_or_none_w(self, w_obj):
         return None if self.is_none(w_obj) else self.text_w(w_obj)
 
+    def text0_or_none_w(self, w_obj):
+        return None if self.is_none(w_obj) else self.text0_w(w_obj)
+
     @specialize.argtype(1)
     def bytes_w(self, w_obj):
         """ Takes an application level :py:class:`bytes`
@@ -1641,7 +1760,7 @@ class ObjSpace(object):
         """
         assert w_obj is not None
         if not self.isinstance_w(w_obj, self.w_unicode):
-            w_obj._typed_unwrap_error(self, "unicode")
+            w_obj._typed_unwrap_error(self, "str")
         return w_obj.text_w(self)
 
     @not_rpython    # tests only; should be replaced with bytes_w or text_w
@@ -1670,7 +1789,6 @@ class ObjSpace(object):
 
     def text0_w(self, w_obj):
         "Like text_w, but rejects strings with NUL bytes."
-        from rpython.rlib import rstring
         result = self.text_w(w_obj)
         if '\x00' in result:
             raise oefmt(self.w_ValueError, "embedded null character")
@@ -1790,7 +1908,7 @@ class ObjSpace(object):
 
     def utf8_len_w(self, w_obj):
         w_obj = self.convert_arg_to_w_unicode(w_obj)
-        return w_obj._utf8, w_obj._len()
+        return w_obj.utf8_w(self), w_obj._len()
 
     def realutf8_w(self, w_obj):
         # Like utf8_w(), but only works if w_obj is really of type
@@ -1866,7 +1984,7 @@ class ObjSpace(object):
         return value
 
     def c_uint_w(self, w_obj):
-        # Like space.gateway_uint_w(), but raises an app-level OverflowError if
+        # Like space.gateway_r_uint_w(), but raises an app-level OverflowError if
         # the integer does not fit in 32 bits.  Here for gateway.py.
         value = self.uint_w(w_obj)
         if value > UINT_MAX:
@@ -1898,8 +2016,8 @@ class ObjSpace(object):
     def c_ushort_w(self, w_obj):
         value = self.int_w(w_obj)
         if value < 0:
-            raise oefmt(self.w_OverflowError,
-                "can't convert negative value to C unsigned short")
+            raise oefmt(self.w_ValueError,
+                "value must be positive")
         elif value > USHRT_MAX:
             raise oefmt(self.w_OverflowError,
                 "Python int too large for C unsigned short")
@@ -1976,12 +2094,28 @@ class ObjSpace(object):
             assert self.issubtype_w(w_warningcls, self.w_Warning)
         do_warn(self, w_msg, w_warningcls, stacklevel - 1)
 
+    def audit(self, event, args_w):
+        from pypy.module.sys.vm import audit
+        audit(self, event, args_w)
+
     def iterator_greenkey(self, w_iterable):
         """ Return something that can be used as a green key in jit drivers
         that iterate over self. by default, it's just the type of self, but
         custom iterators should override it. """
         return w_iterable.iterator_greenkey(self)
 
+    def guess_function_name_parens(self, w_function):
+        """ Returns 'funcname()' from callable w_function. If it's not a
+        function or a method, returns 'Classname object'"""
+        # XXX this is super annoying to compute every time we do a function call!
+        # CPython has a similar function, PyEval_GetFuncName
+        from pypy.interpreter.function import Function, _Method
+        if isinstance(w_function, Function):
+            return w_function.qualname + '()'
+        if isinstance(w_function, _Method):
+            return self.guess_function_name_parens(w_function.w_function)
+        return self.type(w_function).getname(self) + ' object'
+
 
 class AppExecCache(SpaceCache):
     @not_rpython
