@@ -1,12 +1,15 @@
 import sys
-from pypy.interpreter.error import OperationError, get_cleared_operation_error
+from pypy.interpreter.error import OperationError
 from rpython.rlib.unroll import unrolling_iterable
 from rpython.rlib.objectmodel import specialize, not_rpython
 from rpython.rlib import jit, rgc, objectmodel
+from rpython.rlib.rarithmetic import r_uint
 
 TICK_COUNTER_STEP = 100
 
 def app_profile_call(space, w_callable, frame, event, w_arg):
+    # from here on, frame is just a normal w_object
+    frame = jit.hint(frame, access_directly=False)
     space.call_function(w_callable,
                         frame,
                         space.newtext(event), w_arg)
@@ -23,8 +26,9 @@ class ExecutionContext(object):
     # XXX [fijal] but they're not. is_being_profiled is guarded a bit all
     #     over the place as well as w_tracefunc
 
-    _immutable_fields_ = ['profilefunc?', 'w_tracefunc?',
-                          'w_coroutine_wrapper_fn?']
+    _immutable_fields_ = [
+        'profilefunc?', 'w_tracefunc?',
+        'w_asyncgen_firstiter_fn?', 'w_asyncgen_finalizer_fn?']
 
     def __init__(self, space):
         self.space = space
@@ -39,8 +43,16 @@ class ExecutionContext(object):
         self.profilefunc = None
         self.w_profilefuncarg = None
         self.thread_disappeared = False   # might be set to True after os.fork()
-        self.w_coroutine_wrapper_fn = None
-        self.in_coroutine_wrapper = False
+        # an instance of this will be raised the next time we switch to the
+        # thread that self represents
+        self.w_async_exception_type = None
+
+        self.w_asyncgen_firstiter_fn = None
+        self.w_asyncgen_finalizer_fn = None
+        self.contextvar_context = None
+        self.coroutine_origin_tracking_depth = 0
+
+        self.current_gen_or_coroutine = None # A linked list of GeneratorOrCoroutine instances
 
     @staticmethod
     def _mark_thread_disappeared(space):
@@ -117,7 +129,7 @@ class ExecutionContext(object):
             if isinstance(w_func, FunctionWithFixedCode) and args is not None:
                 w_firstarg = args.firstarg()
                 if w_firstarg is not None:
-                    from pypy.interpreter.function import descr_function_get
+                    from pypy.interpreter.gateway import descr_function_get
                     w_func = descr_function_get(self.space, w_func, w_firstarg,
                                                 self.space.type(w_firstarg))
             #
@@ -173,49 +185,18 @@ class ExecutionContext(object):
 
     @jit.unroll_safe
     def run_trace_func(self, frame):
-        code = frame.pycode
+        code = frame.getcode() # promote the frame!
         d = frame.getorcreatedebug()
-        if d.instr_lb <= frame.last_instr < d.instr_ub:
-            if frame.last_instr < d.instr_prev_plus_one:
-                # We jumped backwards in the same line.
+        lastline = d.f_lineno
+        lineno = frame.pycode._get_lineno_for_pc_tracing(frame.last_instr)
+        d.f_lineno = lineno
+        if d.f_trace_lines and lineno != -1:
+            # when we are at a start of a line, or executing a backwards jump,
+            # produce a line event
+            if lastline != lineno or frame.last_instr < d.instr_prev_plus_one:
                 self._trace(frame, 'line', self.space.w_None)
-        else:
-            size = len(code.co_lnotab) / 2
-            addr = 0
-            line = code.co_firstlineno
-            p = 0
-            lineno = code.co_lnotab
-            while size > 0:
-                c = ord(lineno[p])
-                if (addr + c) > frame.last_instr:
-                    break
-                addr += c
-                if c:
-                    d.instr_lb = addr
-                line_offset = ord(lineno[p + 1])
-                if line_offset >= 0x80:
-                    line_offset -= 0x100
-                line += line_offset
-                p += 2
-                size -= 1
-
-            if size > 0:
-                while True:
-                    size -= 1
-                    if size < 0:
-                        break
-                    addr += ord(lineno[p])
-                    if ord(lineno[p + 1]):
-                        break
-                    p += 2
-                d.instr_ub = addr
-            else:
-                d.instr_ub = sys.maxint
-
-            if d.instr_lb == frame.last_instr: # At start of line!
-                d.f_lineno = line
-                self._trace(frame, 'line', self.space.w_None)
-
+        if d.f_trace_opcodes:
+            self._trace(frame, 'opcode', self.space.w_None)
         d.instr_prev_plus_one = frame.last_instr + 1
 
     @objectmodel.try_inline
@@ -238,18 +219,75 @@ class ExecutionContext(object):
     def sys_exc_info(self):
         """Implements sys.exc_info().
         Return an OperationError instance or None.
+        Returns the "top-most" exception in the stack.
 
         # NOTE: the result is not the wrapped sys.exc_info() !!!
 
         """
+        operr = self.sys_exc_operror
+        if operr is not None:
+            return operr
+        if self.current_gen_or_coroutine:
+            return self._get_topmost_exception()
+        return None
+
+    def current_exception(self):
         return self.sys_exc_operror
 
     def set_sys_exc_info(self, operror):
         self.sys_exc_operror = operror
 
+    def set_sys_exc_info3(self, w_value):
+        from pypy.interpreter import pytraceback
+        from pypy.module.exceptions.interp_exceptions import W_BaseException
+
+        space = self.space
+        if space.is_none(w_value):
+            self.sys_exc_operror = None
+            return
+        w_value = space.interp_w(W_BaseException, w_value)
+        w_type = space.type(w_value)
+        tb = w_value.w_traceback
+        operror = OperationError(w_type, w_value, tb)
+        self.set_sys_exc_info(operror)
+
+    def push_gen_or_coroutine(self, gen):
+        current_operr = self.sys_exc_operror
+        self.set_sys_exc_info(gen.saved_operr)
+        gen.saved_operr = current_operr
+        gen.previous_gen_or_coroutine = self.current_gen_or_coroutine
+        self.current_gen_or_coroutine = gen
+
+    def pop_gen_or_coroutine(self, gen):
+        assert self.current_gen_or_coroutine is gen
+        saved_operr = gen.saved_operr
+        self.current_gen_or_coroutine = gen.previous_gen_or_coroutine
+        gen.previous_gen_or_coroutine = None
+        gen.saved_operr = self.sys_exc_operror
+        self.sys_exc_operror = saved_operr
+
+    def save_and_reset_exc_info_state(self):
+        res = self.sys_exc_operror, self.current_gen_or_coroutine
+        self.sys_exc_operror = None
+        self.current_gen_or_coroutine = None
+        return res
+
+    def restore_save_exc_info_state(self, state):
+        self.sys_exc_operror, self.current_gen_or_coroutine = state
+
+    def _get_topmost_exception(self):
+        gen = self.current_gen_or_coroutine
+        while gen:
+            if gen.saved_operr:
+                return gen.saved_operr
+            gen = gen.previous_gen_or_coroutine
+        return None
+
+
     @jit.dont_look_inside
     def settrace(self, w_func):
         """Set the global trace function."""
+        # self.space.audit("sys.settrace", [])
         if self.space.is_w(w_func, self.space.w_None):
             self.w_tracefunc = None
         else:
@@ -264,6 +302,7 @@ class ExecutionContext(object):
 
     def setprofile(self, w_func):
         """Set the global trace function."""
+        # self.space.audit("sys.setprofile", [])
         if self.space.is_w(w_func, self.space.w_None):
             self.profilefunc = None
             self.w_profilefuncarg = None
@@ -318,12 +357,16 @@ class ExecutionContext(object):
 
         if w_callback is not None and event != "leaveframe":
             if operr is not None:
-                operr.normalize_exception(space)
-                w_value = operr.get_w_value(space)
+                w_value = operr.normalize_exception(space)
                 w_arg = space.newtuple([operr.w_type, w_value,
                                         operr.get_w_traceback(space)])
 
-            d = frame.getorcreatedebug()
+            lineno = frame.pycode._get_lineno_for_pc_tracing(frame.last_instr)
+            if frame.last_instr >= 1:
+                init_lineno = lineno
+            else:
+                init_lineno = -1
+            d = frame.getorcreatedebug(init_lineno=init_lineno)
             if d.w_locals is not None:
                 # only update the w_locals dict if it exists
                 # if it does not exist yet and the tracer accesses it via
@@ -331,10 +374,14 @@ class ExecutionContext(object):
                 frame.fast2locals()
             prev_line_tracing = d.is_in_line_tracing
             self.is_tracing += 1
+            old_lineno = d.f_lineno
             try:
                 if event == 'line':
                     d.is_in_line_tracing = True
                 try:
+                    d.f_lineno = lineno
+                    # from here on, frame is just a normal w_object
+                    frame = jit.hint(frame, access_directly=False)
                     w_result = space.call_function(w_callback, frame, space.newtext(event), w_arg)
                     if space.is_w(w_result, space.w_None):
                         # bug-to-bug compatibility with CPython
@@ -347,6 +394,8 @@ class ExecutionContext(object):
                     d.w_f_trace = None
                     raise
             finally:
+                if d.f_lineno == lineno:
+                    d.f_lineno = old_lineno
                 self.is_tracing -= 1
                 d.is_in_line_tracing = prev_line_tracing
                 if d.w_locals is not None:
@@ -418,8 +467,11 @@ class AbstractActionFlag(object):
     def __init__(self):
         self._periodic_actions = []
         self._nonperiodic_actions = []
+        # a bitmask where the ith bit is set if self._nonperiodic_actions[i]
+        # has fired
+        self._fired_bitmask = r_uint(0)
+
         self.has_bytecode_counter = False
-        self._fired_actions_reset()
         # the default value is not 100, unlike CPython 2.7, but a much
         # larger value, because we use a technique that not only allows
         # but actually *forces* another thread to run whenever the counter
@@ -429,30 +481,14 @@ class AbstractActionFlag(object):
 
     def fire(self, action):
         """Request for the action to be run before the next opcode."""
-        if not action._fired:
-            action._fired = True
-            self._fired_actions_append(action)
+        assert action._action_index >= 0 # period actions must not call fire
+        mask = r_uint(1) << action._action_index
+        if not self._fired_bitmask & mask:
+            self._fired_bitmask |= mask
             # set the ticker to -1 in order to force action_dispatcher()
             # to run at the next possible bytecode
             self.reset_ticker(-1)
 
-    def _fired_actions_reset(self):
-        # linked list of actions. We cannot use a normal RPython list because
-        # we want AsyncAction.fire() to be marked as @rgc.collect: this way,
-        # we can call it from e.g. GcHooks or cpyext's dealloc_trigger.
-        self._fired_actions_first = None
-        self._fired_actions_last = None
-
-    @rgc.no_collect
-    def _fired_actions_append(self, action):
-        assert action._next is None
-        if self._fired_actions_first is None:
-            self._fired_actions_first = action
-            self._fired_actions_last = action
-        else:
-            self._fired_actions_last._next = action
-            self._fired_actions_last = action
-
     @not_rpython
     def register_periodic_action(self, action, use_bytecode_counter):
         """
@@ -473,6 +509,13 @@ class AbstractActionFlag(object):
             self._periodic_actions.insert(0, action)
         self._rebuild_action_dispatcher()
 
+    @not_rpython
+    def register_nonperiodic_action(self, action):
+        self._nonperiodic_actions.append(action)
+        assert len(self._nonperiodic_actions) < 32
+        self._rebuild_action_dispatcher()
+        return len(self._nonperiodic_actions) - 1
+
     def getcheckinterval(self):
         return self.checkinterval_scaled // TICK_COUNTER_STEP
 
@@ -497,26 +540,25 @@ class AbstractActionFlag(object):
                 action.perform(ec, frame)
 
             # nonperiodic actions
-            action = self._fired_actions_first
-            if action:
-                self._fired_actions_reset()
+            if self._fired_bitmask:
                 # NB. in case there are several actions, we reset each
-                # 'action._fired' to false only when we're about to call
+                # fired bit to 0  only when we're about to call
                 # 'action.perform()'.  This means that if
                 # 'action.fire()' happens to be called any time before
                 # the corresponding perform(), the fire() has no
                 # effect---which is the effect we want, because
-                # perform() will be called anyway.  All such pending
-                # actions with _fired == True are still inside the old
-                # chained list.  As soon as we reset _fired to False,
-                # we also reset _next to None and we are ready for
-                # another fire().
-                while action is not None:
-                    next_action = action._next
-                    action._next = None
-                    action._fired = False
-                    action.perform(ec, frame)
-                    action = next_action
+                # perform() will be called anyway.
+                for i in range(len(self._nonperiodic_actions)):
+                    mask = r_uint(1) << i
+                    if self._fired_bitmask & mask:
+                        action = self._nonperiodic_actions[i]
+                        self._fired_bitmask  &= ~mask
+                        action.perform(ec, frame)
+                # one of the actions with higher index re-triggered one of the
+                # earlier actions. don't run them in a loop here, execute some
+                # python code first.
+                if self._fired_bitmask:
+                    self.reset_ticker(-1)
 
         self.action_dispatcher = action_dispatcher
 
@@ -548,11 +590,14 @@ class AsyncAction(object):
     asynchronously with regular bytecode execution, but that still need
     to occur between two opcodes, not at a completely random time.
     """
-    _fired = False
-    _next = None
 
+    @not_rpython
     def __init__(self, space):
         self.space = space
+        if not isinstance(self, PeriodicAsyncAction):
+            self._action_index = self.space.actionflag.register_nonperiodic_action(self)
+        else:
+            self._action_index = -1
 
     @rgc.no_collect
     def fire(self):
@@ -636,7 +681,7 @@ class UserDelAction(AsyncAction):
                 try:
                     space.call_function(w_impl)
                 except Exception as e:
-                    report_error(space, e, '', w_impl)
+                    report_error(space, e, '', w_del)
 
         # Call the RPython-level _finalize_() method.
         try:
