@@ -1,4 +1,7 @@
 from pypy.interpreter.pyparser import parsestring
+from pypy.interpreter.pyparser.error import SyntaxError
+from pypy.interpreter.pyparser.parser import Token
+
 import py, sys
 
 class TestParsetring:
@@ -43,13 +46,12 @@ class TestParsetring:
         # you can use escapes to get the non-ASCII ones (note that in the
         # second case we use a raw string, the the parser actually sees the
         # chars '\' 'x' 'e' '9'
-        space.raises_w(space.w_SyntaxError,
-                       parsestring.parsestr, space, None, "b'\xe9'")
+        excinfo = py.test.raises(SyntaxError,
+                       parsestring.parsestr, space, None,
+                       "b'     \xe9'", Token(None, -1, "b'     \xe9'", 5, 0))
         self.parse_and_compare(r"b'\xe9'", chr(0xE9))
 
-
     def test_unicode(self):
-        space = self.space
         for s in ['hello world', 'hello\n world']:
             self.parse_and_compare(repr(s), unicode(s))
 
@@ -106,7 +108,7 @@ class TestParsetring:
         s = s.decode("koi8-u").encode("utf8")
         w_ret = parsestring.parsestr(self.space, 'koi8-u', s)
         ret = space.unwrap(w_ret)
-        assert ret == eval("# -*- coding: koi8-u -*-\nu'\x81\\t'") 
+        assert ret == eval("# -*- coding: koi8-u -*-\nu'\x81\\t'")
 
     def test_multiline_unicode_strings_with_backslash(self):
         space = self.space
@@ -132,3 +134,8 @@ class TestParsetring:
         buf = parsestring.decode_unicode_utf8(self.space,
                                               'u"\xf0\x9f\x92\x8b"', 2, 6)
         assert buf == r"\U0001f48b"
+
+    def test_parsestr_segfault(self):
+        space = self.space
+        space.raises_w(space.w_UnicodeDecodeError,
+                       parsestring.parsestr, space, None, "r'\xc2'")
