@@ -24,7 +24,8 @@ def test_yield_in_nested_try_excepts():
         coro.throw(SubError())
 
 def test_generator_doesnt_retain_old_exc2():
-    #Issue 28884#msg282532
+    # Issue bpo 28884#msg282532
+    # Fixed in CPython via https://github.com/python/cpython/pull/1773
     import sys
     def g():
         try:
@@ -57,4 +58,17 @@ def test_raise_in_generator():
             next(i)
             next(i)
 
+def test_assertion_error_global_ignored():
+    if hasattr(pytest, 'py3k_skip'):
+        pytest.py3k_skip('only untranslated')
+    global AssertionError
 
+    class Foo(Exception):
+        pass
+    OrigAssertionError = AssertionError
+    AssertionError = Foo
+    try:
+        with pytest.raises(OrigAssertionError): # not Foo!
+            exec("assert 0") # to stop the pytest ast rewriting from touching it
+    finally:
+        AssertionError = OrigAssertionError
