@@ -7,7 +7,7 @@ class TestCode:
     def test_code_eq_corner_cases(self):
         space = self.space
         def make_code_with_const(w_obj):
-            return PyCode(space, 0, 0, 0, 1, 0, '', [w_obj], [], [], '', '', 0, '', [], [], False)
+            return PyCode(space, 0, 0, 0, 0, 1, 0, '', [w_obj], [], [], '', '', '', 0, '', [], [], False)
         def cmp_code_consts(w_obj1, w_obj2):
             w_code1 = make_code_with_const(w_obj1)
             w_code2 = make_code_with_const(w_obj2)
@@ -77,6 +77,7 @@ class AppTestCodeIntrospection:
                            'co_names': (),
                            'co_varnames': (),
                            'co_argcount': 0,
+                           'co_posonlyargcount': 0,
                            'co_kwonlyargcount': 0,
                            'co_consts': (None,)
                            }),
@@ -84,6 +85,7 @@ class AppTestCodeIntrospection:
                            'co_names': (),
                            'co_varnames': ('x', 'y', 'z'),
                            'co_argcount': 1,
+                           'co_posonlyargcount': 0,
                            'co_kwonlyargcount': 0,
                            'co_consts': ("docstring", None),
                            }),
@@ -95,6 +97,7 @@ class AppTestCodeIntrospection:
                 (abs.__code__, {'co_name': 'abs',
                                  'co_varnames': ('val',),
                                  'co_argcount': 1,
+                           'co_posonlyargcount': 0,
                                  'co_kwonlyargcount': 0,
                                  'co_flags': 0,
                                  'co_consts': ("abs(number) -> number\n\nReturn the absolute value of the argument.",),
@@ -103,6 +106,7 @@ class AppTestCodeIntrospection:
                                 {#'co_name': '__init__',   XXX getting descr__init__
                                  'co_varnames': ('obj', 'args', 'keywords'),
                                  'co_argcount': 1,
+                                 'co_posonlyargcount': 0,
                                  'co_kwonlyargcount': 0,
                                  'co_flags': 0x000C,  # VARARGS|VARKEYWORDS
                                  }),
@@ -116,6 +120,15 @@ class AppTestCodeIntrospection:
             for key, value in expected.items():
                 assert getattr(code, key) == value
 
+
+    def test_posonlyargcount(self):
+        """
+        def f(a, b, /, c): pass
+        assert f.__code__.co_argcount == 3
+        assert f.__code__.co_posonlyargcount == 2
+        """
+
+
     def test_kwonlyargcount(self):
         """
         def f(*args, a, b, **kw): pass
@@ -138,77 +151,6 @@ class AppTestCodeIntrospection:
 
         assert list(sorted(d['f'].__code__.co_names)) == ['foo', 'g']
 
-    def test_code(self):
-        import sys
-        try: 
-            import new
-        except ImportError: 
-            skip("could not import new module")
-        codestr = "global c\na = 1\nb = 2\nc = a + b\n"
-        ccode = compile(codestr, '<string>', 'exec')
-        co = new.code(ccode.co_argcount,
-                      ccode.co_nlocals,
-                      ccode.co_stacksize,
-                      ccode.co_flags,
-                      ccode.co_code,
-                      ccode.co_consts,
-                      ccode.co_names,
-                      ccode.co_varnames,
-                      ccode.co_filename,
-                      ccode.co_name,
-                      ccode.co_firstlineno,
-                      ccode.co_lnotab,
-                      ccode.co_freevars,
-                      ccode.co_cellvars)
-        d = {}
-        exec(co, d)
-        assert d['c'] == 3
-        # test backwards-compatibility version with no freevars or cellvars
-        co = new.code(ccode.co_argcount,
-                      ccode.co_nlocals,
-                      ccode.co_stacksize,
-                      ccode.co_flags,
-                      ccode.co_code,
-                      ccode.co_consts,
-                      ccode.co_names,
-                      ccode.co_varnames,
-                      ccode.co_filename,
-                      ccode.co_name,
-                      ccode.co_firstlineno,
-                      ccode.co_lnotab)
-        d = {}
-        exec(co, d)
-        assert d['c'] == 3
-        def f(x):
-            y = 1
-        ccode = f.__code__
-        raises(ValueError, new.code,
-              -ccode.co_argcount,
-              ccode.co_nlocals,
-              ccode.co_stacksize,
-              ccode.co_flags,
-              ccode.co_code,
-              ccode.co_consts,
-              ccode.co_names,
-              ccode.co_varnames,
-              ccode.co_filename,
-              ccode.co_name,
-              ccode.co_firstlineno,
-              ccode.co_lnotab)
-        raises(ValueError, new.code,
-              ccode.co_argcount,
-              -ccode.co_nlocals,
-              ccode.co_stacksize,
-              ccode.co_flags,
-              ccode.co_code,
-              ccode.co_consts,
-              ccode.co_names,
-              ccode.co_varnames,
-              ccode.co_filename,
-              ccode.co_name,
-              ccode.co_firstlineno,
-              ccode.co_lnotab)
-
     def test_hash(self):
         d1 = {}
         exec("def f(): pass", d1)
@@ -250,8 +192,29 @@ class AppTestCodeIntrospection:
         assert d['f'](4).__code__.co_flags & 0x10
         assert d['f'].__code__.co_flags & 0x10 == 0
 
+    def test_code_eq_non_code(self):
+        class A(object):
+            def __eq__(self, other):
+                return 23
+            def __ne__(self, other):
+                return 41
+        def f(): pass
+        assert (f.__code__ == A()) == 23
+        assert (f.__code__ != A()) == 41
+
     def test_issue1844(self):
         import types
-        args = (1, 0, 1, 0, 0, b'', (), (), (), '', 'operator', 0, b'')
+        args = (1, 0, 0, 1, 0, 0, b'', (), (), (), '', 'operator', 'operator', 0, b'')
         # previously raised a MemoryError when translated
         types.CodeType(*args)
+
+    def test_replace(self):
+        co = compile("x = x + 1", 'baz', 'exec')
+        co2 = co.replace(co_flags=co.co_flags | 0x100)
+        assert co2.co_name == co.co_name # in theory need to check them all
+        assert co2.co_flags == co.co_flags | 0x100
+
+        with raises(TypeError):
+            co.replace(1)
+        with raises(TypeError):
+            co.replace(abc=123)
