@@ -11,9 +11,9 @@ from pypy.interpreter.nestedscope import Cell
 from pypy.interpreter.function import Function
 
 @unwrap_spec(filename='fsencode', mode='text', flags=int, dont_inherit=int,
-             optimize=int)
+             optimize=int, _feature_version=int)
 def compile(space, w_source, filename, mode, flags=0, dont_inherit=0,
-            optimize=-1):
+            optimize=-1, _feature_version=-1):
     """Compile the source string (a Python module, statement or expression)
 into a code object that can be executed by the exec statement or eval().
 The filename will be used for run-time error messages.
@@ -27,27 +27,47 @@ compile; if absent or zero these statements do influence the compilation,
 in addition to any features explicitly specified.
 """
     from pypy.interpreter.pyopcode import source_as_str
+    # only allow default value of _feature_version for now
+    # we need to support the keyword argument, the ast module passes it (set to
+    # -1, usually)
+    if _feature_version >= 0 and (flags & consts.PyCF_ONLY_AST):
+        feature_version = _feature_version
+    else:
+        feature_version = -1
+
     ec = space.getexecutioncontext()
     if flags & ~(ec.compiler.compiler_flags | consts.PyCF_ONLY_AST |
                  consts.PyCF_DONT_IMPLY_DEDENT | consts.PyCF_SOURCE_IS_UTF8 |
-                 consts.PyCF_ACCEPT_NULL_BYTES):
+                 consts.PyCF_ACCEPT_NULL_BYTES | consts.PyCF_TYPE_COMMENTS |
+                 consts.PyCF_ALLOW_TOP_LEVEL_AWAIT | consts.PyCF_ALLOW_INCOMPLETE_INPUT |
+                 consts.PyCF_IGNORE_COOKIE):
         raise oefmt(space.w_ValueError, "compile() unrecognized flags")
 
+    only_ast = flags & consts.PyCF_ONLY_AST
+
     if not dont_inherit:
         caller = ec.gettopframe_nohidden()
         if caller:
             flags |= ec.compiler.getcodeflags(caller.getcode())
 
-    if mode not in ('exec', 'eval', 'single'):
+    if mode not in ('exec', 'eval', 'single', 'func_type'):
+        if only_ast:
+            raise oefmt(space.w_ValueError,
+                        "compile() mode must be 'exec', 'eval', 'single' or 'func_type'")
+        else:
+            raise oefmt(space.w_ValueError,
+                        "compile() arg 3 must be 'exec', 'eval' or 'single'")
+
+    if mode == "func_type" and not only_ast:
         raise oefmt(space.w_ValueError,
-                    "compile() arg 3 must be 'exec', 'eval' or 'single'")
+                    "compile() mode 'func_type' requires flag PyCF_ONLY_AST")
 
     if optimize < -1 or optimize > 2:
         raise oefmt(space.w_ValueError,
             "compile(): invalid optimize value")
 
     if space.isinstance_w(w_source, space.gettypeobject(ast.W_AST.typedef)):
-        if flags & consts.PyCF_ONLY_AST:
+        if only_ast:
             return w_source
         ast_node = ast.mod.from_object(space, w_source)
         ec.compiler.validate_ast(ast_node)
@@ -58,8 +78,8 @@ in addition to any features explicitly specified.
     source, flags = source_as_str(space, w_source, 'compile',
                                   "string, bytes or AST", flags)
 
-    if flags & consts.PyCF_ONLY_AST:
-        node = ec.compiler.compile_to_ast(source, filename, mode, flags)
+    if only_ast:
+        node = ec.compiler.compile_to_ast(source, filename, mode, flags, feature_version)
         return node.to_object(space)
     else:
         return ec.compiler.compile(source, filename, mode, flags,
@@ -78,6 +98,7 @@ If only globals is given, locals defaults to it.
 
     if space.isinstance_w(w_prog, space.gettypeobject(PyCode.typedef)):
         code = space.interp_w(PyCode, w_prog)
+        space.audit("exec", [w_prog])
     else:
         source, flags = source_as_str(space, w_prog, 'eval',
                                       "string, bytes or code",
@@ -85,31 +106,54 @@ If only globals is given, locals defaults to it.
         ec = space.getexecutioncontext()
         code = ec.compiler.compile(source.lstrip(' \t'), "<string>", 'eval',
                                    flags)
-
-    # XXX: skip adding of __builtins__ to w_globals. it requires a
-    # costly gettopframe_nohidden() here and nobody seems to miss its
-    # absence
+    if not space.contains_w(w_globals, space.newtext("__builtins__")):
+        space.setitem_str(w_globals, "__builtins__", space.builtin)
 
     return code.exec_code(space, w_globals, w_locals)
 
-def exec_(space, w_prog, w_globals=None, w_locals=None):
+def exec_(space, w_prog, w_globals=None, w_locals=None, __posonly__=None, w_closure=None):
+    """
+    exec(source, globals=None, locals=None, *, closure=None)
+
+    Execute the given source in the context of globals and locals.
+
+    The source may be a string representing one or more Python statements
+    or a code object as returned by compile().
+    The globals must be a dictionary and locals can be any mapping,
+    defaulting to the current globals and locals.
+    If only globals is given, locals defaults to it.
+
+    The closure must be a tuple of cellvars, and can only be used when source
+    is a code object requiring exactly that many cellvars.
+    """
+
     frame = space.getexecutioncontext().gettopframe()
-    frame.exec_(w_prog, w_globals, w_locals)
+    frame.exec_(w_prog, w_globals, w_locals, w_closure)
 
 def _update_bases(space, w_bases, bases_w):
-    new_bases_w = []
-    changed = False
-    for w_base in bases_w:
-        w_meth = space.lookup(w_base, '__mro_entries__')
-        if w_meth is not None:
-            new_base_w = space.get_and_call_function(w_meth, w_base, w_bases)
-            if not space.isinstance_w(new_base_w, space.w_tuple):
-                raise oefmt(space.w_TypeError, "__mro_entries__ must return a tuple")
-            new_bases_w.extend(space.fixedview(new_base_w))
-            changed = True
+    new_bases_w = None
+    for i, w_base in enumerate(bases_w):
+        if space.isinstance_w(w_base, space.w_type):
+            if new_bases_w is not None:
+                new_bases_w.append(w_base)
+            continue
+        w_meth = None
+        try:
+            # one of the few places where cpython uses getattr not lookup
+            w_meth = space.getattr(w_base, space.newtext('__mro_entries__'))
+        except OperationError as e:
+            if not e.match(space, space.w_AttributeError):
+                raise
+            if new_bases_w is not None:
+                new_bases_w.append(w_base)
         else:
-            new_bases_w.append(w_base)
-    if not changed:
+            w_new_base = space.call_function(w_meth, w_bases)
+            if not space.isinstance_w(w_new_base, space.w_tuple):
+                raise oefmt(space.w_TypeError, "__mro_entries__ must return a tuple")
+            if new_bases_w is None:
+                new_bases_w = bases_w[:i]
+            new_bases_w.extend(space.fixedview(w_new_base))
+    if new_bases_w is None:
         return bases_w
     return new_bases_w[:]
 
@@ -141,12 +185,13 @@ def build_class(space, w_func, w_name, __args__):
     except OperationError as e:
         if not e.match(space, space.w_AttributeError):
             raise
-        w_namespace = space.newdict()
+        w_namespace = space.newdict(module=True)
     else:
-        keywords = kwds_w.keys()
+        # XXX mess
+        keyword_names_w = [space.newtext(kwd) for kwd in kwds_w.keys()]
         args = Arguments(space,
                          args_w=[w_name, w_bases],
-                         keywords=keywords,
+                         keyword_names_w=keyword_names_w,
                          keywords_w=kwds_w.values())
         w_namespace = space.call_args(w_prep, args)
     if not space.ismapping_w(w_namespace):
@@ -165,10 +210,10 @@ def build_class(space, w_func, w_name, __args__):
     w_cell = frame.run()
     if bases_w is not orig_bases_w:
         space.setitem(w_namespace, space.newtext("__orig_bases__"), w_orig_bases)
-    keywords = kwds_w.keys()
+    keyword_names_w = [space.newtext(kwd) for kwd in kwds_w.keys()]
     args = Arguments(space,
                      args_w=[w_name, w_bases, w_namespace],
-                     keywords=keywords,
+                     keyword_names_w=keyword_names_w,
                      keywords_w=kwds_w.values())
     try:
         w_class = space.call_args(w_meta, args)
@@ -186,15 +231,12 @@ def build_class(space, w_func, w_name, __args__):
             e.get_w_value(space))
     if isinstance(w_cell, Cell) and isinstance(w_class, W_TypeObject):
         if w_cell.empty():
-            # will become an error in Python 3.7
-            space.warn(space.newtext(
-                "__class__ not set defining %s as %s . "
-                "Was __classcell__ propagated to type.__new__?" % (
-                    space.text_w(w_name),
-                    space.text_w(space.str(w_class))
-                )),
-                space.w_DeprecationWarning)
-            w_cell.set(w_class)
+            raise oefmt(space.w_RuntimeError,
+                "__class__ not set defining %S as %S. "
+                "Was __classcell__ propagated to type.__new__?",
+                    w_name,
+                    w_class)
+
         else:
             w_class_from_cell = w_cell.get()
             if not space.is_w(w_class, w_class_from_cell):
