@@ -23,6 +23,8 @@ class W_Super(W_Root):
             w_type = None  # unbound super object
             w_obj_or_type = None
         else:
+            if not space.isinstance_w(w_starttype, space.w_type):
+                raise oefmt(space.w_TypeError, "super() argument 1 must be a type, not %T", w_starttype)
             w_type = _super_check(space, w_starttype, w_obj_or_type)
         self.w_starttype = w_starttype
         self.w_objtype = w_type
@@ -164,7 +166,7 @@ class C(B):
 
 
 class W_Property(W_Root):
-    _immutable_fields_ = ["w_fget", "w_fset", "w_fdel"]
+    _immutable_fields_ = ["w_fget?", "w_fset?", "w_fdel?"]
 
     def __init__(self, space):
         self.w_fget = space.w_None
@@ -172,6 +174,7 @@ class W_Property(W_Root):
         self.w_fdel = space.w_None
         self.w_doc = space.w_None
         self.getter_doc = False
+        self.w_name = None
 
     @unwrap_spec(w_fget=WrappedDefault(None),
                  w_fset=WrappedDefault(None),
@@ -194,22 +197,37 @@ class W_Property(W_Root):
                     space.setattr(self, space.newtext('__doc__'), w_getter_doc)
                 self.getter_doc = True
 
+    def _properror(self, space, w_obj, kind):
+        qualname = space.type(w_obj).qualname
+        if self.w_name is not None:
+            raise oefmt(
+                space.w_AttributeError,
+                "property %R of '%s' object has no %s",
+                self.w_name, qualname, kind)
+        else:
+            raise oefmt(
+                space.w_AttributeError,
+                "property of '%s' object has no %s",
+                qualname, kind)
+
+
+
     def get(self, space, w_obj, w_objtype=None):
         if space.is_w(w_obj, space.w_None):
             return self
         if space.is_w(self.w_fget, space.w_None):
-            raise oefmt(space.w_AttributeError, "unreadable attribute")
+            self._properror(space, w_obj, 'getter')
         return space.call_function(self.w_fget, w_obj)
 
     def set(self, space, w_obj, w_value):
         if space.is_w(self.w_fset, space.w_None):
-            raise oefmt(space.w_AttributeError, "can't set attribute")
+            self._properror(space, w_obj, 'setter')
         space.call_function(self.w_fset, w_obj, w_value)
         return space.w_None
 
     def delete(self, space, w_obj):
         if space.is_w(self.w_fdel, space.w_None):
-            raise oefmt(space.w_AttributeError, "can't delete attribute")
+            self._properror(space, w_obj, 'deleter')
         space.call_function(self.w_fdel, w_obj)
         return space.w_None
 
@@ -241,14 +259,24 @@ class W_Property(W_Root):
         else:
             w_doc = self.w_doc
         w_type = self.getclass(space)
-        return space.call_function(w_type, w_getter, w_setter, w_deleter,
-                                   w_doc)
+        w_res = space.call_function(w_type, w_getter, w_setter, w_deleter,
+                                    w_doc)
+        if isinstance(w_res, W_Property):
+            w_res.w_name = self.w_name
+        return w_res
+
+    def set_name(self, space, w_type, w_name):
+        """ Method to set name of a property. """
+        self.w_name = w_name
 
     def descr_isabstract(self, space):
         return space.newbool(space.isabstractmethod_w(self.w_fget) or
                              space.isabstractmethod_w(self.w_fset) or
                              space.isabstractmethod_w(self.w_fdel))
 
+    def descr_reduce(self, space):
+        raise oefmt(space.w_TypeError, "cannot pickle 'property' object")
+
 W_Property.typedef = TypeDef(
     'property',
     __doc__ = '''\
@@ -268,12 +296,14 @@ class C(object):
     __set__ = interp2app(W_Property.set),
     __delete__ = interp2app(W_Property.delete),
     __isabstractmethod__ = GetSetProperty(W_Property.descr_isabstract),
+    __reduce__ = interp2app(W_Property.descr_reduce),
     fdel = interp_attrproperty_w('w_fdel', W_Property),
     fget = interp_attrproperty_w('w_fget', W_Property),
     fset = interp_attrproperty_w('w_fset', W_Property),
     getter = interp2app(W_Property.getter),
     setter = interp2app(W_Property.setter),
     deleter = interp2app(W_Property.deleter),
+    __set_name__ = interp2app(W_Property.set_name),
 )
 # This allows there to be a __doc__ of the property type and a __doc__
 # descriptor for the instances.
