@@ -38,34 +38,23 @@ def len(space, w_obj):
     return space.len(w_obj)
 
 
-def checkattrname(space, w_name, msg):
-    # This is a check to ensure that getattr/setattr/delattr only pass a
-    # ascii string to the rest of the code.  XXX not entirely sure if these
-    # functions are the only way for non-string objects to reach
-    # space.{get,set,del}attr()...
-    # Note that if w_name is already an exact string it must be ascii encoded
+def checkattrname(space, w_name):
     if not space.isinstance_w(w_name, space.w_text):
-        try:
-            name = space.text_w(w_name)    # typecheck
-        except OperationError as e:
-            if e.match(space, space.w_UnicodeError):
-                raise e
-            raise oefmt(space.w_TypeError,
-                 "%s(): attribute name must be string", msg)
-        w_name = space.newtext(name)
+        raise oefmt(space.w_TypeError,
+             "attribute name must be string, not '%T'", w_name)
     return w_name
 
 def delattr(space, w_object, w_name):
     """Delete a named attribute on an object.
 delattr(x, 'y') is equivalent to ``del x.y''."""
-    w_name = checkattrname(space, w_name, 'delattr')
+    w_name = checkattrname(space, w_name)
     space.delattr(w_object, w_name)
     return space.w_None
 
 def getattr(space, w_object, w_name, w_defvalue=None):
     """Get a named attribute from an object.
 getattr(x, 'y') is equivalent to ``x.y''."""
-    w_name = checkattrname(space, w_name, 'getattr')
+    w_name = checkattrname(space, w_name)
     try:
         return space.getattr(w_object, w_name)
     except OperationError as e:
@@ -77,7 +66,7 @@ getattr(x, 'y') is equivalent to ``x.y''."""
 def hasattr(space, w_object, w_name):
     """Return whether the object has an attribute with the given name.
     (This is done by calling getattr(object, name) and catching exceptions.)"""
-    w_name = checkattrname(space, w_name, 'hasattr')
+    w_name = checkattrname(space, w_name)
     try:
         space.getattr(w_object, w_name)
     except OperationError as e:
@@ -95,7 +84,9 @@ two un-equal objects to have the same hash value."""
 
 def id(space, w_object):
     "Return the identity of an object: id(x) == id(y) if and only if x is y."
-    return space.id(w_object)
+    w_res = space.id(w_object)
+    space.audit("builtins.id", [w_res])
+    return w_res
 
 def divmod(space, w_x, w_y):
     """Return the tuple ((x-x%y)/y, x%y).  Invariant: div*y + mod == x."""
@@ -126,7 +117,10 @@ iter_sentinel = gateway.applevel('''
 
     def iter_generator(callable_, sentinel):
         while 1:
-            result = callable_()
+            try:
+                result = callable_()
+            except StopIteration:
+                return
             if result == sentinel:
                 return
             yield result
@@ -164,12 +158,14 @@ def ord(space, w_val):
     """Return the integer ordinal of a character."""
     return space.ord(w_val)
 
-@unwrap_spec(w_modulus = WrappedDefault(None))
-def pow(space, w_base, w_exponent, w_modulus):
-    """With two arguments, equivalent to ``base**exponent''.
-With three arguments, equivalent to ``(base**exponent) % modulus'',
-but much more efficient for large exponents."""
-    return space.pow(w_base, w_exponent, w_modulus)
+@unwrap_spec(w_mod=WrappedDefault(None))
+def pow(space, w_base, w_exp, w_mod):
+    """With two arguments, equivalent to ``base**exp''.
+With three arguments, equivalent to ``(base**exp) % mod''.
+
+Some types, such as ints, are able to use a more efficient algorithm when
+invoked using the three argument form."""
+    return space.pow(w_base, w_exp, w_mod)
 
 def repr(space, w_object):
     """Return a canonical string representation of the object.
@@ -179,7 +175,7 @@ For simple object types, eval(repr(object)) == object."""
 def setattr(space, w_object, w_name, w_val):
     """Store a named attribute into an object.
 setattr(x, 'y', z) is equivalent to ``x.y = z''."""
-    w_name = checkattrname(space, w_name, 'setattr')
+    w_name = checkattrname(space, w_name)
     space.setattr(w_object, w_name, w_val)
     return space.w_None
 
