@@ -237,7 +237,7 @@ class W_CData(W_Root):
                 w_value.get_array_length() == length):
                 # fast path: copying from exactly the correct type
                 with w_value as source:
-                    source = rffi.cast(rffi.VOIDP, source)
+                    source = rffi.cast(rffi.CONST_VOIDP, source)
                     target = rffi.cast(rffi.VOIDP, target)
                     size = rffi.cast(rffi.SIZE_T, ctitemsize * length)
                     rffi.c_memcpy(target, source, size)
@@ -489,6 +489,9 @@ class W_CData(W_Root):
     def descr_exit(self, args_w):
         self.enter_exit(True)
 
+    def get_maximum_buffer_size(self):
+        return -1
+
 
 class W_CDataMem(W_CData):
     """This is used only by the results of cffi.cast('int', x)
@@ -554,21 +557,28 @@ class W_CDataNewOwning(W_CData):
 
 class W_CDataNewStd(W_CDataNewOwning):
     """Subclass using the standard allocator, lltype.malloc()/lltype.free()"""
-    _attrs_ = ['explicitly_freed']
-    explicitly_freed = False
+    _attrs_ = ['datasize']    # changed to -1 after being explicitly freed
+
+    def __init__(self, space, cdata, ctype, length, datasize):
+        W_CDataNewOwning.__init__(self, space, cdata, ctype, length)
+        assert datasize >= 0
+        self.datasize = datasize
 
     @rgc.must_be_light_finalizer
     def __del__(self):
-        if not self.explicitly_freed:
+        if self.datasize >= 0:
             lltype.free(self._ptr, flavor='raw')
 
     def _do_exit(self):
-        if not self.explicitly_freed:
-            rgc.add_memory_pressure(-self._sizeof(), self)
-            self.explicitly_freed = True
+        if self.datasize >= 0:
+            rgc.add_memory_pressure(-self.datasize, self)
+            self.datasize = -1
             rgc.may_ignore_finalizer(self)
             lltype.free(self._ptr, flavor='raw')
 
+    def get_maximum_buffer_size(self):
+        return self.datasize
+
 
 class W_CDataNewNonStd(W_CDataNewOwning):
     """Subclass using a non-standard allocator"""
@@ -620,6 +630,9 @@ class W_CDataPtrToStructOrUnion(W_CData):
             if isinstance(structobj, W_CDataNewOwning):
                 structobj._do_exit()
 
+    def get_maximum_buffer_size(self):
+        return self.structobj._sizeof()
+
 
 class W_CDataSliced(W_CData):
     """Subclass with an explicit length, for slices."""
@@ -669,6 +682,14 @@ class W_CDataFromBuffer(W_CData):
     def get_array_length(self):
         return self.length
 
+    def _sizeof(self):
+        from pypy.module._cffi_backend import ctypearray
+        ctype = self.ctype
+        if isinstance(ctype, ctypearray.W_CTypeArray):
+            return self.length * ctype.ctitem.size
+        else:
+            return W_CData._sizeof(self)
+
     def _repr_extra(self):
         from pypy.module._cffi_backend import ctypearray
         if self.w_keepalive is None:
