@@ -4,11 +4,12 @@ Pointers.
 
 import os
 
-from rpython.rlib import rposix
+from rpython.rlib import rposix, jit, rgc
+from rpython.rlib.objectmodel import we_are_translated
 from rpython.rlib.rarithmetic import ovfcheck
 from rpython.rtyper.annlowlevel import llstr
 from rpython.rtyper.lltypesystem import lltype, rffi
-from rpython.rtyper.lltypesystem.rstr import copy_string_to_raw
+from rpython.rtyper.lltypesystem.rstr import copy_string_to_raw, STR
 
 from pypy.interpreter.error import OperationError, oefmt, wrap_oserror
 from pypy.module._cffi_backend import cdataobj, misc, ctypeprim, ctypevoid
@@ -48,9 +49,13 @@ class W_CTypePtrOrArray(W_CType):
         if self.size < 0:
             return W_CType.cast(self, w_ob)
         space = self.space
+        from pypy.module._cffi_backend import wrapper
         if (isinstance(w_ob, cdataobj.W_CData) and
                 isinstance(w_ob.ctype, W_CTypePtrOrArray)):
             value = w_ob.unsafe_escaping_ptr()
+        elif (isinstance(w_ob, wrapper.W_FunctionWrapper) and
+                w_ob.directfnptr):
+            value = w_ob.directfnptr
         else:
             value = misc.as_unsigned_long(space, w_ob, strict=False)
             value = rffi.cast(rffi.CCHARP, value)
@@ -166,7 +171,11 @@ class W_CTypePtrBase(W_CTypePtrOrArray):
 
     def convert_from_object(self, cdata, w_ob):
         if not isinstance(w_ob, cdataobj.W_CData):
-            raise self._convert_error("cdata pointer", w_ob)
+            from pypy.module._cffi_backend import wrapper
+            if isinstance(w_ob, wrapper.W_FunctionWrapper):
+                w_ob = w_ob.try_extract_direct_fnptr_as_cdata(self.space)
+            if not isinstance(w_ob, cdataobj.W_CData):
+                raise self._convert_error("cdata pointer", w_ob)
         other = w_ob.ctype
         if not isinstance(other, W_CTypePtrBase):
             from pypy.module._cffi_backend import ctypearray
@@ -305,16 +314,8 @@ class W_CTypePointer(W_CTypePtrBase):
         else:
             return lltype.nullptr(rffi.CCHARP.TO)
 
-    def _prepare_pointer_call_argument(self, w_init, cdata, keepalives, i):
+    def _prepare_pointer_call_argument(self, w_init, cdata):
         space = self.space
-        if self.accept_str and space.isinstance_w(w_init, space.w_bytes):
-            # special case to optimize strings passed to a "char *" argument
-            value = space.bytes_w(w_init)
-            if isinstance(self.ctitem, ctypeprim.W_CTypePrimitiveBool):
-                self._must_be_string_of_zero_or_one(value)
-            keepalives[i] = value
-            return misc.write_string_as_charp(cdata, value)
-        #
         if (space.isinstance_w(w_init, space.w_list) or
             space.isinstance_w(w_init, space.w_tuple)):
             length = space.int_w(space.len(w_init))
@@ -360,14 +361,68 @@ class W_CTypePointer(W_CTypePtrBase):
         return 1
 
     def convert_argument_from_object(self, cdata, w_ob, keepalives, i):
+        # writes the pointer to cdata[0], writes the must-free flag in
+        # the byte just before cdata[0], and returns >0 if something
+        # must be done later to free. Description of flags in set_mustfree_flag
         from pypy.module._cffi_backend.ctypefunc import set_mustfree_flag
-        result = (not isinstance(w_ob, cdataobj.W_CData) and
-                  self._prepare_pointer_call_argument(w_ob, cdata,
-                                                      keepalives, i))
+        from pypy.module._cffi_backend.func import OffsetInBytes
+
+        if isinstance(w_ob, cdataobj.W_CData):
+            result = 0
+        else:
+            space = self.space
+            if self.accept_str and isinstance(w_ob, OffsetInBytes):
+                return self.process_str_from_offset_in_bytes(cdata, w_ob, keepalives, i)
+            if self.accept_str and space.isinstance_w(w_ob, space.w_bytes):
+                # special case to optimize strings passed to a "char *" argument
+                value = space.bytes_w(w_ob)
+                return self.accept_movable_str(cdata, value, keepalives, i)
+            result = self._prepare_pointer_call_argument(w_ob, cdata)
+
         if result == 0:
             self.convert_from_object(cdata, w_ob)
         set_mustfree_flag(cdata, result)
-        return result
+        return result == 1      # 0 or 2 => False, nothing to do later
+
+    @jit.dont_look_inside
+    def process_str_from_offset_in_bytes(self, cdata, w_ob, keepalives, i):
+        if not rgc.must_split_gc_address_space():
+            lldata = llstr(w_ob.bytes)
+            if not rgc.can_move(lldata):
+                # easy case - it can't move, pass it directly
+                self.accept_str_from_offset_in_bytes(cdata, lldata,
+                                           keepalives, i, w_ob.offset)
+                if not we_are_translated():
+                    keepalives[i] = lldata
+                    return 1
+                return 0
+            elif rgc.pin(lldata):
+                return self.accept_str_from_offset_in_bytes(cdata, lldata,
+                                               keepalives, i, w_ob.offset)
+        # we failed to pin, need to make a copy
+        value = w_ob.bytes
+        return self.accept_movable_str(cdata, value, keepalives, i)
+
+
+    def accept_movable_str(self, cdata, value, keepalives, i):
+        if isinstance(self.ctitem, ctypeprim.W_CTypePrimitiveBool):
+            self._must_be_string_of_zero_or_one(value)
+        keepalives[i] = misc.write_string_as_charp(cdata, value)
+        return 1
+
+    @jit.dont_look_inside
+    def accept_str_from_offset_in_bytes(self, cdata, lldata, keepalives, i, offset):
+        from rpython.rtyper.lltypesystem import llmemory
+        from rpython.rtyper.lltypesystem.rffi import offsetof, VOIDP, VOIDPP
+
+        addr = (llmemory.cast_ptr_to_adr(lldata) +
+              offsetof(STR, 'chars') +
+              llmemory.itemoffsetof(STR.chars, 0) + llmemory.sizeof(lltype.Char) * offset)
+        rffi.cast(VOIDPP, cdata)[0] = rffi.cast(VOIDP, addr)
+        if not we_are_translated():
+            keepalives[i] = lldata
+            return 3
+        return 0
 
     def getcfield(self, attr):
         from pypy.module._cffi_backend.ctypestruct import W_CTypeStructOrUnion
