@@ -72,8 +72,8 @@ class RealizeCache:
         "uint_fast64_t",
         "intmax_t",
         "uintmax_t",
-        "float _Complex",
-        "double _Complex",
+        "_cffi_float_complex_t",
+        "_cffi_double_complex_t",
         "char16_t",
         "char32_t",
         ]
@@ -83,12 +83,42 @@ class RealizeCache:
         self.space = space
         self.all_primitives = [None] * cffi_opcode._NUM_PRIM
         self.file_struct = None
+        self.lock = None
+        self.lock_owner = 0
+        self.rec_level = 0
 
     def get_file_struct(self):
         if self.file_struct is None:
             self.file_struct = ctypestruct.W_CTypeStruct(self.space, "FILE")
         return self.file_struct
 
+    def __enter__(self):
+        # This is a simple recursive lock implementation
+        if self.space.config.objspace.usemodules.thread:
+            from rpython.rlib import rthread
+            #
+            tid = rthread.get_ident()
+            if tid != self.lock_owner:
+                if self.lock is None:
+                    self.lock = self.space.allocate_lock()
+                self.lock.acquire(True)
+                assert self.lock_owner == 0
+                assert self.rec_level == 0
+                self.lock_owner = tid
+        self.rec_level += 1
+
+    def __exit__(self, *args):
+        assert self.rec_level > 0
+        self.rec_level -= 1
+        if self.space.config.objspace.usemodules.thread:
+            from rpython.rlib import rthread
+            #
+            tid = rthread.get_ident()
+            assert tid == self.lock_owner
+            if self.rec_level == 0:
+                self.lock_owner = 0
+                self.lock.release()
+
 
 def get_primitive_type(ffi, num):
     space = ffi.space
@@ -153,7 +183,7 @@ def realize_global_int(ffi, g, gindex):
         got = "%d" % (rffi.cast(rffi.LONGLONG, value),)
     raise oefmt(ffi.w_FFIError,
                 "the C compiler says '%s' is equal to %s, "
-                "but the cdef disagrees", rffi.charp2str(g.c_name), got)
+                "but the cdef disagrees", rffi.constcharp2str(g.c_name), got)
 
 
 class W_RawFuncType(W_Root):
@@ -281,9 +311,9 @@ def _realize_name(prefix, charp_src_name):
     # "$1" => "struct $1"
     if (charp_src_name[0] == '$' and charp_src_name[1] != '$'
             and not ('0' <= charp_src_name[1] <= '9')):
-        return rffi.charp2str(rffi.ptradd(charp_src_name, 1))
+        return rffi.constcharp2str(rffi.ptradd(charp_src_name, 1))
     else:
-        return prefix + rffi.charp2str(charp_src_name)
+        return prefix + rffi.constcharp2str(charp_src_name)
 
 
 def _realize_c_struct_or_union(ffi, sindex):
@@ -291,7 +321,8 @@ def _realize_c_struct_or_union(ffi, sindex):
         # returns a single global cached opaque type
         return ffi.space.fromcache(RealizeCache).get_file_struct()
 
-    s = ffi.ctxobj.ctx.c_struct_unions[sindex]
+    unions = rffi.cast(rffi.CArrayPtr(parse_c_type.STRUCT_UNION_S), ffi.ctxobj.ctx.c_struct_unions)
+    s = unions[sindex]
     type_index = rffi.getintfield(s, 'c_type_index')
     if ffi.cached_types[type_index] is not None:
         return ffi.cached_types[type_index] #found already in the "primary" slot
@@ -328,11 +359,11 @@ def _realize_c_struct_or_union(ffi, sindex):
             raise oefmt(ffi.w_FFIError,
                     "'%s %s' should come from ffi.include() but was not found",
                     "union" if c_flags & cffi_opcode.F_UNION else "struct",
-                    rffi.charp2str(s.c_name))
+                    rffi.constcharp2str(s.c_name))
         assert isinstance(x, ctypestruct.W_CTypeStructOrUnion)
         if (c_flags & cffi_opcode.F_OPAQUE) == 0 and x.size < 0:
             prefix = "union" if c_flags & cffi_opcode.F_UNION else "struct"
-            name = rffi.charp2str(s.c_name)
+            name = rffi.constcharp2str(s.c_name)
             raise oefmt(space.w_NotImplementedError,
                     "'%s %s' is opaque in the ffi.include(), but no "
                     "longer in the ffi doing the include (workaround: don't "
@@ -355,7 +386,9 @@ def _realize_c_struct_or_union(ffi, sindex):
     return x
 
 def _realize_c_enum(ffi, eindex):
-    e = ffi.ctxobj.ctx.c_enums[eindex]
+    enums = rffi.cast(rffi.CArrayPtr(parse_c_type.ENUM_S), ffi.ctxobj.ctx.c_enums)
+    globals = rffi.cast(rffi.CArrayPtr(parse_c_type.GLOBAL_S), ffi.ctxobj.ctx.c_globals)
+    e = enums[eindex]
     type_index = rffi.getintfield(e, 'c_type_index')
     if ffi.cached_types[type_index] is not None:
         return ffi.cached_types[type_index] #found already in the "primary" slot
@@ -371,12 +404,12 @@ def _realize_c_enum(ffi, eindex):
             j = 0
             while p[j] != ',' and p[j] != '\x00':
                 j += 1
-            enname = rffi.charpsize2str(p, j)
+            enname = rffi.constcharpsize2str(p, j)
             enumerators_w.append(space.newtext(enname))
 
             gindex = parse_c_type.search_in_globals(ffi.ctxobj.ctx, enname)
             assert gindex >= 0
-            g = ffi.ctxobj.ctx.c_globals[gindex]
+            g = globals[gindex]
             assert getop(g.c_type_op) == cffi_opcode.OP_ENUM
             assert getarg(g.c_type_op) == -1
 
@@ -406,16 +439,26 @@ def realize_c_type_or_func(ffi, opcodes, index):
     if from_ffi and ffi.cached_types[index] is not None:
         return ffi.cached_types[index]
 
-    opcodes[index] = rffi.cast(rffi.VOIDP, 255)
-    try:
-        x = realize_c_type_or_func_now(ffi, op, opcodes, index)
-    finally:
-        if opcodes[index] == rffi.cast(rffi.VOIDP, 255):
-            opcodes[index] = op
+    realize_cache = ffi.space.fromcache(RealizeCache)
+    with realize_cache:
+        #
+        # check again cached_types, which might have been filled while
+        # we were waiting for the recursive lock
+        if from_ffi and ffi.cached_types[index] is not None:
+            return ffi.cached_types[index]
 
-    if from_ffi:
-        assert ffi.cached_types[index] is None or ffi.cached_types[index] is x
-        ffi.cached_types[index] = x
+        if realize_cache.rec_level > 1000:
+            raise oefmt(ffi.space.w_RuntimeError,
+                "type-building recursion too deep or infinite.  "
+                "This is known to occur e.g. in ``struct s { void(*callable)"
+                "(struct s); }''.  Please report if you get this error and "
+                "really need support for your case.")
+        x = realize_c_type_or_func_now(ffi, op, opcodes, index)
+
+        if from_ffi:
+            old = ffi.cached_types[index]
+            assert old is None or old is x
+            ffi.cached_types[index] = x
 
     return x
 
@@ -457,17 +500,10 @@ def realize_c_type_or_func_now(ffi, op, opcodes, index):
         # essential: the TYPENAME opcode resolves the type index looked
         # up in the 'ctx.c_typenames' array, but it does so in 'ctx.c_types'
         # instead of in 'opcodes'!
-        type_index = rffi.getintfield(ffi.ctxobj.ctx.c_typenames[getarg(op)],
-                                      'c_type_index')
+        typenames = rffi.cast(rffi.CArrayPtr(parse_c_type.TYPENAME_S), ffi.ctxobj.ctx.c_typenames)
+        type_index = rffi.getintfield(typenames[getarg(op)], 'c_type_index')
         x = realize_c_type_or_func(ffi, ffi.ctxobj.ctx.c_types, type_index)
 
-    elif case == 255:
-        raise oefmt(ffi.space.w_RuntimeError,
-            "found a situation in which we try to build a type recursively.  "
-            "This is known to occur e.g. in ``struct s { void(*callable)"
-            "(struct s); }''.  Please report if you get this error and "
-            "really need support for your case.")
-
     else:
         raise oefmt(ffi.space.w_NotImplementedError, "op=%d", case)
 
@@ -488,9 +524,10 @@ def do_realize_lazy_struct(w_ctype):
     num_fields = rffi.getintfield(s, 'c_num_fields')
     fields_w = [None] * num_fields
 
+    fields = rffi.cast(rffi.CArrayPtr(parse_c_type.FIELD_S), ffi.ctxobj.ctx.c_fields)
     for i in range(num_fields):
-        fld = ffi.ctxobj.ctx.c_fields[first_field + i]
-        field_name   = rffi.charp2str(fld.c_name)
+        fld = fields[first_field + i]
+        field_name   = rffi.constcharp2str(fld.c_name)
         field_size   = rffi.getintfield(fld, 'c_field_size')
         field_offset = rffi.getintfield(fld, 'c_field_offset')
         op = rffi.getintfield(fld, 'c_field_type_op')
@@ -530,8 +567,13 @@ def do_realize_lazy_struct(w_ctype):
     if c_flags & cffi_opcode.F_PACKED:
         sflags |= newtype.SF_PACKED
 
-    assert w_ctype.size      == rffi.getintfield(s, 'c_size')
-    assert w_ctype.alignment == rffi.getintfield(s, 'c_alignment')
+    if rffi.getintfield(s, 'c_size') == -2:
+        # Before the loop, the struct was "lazy", now it should be complete
+        # so the pre-loop values do not match
+        pass
+    else:
+        assert w_ctype.size == rffi.getintfield(s, 'c_size')
+        assert w_ctype.alignment == rffi.getintfield(s, 'c_alignment')
     try:
         w_ctype.size = -1              # make opaque again
         newtype.complete_struct_or_union(
@@ -555,7 +597,7 @@ def do_realize_lazy_struct(w_ctype):
 
 
 def _fetch_external_struct_or_union(s, included_ffis_libs):
-    name = rffi.charp2str(s.c_name)
+    name = rffi.constcharp2str(s.c_name)
     #
     for ffi1, _ in included_ffis_libs:
         ctx1 = ffi1.ctxobj.ctx
@@ -563,7 +605,8 @@ def _fetch_external_struct_or_union(s, included_ffis_libs):
         if sindex < 0:   # not found at all
             continue
 
-        s1 = ctx1.c_struct_unions[sindex]
+        unions = rffi.cast(rffi.CArrayPtr(parse_c_type.STRUCT_UNION_S), ctx1.c_struct_unions)
+        s1 = unions[sindex]
         s1_flags = rffi.getintfield(s1, 'c_flags')
         s_flags  = rffi.getintfield(s,  'c_flags')
         if ((s1_flags & (cffi_opcode.F_EXTERNAL | cffi_opcode.F_UNION))
