@@ -30,7 +30,7 @@ class CodecState(object):
         self.modified()
 
     def _make_errorhandler(self, space, decode):
-        def call_errorhandler(errors, encoding, reason, input, startpos,
+        def call_errorhandler(errors, encoding, reason, w_input, startpos,
                               endpos):
             """Generic wrapper for calling into error handlers.
 
@@ -42,16 +42,11 @@ class CodecState(object):
             handlers return utf8 so we add whether they used unicode or bytes
             """
             w_errorhandler = lookup_error(space, errors)
+            length = space.len_w(w_input)
             if decode:
                 w_cls = space.w_UnicodeDecodeError
-                assert isinstance(input, str)
-                w_input = space.newbytes(input)
-                length = len(input)
             else:
                 w_cls = space.w_UnicodeEncodeError
-                assert isinstance(input, str)
-                length = rutf8.codepoints_in_utf8(input)
-                w_input = space.newtext(input, length)
             w_exc =  space.call_function(
                 w_cls,
                 space.newtext(encoding),
@@ -101,12 +96,16 @@ class CodecState(object):
                 if not space.isinstance_w(w_obj, space.w_bytes):
                     raise oefmt(space.w_ValueError,
                             "error handler modified exc.object must be bytes")
+                obj = space.bytes_w(w_obj)
             else:
                 if not space.isinstance_w(w_obj, space.w_unicode):
                     raise oefmt(space.w_ValueError,
                             "error handler modified exc.object must be str")
-            obj = space.utf8_w(w_obj) 
-            return space.utf8_w(w_replace), newpos, rettype, obj
+                obj = space.utf8_w(w_obj)
+            if rettype == 'b':
+                return space.bytes_w(w_replace), newpos, rettype, obj, w_obj
+            else:
+                return space.utf8_w(w_replace), newpos, rettype, obj, w_obj
         return call_errorhandler
 
     def make_decode_errorhandler(self, space):
@@ -154,8 +153,49 @@ def register_codec(space, w_search_function):
     else:
         raise oefmt(space.w_TypeError, "argument must be callable")
 
+def unregister(space, w_search_function):
+    """unregister(search_function)
 
-@unwrap_spec(encoding='text')
+    Unregister a codec search function and clear the registry's cache.
+    If the search function is not registered, do nothing."""
+    state = space.fromcache(CodecState)
+    try:
+        state.codec_search_path.remove(w_search_function)
+        state.codec_search_cache.clear()
+        state.modified()
+        return space.newint(0)
+    except ValueError:
+        return space.newint(-1)
+
+@jit.elidable
+def normalize(encoding):
+    """ Normalize an encoding name.
+
+        Normalization works as follows: all non-alphanumeric
+        characters except the dot used for Python package names are
+        collapsed and replaced with a single underscore, e.g. '  -;#'
+        becomes '_'. Leading and trailing underscores are removed.
+        Also convert to lower case
+
+        Note that encoding names should be ASCII only.
+
+    """
+    chars = []
+    punct = False
+    for c in encoding:
+        if c.isalnum() or c == '.':
+            if punct and chars:
+                chars.append('_')
+            if ord(c) <= 127:
+                chars.append(c.lower())
+            punct = False
+        else:
+            punct = True
+    return ''.join(chars)
+
+
+
+@unwrap_spec(encoding='text0')
 def lookup_codec(space, encoding):
     """lookup(encoding) -> (encoder, decoder, stream_reader, stream_writer)
     Looks up a codec tuple in the Python codec registry and returns
@@ -164,7 +204,7 @@ def lookup_codec(space, encoding):
     assert not (space.config.translating and not we_are_translated()), \
         "lookup_codec() should not be called during translation"
     state = space.fromcache(CodecState)
-    normalized_encoding = encoding.replace(" ", "-").lower()
+    normalized_encoding = normalize(encoding)
     w_result = state.get_codec_from_cache(normalized_encoding)
     if w_result is not None:
         return w_result
@@ -189,9 +229,9 @@ def _lookup_codec_loop(space, encoding, normalized_encoding):
             raise oefmt(space.w_LookupError,
                         "no codec search functions registered: can't find "
                         "encoding")
+    w_v = space.newtext(normalized_encoding)
     for w_search in state.codec_search_path:
-        w_result = space.call_function(w_search,
-                                       space.newtext(normalized_encoding))
+        w_result = space.call_function(w_search, w_v)
         if not space.is_w(w_result, space.w_None):
             if not (space.isinstance_w(w_result, space.w_tuple) and
                     space.len_w(w_result) == 4):
@@ -231,7 +271,7 @@ def strict_errors(space, w_exc):
 def ignore_errors(space, w_exc):
     check_exception(space, w_exc)
     w_end = space.getattr(w_exc, space.newtext('end'))
-    return space.newtuple([space.newutf8('', 0), w_end])
+    return space.newtuple2(space.newutf8('', 0), w_end)
 
 REPLACEMENT = u'\ufffd'.encode('utf8')
 
@@ -244,13 +284,13 @@ def replace_errors(space, w_exc):
         size = 0
     if space.isinstance_w(w_exc, space.w_UnicodeEncodeError):
         text = '?' * size
-        return space.newtuple([space.newutf8(text, size), w_end])
+        return space.newtuple2(space.newutf8(text, size), w_end)
     elif space.isinstance_w(w_exc, space.w_UnicodeDecodeError):
         text = REPLACEMENT
-        return space.newtuple([space.newutf8(text, 1), w_end])
+        return space.newtuple2(space.newutf8(text, 1), w_end)
     elif space.isinstance_w(w_exc, space.w_UnicodeTranslateError):
         text = REPLACEMENT * size
-        return space.newtuple([space.newutf8(text, size), w_end])
+        return space.newtuple2(space.newutf8(text, size), w_end)
     else:
         raise oefmt(space.w_TypeError,
                     "don't know how to handle %T in error callback", w_exc)
@@ -278,7 +318,7 @@ def xmlcharrefreplace_errors(space, w_exc):
             pos = rutf8.next_codepoint_pos(obj, pos)
         r = builder.build()
         lgt = rutf8.check_utf8(r, True)
-        return space.newtuple([space.newutf8(r, lgt), w_end])
+        return space.newtuple2(space.newutf8(r, lgt), w_end)
     else:
         raise oefmt(space.w_TypeError,
                     "don't know how to handle %T in error callback", w_exc)
@@ -347,7 +387,7 @@ def namereplace_errors(space, w_exc):
             pos = rutf8.next_codepoint_pos(obj, pos)
         r = builder.build()
         lgt = rutf8.check_utf8(r, True)
-        return space.newtuple([space.newutf8(r, lgt), w_end])
+        return space.newtuple2(space.newutf8(r, lgt), w_end)
     else:
         raise oefmt(space.w_TypeError,
                     "don't know how to handle %T in error callback", w_exc)
@@ -554,7 +594,7 @@ def _call_codec(space, w_coder, w_obj, action, encoding, errors):
                     "%s must return a tuple (object, integer)", action)
     return space.getitem(w_res, space.newint(0))
 
-@unwrap_spec(errors='text')
+@unwrap_spec(errors='text0')
 def lookup_error(space, errors):
     """lookup_error(errors) -> handler
 
@@ -590,7 +630,7 @@ def encode(space, w_obj, encoding=None, errors=None):
 @unwrap_spec(errors='text_or_none')
 def readbuffer_encode(space, w_data, errors='strict'):
     s = space.getarg_w('s#', w_data)
-    return space.newtuple([space.newbytes(s), space.newint(len(s))])
+    return space.newtuple2(space.newbytes(s), space.newint(len(s)))
 
 @unwrap_spec(encoding='text_or_none', errors='text_or_none')
 def decode(space, w_obj, encoding=None, errors=None):
@@ -663,9 +703,9 @@ def make_encoder_wrapper(name):
             allow_surrogates = True
         state = space.fromcache(CodecState)
         ulen = w_arg._length
-        result = func(w_arg._utf8, errors, state.encode_error_handler,
+        result = func(space, w_arg._utf8, w_arg, errors, state.encode_error_handler,
                       allow_surrogates=allow_surrogates)
-        return space.newtuple([space.newbytes(result), space.newint(ulen)])
+        return space.newtuple2(space.newbytes(result), space.newint(ulen))
     wrap_encoder.__name__ = func.__name__
     globals()[name] = wrap_encoder
 
@@ -679,9 +719,10 @@ def make_decoder_wrapper(name):
             errors = 'strict'
         final = space.is_true(w_final)
         state = space.fromcache(CodecState)
-        result, length, pos = func(string, errors, final, state.decode_error_handler)
+        w_str = space.newbytes(string)
+        result, length, pos = func(space, string, w_str, errors, final, state.decode_error_handler)
         # must return bytes, pos
-        return space.newtuple([space.newutf8(result, length), space.newint(pos)])
+        return space.newtuple2(space.newutf8(result, length), space.newint(pos))
     wrap_decoder.__name__ = func.__name__
     globals()[name] = wrap_decoder
 
@@ -697,7 +738,6 @@ for encoder in [
          "utf_32_le_encode",
          "unicode_escape_encode",
          "raw_unicode_escape_encode",
-         "unicode_internal_encode",
         ]:
     make_encoder_wrapper(encoder)
 
@@ -738,7 +778,8 @@ if getattr(unicodehelper, '_WIN32', False):
         state = space.fromcache(CodecState)
         ulen = w_arg._length
         try:
-            result = unicodehelper.utf8_encode_code_page(code_page, w_arg._utf8,
+            result = unicodehelper.utf8_encode_code_page(space, code_page,
+                      w_arg._utf8, w_arg,
                       errors, state.encode_error_handler,
                       allow_surrogates=allow_surrogates)
         except OSError as e:
@@ -756,14 +797,19 @@ if getattr(unicodehelper, '_WIN32', False):
         final = space.is_true(w_final)
         state = space.fromcache(CodecState)
         try:
-            result, length, pos = unicodehelper.str_decode_code_page(code_page,
-                                   string, errors, final,
+            w_s = space.newbytes(string)
+            result, pos, length = unicodehelper.str_decode_code_page(space, code_page,
+                                   string, w_s, errors, final,
                                    state.decode_error_handler)
         except OSError as e:
             raise wrap_oserror(space, e)
         # must return bytes, pos
-        return space.newtuple([space.newutf8(result, length), space.newint(pos)])
+        return space.newtuple([space.newutf8(result, length), space.newint(len(string))])
 
+def utf8_encode_wrapper(space, utf8, w_obj, errors):
+    state = space.fromcache(CodecState)
+    return unicodehelper.utf8_encode_utf_8(space, utf8, w_obj, errors,
+                 state.encode_error_handler, allow_surrogates=False)
 
 # utf-8 functions are not regular, because we have to pass
 # "allow_surrogates=False"
@@ -771,19 +817,15 @@ if getattr(unicodehelper, '_WIN32', False):
 def utf_8_encode(space, w_obj, errors="strict"):
     utf8, lgt = space.utf8_len_w(w_obj)
     if lgt == len(utf8): # ascii
-        return space.newtuple([space.newbytes(utf8), space.newint(lgt)])
+        return space.newtuple2(space.newbytes(utf8), space.newint(lgt))
     if errors is None:
         errors = 'strict'
-    state = space.fromcache(CodecState)
-    try:
-        result = unicodehelper.utf8_encode_utf_8(utf8, errors,
-                     state.encode_error_handler, allow_surrogates=False)
-    except unicodehelper.ErrorHandlerError as e:
-        raise oefmt(space.w_IndexError,
-                   "position %d from error handler invalid, already encoded %d",
-                    e.new,e.old)
-
-    return space.newtuple([space.newbytes(result), space.newint(lgt)])
+    result = utf8_encode_wrapper(space, utf8, w_obj, errors)
+    if errors == 'strict':
+        # since we did strict checking and surrogates are forbidden, we know
+        # the result of the _check_utf8 call with allow_surrogates=False
+        jit.record_known_result(lgt, rutf8._check_utf8, utf8, False, 0, -1)
+    return space.newtuple2(space.newbytes(result), space.newint(lgt))
 
 @unwrap_spec(string='bufferstr', errors='text_or_none',
              w_final = WrappedDefault(False))
@@ -792,16 +834,19 @@ def utf_8_decode(space, string, errors="strict", w_final=None):
         errors = 'strict'
     final = space.is_true(w_final)
     state = space.fromcache(CodecState)
-    res, lgt, pos = unicodehelper.str_decode_utf8(string,
-        errors, final, state.decode_error_handler)
-    return space.newtuple([space.newutf8(res, lgt),
-                           space.newint(pos)])
+    if errors == 'surrogatepass':
+        allow_surrogates = True
+    else:
+        allow_surrogates = False
+    w_string = space.newbytes(string)
+    res, lgt, pos = unicodehelper.str_decode_utf8(space, string, w_string,
+        errors, final, state.decode_error_handler, allow_surrogates=allow_surrogates)
+    return space.newtuple2(space.newutf8(res, lgt),
+                           space.newint(pos))
 
 @unwrap_spec(data='bufferstr', errors='text_or_none', byteorder=int,
              w_final=WrappedDefault(False))
 def utf_16_ex_decode(space, data, errors='strict', byteorder=0, w_final=None):
-    from pypy.interpreter.unicodehelper import str_decode_utf_16_helper
-
     if errors is None:
         errors = 'strict'
     final = space.is_true(w_final)
@@ -812,8 +857,9 @@ def utf_16_ex_decode(space, data, errors='strict', byteorder=0, w_final=None):
         _byteorder = 'little'
     else:
         _byteorder = 'big'
-    res, lgt, pos, bo = str_decode_utf_16_helper(
-        data, errors, final,
+    w_data = space.newbytes(data)
+    res, lgt, pos, bo = unicodehelper.str_decode_utf_16_helper(
+        space, data, w_data, errors, final,
         state.decode_error_handler, _byteorder)
     # return result, consumed, byteorder for buffered incremental encoders
     return space.newtuple([space.newutf8(res, lgt),
@@ -822,8 +868,6 @@ def utf_16_ex_decode(space, data, errors='strict', byteorder=0, w_final=None):
 @unwrap_spec(data='bufferstr', errors='text_or_none', byteorder=int,
              w_final=WrappedDefault(False))
 def utf_32_ex_decode(space, data, errors='strict', byteorder=0, w_final=None):
-    from pypy.interpreter.unicodehelper import str_decode_utf_32_helper
-
     final = space.is_true(w_final)
     state = space.fromcache(CodecState)
     if byteorder == 0:
@@ -832,8 +876,9 @@ def utf_32_ex_decode(space, data, errors='strict', byteorder=0, w_final=None):
         _byteorder = 'little'
     else:
         _byteorder = 'big'
-    res, lgt, pos, bo = str_decode_utf_32_helper(
-        data, errors, final,
+    w_data = space.newbytes(data)
+    res, lgt, pos, bo = unicodehelper.str_decode_utf_32_helper(
+        space, data, w_data, errors, final,
         state.decode_error_handler, _byteorder)
     # return result, consumed, byteorder for buffered incremental encoders
     return space.newtuple([space.newutf8(res, lgt),
@@ -923,8 +968,8 @@ def charmap_decode(space, string, errors="strict", w_mapping=None):
     if errors is None:
         errors = 'strict'
     if len(string) == 0:
-        return space.newtuple([space.newutf8('', 0),
-                               space.newint(0)])
+        return space.newtuple2(space.newutf8('', 0),
+                               space.newint(0))
 
     if space.is_none(w_mapping):
         mapping = None
@@ -933,10 +978,12 @@ def charmap_decode(space, string, errors="strict", w_mapping=None):
 
     final = True
     state = space.fromcache(CodecState)
+    w_string = space.newbytes(string)
     result, lgt, pos = unicodehelper.str_decode_charmap(
-        string, errors, final, state.decode_error_handler, mapping)
-    return space.newtuple([space.newutf8(result, lgt),
-                           space.newint(len(string))])
+        space, string, w_string, errors, final,
+        state.decode_error_handler, mapping)
+    return space.newtuple2(space.newutf8(result, lgt),
+                           space.newint(len(string)))
 
 @unwrap_spec(errors='text_or_none')
 def charmap_encode(space, w_unicode, errors="strict", w_mapping=None):
@@ -950,8 +997,9 @@ def charmap_encode(space, w_unicode, errors="strict", w_mapping=None):
     state = space.fromcache(CodecState)
     w_uni = space.convert_arg_to_w_unicode(w_unicode)
     result = unicodehelper.utf8_encode_charmap(
-        space.utf8_w(w_uni), errors, state.encode_error_handler, mapping)
-    return space.newtuple([space.newbytes(result), space.newint(w_uni._len())])
+        space, space.utf8_w(w_uni), w_uni, errors,
+        state.encode_error_handler, mapping)
+    return space.newtuple2(space.newbytes(result), space.newint(w_uni._len()))
 
 
 @unwrap_spec(chars='utf8')
@@ -985,7 +1033,7 @@ class UnicodeData_Handler:
             return -1
         return space.int_w(w_code)
 
-@unwrap_spec(errors='text_or_none', w_final=WrappedDefault(False))
+@unwrap_spec(errors='text_or_none', w_final=WrappedDefault(True))
 def unicode_escape_decode(space, w_string, errors="strict", w_final=None):
     string = space.getarg_w('s*', w_string).as_str()
 
@@ -997,8 +1045,9 @@ def unicode_escape_decode(space, w_string, errors="strict", w_final=None):
 
     unicode_name_handler = state.get_unicodedata_handler(space)
 
+    w_s = space.newbytes(string)
     result, u_len, lgt, first_escape_error_char = unicodehelper.str_decode_unicode_escape(
-        string, errors,
+        space, string, w_s, errors,
         final, state.decode_error_handler,
         unicode_name_handler)
 
@@ -1006,7 +1055,9 @@ def unicode_escape_decode(space, w_string, errors="strict", w_final=None):
         # Here, 'first_escape_error_char' is a single string character.
         # Careful, it might be >= '\x80'.  If it is, it would made an
         # invalid utf-8 string when pasted directory in it.
-        if ' ' <= first_escape_error_char < '\x7f':
+        if len(first_escape_error_char) == 3:
+            msg = "invalid octal escape sequence '\\%s'" % (first_escape_error_char,)
+        elif ' ' <= first_escape_error_char <= '\x7f':
             msg = "invalid escape sequence '\\%s'" % (first_escape_error_char,)
         else:
             msg = "invalid escape sequence: '\\' followed by %s" % (
@@ -1016,67 +1067,26 @@ def unicode_escape_decode(space, w_string, errors="strict", w_final=None):
             space.newtext(msg),
             space.w_DeprecationWarning
         )
-    return space.newtuple([space.newutf8(result, u_len), space.newint(lgt)])
+    return space.newtuple2(space.newutf8(result, u_len), space.newint(lgt))
 
 # ____________________________________________________________
 # Raw Unicode escape (accepts bytes or str)
 
-@unwrap_spec(errors='text_or_none', w_final=WrappedDefault(False))
+@unwrap_spec(errors='text_or_none', w_final=WrappedDefault(True))
 def raw_unicode_escape_decode(space, w_string, errors="strict", w_final=None):
     string = space.getarg_w('s*', w_string).as_str()
     if errors is None:
         errors = 'strict'
     final = space.is_true(w_final)
     state = space.fromcache(CodecState)
+    w_s = space.newbytes(string)
     result, u_len, lgt = unicodehelper.str_decode_raw_unicode_escape(
-        string, errors, final, state.decode_error_handler)
-    return space.newtuple([space.newtext(result), space.newint(lgt)])
+        space, string, w_s, errors, final, state.decode_error_handler)
+    return space.newtuple2(space.newtext(result), space.newint(lgt))
 
 # ____________________________________________________________
 # Unicode-internal
 
-@unwrap_spec(errors='text_or_none')
-def unicode_internal_decode(space, w_string, errors="strict"):
-    if errors is None:
-        errors = 'strict'
-    # special case for this codec: unicodes are returned as is
-    if space.isinstance_w(w_string, space.w_unicode):
-        return space.newtuple([w_string, space.len(w_string)])
-
-    string = space.charbuf_w(w_string)
-    space.warn(space.newtext("unicode_internal codec has been deprecated"),
-               space.w_DeprecationWarning)
-
-    if len(string) == 0:
-        return space.newtuple([space.newutf8('', 0),
-                               space.newint(0)])
-
-    final = True
-    state = space.fromcache(CodecState)
-    result, lgt = unicodehelper.str_decode_unicode_internal(
-        string, errors,
-        final, state.decode_error_handler)
-    return space.newtuple([space.newutf8(result, lgt),
-                           space.newint(len(string))])
-
-@unwrap_spec(errors='text_or_none')
-def unicode_internal_encode(space, w_uni, errors="strict"):
-    space.warn(space.newtext("unicode_internal codec has been deprecated"),
-               space.w_DeprecationWarning)
-    if errors is None:
-        errors = 'strict'
-    if space.isinstance_w(w_uni, space.w_unicode):
-        utf8 = space.utf8_w(w_uni)
-        state = space.fromcache(CodecState)
-        result = unicodehelper.utf8_encode_unicode_internal(
-            utf8, errors, state.encode_error_handler)
-        w_lgt = space.newint(space.len_w(w_uni))
-        return space.newtuple([space.newbytes(result), w_lgt])
-    else:
-        # special case for this codec: bytes are returned as is
-        string = space.charbuf_w(w_uni)
-        return space.newtuple([space.newbytes(string), space.newint(len(string))])
-
 # ____________________________________________________________
 # support for the "string escape" translation
 # This is a bytes-to bytes transformation
@@ -1093,4 +1103,4 @@ def escape_decode(space, w_data, errors='strict'):
     from pypy.interpreter.pyparser.parsestring import PyString_DecodeEscape
     result, _ = PyString_DecodeEscape(space, data, errors, None)
 
-    return space.newtuple([space.newbytes(result), space.newint(len(data))])
+    return space.newtuple2(space.newbytes(result), space.newint(len(data)))
