@@ -1,5 +1,3 @@
-from __future__ import with_statement
-
 from rpython.rlib.signature import signature
 from rpython.rlib import types
 
@@ -14,6 +12,9 @@ from rpython.rlib.rstring import StringBuilder
 from rpython.rlib.rarithmetic import r_longlong, intmask
 from rpython.rlib import rposix
 from rpython.tool.sourcetools import func_renamer
+from rpython.rlib.objectmodel import import_from_mixin, try_inline, keepalive_until_here
+from rpython.rtyper.lltypesystem import lltype, rffi
+
 from pypy.module._io.interp_iobase import (
     W_IOBase, DEFAULT_BUFFER_SIZE, convert_size, trap_eintr,
     check_readable_w, check_writable_w, check_seekable_w)
@@ -160,8 +161,6 @@ implementation, but wrap one.
 )
 
 class BufferedMixin:
-    _mixin_ = True
-
     def __init__(self, space):
         W_IOBase.__init__(self, space)
         self.state = STATE_ZERO
@@ -185,6 +184,8 @@ class BufferedMixin:
         self.readable = False
         self.writable = False
 
+        self._fast_closed_check = False
+
     def _reader_reset_buf(self):
         self.read_end = -1
 
@@ -192,19 +193,22 @@ class BufferedMixin:
         self.write_pos = 0
         self.write_end = -1
 
+    def _make_buffer(self, space, size):
+        if space.config.translation.split_gc_address_space:
+            # When using split GC address space, it is not possible to get the
+            # raw address of a GC buffer. Therefore we use a buffer backed by
+            # raw memory.
+            return RawByteBuffer(size)
+        else:
+            # TODO: test whether using the raw buffer is faster
+            return ByteBuffer(size)
+
     def _init(self, space):
         if self.buffer_size <= 0:
             raise oefmt(space.w_ValueError,
                         "buffer size must be strictly positive")
 
-        if space.config.translation.split_gc_address_space:
-            # When using split GC address space, it is not possible to get the
-            # raw address of a GC buffer. Therefore we use a buffer backed by
-            # raw memory.
-            self.buffer = RawByteBuffer(self.buffer_size)
-        else:
-            # TODO: test whether using the raw buffer is faster
-            self.buffer = ByteBuffer(self.buffer_size)
+        self.buffer = self._make_buffer(space, self.buffer_size)
 
         self.lock = TryLock(space)
 
@@ -220,13 +224,26 @@ class BufferedMixin:
         elif self.state == STATE_DETACHED:
             raise oefmt(space.w_ValueError, "raw stream has been detached")
 
+    @try_inline
     def _check_closed(self, space, message=None):
+        from pypy.module._io.interp_fileio import W_FileIO
+        if self._fast_closed_check:
+            w_raw = self.w_raw
+            assert isinstance(w_raw, W_FileIO)
+            if w_raw.fd >= 0:
+                return
         self._check_init(space)
         W_IOBase._check_closed(self, space, message)
 
     def _raw_tell(self, space):
-        w_pos = space.call_method(self.w_raw, "tell")
-        pos = space.r_longlong_w(w_pos)
+        from pypy.module._io.interp_fileio import W_FileIO
+        if self._fast_closed_check:
+            w_raw = self.w_raw
+            assert isinstance(w_raw, W_FileIO)
+            pos = r_longlong(w_raw._raw_tell(space))
+        else:
+            w_pos = space.call_method(self.w_raw, "tell")
+            pos = space.r_longlong_w(w_pos)
         if pos < 0:
             raise oefmt(space.w_IOError,
                         "raw stream returned invalid position")
@@ -293,8 +310,11 @@ class BufferedMixin:
 
     def tell_w(self, space):
         self._check_init(space)
-        pos = self._raw_tell(space) - self._raw_offset()
-        return space.newint(pos)
+        raw_tell = self._raw_tell(space)
+        raw_offset = self._raw_offset()
+        if raw_tell < raw_offset:
+            return space.newint(0)
+        return space.newint(raw_tell - raw_offset)
 
     @unwrap_spec(pos=r_longlong, whence=int)
     def seek_w(self, space, pos, whence=0):
@@ -320,8 +340,10 @@ class BufferedMixin:
                     newpos = self.pos + int(offset)
                     assert newpos >= 0
                     self.pos = newpos
-                    return space.newint(current - available + offset)
-
+                    res = current - available + offset
+                    if res < 0:
+                        return space.newint(0)
+                    return space.newint(res)
         # Fallback: invoke raw seek() method and clear buffer
         with self.lock:
             if self.writable:
@@ -367,9 +389,15 @@ class BufferedMixin:
                     if flush_operr:
                         e.chain_exceptions(space, flush_operr)
                     raise
+        self.buffer = None # free buffer memory
+        self.maybe_unregister_rpython_finalizer_io(space)
 
     def _dealloc_warn_w(self, space, w_source):
-        space.call_method(self.w_raw, "_dealloc_warn", w_source)
+        if self.w_raw:
+            space.call_method(self.w_raw, "_dealloc_warn", w_source)
+        else:
+            raise oefmt(space.w_RuntimeError,
+                 "calling _dealloc_warn' on a closed or detached reader")
 
     def simple_flush_w(self, space):
         self._check_init(space)
@@ -433,6 +461,7 @@ class BufferedMixin:
         w_raw = self.w_raw
         self.w_raw = None
         self.state = STATE_DETACHED
+        self._fast_closed_check = False
         return w_raw
 
     def fileno_w(self, space):
@@ -443,9 +472,10 @@ class BufferedMixin:
     def truncate_w(self, space, w_size):
         self._check_init(space)
         self._check_closed(space, "truncate of closed file")
+        if not self.writable:
+            self._unsupportedoperation(space, "truncate")
         with self.lock:
-            if self.writable:
-                self._flush_and_rewind_unlocked(space)
+            self._flush_and_rewind_unlocked(space)
             # invalidate cached position
             self.abs_pos = -1
 
@@ -470,6 +500,8 @@ class BufferedMixin:
         else:
             raise oefmt(space.w_ValueError,
                         "read length must be positive or -1")
+        if res is None:
+            return space.w_None
         return space.newbytes(res)
 
     @unwrap_spec(size=int)
@@ -570,19 +602,36 @@ class BufferedMixin:
         return space.newbytes(builder.build())
 
     def _raw_read(self, space, buffer, start, length):
+        from pypy.module._io.interp_fileio import W_FileIO
         assert buffer is not None
+
         length = intmask(length)
         start = intmask(start)
-        w_view = SimpleView(SubBuffer(buffer, start, length)).wrap(space)
-        while True:
+        w_raw = self.w_raw
+        target_address = lltype.nullptr(rffi.CCHARP.TO)
+        if type(w_raw) is W_FileIO:
             try:
-                w_size = space.call_method(self.w_raw, "readinto", w_view)
-            except OperationError as e:
-                if trap_eintr(space, e):
-                    continue  # try again
-                raise
+                raw_address = buffer.get_raw_address()
+            except ValueError:
+                pass
             else:
-                break
+                target_address = rffi.ptradd(raw_address, start)
+
+        if target_address:
+            assert type(w_raw) is W_FileIO
+            w_size = w_raw._readinto_raw(space, target_address, length)
+            keepalive_until_here(buffer)
+        else:
+            w_view = SimpleView(SubBuffer(buffer, start, length), w_obj=self).wrap(space)
+            while True:
+                try:
+                    w_size = space.call_method(self.w_raw, "readinto", w_view)
+                except OperationError as e:
+                    if trap_eintr(space, e):
+                        continue  # try again
+                    raise
+                else:
+                    break
 
         if space.is_w(w_size, space.w_None):
             raise BlockingIOError()
@@ -596,6 +645,7 @@ class BufferedMixin:
         return size
 
     def _fill_buffer(self, space):
+        self._check_closed(space)
         start = self.read_end
         if start == -1:
             start = 0
@@ -613,7 +663,7 @@ class BufferedMixin:
         if n <= current_size:
             return self._read_fast(n)
 
-        result_buffer = ByteBuffer(n)
+        result_buffer = self._make_buffer(space, n)
         remaining = n
         written = 0
         if current_size:
@@ -692,11 +742,21 @@ class BufferedMixin:
         have = self._readahead()
         if limit >= 0 and have > limit:
             have = limit
-        for pos in range(self.pos, self.pos+have):
-            if self.buffer[pos] == '\n':
-                break
+        buffer = self.buffer
+        if isinstance(buffer, ByteBuffer):
+            # hack at the internals of self.buffer, otherwise this loop is
+            # extremely slow
+            for pos in range(self.pos, self.pos+have):
+                if buffer.data[pos] == '\n':
+                    break
+            else:
+                pos = -1
         else:
-            pos = -1
+            for pos in range(self.pos, self.pos+have):
+                if buffer[pos] == '\n':
+                    break
+            else:
+                pos = -1
         if pos >= 0:
             w_res = space.newbytes(self.buffer[self.pos:pos+1])
             self.pos = pos + 1
@@ -879,8 +939,8 @@ class BufferedMixin:
             finally:
                 self._reader_reset_buf()
 
-class BufferedReaderMixin(BufferedMixin):
-    _mixin_ = True
+class BufferedReaderMixin(object):
+    import_from_mixin(BufferedMixin)
 
     def readinto_w(self, space, w_buffer):
         return self._readinto(space, w_buffer, read_once=False)
@@ -889,6 +949,8 @@ class BufferedReaderMixin(BufferedMixin):
         return self._readinto(space, w_buffer, read_once=True)
 
     def _readinto(self, space, w_buffer, read_once):
+        self._check_init(space)
+        self._check_closed(space, "readinto of closed file")
         rwbuffer = space.writebuf_w(w_buffer)
         length = rwbuffer.getlength()
         with self.lock:
@@ -938,9 +1000,12 @@ class BufferedReaderMixin(BufferedMixin):
             return space.newint(written)
 
 
-class W_BufferedReader(BufferedReaderMixin, W_BufferedIOBase):
+class W_BufferedReader(W_BufferedIOBase):
+    import_from_mixin(BufferedReaderMixin)
+
     @unwrap_spec(buffer_size=int)
     def descr_init(self, space, w_raw, buffer_size=DEFAULT_BUFFER_SIZE):
+        from pypy.module._io.interp_fileio import W_FileIO
         self.state = STATE_ZERO
         check_readable_w(space, w_raw)
 
@@ -951,6 +1016,9 @@ class W_BufferedReader(BufferedReaderMixin, W_BufferedIOBase):
         self._init(space)
         self._reader_reset_buf()
         self.state = STATE_OK
+        self._fast_closed_check = (type(self) is W_BufferedReader and
+                type(w_raw) is W_FileIO)
+
 
 W_BufferedReader.typedef = TypeDef(
     '_io.BufferedReader', W_BufferedIOBase.typedef,
@@ -984,9 +1052,12 @@ W_BufferedReader.typedef = TypeDef(
     mode = GetSetProperty(W_BufferedReader.mode_get_w),
 )
 
-class W_BufferedWriter(BufferedMixin, W_BufferedIOBase):
+class W_BufferedWriter(W_BufferedIOBase):
+    import_from_mixin(BufferedMixin)
+
     @unwrap_spec(buffer_size=int)
     def descr_init(self, space, w_raw, buffer_size=DEFAULT_BUFFER_SIZE):
+        from pypy.module._io.interp_fileio import W_FileIO
         self.state = STATE_ZERO
         check_writable_w(space, w_raw)
 
@@ -997,6 +1068,8 @@ class W_BufferedWriter(BufferedMixin, W_BufferedIOBase):
         self._init(space)
         self._writer_reset_buf()
         self.state = STATE_OK
+        self._fast_closed_check = (type(self) is W_BufferedWriter and
+                type(w_raw) is W_FileIO)
 
 W_BufferedWriter.typedef = TypeDef(
     '_io.BufferedWriter', W_BufferedIOBase.typedef,
@@ -1124,9 +1197,12 @@ W_BufferedRWPair.typedef = TypeDef(
     **methods
 )
 
-class W_BufferedRandom(BufferedReaderMixin, W_BufferedIOBase):
+class W_BufferedRandom(W_BufferedIOBase):
+    import_from_mixin(BufferedReaderMixin)
+
     @unwrap_spec(buffer_size=int)
     def descr_init(self, space, w_raw, buffer_size=DEFAULT_BUFFER_SIZE):
+        from pypy.module._io.interp_fileio import W_FileIO
         self.state = STATE_ZERO
         check_readable_w(space, w_raw)
         check_writable_w(space, w_raw)
@@ -1141,6 +1217,8 @@ class W_BufferedRandom(BufferedReaderMixin, W_BufferedIOBase):
         self._writer_reset_buf()
         self.pos = 0
         self.state = STATE_OK
+        self._fast_closed_check = (type(self) is W_BufferedWriter and
+                type(w_raw) is W_FileIO)
 
 W_BufferedRandom.typedef = TypeDef(
     '_io.BufferedRandom', W_BufferedIOBase.typedef,
