@@ -1,24 +1,25 @@
 import sys
 import os
 
+
 from pypy.interpreter.error import OperationError, oefmt
 from pypy.interpreter.typedef import (
     TypeDef, generic_new_descr, GetSetProperty)
 from pypy.interpreter.gateway import WrappedDefault, interp2app, unwrap_spec
 from pypy.module._io.interp_iobase import (W_RawIOBase, convert_size,
         DEFAULT_BUFFER_SIZE)
-from pypy.module.time.interp_time import sleep
-from pypy.interpreter.unicodehelper import (fsdecode, str_decode_utf_16,
+from pypy.module.time.interp_time import time_sleep
+from pypy.interpreter.unicodehelper import (str_decode_utf_16,
         utf8_encode_utf_16)
 from pypy.module._codecs.interp_codecs import CodecState
 from rpython.rtyper.lltypesystem import lltype, rffi
-from rpython.rlib._os_support import _preferred_traits
+from rpython.rlib._os_support import utf8_traits
 from rpython.rlib import rwin32
 from rpython.rlib.rstring import StringBuilder
-from rpython.rlib.runicode import WideCharToMultiByte, MultiByteToWideChar
 from rpython.rlib.rwin32file import make_win32_traits
 from rpython.rlib.buffer import ByteBuffer
 from rpython.rlib.rarithmetic import intmask
+from rpython.rlib.rposix import getfullpathname
 
 # SMALLBUF determines how many utf-8 characters will be
 # buffered within the stream, in order to support reads
@@ -75,7 +76,7 @@ def read_console_wide(space, handle, maxlen):
                     break
                 err = 0
                 # This will only catch CTRL-C on the main thread
-                sleep(space, space.newfloat(0.1))
+                time_sleep(space, space.newfloat(0.1))
                 continue
             readlen += nread
             readlen_b = 2 * readlen
@@ -125,8 +126,17 @@ def _get_console_type(handle):
 
 def _pyio_get_console_type(space, w_path_or_fd):
 
-    # XXX 2020-07-22 Disable WinConsoleIO since it is flaky
-    return '\0'
+    # XXX 2021-01-10 Disable WinConsoleIO (again) it is flaky. Some interaction
+    # with pytest in running numpy's tests makes the handle invalid.
+    # TODO: refactor the w_path_or_fd handling to be more like interp_posix
+    #       and use the path_or_fd() unwrap_spec all through the _io module
+    #       Then this will recieve a already-processed Path object
+    # Another alternative to this whole mess would be to adapt the ctypes-based
+    # https://pypi.org/project/win_unicode_console/ which also implements PEP 528
+
+    enable_winconsoleio = os.environ.get("PYPY_ENABLE_WINCONSOLEIO")
+    if not enable_winconsoleio:
+        return '\0'
 
     if space.isinstance_w(w_path_or_fd, space.w_int):
         fd = space.int_w(w_path_or_fd)
@@ -135,14 +145,14 @@ def _pyio_get_console_type(space, w_path_or_fd):
             return '\0'
         return _get_console_type(handle)
 
-    decoded = space.fsdecode_w(w_path_or_fd)
-    if not decoded:
+    encoded = space.fsencode_w(w_path_or_fd)
+    if not encoded:
         return '\0'
  
     m = '\0'
 
     # In CPython the _wcsicmp function is used to perform case insensitive comparison
-    dlower = decoded.lower()
+    dlower = encoded.lower()
     if len(dlower) >=4:
         if dlower[:4] == '\\\\.\\' or dlower[:4] == '\\\\?\\':
             dlower = dlower[4:]
@@ -156,8 +166,19 @@ def _pyio_get_console_type(space, w_path_or_fd):
         m = 'w'
     elif dlower == 'con':
         m = 'x'
+    if m != '\0':
+        return m
 
-    # TODO: call GetFullPathNameW to deal with C:\Program Files\CONOUT$
+    # Handle things like 'c:\users\user\appdata\local\temp\usession\CONOUT$
+    dlower = getfullpathname(encoded).lower()
+    if dlower[:4] == '\\\\.\\' or dlower[:4] == '\\\\?\\':
+        dlower = dlower[4:]
+    if  dlower == 'conin$':
+        m = 'r'
+    elif dlower == 'conout$':
+        m = 'w'
+    elif dlower == 'con':
+        m = 'x'
     return m
 
 
@@ -277,26 +298,24 @@ class W_WinConsoleIO(W_RawIOBase):
             if self.writable:
                 access = rwin32.GENERIC_WRITE
         
-            traits = _preferred_traits(space.realunicode_w(w_path))
-            if not (traits.str is unicode):
-                raise oefmt(space.w_ValueError,
-                            "Non-unicode string name %s", traits.str)
+            traits = utf8_traits
             win32traits = make_win32_traits(traits)
             
             pathlen = space.len_w(w_path)
             name = rffi.utf82wcharp(space.utf8_w(w_path), pathlen)
-            self.handle = win32traits.CreateFile(name, 
-                rwin32.GENERIC_READ | rwin32.GENERIC_WRITE,
-                rwin32.FILE_SHARE_READ | rwin32.FILE_SHARE_WRITE,
-                rffi.NULL, win32traits.OPEN_EXISTING,
-                0, rffi.cast(rwin32.HANDLE, 0))
-            if self.handle == rwin32.INVALID_HANDLE_VALUE:
+            try:
                 self.handle = win32traits.CreateFile(name, 
-                    access,
-                    rwin32.FILE_SHARE_READ | rwin32.FILE_SHARE_WRITE,
+                    rwin32.ALL_READ_WRITE, rwin32.SHARE_READ_WRITE,
                     rffi.NULL, win32traits.OPEN_EXISTING,
                     0, rffi.cast(rwin32.HANDLE, 0))
-            lltype.free(name, flavor='raw')
+                if self.handle == rwin32.INVALID_HANDLE_VALUE:
+                    self.handle = win32traits.CreateFile(name, 
+                        access,
+                        rwin32.SHARE_READ_WRITE,
+                        rffi.NULL, win32traits.OPEN_EXISTING,
+                        0, rffi.cast(rwin32.HANDLE, 0))
+            finally:
+                lltype.free(name, flavor='raw')
             
             if self.handle == rwin32.INVALID_HANDLE_VALUE:
                 raise WindowsError(rwin32.GetLastError_saved(),
@@ -347,7 +366,7 @@ class W_WinConsoleIO(W_RawIOBase):
             return space.newtext("<%s name=%s>" % (typename, name_repr))
             
     def fileno_w(self, space):
-        traits = _preferred_traits(u"")
+        traits = utf8_traits
         win32traits = make_win32_traits(traits)
         if self.fd < 0 and self.handle != rwin32.INVALID_HANDLE_VALUE:
             if self.writable:
@@ -396,12 +415,13 @@ class W_WinConsoleIO(W_RawIOBase):
         # first copy any remaining buffered utf16 data
         builder = StringBuilder(length)
         wbuf = self._getbuffer(length * 2)
+        w_wbuf = space.newbytes(wbuf)
 
         state = space.fromcache(CodecState)
         errh = state.decode_error_handler
         outlen = 0
         if len(wbuf) > 0:
-            utf8, lgt, pos = str_decode_utf_16(wbuf, 'strict', final=True, errorhandler=errh)
+            utf8, lgt, pos = str_decode_utf_16(space, wbuf, w_wbuf, 'strict', final=True, errorhandler=errh)
             if self.mode == 'u':
                 length -= lgt
                 outlen += lgt
@@ -412,7 +432,7 @@ class W_WinConsoleIO(W_RawIOBase):
         
         if length > 0:
             wbuf = read_console_wide(space, self.handle, length)
-            utf8, lgt, pos = str_decode_utf_16(wbuf, 'strict', final=True, errorhandler=errh)
+            utf8, lgt, pos = str_decode_utf_16(space, wbuf, w_wbuf, 'strict', final=True, errorhandler=errh)
             if 1 or self.mode == 'u':
                 length -= lgt
                 outlen += lgt
@@ -461,7 +481,8 @@ class W_WinConsoleIO(W_RawIOBase):
         wbuf = result.build()
         state = space.fromcache(CodecState)
         errh = state.decode_error_handler
-        utf8, lgt, pos = str_decode_utf_16(wbuf, 'strict', final=True, errorhandler=errh)
+        w_wbuf = space.newbytes(wbuf)
+        utf8, lgt, pos = str_decode_utf_16(space, wbuf, w_wbuf, 'strict', final=True, errorhandler=errh)
 
         return space.newtext(utf8, lgt)
 
@@ -472,14 +493,17 @@ class W_WinConsoleIO(W_RawIOBase):
         if not self.writable:
             raise err_mode(space,"writing")
         
-        utf8 = space.utf8_w(w_data)
+        if space.isinstance_w(w_data, space.w_text):
+            utf8 = space.utf8_w(w_data)
+        else:
+            utf8 = space.bytes_w(w_data)
         if not len(utf8):
             return space.newint(0)
         
         # TODO: break up the encoding into chunks to save memory
         state = space.fromcache(CodecState)
         errh = state.encode_error_handler
-        utf16 = utf8_encode_utf_16(utf8, 'strict', errh, allow_surrogates=False)
+        utf16 = utf8_encode_utf_16(space, utf8, w_data, 'strict', errh, allow_surrogates=False)
         wlen = len(utf16) // 2
     
         with lltype.scoped_alloc(rwin32.LPDWORD.TO, 1) as n:
@@ -495,7 +519,7 @@ class W_WinConsoleIO(W_RawIOBase):
                         raise OperationError(space.w_WindowsError, space.newint(err))
                     nwrote = intmask(n[0])
                     offset += nwrote
-                return space.newint(offset - 1)
+        return space.newint(len(utf8))
             
     def get_blksize(self,space):
         return space.newint(self.blksize)
