@@ -13,7 +13,7 @@ class AppTestBufferedReader:
     spaceconfig = dict(usemodules=['_io'])
     if os.name != 'nt':
         spaceconfig['usemodules'].append('fcntl')
-        
+
     def setup_class(cls):
         tmpfile = udir.join('tmpfile')
         tmpfile.write("a\nb\nc", mode='wb')
@@ -188,12 +188,12 @@ class AppTestBufferedReader:
             exc = raises(TypeError, readinto, u"hello")
             msg = str(exc.value)
             # print(msg)
-            assert " read-write b" in msg and msg.endswith(", not str")
+            assert " read-write b" in msg and msg.endswith(", not 'str'")
             #
             exc = raises(TypeError, readinto, memoryview(b"hello"))
             msg = str(exc.value)
             # print(msg)
-            assert " read-write b" in msg and msg.endswith(", not memoryview")
+            assert " read-write b" in msg and msg.endswith(", not 'memoryview'")
             #
             f.close()
             assert a == b'a\nb\ncxxxxx'
@@ -307,6 +307,17 @@ class AppTestBufferedReader:
         raises(ValueError, buf.isatty)
         repr(buf)  # Should still work
 
+    def test_detach_warn(self):
+        import _io, warnings
+        warnings.resetwarnings()
+        raw = _io.FileIO(self.tmpfile)
+        f = _io.BufferedReader(raw)
+        raw2 = f.detach()
+        assert raw2 == raw
+        raw.close()
+        with raises(RuntimeError):
+            f._dealloc_warn(1)
+
     def test_tell(self):
         import _io
         raw = _io.FileIO(self.tmpfile)
@@ -328,6 +339,7 @@ class AppTestBufferedReader:
         raw = _io.FileIO(self.tmpfile)
         f = _io.BufferedReader(raw)
         assert repr(f) == '<_io.BufferedReader name=%r>' % (self.tmpfile,)
+        f.close()
 
     def test_read_interrupted(self):
         import _io, errno
@@ -413,12 +425,49 @@ class AppTestBufferedReader:
         f.close()
         self.posix.close(fdout)
 
+    def test_read_nonblocking_crash(self):
+        import _io as io
+        try:
+            import fcntl
+        except ImportError:
+            skip('fcntl missing')
+        fdin, fdout = self.posix.pipe()
+        f = io.open(fdin, "rb")
+        fl = fcntl.fcntl(f, fcntl.F_GETFL)
+        fcntl.fcntl(f, fcntl.F_SETFL, fl | self.posix.O_NONBLOCK)
+        s = f.read(12)
+        assert s == None
+        self.posix.close(fdout)
+
+        s = f.read(12)
+        assert s == b''
+        f.close()
+
+    def test_device_like(self):
+        # CPython GH-95782
+        import _io
+        class MockDevice(_io.BytesIO):
+            def seek(self, *args):
+                return 0
+
+            def tell(self, *args):
+                return 0
+
+        fd = MockDevice(b"aaaa")
+        bufio = _io.BufferedReader(fd, 4)
+        textio = _io.TextIOWrapper(bufio)
+        val = textio.read(1)
+        assert val == "a", val
+        t = textio.tell()
+        assert t == 0, t
+        assert textio.seek(0) == 0
+
 
 class AppTestBufferedReaderWithThreads(AppTestBufferedReader):
     spaceconfig = dict(usemodules=['_io', 'thread', 'time'])
     if os.name != 'nt':
         spaceconfig['usemodules'].append('fcntl')
-        
+
 
     def test_readinto_small_parts(self):
         import _io, os, _thread, time
@@ -552,7 +601,7 @@ class AppTestBufferedWriter:
         import _io
         raw = _io.FileIO(self.tmpfile, 'w+')
         raw.write(b'x' * 20)
-        b = _io.BufferedReader(raw)
+        b = _io.BufferedRandom(raw)
         assert b.seek(8) == 8
         assert b.truncate() == 8
         assert b.tell() == 8
@@ -1015,3 +1064,13 @@ class TestNonReentrantLock:
         with lock:
             exc = py.test.raises(OperationError, "with lock: pass")
         assert exc.value.match(space, space.w_RuntimeError)
+
+    def test_fast_closed_check(self, space):
+        from pypy.module._io.interp_fileio import W_FileIO
+        from pypy.module._io.interp_bufferedio import W_BufferedRandom
+        tmpfile = udir.join('tmpfile')
+        tmpfile.write("a\nb\nc", mode='wb')
+        w_fn = space.appexec([space.newtext(str(tmpfile))], """(fn):
+            return open(fn, "rb")""")
+        assert w_fn._fast_closed_check == True
+
