@@ -190,7 +190,8 @@ class W_ArrayInstance(W_DataInstance):
 
     def buffer_w(self, space, flags):
         return RawBufferView(
-            RawFFIBuffer(self), self.shape.itemcode, self.shape.size)
+            RawFFIBuffer(self), self.shape.itemcode, self.shape.size,
+            w_obj=self)
 
 
 W_ArrayInstance.typedef = TypeDef(
