@@ -14,7 +14,7 @@ from rpython.rtyper.lltypesystem import lltype, rffi
 from rpython.rtyper.tool import rffi_platform
 from rpython.rlib.unroll import unrolling_iterable
 from rpython.rlib import rutf8
-from rpython.rlib.objectmodel import specialize
+from rpython.rlib.objectmodel import specialize, dict_to_switch
 import rpython.rlib.rposix as rposix
 
 _MS_WINDOWS = os.name == "nt"
@@ -52,17 +52,18 @@ TYPEMAP = {
     'O' : ffi_type_pointer,
     'Z' : ffi_type_pointer,
     '?' : cast_type_to_ffitype(lltype.Bool),
+    'v' : ffi_type_sshort
 }
 TYPEMAP_PTR_LETTERS = "POszZ"
-TYPEMAP_NUMBER_LETTERS = "bBhHiIlLqQ?"
+TYPEMAP_NUMBER_LETTERS = "bBhHiIlLqQ?v"
 TYPEMAP_FLOAT_LETTERS = "fd" # XXX long doubles are not propperly supported in
                              # rpython, so we ignore then here
 
 if _MS_WINDOWS:
     TYPEMAP['X'] = ffi_type_pointer
-    TYPEMAP['v'] = ffi_type_sshort
     TYPEMAP_PTR_LETTERS += 'X'
-    TYPEMAP_NUMBER_LETTERS += 'v'
+
+TYPEMAP_FUNCTION = dict_to_switch(TYPEMAP)
 
 def size_alignment(ffi_type):
     return intmask(ffi_type.c_size), intmask(ffi_type.c_alignment)
@@ -89,11 +90,11 @@ LL_TYPEMAP = {
     'O' : rffi.VOIDP,
     'P' : rffi.VOIDP,
     '?' : lltype.Bool,
+    'v' : rffi.SHORT
 }
 
 if _MS_WINDOWS:
     LL_TYPEMAP['X'] = rffi.CCHARP
-    LL_TYPEMAP['v'] = rffi.SHORT
 
 def letter2tp(space, key):
     from pypy.module._rawffi.interp_array import PRIMITIVE_ARRAY_TYPES
@@ -167,8 +168,8 @@ class W_CDLL(W_Root):
         self.w_cache = space.newdict()
         self.space = space
 
-    @unwrap_spec(flags=int)
-    def ptr(self, space, w_name, w_argtypes, w_restype, flags=FUNCFLAG_CDECL):
+    @unwrap_spec(flags=int, variadic_args=int)
+    def ptr(self, space, w_name, w_argtypes, w_restype, flags=FUNCFLAG_CDECL, variadic_args=0):
         """ Get a pointer for function name with provided argtypes
         and restype
         """
@@ -201,7 +202,7 @@ class W_CDLL(W_Root):
 
             try:
                 ptr = self.cdll.getrawpointer(name, ffi_argtypes, ffi_restype,
-                                              flags)
+                                              flags, variadic_args)
             except KeyError:
                 raise oefmt(space.w_AttributeError,
                             "No symbol %s found in library %s",
@@ -213,7 +214,7 @@ class W_CDLL(W_Root):
             ordinal = space.int_w(w_name)
             try:
                 ptr = self.cdll.getrawpointer_byordinal(ordinal, ffi_argtypes,
-                                                        ffi_restype, flags)
+                                                        ffi_restype, flags, variadic_args)
             except KeyError:
                 raise oefmt(space.w_AttributeError,
                             "No symbol %d found in library %s",
@@ -291,7 +292,7 @@ def read_ptr(ptr, ofs, TP):
             with lltype.scoped_alloc(T.TO, 1) as t_array:
                 rffi.c_memcpy(
                     rffi.cast(rffi.VOIDP, t_array),
-                    rffi.cast(rffi.VOIDP, ptr),
+                    rffi.cast(rffi.CONST_VOIDP, ptr),
                     rffi.sizeof(TP))
                 ptr_val = t_array[0]
                 return ptr_val
@@ -314,13 +315,16 @@ def write_ptr(ptr, ofs, value):
                 s_array[0] = value
                 rffi.c_memcpy(
                     rffi.cast(rffi.VOIDP, ptr),
-                    rffi.cast(rffi.VOIDP, s_array),
+                    rffi.cast(rffi.CONST_VOIDP, s_array),
                     rffi.sizeof(TP))
                 return
     else:
         rffi.cast(T, ptr)[ofs] = value
 
 def segfault_exception(space, reason):
+    # This will cause _rawffi to be pre-installed in sys.modules, and it will
+    # have  __loader__ and __spec__ set to None, so will not be marked as
+    # "built-in"
     w_mod = space.getbuiltinmodule("_rawffi")
     w_exception = space.getattr(w_mod, space.newtext("SegfaultException"))
     return OperationError(w_exception, space.newtext(reason))
@@ -343,8 +347,8 @@ class W_DataShape(W_Root):
 
     @unwrap_spec(n=int)
     def descr_size_alignment(self, space, n=1):
-        return space.newtuple([space.newint(self.size * n),
-                               space.newint(self.alignment)])
+        return space.newtuple2(space.newint(self.size * n),
+                               space.newint(self.alignment))
 
 
 class W_DataInstance(W_Root):
@@ -387,7 +391,7 @@ class W_DataInstance(W_Root):
         self._ll_buffer = self.ll_buffer
 
     def buffer_w(self, space, flags):
-        return SimpleView(RawFFIBuffer(self))
+        return SimpleView(RawFFIBuffer(self), w_obj=self)
 
     def getrawsize(self):
         raise NotImplementedError("abstract base class")
@@ -584,7 +588,7 @@ def _create_new_accessor(func_name, name):
             raise oefmt(space.w_ValueError, "Expecting string of length one")
         tp_letter = tp_letter[0] # fool annotator
         try:
-            return space.newint(intmask(getattr(TYPEMAP[tp_letter], name)))
+            return space.newint(intmask(getattr(TYPEMAP_FUNCTION(tp_letter), name)))
         except KeyError:
             raise oefmt(space.w_ValueError, "Unknown type specification %s",
                         tp_letter)
@@ -667,7 +671,7 @@ def set_errno(space, w_errno):
 
 if sys.platform == 'win32':
     # see also
-    # https://bitbucket.org/pypy/pypy/issue/1944/ctypes-on-windows-getlasterror
+    # issue #1944 ctypes-on-windows-getlasterror
     def get_last_error(space):
         return space.newint(rwin32.GetLastError_alt_saved())
     @unwrap_spec(error=int)
@@ -675,7 +679,7 @@ if sys.platform == 'win32':
         rwin32.SetLastError_alt_saved(error)
 else:
     # always have at least a dummy version of these functions
-    # (https://bugs.pypy.org/issue1242)
+    # issue 1242
     def get_last_error(space):
         return space.newint(0)
     @unwrap_spec(error=int)
