@@ -6,7 +6,7 @@ from pypy.module._rawffi.tracker import Tracker
 
 import os, sys, py
 
-class AppTestFfi:
+class AppTestFfi(object):
     spaceconfig = dict(usemodules=['_rawffi', 'struct'])
 
     def prepare_c_example():
@@ -16,6 +16,7 @@ class AppTestFfi:
         #include "src/precommondefs.h"
         #include <stdlib.h>
         #include <stdio.h>
+        #include <stdarg.h>
         #include <errno.h>
 
         struct x
@@ -57,11 +58,11 @@ class AppTestFfi:
         }
 
         const char *static_str = "xxxxxx";
-        RPY_EXPORTED
+        RPY_EXPORTED long static_int;
         long static_int = 42;
-        RPY_EXPORTED
+        RPY_EXPORTED double static_double;
         double static_double = 42.42;
-        RPY_EXPORTED
+        RPY_EXPORTED long double static_longdouble;
         long double static_longdouble = 42.42;
 
         RPY_EXPORTED
@@ -223,6 +224,22 @@ class AppTestFfi:
             return old_errno;
         }
         #endif
+
+        RPY_EXPORTED
+        long variadic_sum(long n, ...) {
+            va_list ptr;
+            int sum = 0;
+            printf("n: %ld\\n", n);
+
+            va_start(ptr, n);
+            for (int i = 0; i < n; i++) {
+                long foo = va_arg(ptr, long);
+                sum += foo;
+                printf("Arg %d: %ld\\n", i, foo);
+            }
+            va_end(ptr);
+            return sum;
+        }
         '''))
         eci = ExternalCompilationInfo(include_dirs=[cdir])
         return str(platform.compile([c_file], eci, 'x', standalone=False))
@@ -245,6 +262,8 @@ class AppTestFfi:
         cls.w_sizes_and_alignments = space.wrap(dict(
             [(k, (v.c_size, v.c_alignment)) for k,v in TYPEMAP.iteritems()]))
         cls.w_float_typemap = space.wrap(TYPEMAP_FLOAT_LETTERS)
+        cls.w_is64bit = space.wrap(sys.maxint > 2147483647)
+        cls.w_runappdirect = space.wrap(cls.runappdirect)
 
     def test_libload(self):
         import _rawffi
@@ -262,14 +281,17 @@ class AppTestFfi:
             raise AssertionError("did not fail??")
 
     def test_libload_None(self):
-        if self.iswin32:
-            skip("unix specific")
         import _rawffi
         # this should return *all* loaded libs, dlopen(NULL)
-        dll = _rawffi.CDLL(None)
-        func = dll.ptr('rand', [], 'i')
-        res = func()
-        assert res[0] != 0
+        try:
+            dll = _rawffi.CDLL(None)
+        except TypeError:
+            if not self.iswin32:
+                raise
+        else:
+            func = dll.ptr('rand', [], 'i')
+            res = func()
+            assert res[0] != 0
 
     def test_libc_load(self):
         import _rawffi
@@ -348,10 +370,29 @@ class AppTestFfi:
         arg2.free()
         a.free()
 
+    def test_unicode_array(self):
+        import _rawffi
+        A = _rawffi.Array('u')
+        a = A(6, u'\u1234')
+        assert a[0] == u'\u1234'
+        a[0] = u'\ud800'
+        assert a[0] == u'\ud800'
+        if _rawffi.sizeof('u') == 4:
+            a[0] = u'\U00012345'
+            assert a[0] == u'\U00012345'
+            B = _rawffi.Array('i')
+            b = B.fromaddress(a.itemaddress(0), 1)
+            b[0] = 0xffffffff
+            if self.runappdirect:
+                # Fails to run untranslated: ll2ctypes is not up to the task of
+                # creating a unicode char w/value 0xffffffff
+                raises(ValueError, "a[0]")
+        a.free()
+
     def test_returning_unicode(self):
         import _rawffi
         A = _rawffi.Array('u')
-        a = A(6, 'xx\x00\x00xx')
+        a = A(6, u'xx\x00\x00xx')
         for i in (-1, 6):
             res = _rawffi.wcharp2unicode(a.buffer, i)
             assert isinstance(res, str)
@@ -704,7 +745,7 @@ class AppTestFfi:
         ll_to_sort = _rawffi.Array('i')(4)
         for i in range(4):
             ll_to_sort[i] = 4-i
-        qsort = libc.ptr('qsort', ['P', 'l', 'l', 'P'], None)
+        qsort = libc.ptr('qsort', ['P', 'Z', 'Z', 'P'], None)
         bogus_args = []
         def compare(a, b):
             a1 = _rawffi.Array('i').fromaddress(_rawffi.Array('P').fromaddress(a, 1)[0], 1)
@@ -715,11 +756,11 @@ class AppTestFfi:
                 return 1
             return -1
         a1 = ll_to_sort.byptr()
-        a2 = _rawffi.Array('l')(1)
+        a2 = _rawffi.Array('Z')(1)
         a2[0] = len(ll_to_sort)
-        a3 = _rawffi.Array('l')(1)
+        a3 = _rawffi.Array('Z')(1)
         a3[0] = struct.calcsize('i')
-        cb = _rawffi.CallbackPtr(compare, ['P', 'P'], 'l')
+        cb = _rawffi.CallbackPtr(compare, ['P', 'P'], 'i')
         a4 = cb.byptr()
         qsort(a1, a2, a3, a4)
         res = [ll_to_sort[i] for i in range(len(ll_to_sort))]
@@ -892,12 +933,12 @@ class AppTestFfi:
         import _rawffi, sys
         A = _rawffi.Array('u')
         a = A(3)
-        a[0] = 'x'
-        a[1] = 'y'
-        a[2] = 'z'
-        assert a[0] == 'x'
+        a[0] = u'x'
+        a[1] = u'y'
+        a[2] = u'z'
+        assert a[0] == u'x'
         b = _rawffi.Array('c').fromaddress(a.buffer, 38)
-        if sys.maxunicode > 65535:
+        if _rawffi.sizeof('u') == 4:
             # UCS4 build
             if sys.byteorder == 'big':
                 assert b[0:8] == b'\x00\x00\x00x\x00\x00\x00y'
@@ -905,7 +946,8 @@ class AppTestFfi:
                 assert b[0:5] == b'x\x00\x00\x00y'
         else:
             # UCS2 build
-            assert b[0:2] == b'x\x00y'
+            print(b[0:4])
+            assert b[0:4] == b'x\x00y\x00'
         a.free()
 
     def test_truncate(self):
@@ -1006,8 +1048,8 @@ class AppTestFfi:
         raises(_rawffi.SegfaultException, a.__setitem__, 3, 3)
 
     def test_stackcheck(self):
-        if self.platform != "msvc":
-            skip("win32 msvc specific")
+        if self.platform != "msvc" or self.is64bit:
+            skip("32-bit win32 msvc specific")
 
         # Even if the call corresponds to the specified signature,
         # the STDCALL calling convention may detect some errors
@@ -1039,13 +1081,10 @@ class AppTestFfi:
         X_Y = _rawffi.Structure([('x', 'l'), ('y', 'l')])
         x_y = X_Y()
         lib = _rawffi.CDLL(self.lib_name)
-        print("getting...")
         sum_x_y = lib.ptr('sum_x_y', [(X_Y, 1)], 'l')
         x_y.x = 200
         x_y.y = 220
-        print("calling...")
         res = sum_x_y(x_y)
-        print("done")
         assert res[0] == 420
         x_y.free()
 
@@ -1131,8 +1170,8 @@ class AppTestFfi:
         a[3] = b'x'
         b = memoryview(a)
         assert len(b) == 10
-        assert b[3] == ord(b'x')
-        b[6] = ord(b'y')
+        assert b[3] == b'x'
+        b[6] = b'y'
         assert a[6] == b'y'
         b[3:5] = b'zt'
         assert a[3] == b'z'
@@ -1140,9 +1179,9 @@ class AppTestFfi:
 
         b = memoryview(a)
         assert len(b) == 10
-        assert b[3] == ord(b'z')
-        b[3] = ord(b'x')
-        assert b[3] == ord(b'x')
+        assert b[3] == b'z'
+        b[3] = b'x'
+        assert b[3] == b'x'
 
     def test_pypy_raw_address(self):
         import _rawffi
@@ -1235,13 +1274,33 @@ class AppTestFfi:
         u = _rawffi.wcharp2rawunicode(arg.itemaddress(0), 1)
         assert u == u'\u1234'
         arg[0] = -1
-        raises(ValueError, _rawffi.wcharp2rawunicode, arg.itemaddress(0))
-        raises(ValueError, _rawffi.wcharp2rawunicode, arg.itemaddress(0), 1)
-        arg[0] = 0x110000
-        raises(ValueError, _rawffi.wcharp2rawunicode, arg.itemaddress(0))
-        raises(ValueError, _rawffi.wcharp2rawunicode, arg.itemaddress(0), 1)
+        if _rawffi.sizeof('u') == 4 and self.runappdirect:
+            # Fails to run untranslated: ll2ctypes is not up to the task of
+            # creating a unicode char w/value > maxunicode
+            raises(ValueError, _rawffi.wcharp2rawunicode, arg.itemaddress(0))
+            raises(ValueError, _rawffi.wcharp2rawunicode, arg.itemaddress(0), 1)
+            arg[0] = 0x110000
+            raises(ValueError, _rawffi.wcharp2rawunicode, arg.itemaddress(0))
+            raises(ValueError, _rawffi.wcharp2rawunicode, arg.itemaddress(0), 1)
         arg.free()
 
+    def test_variadic_sum(self):
+        import _rawffi
+        lib = _rawffi.CDLL(self.lib_name)
+
+        A = _rawffi.Array('l')
+        n = A(1)
+        arg1 = A(1)
+        arg2 = A(1)
+        n[0] = 2
+        arg1[0] = 3
+        arg2[0] = 15
+        variadic_sum = lib.ptr('variadic_sum', ['l', 'l', 'l'], 'l', variadic_args=2)
+        res = variadic_sum(n, arg1, arg2)
+        assert res[0] == 3 + 15
+        arg1.free()
+        arg2.free()
+        n.free()
 
 class AppTestAutoFree:
     spaceconfig = dict(usemodules=['_rawffi', 'struct'])
