@@ -4,6 +4,7 @@ from pypy.interpreter.error import oefmt
 from pypy.interpreter.gateway import interp2app, ObjSpace
 from pypy.interpreter.typedef import TypeDef, GetSetProperty
 from pypy.interpreter.executioncontext import AsyncAction, report_error
+from pypy.objspace.std.util import generic_alias_class_getitem
 from rpython.rlib import jit, rgc
 from rpython.rlib.rshrinklist import AbstractShrinkList
 from rpython.rlib.objectmodel import specialize
@@ -157,6 +158,8 @@ class WeakrefLifeline(W_Root):
 
 
 class W_WeakrefBase(W_Root):
+    exact_class_applevel_name = 'weakref-or-proxy'
+
     def __init__(self, space, w_obj, w_callable):
         assert w_callable is not space.w_None    # should be really None
         self.space = space
@@ -203,14 +206,8 @@ class W_Weakref(W_WeakrefBase):
         # to a frozen object, so it will never die.
         self.w_hash = None
 
-    def descr__init__weakref(self, space, w_obj, w_callable=None,
-                             __args__=None):
-        if __args__.arguments_w:
-            raise oefmt(space.w_TypeError,
-                        "__init__ expected at most 2 arguments")
-        if __args__.keywords:
-            raise oefmt(space.w_TypeError,
-                        "ref() does not take keyword arguments")
+    def descr__init__weakref(self, space, w_obj, w_callable=None):
+        pass
 
     def descr_hash(self):
         if self.w_hash is not None:
@@ -269,7 +266,7 @@ def descr__new__weakref(space, w_subtype, w_obj, w_callable=None,
     else:
         return lifeline.make_weakref_with_callback(w_subtype, w_obj, w_callable)
 
-W_Weakref.typedef = TypeDef("weakref",
+W_Weakref.typedef = TypeDef("weakref.ReferenceType",
     __doc__ = """A weak reference to an object 'obj'.  A 'callback' can be given,
 which is called with 'obj' as an argument when it is about to be finalized.""",
     __new__ = interp2app(descr__new__weakref),
@@ -280,6 +277,8 @@ which is called with 'obj' as an argument when it is about to be finalized.""",
     __call__ = interp2app(W_Weakref.descr_call),
     __repr__ = interp2app(W_WeakrefBase.descr__repr__),
     __callback__ = GetSetProperty(W_Weakref.descr_callback),
+    __class_getitem__ = interp2app(
+        generic_alias_class_getitem, as_classmethod=True),
 )
 
 
@@ -314,15 +313,20 @@ def getweakrefs(space, w_obj):
 #_________________________________________________________________
 # Proxy
 
-class W_Proxy(W_WeakrefBase):
+class W_AbstractProxy(W_WeakrefBase):
+    pass
+
+class W_Proxy(W_AbstractProxy):
     def descr__hash__(self, space):
         raise oefmt(space.w_TypeError, "unhashable type")
 
-class W_CallableProxy(W_Proxy):
+class W_CallableProxy(W_AbstractProxy):
     def descr__call__(self, space, __args__):
         w_obj = force(space, self)
         return space.call_args(w_obj, __args__)
 
+    def descr__hash__(self, space):
+        raise oefmt(space.w_TypeError, "unhashable type")
 
 def proxy(space, w_obj, w_callable=None):
     """Create a proxy object that weakly references 'obj'.
@@ -343,7 +347,7 @@ def descr__new__callableproxy(space, w_subtype, w_obj, w_callable=None):
 
 
 def force(space, proxy):
-    if not isinstance(proxy, W_Proxy):
+    if not isinstance(proxy, W_AbstractProxy):
         return proxy
     w_obj = proxy.dereference()
     if w_obj is None:
@@ -371,28 +375,51 @@ for opname, _, arity, special_methods in ObjSpace.MethodTable:
     exec py.code.Source(code).compile()
 
     func.func_name = opname
-    for special_method in special_methods:
-        proxy_typedef_dict[special_method] = interp2app(func)
-        callable_proxy_typedef_dict[special_method] = interp2app(func)
+    if len(special_methods) == 2 and special_methods[1].startswith("__r"):
+        proxy_typedef_dict[special_methods[0]] = interp2app(func)
+        callable_proxy_typedef_dict[special_methods[0]] = interp2app(func)
+
+        # HACK: need to call the space method with arguments in the reverse
+        # order!
+        code = code.replace("(w_obj0, w_obj1)", "(w_obj1, w_obj0)")
+        code = code.replace("func", "rfunc")
+
+        exec py.code.Source(code).compile()
+        rfunc.func_name = special_methods[1][2:-2]
+
+        proxy_typedef_dict[special_methods[1]] = interp2app(rfunc)
+        callable_proxy_typedef_dict[special_methods[1]] = interp2app(rfunc)
+    elif opname in ["lt", "le", "gt", "ge", "eq", "ne"]:
+        proxy_typedef_dict[special_methods[0]] = interp2app(func)
+    else:
+        for special_method in special_methods:
+            proxy_typedef_dict[special_method] = interp2app(func)
+            callable_proxy_typedef_dict[special_method] = interp2app(func)
 
 # __bytes__ is not yet a space operation
-def proxy_unicode(space, w_obj):
+def proxy_bytes(space, w_obj):
     w_obj = force(space, w_obj)
     return space.call_method(w_obj, '__bytes__')
-proxy_typedef_dict['__bytes__'] = interp2app(proxy_unicode)
-callable_proxy_typedef_dict['__bytes__'] = interp2app(proxy_unicode)
+proxy_typedef_dict['__bytes__'] = interp2app(proxy_bytes)
+callable_proxy_typedef_dict['__bytes__'] = interp2app(proxy_bytes)
 
+# neither is __reversed__
+def proxy_reversed(space, w_obj):
+    w_obj = force(space, w_obj)
+    return space.call_method(w_obj, '__reversed__')
+proxy_typedef_dict['__reversed__'] = interp2app(proxy_reversed)
+callable_proxy_typedef_dict['__reversed__'] = interp2app(proxy_reversed)
 
-W_Proxy.typedef = TypeDef("weakproxy",
+W_Proxy.typedef = TypeDef("weakref.ProxyType",
     __new__ = interp2app(descr__new__proxy),
     __hash__ = interp2app(W_Proxy.descr__hash__),
     __repr__ = interp2app(W_WeakrefBase.descr__repr__),
     **proxy_typedef_dict)
 W_Proxy.typedef.acceptable_as_base_class = False
 
-W_CallableProxy.typedef = TypeDef("weakcallableproxy",
+W_CallableProxy.typedef = TypeDef("weakref.CallableProxyType",
     __new__ = interp2app(descr__new__callableproxy),
-    __hash__ = interp2app(W_Proxy.descr__hash__),
+    __hash__ = interp2app(W_CallableProxy.descr__hash__),
     __repr__ = interp2app(W_WeakrefBase.descr__repr__),
     __call__ = interp2app(W_CallableProxy.descr__call__),
     **callable_proxy_typedef_dict)
