@@ -1,4 +1,4 @@
-from pypy.interpreter.error import oefmt
+from pypy.interpreter.error import oefmt, OperationError
 from pypy.interpreter.astcompiler import consts
 from rpython.rtyper.lltypesystem import rffi, lltype
 from rpython.rlib.objectmodel import we_are_translated
@@ -8,12 +8,12 @@ from pypy.module.cpyext.api import (
     cpython_struct, ferror)
 from pypy.module.cpyext.pyobject import PyObject
 from pypy.module.cpyext.pyerrors import PyErr_SetFromErrno
-from pypy.module.cpyext.funcobject import PyCodeObject
 from pypy.module.cpyext.frameobject import PyFrameObject
 from pypy.module.__builtin__ import compiling
 
 PyCompilerFlags = cpython_struct(
-    "PyCompilerFlags", (("cf_flags", rffi.INT),))
+    "PyCompilerFlags", (("cf_flags", rffi.INT),
+                        ("cf_feature_version", rffi.INT)))
 PyCompilerFlagsPtr = lltype.Ptr(PyCompilerFlags)
 
 PyCF_MASK = (consts.CO_FUTURE_DIVISION |
@@ -64,7 +64,7 @@ def PyEval_GetFrame(space):
     caller = space.getexecutioncontext().gettopframe_nohidden()
     return caller    # borrowed ref, may be null
 
-@cpython_api([PyCodeObject, PyObject, PyObject], PyObject)
+@cpython_api([PyObject, PyObject, PyObject], PyObject)
 def PyEval_EvalCode(space, w_code, w_globals, w_locals):
     """This is a simplified interface to PyEval_EvalCodeEx(), with just
     the code object, and the dictionaries of global and local variables.
@@ -85,6 +85,24 @@ def PyObject_CallObject(space, w_obj, w_arg):
     callable_object(*args)."""
     return space.call(w_obj, w_arg)
 
+@cpython_api([PyObject], PyObject)
+def PyObject_CallNoArgs(space, w_obj):
+    return space.call_function(w_obj)
+
+@cpython_api([PyObject, PyObject], PyObject)
+def PyObject_CallOneArg(space, w_obj, w_arg):
+    return space.call_function(w_obj, w_arg)
+
+@cpython_api([PyObject, PyObject], PyObject)
+def PyObject_CallMethodNoArgs(space, w_self, w_name):
+    name = space.text_w(w_name)
+    return space.call_method(w_self, name)
+
+@cpython_api([PyObject, PyObject, PyObject], PyObject)
+def PyObject_CallMethodOneArg(space, w_self, w_name, w_arg):
+    name = space.text_w(w_name)
+    return space.call_method(w_self, name, w_arg)
+
 @cpython_api([PyObject, PyObject, PyObject], PyObject)
 def PyObject_Call(space, w_obj, w_args, w_kw):
     """
@@ -105,8 +123,9 @@ def PyCFunction_Call(space, w_obj, w_args, w_kw):
 Py_single_input = 256
 Py_file_input = 257
 Py_eval_input = 258
+Py_func_type_input = 345
 
-def compile_string(space, source, filename, start, flags=0):
+def compile_string(space, source, filename, start, flags=0, feature_version=-1):
     w_source = space.newbytes(source)
     start = rffi.cast(lltype.Signed, start)
     if start == Py_file_input:
@@ -115,10 +134,13 @@ def compile_string(space, source, filename, start, flags=0):
         mode = 'eval'
     elif start == Py_single_input:
         mode = 'single'
+    elif start == Py_func_type_input:
+        mode = 'func_type'
     else:
         raise oefmt(space.w_ValueError,
                     "invalid mode parameter for compilation")
-    return compiling.compile(space, w_source, filename, mode, flags)
+    return compiling.compile(space, w_source, filename, mode, flags,
+                             _feature_version=feature_version)
 
 def run_string(space, source, filename, start, w_globals, w_locals):
     w_code = compile_string(space, source, filename, start)
@@ -154,9 +176,12 @@ def PyRun_StringFlags(space, source, start, w_globals, w_locals, flagsptr):
     source = rffi.charp2str(source)
     if flagsptr:
         flags = rffi.cast(lltype.Signed, flagsptr.c_cf_flags)
+        feature_version = rffi.cast(lltype.Signed, flagsptr.c_cf_feature_version)
     else:
         flags = 0
-    w_code = compile_string(space, source, "<string>", start, flags)
+        feature_version = -1
+    w_code = compile_string(space, source, "<string>", start, flags=flags,
+                            feature_version=feature_version)
     return compiling.eval(space, w_code, w_globals, w_locals)
 
 @cpython_api([FILEP, CONST_STRING, rffi.INT_real, PyObject, PyObject], PyObject)
@@ -288,3 +313,31 @@ def Py_LeaveRecursiveCall(space):
     # A NOP in PyPy
     if not we_are_translated():
         limit = 0
+
+@cpython_api([], PyObject, result_borrowed=True)
+def _PyEval_GetAsyncGenFirstiter(space):
+    # sys.get_asyncgen_hooks().firstiter
+    ec = space.getexecutioncontext()
+    return ec.w_asyncgen_firstiter_fn
+
+@cpython_api([], PyObject, result_borrowed=True)
+def _PyEval_GetAsyncGenFinalizer(space):
+    # sys.get_asyncgen_hooks().finalizer
+    ec = space.getexecutioncontext()
+    return ec.w_asyncgen_finalizer_fn
+
+@cpython_api([PyObject], rffi.CONST_CCHARP)
+def PyEval_GetFuncName(space, w_obj):
+    try:
+        w_name = space.getattr(w_obj, space.newtext('__name__'))
+        name = space.utf8_w(w_name)
+    except OperationError as e:
+        e.write_unraisable(space, "PyEval_GetFuncName")
+        name = "unknown"
+        
+    # leak the allocation. We could track this in a cache, and free it at
+    # shutdown.
+    return rffi.str2constcharp(name, track_allocation=False)
+
+
+
