@@ -1,6 +1,6 @@
 from pypy.interpreter.error import oefmt
 from pypy.module.cpyext.api import (
-    cpython_api, CANNOT_FAIL, Py_MAX_FMT, Py_MAX_NDIMS, build_type_checkers,
+    cpython_api, CANNOT_FAIL, PyBUF_MAX_NDIM, build_type_checkers,
     Py_ssize_tP, cts, parse_dir, bootstrap_function, Py_bufferP, slot_function,
     PyBUF_READ, PyBUF_WRITE)
 from pypy.module.cpyext.pyobject import (
@@ -19,6 +19,7 @@ PyMemoryViewObject = cts.gettype('PyMemoryViewObject*')
 
 PyMemoryView_Check, PyMemoryView_CheckExact = build_type_checkers("MemoryView")
 
+FORMAT_ALLOCATED = 0x04
 
 @bootstrap_function
 def init_memoryobject(space):
@@ -38,15 +39,16 @@ def memory_attach(space, py_obj, w_obj, w_userdata=None):
     py_obj = rffi.cast(PyMemoryViewObject, py_obj)
     view = py_obj.c_view
     ndim = w_obj.getndim()
-    if ndim >= Py_MAX_NDIMS:
-        # XXX warn?
-        return
+    if ndim >= PyBUF_MAX_NDIM:
+        raise oefmt(space.w_RuntimeError,
+                    "cannot create PyMemoryViewObject with ndim (%d) > %d",
+                    ndim, PyBUF_MAX_NDIM)
     fill_Py_buffer(space, w_obj.view, view)
     try:
         view.c_buf = rffi.cast(rffi.VOIDP, w_obj.view.get_raw_address())
-        # not used in PyPy to keep something alive,
-        # but some c-extensions check the type without checking for NULL
-        view.c_obj = make_ref(space, space.w_None)
+        # In CPython, this is used to keep w_obj alive. We don't need that,
+        # but do it anyway for compatibility when checking mview.obj
+        view.c_obj = make_ref(space, w_obj)
         rffi.setintfield(view, 'c_readonly',
                          rffi.cast(rffi.INT_real, w_obj.view.readonly))
     except ValueError:
@@ -88,9 +90,13 @@ def memory_realize(space, obj):
 @slot_function([PyObject], lltype.Void)
 def memory_dealloc(space, py_obj):
     mem_obj = rffi.cast(PyMemoryViewObject, py_obj)
-    if mem_obj.c_view.c_obj:
-        decref(space, mem_obj.c_view.c_obj)
-    mem_obj.c_view.c_obj = rffi.cast(PyObject, 0)
+    view = mem_obj.c_view
+    if view.c_obj:
+        decref(space, view.c_obj)
+    view.c_obj = rffi.cast(PyObject, 0)
+    flags = widen(view.c_flags)
+    if flags & FORMAT_ALLOCATED == FORMAT_ALLOCATED:
+        lltype.free(view.c_format, flavor='raw')
     _dealloc(space, py_obj)
 
 def fill_Py_buffer(space, buf, view):
@@ -99,19 +105,13 @@ def fill_Py_buffer(space, buf, view):
     view.c_len = buf.getlength()
     view.c_itemsize = buf.getitemsize()
     rffi.setintfield(view, 'c_ndim', ndim)
-    view.c_format = rffi.cast(rffi.CCHARP, view.c__format)
     fmt = buf.getformat()
-    n = Py_MAX_FMT - 1 # NULL terminated buffer
-    if len(fmt) > n:
-        w_message = space.newbytes("PyPy specific Py_MAX_FMT is %d which is too "
-                           "small for buffer format, %d needed" % (
-                           Py_MAX_FMT, len(fmt)))
-        w_stacklevel = space.newint(1)
-        w_module = PyImport_Import(space, space.newbytes("warnings"))
-        w_warn = space.getattr(w_module, space.newbytes("warn"))
-        space.call_function(w_warn, w_message, space.w_None, w_stacklevel)
-    else:
-        n = len(fmt)
+    n = len(fmt)
+    view.c_format = lltype.malloc(rffi.CCHARP.TO, n + 1, flavor='raw',
+                                  add_memory_pressure=True)
+    flags = widen(view.c_flags)
+    flags |= FORMAT_ALLOCATED
+    view.c_flags = rffi.cast(rffi.INT_real, flags)
     for i in range(n):
         view.c_format[i] = fmt[i]
     view.c_format[n] = '\x00'
@@ -194,7 +194,7 @@ def PyMemoryView_FromMemory(space, mem, size, flags):
     PyBUF_READ or PyBUF_WRITE. view->format is set to "B" (unsigned bytes).
     The memoryview has complete buffer information.
     """
-    readonly = int(widen(flags) == PyBUF_WRITE)
+    readonly = int(widen(flags) == PyBUF_READ)
     view = CPyBuffer(space, cts.cast('void*', mem), size, None,
             readonly=readonly)
     w_mview = W_MemoryView(view)
@@ -218,7 +218,8 @@ def PyMemoryView_FromBuffer(space, view):
     py_mem = rffi.cast(PyMemoryViewObject, py_obj)
     mview = py_mem.c_view
     mview.c_buf = view.c_buf
-    mview.c_obj = view.c_obj
+    # like CPython, do not take a reference to the object
+    mview.c_obj = rffi.cast(PyObject, 0)
     mview.c_len = view.c_len
     mview.c_itemsize = view.c_itemsize
     mview.c_readonly = view.c_readonly
