@@ -17,7 +17,7 @@ from rpython.rlib.objectmodel import specialize, we_are_translated
 from rpython.rlib.objectmodel import keepalive_until_here
 from rpython.rtyper.annlowlevel import llhelper, cast_instance_to_base_ptr
 from rpython.rlib import rawrefcount, jit
-from rpython.rlib.debug import ll_assert, fatalerror
+from rpython.rlib.debug import ll_assert, fatalerror, check_annotation
 
 
 #________________________________________________________
@@ -33,15 +33,13 @@ def check_true(s_arg, bookeeper):
     assert s_arg.const is True
 
 def w_root_as_pyobj(w_obj, space):
-    from rpython.rlib.debug import check_annotation
-    # make sure that translation crashes if we see this while not translating
-    # with cpyext
+    # make sure that translation crashes if we see this while translating
+    # without cpyext
     check_annotation(space.config.objspace.usemodules.cpyext, check_true)
     # default implementation of _cpyext_as_pyobj
     return rawrefcount.from_obj(PyObject, w_obj)
 
 def w_root_attach_pyobj(w_obj, space, py_obj):
-    from rpython.rlib.debug import check_annotation
     check_annotation(space.config.objspace.usemodules.cpyext, check_true)
     assert space.config.objspace.usemodules.cpyext
     # default implementation of _cpyext_attach_pyobj
@@ -78,7 +76,7 @@ class BaseCpyTypedescr(object):
         return state.C._PyPy_subtype_dealloc
 
     # CCC port to C
-    def allocate(self, space, w_type, itemcount=0, immortal=False):
+    def allocate(self, space, w_type, itemcount=0, immortal=False, itemsize=-1):
         # typically called from PyType_GenericAlloc via typedescr.allocate
         # this returns a PyObject with ob_refcnt == 1.
 
@@ -90,18 +88,17 @@ class BaseCpyTypedescr(object):
         if flags & Py_TPFLAGS_HEAPTYPE:
             incref(space, pytype)
 
-        if pytype:
-            size = pytype.c_tp_basicsize
-        else:
-            size = rffi.sizeof(self.basestruct)
-        if pytype.c_tp_itemsize:
-            size += itemcount * pytype.c_tp_itemsize
+        size = pytype.c_tp_basicsize
+        if itemsize < 0:
+            itemsize = pytype.c_tp_itemsize
+        if itemsize:
+            size += itemcount * itemsize
         assert size >= rffi.sizeof(PyObject.TO)
         buf = lltype.malloc(rffi.VOIDP.TO, size,
                             flavor='raw', zero=True,
                             add_memory_pressure=True, immortal=immortal)
         pyobj = rffi.cast(PyObject, buf)
-        if pytype.c_tp_itemsize:
+        if itemsize or space.issubtype_w(w_type, space.w_list):
             pyvarobj = rffi.cast(PyVarObject, pyobj)
             pyvarobj.c_ob_size = itemcount
         pyobj.c_ob_refcnt = 1
@@ -156,7 +153,7 @@ def make_typedescr(typedef, **kw):
         basestruct = tp_basestruct
 
         if tp_alloc:
-            def allocate(self, space, w_type, itemcount=0, immortal=False):
+            def allocate(self, space, w_type, itemcount=0, immortal=False, itemsize=-1):
                 return tp_alloc(self, space, w_type, itemcount)
 
         if hasattr(tp_dealloc, 'api_func'):
@@ -217,8 +214,38 @@ def create_ref(space, w_obj, w_userdata=None, immortal=False):
     w_type = space.type(w_obj)
     pytype = rffi.cast(PyTypeObjectPtr, as_pyobj(space, w_type))
     typedescr = get_typedescr(w_obj.typedef)
-    if pytype.c_tp_itemsize != 0:
-        itemcount = space.len_w(w_obj) # PyBytesObject and subclasses
+    if space.is_w(w_type, space.w_text):
+        # These PyUnicodeObjects will always take the compact form
+        # since they come from uint8-encoded strings, so we must
+        # override the default tp_itemsize (see also unicode_alloc)
+        # Maybe this snippet should use the ptype.tp_alloc to allocate the py_obj?
+        itemsize = 1
+    elif space.is_w(w_type, space.w_type):
+        # Subclasses of "type" (which has space for PyMemberDef) do not
+        # need space for tp_member like heap types do
+        itemsize = 0
+    else:
+        itemsize = pytype.c_tp_itemsize
+    if itemsize != 0 or space.issubtype_w(w_type, space.w_list):
+        # PyBytesObject, compact PyUnicodeObject and subclasses
+        try:
+            # Can cause infinite recursion if w_obj.__len__ is a c function
+            # so the call will try to convert w_obj to a pyobj via create_ref
+            from pypy.objspace.std.listobject import W_ListObject
+            if  isinstance(w_obj, W_ListObject):
+                itemcount = w_obj.length()
+            else:
+                itemcount = space.len_w(w_obj)
+        except OperationError as e:
+            if e.match(space, space.w_TypeError):
+                # issue 4013: is this correct?
+                itemcount = 0
+                # raise oefmt(space.w_SystemError,
+                #            "cpyext: Failure to allocate '%N' (with a non-zero "
+                #            "tp_itemsize) when len(obj) cannot be calculated",
+                #            w_type)
+            else:
+                raise
     else:
         itemcount = 0
     py_obj = typedescr.allocate(space, w_type, itemcount=itemcount, immortal=immortal)
@@ -231,6 +258,22 @@ def create_ref(space, w_obj, w_userdata=None, immortal=False):
     py_obj.c_ob_refcnt -= 1
     #
     typedescr.attach(space, py_obj, w_obj, w_userdata)
+    # This is probably not the best place for this stanza, but we need the
+    # w_obj so it cannot be done in allocate()
+    # It should probably also be in track_reference
+    if pytype.c_tp_dictoffset:
+        dictoffset = pytype.c_tp_dictoffset
+        try:
+            w_dict = space.getattr(w_obj, space.newtext("__dict__"))
+        except OperationError as e:
+            dictref = make_ref(space, space.w_None)
+        else:
+            dictref = make_ref(space, w_dict)
+        if dictoffset < 0:
+            dictoffset += pytype.c_tp_basicsize
+        loc = rffi.ptradd(cts.cast("char *", py_obj), dictoffset)
+        dictloc = cts.cast("PyObject **", loc)
+        dictloc[0] = dictref
     return py_obj
 
 def track_reference(space, py_obj, w_obj):
@@ -257,7 +300,7 @@ def from_ref(space, ref):
     assert is_pyobj(ref)
     if not ref:
         return None
-    w_obj = rawrefcount.to_obj(W_Root, ref)
+    w_obj = rawrefcount.to_obj(W_Root, rffi.cast(PyObject, ref))
     if w_obj is not None:
         if w_obj is not w_marker_deallocating:
             return w_obj
@@ -362,7 +405,7 @@ def make_ref(space, w_obj, w_userdata=None, immortal=False):
         # XXX: adapt for pypy3
         state = space.fromcache(State)
         intval = space.int_w(w_obj)
-        return state.ccall("PyInt_FromLong", intval)
+        return state.ccall("PyLong_FromLong", intval)
     return get_pyobj_and_incref(space, w_obj, w_userdata, immortal=False)
 
 @specialize.ll()
