@@ -188,12 +188,29 @@ class AppTestDictObject(AppTestCpythonExtensionBase):
                 Py_RETURN_NONE;
              Py_XINCREF(result);
              return result;
+             """),
+            ("dict_getitem_string", "METH_VARARGS",
+             """
+             PyObject *d, *result;
+             char * key;
+             if (!PyArg_ParseTuple(args, "Os", &d, &key)) {
+                return NULL;
+             }
+             result = _PyDict_GetItemStringWithError(d, key);
+             if (result == NULL && !PyErr_Occurred())
+                Py_RETURN_NONE;
+             Py_XINCREF(result);
+             return result;
              """)])
         d = {'foo': 'bar'}
         assert module.dict_getitem(d, 'foo') == 'bar'
+        assert module.dict_getitem_string(d, 'foo') == 'bar'
         assert module.dict_getitem(d, 'missing') is None
+        assert module.dict_getitem_string(d, 'missing') is None
         with raises(TypeError):
             module.dict_getitem(d, [])
+        with raises(TypeError):
+            module.dict_getitem_string(d, [])
 
     def test_setdefault(self):
         module = self.import_extension('foo', [
@@ -344,6 +361,14 @@ class AppTestDictObject(AppTestCpythonExtensionBase):
                 }
                 return PyLong_FromLong(ret);
             '''),
+            ("exhaust_pos", "METH_O",
+             """
+                PyObject *key, *val;
+                Py_ssize_t pos = 0;
+                while (PyDict_Next(args, &pos, &key, &val)) {
+                }
+                return PyLong_FromLong(pos);
+             """),
             ])
         d = {'a': 1, 'b':2}
         assert module.dict_len(d) == 2
@@ -352,6 +377,7 @@ class AppTestDictObject(AppTestCpythonExtensionBase):
         assert module.dict_delitem(d, 'a') == 0
         r = module.dict_next({'a': 1, 'b': 2})
         assert r == 2
+        assert module.exhaust_pos({'a': 1}) == 1
 
     def test_subclassing(self):
         module = self.import_extension('foo', [
@@ -433,3 +459,29 @@ class AppTestDictObject(AppTestCpythonExtensionBase):
         ])
         assert module.dict_getitem(42, 43) is None
         assert module.dict_getitem({}, []) is None
+
+    def test_contains(self):
+        module = self.import_extension('foo', [
+            ("contains", "METH_VARARGS",
+             """
+             PyObject *d, *key;
+             if (!PyArg_ParseTuple(args, "OO", &d, &key)) {
+                return NULL;
+             }
+             int ret = PyDict_Contains(d, key);
+             if (ret < 0) return NULL;
+             return PyLong_FromLong(ret);
+             """),
+        ])
+
+        class Unhashable():
+            __hash__ = None
+
+        raises(TypeError, module.contains, {}, Unhashable())
+
+        class C(dict):
+            def __contains__(self, key):
+                return dict.__contains__(self, key)
+
+        ret = module.contains(C(), 1)
+        assert ret == 0
