@@ -4,8 +4,7 @@ from pypy.module.cpyext.test.test_api import BaseApiTest
 from pypy.module.cpyext.test.test_cpyext import AppTestCpythonExtensionBase
 from rpython.rtyper.lltypesystem import rffi
 from pypy.module.cpyext.floatobject import (
-    PyFloat_FromDouble, PyFloat_AsDouble, PyFloat_AS_DOUBLE, PyNumber_Float,
-    _PyFloat_Unpack4, _PyFloat_Unpack8)
+    PyFloat_FromDouble, PyFloat_AsDouble, PyFloat_AS_DOUBLE, PyNumber_Float)
 
 class TestFloatObject(BaseApiTest):
     def test_floatobject(self, space):
@@ -26,16 +25,6 @@ class TestFloatObject(BaseApiTest):
             return Coerce()""")
         assert space.eq_w(PyNumber_Float(space, w_obj), space.wrap(42.5))
 
-    def test_unpack(self, space):
-        with rffi.scoped_str2charp("\x9a\x99\x99?") as ptr:
-            assert abs(_PyFloat_Unpack4(space, ptr, 1) - 1.2) < 1e-7
-        with rffi.scoped_str2charp("?\x99\x99\x9a") as ptr:
-            assert abs(_PyFloat_Unpack4(space, ptr, 0) - 1.2) < 1e-7
-        with rffi.scoped_str2charp("\x1f\x85\xebQ\xb8\x1e\t@") as ptr:
-            assert abs(_PyFloat_Unpack8(space, ptr, 1) - 3.14) < 1e-15
-        with rffi.scoped_str2charp("@\t\x1e\xb8Q\xeb\x85\x1f") as ptr:
-            assert abs(_PyFloat_Unpack8(space, ptr, 0) - 3.14) < 1e-15
-
 class AppTestFloatObject(AppTestCpythonExtensionBase):
     def test_fromstring(self):
         module = self.import_extension('foo', [
@@ -131,49 +120,73 @@ class AppTestFloatMacros(AppTestCpythonExtensionBase):
             ])
             assert module.test() == const_strval
 
-    def test_Py_IS_NAN(self):
+    def test_Py_IS(self):
         module = self.import_extension('foo', [
-            ("test", "METH_O",
+            ("test_nan", "METH_O",
              """
                  double d = PyFloat_AsDouble(args);
                  return PyBool_FromLong(Py_IS_NAN(d));
              """),
-            ])
-        assert not module.test(0)
-        assert not module.test(1)
-        assert not module.test(-1)
-        assert not module.test(float('inf'))
-        assert module.test(float('nan'))
-
-    def test_Py_IS_INFINITY(self):
-        module = self.import_extension('foo', [
-            ("test", "METH_O",
+            ("test_inf", "METH_O",
              """
                  double d = PyFloat_AsDouble(args);
                  return PyBool_FromLong(Py_IS_INFINITY(d));
              """),
-            ])
-        assert not module.test(0)
-        assert not module.test(1)
-        assert not module.test(-1)
-        assert not module.test(float('nan'))
-        assert module.test(float('inf'))
-        assert module.test(float('-inf'))
-
-    def test_Py_IS_FINITE(self):
-        module = self.import_extension('foo', [
-            ("test", "METH_O",
+            ("test_finite", "METH_O",
              """
                  double d = PyFloat_AsDouble(args);
                  return PyBool_FromLong(Py_IS_FINITE(d));
              """),
             ])
-        assert module.test(0)
-        assert module.test(1)
-        assert module.test(-1)
-        assert not module.test(float('nan'))
-        assert not module.test(float('inf'))
-        assert not module.test(float('-inf'))
+        assert not module.test_nan(0)
+        assert not module.test_nan(1)
+        assert not module.test_nan(-1)
+        assert not module.test_nan(float('inf'))
+        assert module.test_nan(float('nan'))
+        assert not module.test_inf(0)
+        assert not module.test_inf(1)
+        assert not module.test_inf(-1)
+        assert not module.test_inf(float('nan'))
+        assert module.test_inf(float('inf'))
+        assert module.test_inf(float('-inf'))
+        assert module.test_finite(0)
+        assert module.test_finite(1)
+        assert module.test_finite(-1)
+        assert not module.test_finite(float('nan'))
+        assert not module.test_finite(float('inf'))
+        assert not module.test_finite(float('-inf'))
+
+    def test_Py_Float_AsDouble(self):
+        import warnings
+        module = self.import_extension('foo', [
+            ("test", "METH_O",
+             """
+                double d = PyFloat_AsDouble(args);
+                if (PyErr_Occurred()) {
+                    return NULL;
+                }
+                return PyFloat_FromDouble(d);
+             """),
+            ])
+        try:
+            module.test([])
+        except Exception as e:
+            print(str(e))
+            assert str(e) == 'must be real number, not list'
+        else:
+            assert False
+
+        class Floatsub(float): pass
+        class BadFloat:
+            def __float__(self):
+                return Floatsub(4.25)
+
+        with warnings.catch_warnings(record=True) as w:
+            warnings.simplefilter("always", DeprecationWarning)
+            res = module.test(BadFloat())
+        assert res == 4.25
+        assert len(w) == 1
+        assert "strict subclass" in str(w[0]), str(w[0])
 
     def test_Py_HUGE_VAL(self):
         module = self.import_extension('foo', [
@@ -185,12 +198,125 @@ class AppTestFloatMacros(AppTestCpythonExtensionBase):
         assert module.test() == float('inf')
 
     def test_Py_NAN(self):
+        import sys
         module = self.import_extension('foo', [
             ("test", "METH_NOARGS",
              """
                  return PyFloat_FromDouble(Py_NAN);
              """),
             ])
-        import struct
-        float_bits = struct.Struct('d').pack
-        assert float_bits(module.test()) == float_bits(float('nan'))
+        if sys.platform == 'win32':
+            # CPython does not enforce bit-compatibility between the NANs
+            import math
+            assert math.isnan(module.test())
+        else:
+            import struct
+            float_bits = struct.Struct('d').pack
+            assert float_bits(module.test()) == float_bits(float('nan'))
+
+    def test_roundtrip(self):
+        module = self.import_extension('foo', [
+            ("float_pack", "METH_VARARGS",
+             """
+                int size;
+                double d;
+                int le;
+                if (!PyArg_ParseTuple(args, "idi", &size, &d, &le)) {
+                    return NULL;
+                }
+                switch (size)
+                {
+                case 2:
+                {
+                    char data[2];
+                    if (PyFloat_Pack2(d, data, le) < 0) {
+                        return NULL;
+                    }
+                    return PyBytes_FromStringAndSize(data, Py_ARRAY_LENGTH(data));
+                }
+                case 4:
+                {
+                    char data[4];
+                    if (PyFloat_Pack4(d, data, le) < 0) {
+                        return NULL;
+                    }
+                    return PyBytes_FromStringAndSize(data, Py_ARRAY_LENGTH(data));
+                }
+                case 8:
+                {
+                    char data[8];
+                    if (PyFloat_Pack8(d, data, le) < 0) {
+                        return NULL;
+                    }
+                    return PyBytes_FromStringAndSize(data, Py_ARRAY_LENGTH(data));
+                }
+                default: break;
+                }
+
+                PyErr_SetString(PyExc_ValueError, "size must 2, 4 or 8");
+                return NULL;
+             """),
+            ("float_unpack", "METH_VARARGS",
+             """
+                assert(!PyErr_Occurred());
+                const char *data;
+                Py_ssize_t size;
+                int le;
+                if (!PyArg_ParseTuple(args, "y#i", &data, &size, &le)) {
+                    return NULL;
+                }
+                double d;
+                switch (size)
+                {
+                case 2:
+                    d = PyFloat_Unpack2(data, le);
+                    break;
+                case 4:
+                    d = PyFloat_Unpack4(data, le);
+                    break;
+                case 8:
+                    d = PyFloat_Unpack8(data, le);
+                    break;
+                default:
+                    PyErr_SetString(PyExc_ValueError, "data length must 2, 4 or 8 bytes");
+                    return NULL;
+                }
+
+                if (d == -1.0 && PyErr_Occurred()) {
+                    return NULL;
+                }
+                return PyFloat_FromDouble(d);
+             """),
+        ], PY_SSIZE_T_CLEAN=1)
+        import math
+        HAVE_IEEE_754 = float.__getformat__("double").startswith("IEEE")
+        INF = float("inf")
+        NAN = float("nan")
+        BIG_ENDIAN = 0
+        LITTLE_ENDIAN = 1
+        EPSILON = {
+            2: 2.0 ** -11,  # binary16
+            4: 2.0 ** -24,  # binary32
+            8: 2.0 ** -53,  # binary64
+}
+        large = 2.0 ** 100
+        values = [1.0, 1.5, large, 1.0/7, math.pi]
+        if HAVE_IEEE_754:
+            values.extend((INF, NAN))
+        for value in values:
+            for size in (2, 4, 8,):
+                if size == 2 and value == large:
+                    # too large for 16-bit float
+                    continue
+                rel_tol = EPSILON[size]
+                for endian in (BIG_ENDIAN, LITTLE_ENDIAN):
+                    data = module.float_pack(size, value, endian)
+                    value2 = module.float_unpack(data, endian)
+                    if math.isnan(value):
+                        assert math.isnan(value2), (value, value2)
+                    elif size < 8:
+                        assert math.isclose(value2, value, rel_tol=rel_tol), (value, value2)
+                    else:
+                        assert value2 == value
+
+
