@@ -1,5 +1,8 @@
 # -*- coding: utf-8 -*-
 
+from pypy.interpreter.gateway import interp2app_temp
+from pypy.interpreter.error import OperationError
+
 
 class AppTestExc(object):
     def setup_class(cls):
@@ -121,8 +124,19 @@ class AppTestExc(object):
         s = SyntaxError("a", (1, 2, 3, 4))
         assert s.msg == "a"
         assert s.filename == 1
+        assert s.offset == 3
+        assert s.text == 4
         assert str(SyntaxError("msg", ("file.py", 2, 3, 4))) == "msg (file.py, line 2)"
 
+        s = SyntaxError("a", (1, 2, 3, 4, 5, 6))
+        assert s.msg == "a"
+        assert s.filename == 1
+        assert s.lineno == 2
+        assert s.offset == 3
+        assert s.text == 4
+        assert s.end_lineno == 5
+        assert s.end_offset == 6
+
     def test_system_exit(self):
         assert issubclass(SystemExit, BaseException)
         assert SystemExit().code is None
@@ -205,7 +219,8 @@ class AppTestExc(object):
         import builtins
         for name, e in builtins.__dict__.items():
             if isinstance(e, type) and issubclass(e, BaseException):
-                assert e.__doc__, e
+                if e not in (ExceptionGroup,):
+                    assert e.__doc__, e
                 assert e.__module__ == 'builtins', e
         assert 'run-time' in RuntimeError.__doc__
 
@@ -307,7 +322,9 @@ class AppTestExc(object):
         assert ImportError("message").path is None
         assert ImportError("message", name="x").name == "x"
         assert ImportError("message", path="y").path == "y"
-        raises(TypeError, ImportError, invalid="z")
+        with raises(TypeError) as e:
+            ImportError(invalid="z")
+        assert "__init__() got an unexpected keyword argument 'invalid'" in str(e.value)
         assert ImportError("message").msg == "message"
         assert ImportError("message").args == ("message", )
         assert ImportError("message", "foo").msg is None
@@ -339,6 +356,9 @@ class AppTestExc(object):
         assert e.characters_written == 3
         e.characters_written = 5
         assert e.characters_written == 5
+        del e.characters_written
+        with raises(AttributeError):
+            e.characters_written
 
     def test_errno_mapping(self):
         # The OSError constructor maps errnos to subclasses
@@ -458,8 +478,11 @@ class AppTestExc(object):
 
                 source = source.replace("foo", "(foo.)")
                 exc = raises(SyntaxError, exec_, source)
+                print('source "%s"' % source)
+                print('custom_msg "%s"' % custom_msg)
+                print('exc.value.msg "%s"' % exc.value.msg)
                 assert (custom_msg not in exc.value.msg) == (
-                    ('print (' in source or 'exec (' in source))
+                    ('print ' in source or 'exec ' in source))
 
     def test_bug_print_heuristic_shadows_better_message(self):
         def exec_(s): exec(s)
@@ -487,7 +510,7 @@ class AppTestExc(object):
             "Missing parentheses in call to 'print'. Did you mean print(p)?")
         check("print %", "invalid syntax")
         check("print 1 1",
-            "Missing parentheses in call to 'print'")
+            "Missing parentheses in call to 'print'. Did you mean print(...)?")
 
     def test_print_and_operators(self):
         with raises(TypeError) as excinfo:
@@ -497,24 +520,55 @@ class AppTestExc(object):
             print -1
         assert 'Did you mean "print(<-number>)"?' in str(excinfo.value)
 
-    def test_importerror_kwarg_error(self):
-        msg = "'invalid' is an invalid keyword argument for this function"
-        exc = raises(TypeError,
-                     ImportError,
-                     'test', invalid='keyword', another=True)
-        assert str(exc.value) == msg
 
-        exc = raises(TypeError, ImportError, 'test', invalid='keyword')
-        assert str(exc.value) == msg
+    def test_attribute_error_name_obj_attributes(self):
+        exc = AttributeError("'a' not found", name="a", obj=7)
+        assert exc.name == "a"
+        assert exc.obj == 7
 
-        exc = raises(TypeError,
-                     ImportError,
-                     'test', name='name', invalid='keyword')
-        assert str(exc.value) == msg
+    def test_attribute_errorr_name_obj_attributes_are_filled(self):
+        class A:
+            pass
+        a = A()
+        with raises(AttributeError) as info:
+            a.blub
+        exc = info.value
+        assert exc.name == "blub"
+        assert exc.obj is a
 
-        exc = raises(TypeError,
-                     ImportError,
-                     'test', path='path', invalid='keyword')
-        assert str(exc.value) == msg
+    def test_name_error_name_attribute(self):
+        exc = NameError("'a' not found", name="a")
+        assert exc.name == "a"
 
+    def test_name_error_name_attributes_are_filled(self):
+        class A:
+            pass
+        a = A()
+        with raises(NameError) as info:
+            blub
+        exc = info.value
+        assert exc.name == "blub"
 
+    def test_multiple_inheritance_bug(self):
+        class OptionError(AttributeError, KeyError):
+            pass # does not crash
+        assert issubclass(OptionError, Exception)
+        assert issubclass(OptionError, AttributeError)
+        assert issubclass(OptionError, KeyError)
+        assert issubclass(OptionError, LookupError)
+        OptionError() # check that instantiation works
+
+    def test_multiple_inheritance_bug2(self):
+        class E(EnvironmentError, KeyError): pass
+        assert KeyError.__str__(E(6)) == '6'
+        # harder
+        class E2(KeyError, EnvironmentError): pass
+        assert str(E2(6)) == '6'
+
+    def test_keyerror_subclass(self):
+        class E(KeyError):
+            pass
+        assert str(E(1)) == '1'
+
+    def test_encodingwarning(self):
+        assert isinstance(EncodingWarning('abc'), Warning)
