@@ -17,6 +17,7 @@ from rpython.rlib.objectmodel import we_are_translated, specialize
 from rpython.rlib.signature import signature
 from rpython.rlib import rposix_stat, types
 from pypy.module.sys.version import PYPY_VERSION, CPYTHON_VERSION
+from pypy.module.__pypy__.interp_os import _multiarch
 
 _WIN32 = sys.platform == 'win32'
 
@@ -43,17 +44,12 @@ def get_so_extension(space):
 
     platform_name = sys.platform
     if platform_name.startswith('linux'):
-        if re.match('(i[3-6]86|x86_64)$', platform.machine()):
-            if sys.maxsize < 2**32:
-                platform_name = 'i686-linux-gnu'
-                # xxx should detect if we are inside 'x32', but not for now
-                # because it's not supported anyway by PyPy.  (Relying
-                # on platform.machine() does not work, it may return x86_64
-                # anyway)
-            else:
-                platform_name = 'x86_64-linux-gnu'
-        else:
-            platform_name = 'linux-gnu'
+        platform_name = _multiarch
+    elif platform_name == 'win32' and sys.maxsize > 2**32:
+        platform_name = 'win_amd64'
+    else:
+        # darwin?
+        pass
 
     soabi += '-' + platform_name
 
@@ -94,6 +90,8 @@ def _readall(space, filename):
 @unwrap_spec(modulename='fsencode', level=int)
 def importhook(space, modulename, w_globals=None, w_locals=None, w_fromlist=None, level=0):
     # A minimal version, that can only import builtin and lib_pypy modules!
+    # The actual __import__ is
+    # pypy.module._frozenimportlib.interp_import.interp___import__
     assert w_locals is w_globals
     assert level == 0
 
@@ -110,7 +108,10 @@ def importhook(space, modulename, w_globals=None, w_locals=None, w_fromlist=None
         ec = space.getexecutioncontext()
         source = _readall(space, os.path.join(lib_pypy, modulename + '.py'))
         pathname = "<frozen %s>" % modulename
-        code_w = ec.compiler.compile(source, pathname, 'exec', 0)
+        # *must* pass optimize here, otherwise can get strange bootstrapping
+        # problems, because compile would try to get the sys.flags, which might
+        # not be there yet
+        code_w = ec.compiler.compile(source, pathname, 'exec', 0, optimize=0)
         w_mod = add_module(space, space.newtext(modulename))
         assert isinstance(w_mod, Module) # XXX why is that necessary?
         space.setitem(space.sys.get('modules'), w_mod.w_name, w_mod)
@@ -425,3 +426,46 @@ def load_compiled_module(space, w_modulename, w_mod, cpathname, magic,
     exec_code_module(space, w_mod, code_w, cpathname, cpathname, write_paths)
 
     return w_mod
+
+def get_spec(space, w_module):
+    try:
+        return space.getattr(w_module, space.newtext('__spec__'))
+    except OperationError as e:
+        if not e.match(space, space.w_AttributeError):
+            raise
+        return space.w_None
+
+def is_spec_initializing(space, w_spec):
+    if space.is_none(w_spec):
+        return False
+
+    try:
+        w_initializing = space.getattr(w_spec, space.newtext("_initializing"))
+    except OperationError as e:
+        if not e.match(space, space.w_AttributeError):
+            raise
+
+        return False
+    else:
+        return space.is_true(w_initializing)
+
+def is_spec_uninitialized_submodule(space, w_spec, w_name):
+    # Check if the submodule name is in the "_uninitialized_submodules" attribute
+    # of the module spec.
+    if space.is_none(w_spec):
+        return False
+    w_value = space.getattr(w_spec, space.newtext("_uninitialized_submodules"))
+    return space.contains_w(w_value, w_name)
+
+def get_path(space, w_module):
+    default = space.newtext("unknown location")
+    try:
+        w_ret = space.getattr(w_module, space.newtext('__file__'))
+    except OperationError as e:
+        if not e.match(space, space.w_AttributeError):
+            raise
+        return default
+    if w_ret is space.w_None:
+        return default
+    return w_ret
+
