@@ -2,9 +2,27 @@ from pypy.interpreter.baseobjspace import W_Root
 from pypy.interpreter.error import OperationError, oefmt
 from pypy.interpreter.typedef import TypeDef, make_weakref_descr
 from pypy.interpreter.gateway import interp2app, unwrap_spec, WrappedDefault
+from pypy.objspace.std.util import generic_alias_class_getitem, builtinclass_new_args_check
 from rpython.rlib import jit
+from rpython.rtyper.lltypesystem import rffi
+
+from pypy.module.__builtin__.functional import W_Filter, build_iterators_from_args
+
+def W_Twoarg__new__(space, w_subtype, W_Base, name, __args__):
+    args_w = __args__.arguments_w
+    w_type = space.gettypeobject(W_Base.typedef)
+    w_init = space.newtext("__init__")
+    if (space.is_w(w_subtype, w_type) or 
+        space.is_w(space.getattr(w_subtype, w_init), space.getattr(w_type, w_init))):
+        if __args__.keyword_names_w:
+            raise oefmt(space.w_TypeError,
+                    "%s() takes no keyword arguments", name)
+    length = len(args_w) if args_w else 0
+    if length != 2:
+        raise oefmt(space.w_TypeError,
+                    "%s() expected 2 arguments, got %d", name, length)
+    return args_w
 
-from pypy.module.__builtin__.functional import W_Filter, W_Map
 
 class W_Count(W_Root):
     def __init__(self, space, w_firstval, w_step):
@@ -42,12 +60,14 @@ class W_Count(W_Root):
             args_w = [self.w_c]
         else:
             args_w = [self.w_c, self.w_step]
-        return space.newtuple([space.gettypefor(W_Count),
-                               space.newtuple(args_w)])
+        return space.newtuple2(space.gettypefor(W_Count),
+                               space.newtuple(args_w))
 
 def check_number(space, w_obj):
     if (space.lookup(w_obj, '__int__') is None and
-        space.lookup(w_obj, '__float__') is None):
+        space.lookup(w_obj, '__float__') is None and
+        not space.isinstance_w(w_obj, space.w_complex)
+    ):
         raise oefmt(space.w_TypeError, "expected a number")
 
 @unwrap_spec(w_start=WrappedDefault(0), w_step=WrappedDefault(1))
@@ -188,12 +208,12 @@ class W_TakeWhile(W_Root):
     def descr_setstate(self, space, w_state):
         self.stopped = space.bool_w(w_state)
 
-def W_TakeWhile___new__(space, w_subtype, w_predicate, w_iterable):
+def W_TakeWhile___new__(space, w_subtype, __args__):
+    args_w = W_Twoarg__new__(space, w_subtype, W_TakeWhile, "takewhile", __args__)
     r = space.allocate_instance(W_TakeWhile, w_subtype)
-    r.__init__(space, w_predicate, w_iterable)
+    r.__init__(space, args_w[0], args_w[1])
     return r
 
-
 W_TakeWhile.typedef = TypeDef(
         'itertools.takewhile',
         __new__  = interp2app(W_TakeWhile___new__),
@@ -248,9 +268,10 @@ class W_DropWhile(W_Root):
     def descr_setstate(self, space, w_state):
         self.started = space.bool_w(w_state)
 
-def W_DropWhile___new__(space, w_subtype, w_predicate, w_iterable):
+def W_DropWhile___new__(space, w_subtype, __args__):
+    args_w = W_Twoarg__new__(space, w_subtype, W_DropWhile, "dropwhile", __args__)
     r = space.allocate_instance(W_DropWhile, w_subtype)
-    r.__init__(space, w_predicate, w_iterable)
+    r.__init__(space, args_w[0], args_w[1])
     return r
 
 
@@ -281,13 +302,14 @@ W_DropWhile.typedef = TypeDef(
 class W_FilterFalse(W_Filter):
     reverse = True
     def descr_reduce(self, space):
-        args_w = [space.w_None if self.no_predicate else self.w_predicate,
-                  self.iterable]
+        args_w = [space.w_None if self.w_predicate is None else self.w_predicate,
+                  self.w_iterable]
         return space.newtuple([space.type(self), space.newtuple(args_w)])
 
-def W_FilterFalse___new__(space, w_subtype, w_predicate, w_iterable):
+def W_FilterFalse___new__(space, w_subtype, __args__):
+    args_w = W_Twoarg__new__(space, w_subtype, W_FilterFalse, 'filterfalse', __args__)
     r = space.allocate_instance(W_FilterFalse, w_subtype)
-    r.__init__(space, w_predicate, w_iterable)
+    r.__init__(space, args_w[0], args_w[1])
     return r
 
 W_FilterFalse.typedef = TypeDef(
@@ -311,9 +333,12 @@ W_FilterFalse.typedef = TypeDef(
     """)
 
 
+def get_printable_location(greenkey):
+    return "islice_ignore_items [%s]" % (greenkey.iterator_greenkey_printable(), )
 islice_ignore_items_driver = jit.JitDriver(name='islice_ignore_items',
-                                           greens=['tp'],
-                                           reds=['w_islice', 'w_iterator'])
+                                           greens=['greenkey'],
+                                           reds=['w_islice', 'w_iterator'],
+                                           get_printable_location=get_printable_location)
 
 class W_ISlice(W_Root):
     def __init__(self, space, w_iterable, w_startstop, args_w):
@@ -354,10 +379,10 @@ class W_ISlice(W_Root):
         else:
             step = 1
 
-        self.count = 0
-        self.next = start
+        self.count = rffi.cast(rffi.UNSIGNED, 0)
+        self.next = rffi.cast(rffi.UNSIGNED, start)
         self.stop = stop
-        self.step = step
+        self.step = rffi.cast(rffi.UNSIGNED, step)
 
     def arg_int_w(self, w_obj, minimum, errormsg):
         space = self.space
@@ -377,7 +402,8 @@ class W_ISlice(W_Root):
     def next_w(self):
         if self.iterable is None:
             raise OperationError(self.space.w_StopIteration, self.space.w_None)
-        self._ignore_items()
+        if self.count != self.next:
+            self._ignore_items()
         stop = self.stop
         if 0 <= stop <= self.count:
             self.iterable = None
@@ -391,17 +417,19 @@ class W_ISlice(W_Root):
             raise
         self.count += 1
         oldnext = self.next
-        self.next += self.step
-        if self.next < oldnext or self.next > stop >= 0:
-            self.next = stop
+        new_next = self.next + self.step
+        if new_next < oldnext or (stop >=0 and new_next > rffi.cast(rffi.UNSIGNED,stop)):
+            self.next = rffi.cast(rffi.UNSIGNED, stop)
+        else:
+            self.next = new_next
         return item
 
     def _ignore_items(self):
         w_iterator = self.iterable
-        tp = self.space.type(w_iterator)
+        greenkey = self.space.iterator_greenkey(w_iterator)
         while True:
             islice_ignore_items_driver.jit_merge_point(
-                tp=tp, w_islice=self, w_iterator=w_iterator)
+                greenkey=greenkey, w_islice=self, w_iterator=w_iterator)
             if self.count >= self.next:
                 break
             try:
@@ -435,9 +463,17 @@ class W_ISlice(W_Root):
         ])
 
     def descr_setstate(self, space, w_state):
-        self.count = space.int_w(w_state)
+        self.count = rffi.cast(rffi.UNSIGNED, space.int_w(w_state))
 
-def W_ISlice___new__(space, w_subtype, w_iterable, w_startstop, args_w):
+def W_ISlice___new__(space, w_subtype, w_iterable, w_startstop, __args__):
+    args_w = __args__.arguments_w
+    w_islice = space.gettypeobject(W_ISlice.typedef)
+    w_init = space.newtext("__init__")
+    if (space.is_w(w_subtype, w_islice) or 
+        space.is_w(space.getattr(w_subtype, w_init), space.getattr(w_islice, w_init))):
+        if __args__.keyword_names_w:
+            raise oefmt(space.w_TypeError,
+                    "islice() takes no keyword arguments")
     r = space.allocate_instance(W_ISlice, w_subtype)
     r.__init__(space, w_iterable, w_startstop, args_w)
     return r
@@ -532,7 +568,15 @@ class W_Chain(W_Root):
                         "function takes at most 2 arguments (%d given)",
                         num_args)
 
-def W_Chain___new__(space, w_subtype, args_w):
+def W_Chain___new__(space, w_subtype, __args__):
+    args_w = __args__.arguments_w
+    w_chain = space.gettypeobject(W_Chain.typedef)
+    w_init = space.newtext("__init__")
+    if (space.is_w(w_subtype, w_chain) or 
+        space.is_w(space.getattr(w_subtype, w_init), space.getattr(w_chain, w_init))):
+        if __args__.keyword_names_w:
+            raise oefmt(space.w_TypeError,
+                    "chain() takes no keyword arguments")
     r = space.allocate_instance(W_Chain, w_subtype)
     w_args = space.newtuple(args_w)
     r.__init__(space, space.iter(w_args))
@@ -548,14 +592,16 @@ def chain_from_iterable(space, w_cls, w_arg):
     return r
 
 W_Chain.typedef = TypeDef(
-        'itertools.chain',
-        __new__  = interp2app(W_Chain___new__),
-        __iter__ = interp2app(W_Chain.iter_w),
-        __next__ = interp2app(W_Chain.next_w),
-        __reduce__ = interp2app(W_Chain.descr_reduce),
-        __setstate__ = interp2app(W_Chain.descr_setstate),
-        from_iterable = interp2app(chain_from_iterable, as_classmethod=True),
-        __doc__  = """Make an iterator that returns elements from the first iterable
+    'itertools.chain',
+    __new__  = interp2app(W_Chain___new__),
+    __iter__ = interp2app(W_Chain.iter_w),
+    __next__ = interp2app(W_Chain.next_w),
+    __reduce__ = interp2app(W_Chain.descr_reduce),
+    __setstate__ = interp2app(W_Chain.descr_setstate),
+    from_iterable = interp2app(chain_from_iterable, as_classmethod=True),
+    __class_getitem__ = interp2app(
+        generic_alias_class_getitem, as_classmethod=True),
+    __doc__  = """Make an iterator that returns elements from the first iterable
     until it is exhausted, then proceeds to the next iterable, until
     all of the iterables are exhausted. Used for treating consecutive
     sequences as a single sequence.
@@ -566,12 +612,19 @@ W_Chain.typedef = TypeDef(
         for it in iterables:
             for element in it:
                 yield element
-    """)
+""")
 
 
-class W_ZipLongest(W_Map):
-    _error_name = "zip_longest"
-    _immutable_fields_ = ["w_fillvalue"]
+class W_ZipLongest(W_Root):
+    _immutable_fields_ = ["w_fillvalue", "iterators"]
+
+    def __init__(self, space, w_fun, args_w):
+        self.space = space
+        self.w_fun = w_fun
+        self.iterators_w = build_iterators_from_args(space, args_w)
+
+    def iter_w(self):
+        return self
 
     def _fetch(self, index):
         w_iter = self.iterators_w[index]
@@ -592,10 +645,10 @@ class W_ZipLongest(W_Map):
     def next_w(self):
         # common case: 2 arguments
         if len(self.iterators_w) == 2:
-            objects = [self._fetch(0), self._fetch(1)]
+            return self.space.newtuple2(self._fetch(0), self._fetch(1))
         else:
             objects = self._get_objects()
-        return self.space.newtuple(objects)
+            return self.space.newtuple(objects)
 
     def _get_objects(self):
         # the loop is out of the way of the JIT
@@ -621,6 +674,14 @@ class W_ZipLongest(W_Map):
     def descr_setstate(self, space, w_state):
         self.w_fillvalue = w_state
 
+    def iterator_greenkey(self, space):
+        # XXX in theory we should tupleize the greenkeys of all the
+        # sub-iterators, but much more work
+        if len(self.iterators_w) > 0:
+            return space.iterator_greenkey(self.iterators_w[0])
+        return None
+
+
 def W_ZipLongest___new__(space, w_subtype, __args__):
     arguments_w, kwds_w = __args__.unpack()
     w_fillvalue = space.w_None
@@ -701,18 +762,23 @@ class W_Cycle(W_Root):
             space.newtuple([
                 space.newlist(self.saved_w),
                 space.newint(self.index),
-                space.newbool(self.index > 0),
+                # space.newbool(self.index > 0),
             ]),
         ])
 
     def descr_setstate(self, space, w_state):
-        w_saved, w_index, w_exhausted = space.unpackiterable(w_state, 3)
+        state_w = space.unpackiterable(w_state, 2)
+        w_saved = state_w[0]
         self.saved_w = space.unpackiterable(w_saved)
+        w_index = state_w[1]
         self.index = space.int_w(w_index)
         # w_exhausted ignored
+        # w_exhausted = state_w[2]
 
 
-def W_Cycle___new__(space, w_subtype, w_iterable):
+def W_Cycle___new__(space, w_subtype, w_iterable, __posonly__, __args__):
+    w_type = space.gettypeobject(W_Cycle.typedef)
+    builtinclass_new_args_check(space, "cycle", w_type, w_subtype, __args__)
     r = space.allocate_instance(W_Cycle, w_subtype)
     r.__init__(space, w_iterable)
     return r
@@ -761,9 +827,10 @@ class W_StarMap(W_Root):
                                         self.w_iterable])
                                     ])
 
-def W_StarMap___new__(space, w_subtype, w_fun, w_iterable):
+def W_StarMap___new__(space, w_subtype, __args__):
+    args_w = W_Twoarg__new__(space, w_subtype, W_StarMap, "starmap", __args__)
     r = space.allocate_instance(W_StarMap, w_subtype)
-    r.__init__(space, w_fun, w_iterable)
+    r.__init__(space, args_w[0], args_w[1])
     return r
 
 W_StarMap.typedef = TypeDef(
@@ -821,17 +888,16 @@ def tee(space, w_iterable, n=2):
     if n < 0:
         raise oefmt(space.w_ValueError, "n must be >= 0")
 
-    if space.findattr(w_iterable, space.newtext("__copy__")) is not None:
+    w_iterator = space.iter(w_iterable)
+    if space.findattr(w_iterator, space.newtext("__copy__")) is not None:
         # In this case, we don't instantiate any W_TeeIterable.
         # We just rely on doing repeated __copy__().  This case
         # includes the situation where w_iterable is already
         # a W_TeeIterable itself.
-        iterators_w = [w_iterable] * n
+        iterators_w = [w_iterator] * n
         for i in range(1, n):
-            w_iterable = space.call_method(w_iterable, "__copy__")
-            iterators_w[i] = w_iterable
+            iterators_w[i] = space.call_method(w_iterator, "__copy__")
     else:
-        w_iterator = space.iter(w_iterable)
         w_chained_list = W_TeeChainedListNode(space)
         iterators_w = [W_TeeIterable(space, w_iterator, w_chained_list)
                        for x in range(n)]
@@ -984,7 +1050,7 @@ class W_GroupBy(W_Root):
         self._skip_to_next_iteration_group()
         w_key = self.w_tgtkey = self.w_currkey
         w_grouper = W_GroupByIterator(self, w_key)
-        return self.space.newtuple([w_key, w_grouper])
+        return self.space.newtuple2(w_key, w_grouper)
 
     def _skip_to_next_iteration_group(self):
         space = self.space
@@ -1641,31 +1707,46 @@ permutations(range(3), 2) --> (0,1), (0,2), (1,0), (1,2), (2,0), (2,1)""",
 
 
 class W_Accumulate(W_Root):
-    def __init__(self, space, w_iterable, w_func=None):
+    'Return series of accumulated sums (or other binary function results).'
+
+    def __init__(self, space, w_iterable, w_func, w_initial):
         self.space = space
         self.w_iterable = w_iterable
         self.w_func = w_func if not space.is_w(w_func, space.w_None) else None
         self.w_total = None
+        self.w_initial = w_initial
 
     def iter_w(self):
         return self
 
     def next_w(self):
-        w_value = self.space.next(self.w_iterable)
+        space = self.space
+        if not space.is_w(self.w_initial, space.w_None):
+            w_res = self.w_total = self.w_initial
+            self.w_initial = space.w_None
+            return w_res
+        w_value = space.next(self.w_iterable)
         if self.w_total is None:
             self.w_total = w_value
             return w_value
 
         if self.w_func is None:
-            self.w_total = self.space.add(self.w_total, w_value)
+            self.w_total = space.add(self.w_total, w_value)
         else:
-            self.w_total = self.space.call_function(self.w_func, self.w_total, w_value)
+            self.w_total = space.call_function(self.w_func, self.w_total, w_value)
         return self.w_total
 
     def reduce_w(self):
         space = self.space
         w_func = space.w_None if self.w_func is None else self.w_func
-        if self.w_total is space.w_None:      # :-(
+        if not space.is_w(self.w_initial, space.w_None):
+            w_it = W_Chain(space, space.iter(space.newlist([
+                space.newtuple([self.w_initial]),
+                self.w_iterable])))
+            return space.newtuple([space.gettypefor(W_Accumulate),
+                space.newtuple([w_it, w_func]),
+                space.w_None])
+        if self.w_total is space.w_None: # :-(
             w_it = W_Chain(space, space.iter(space.newlist([
                                      space.newtuple([self.w_total]),
                                      self.w_iterable])))
@@ -1681,9 +1762,10 @@ class W_Accumulate(W_Root):
     def setstate_w(self, space, w_state):
         self.w_total = w_state if not space.is_w(w_state, space.w_None) else None
 
-def W_Accumulate__new__(space, w_subtype, w_iterable, w_func=None):
+@unwrap_spec(w_initial=WrappedDefault(None))
+def W_Accumulate__new__(space, w_subtype, w_iterable, w_func=None, __kwonly__=None, w_initial=None):
     r = space.allocate_instance(W_Accumulate, w_subtype)
-    r.__init__(space, space.iter(w_iterable), w_func)
+    r.__init__(space, space.iter(w_iterable), w_func, w_initial)
     return r
 
 W_Accumulate.typedef = TypeDef("itertools.accumulate",
@@ -1696,3 +1778,41 @@ W_Accumulate.typedef = TypeDef("itertools.accumulate",
 "accumulate(iterable) --> accumulate object
 
 Return series of accumulated sums.""")
+
+
+class W_Pairwise(W_Root):
+    'Return series of accumulated sums (or other binary function results).'
+
+    def __init__(self, space, w_iterator):
+        self.space = space
+        w_prev = None
+        self.w_iterator = w_iterator
+        self.w_prev = None
+
+    def iter_w(self):
+        return self
+
+    def next_w(self):
+        space = self.space
+        w_prev = self.w_prev
+        if w_prev is None:
+            w_prev = space.next(self.w_iterator)
+        w_next = space.next(self.w_iterator)
+        w_res = space.newtuple2(w_prev, w_next)
+        self.w_prev = w_next
+        return w_res
+
+def W_Pairwise__new__(space, w_subtype, w_iterable, __posonly__=None):
+    r = space.allocate_instance(W_Pairwise, w_subtype)
+    r.__init__(space, space.iter(w_iterable))
+    return r
+
+W_Pairwise.typedef = TypeDef("itertools.pairwise",
+    __new__  = interp2app(W_Pairwise__new__),
+    __iter__ = interp2app(W_Pairwise.iter_w),
+    __next__ = interp2app(W_Pairwise.next_w),
+    __doc__  = """\
+Return an iterator of overlapping pairs taken from the input iterator.
+
+    s -> (s0, s1), (s1, s2), (s2, s3), ...
+""")
