@@ -115,10 +115,13 @@ class FakeSpace(ObjSpace):
         return isinstance(w_obj, ListObject) or isinstance(w_obj, W_NDimArray)
 
     def len(self, w_obj):
+        return self.wrap(self.len_w(w_obj))
+
+    def len_w(self, w_obj):
         if isinstance(w_obj, ListObject):
-            return self.wrap(len(w_obj.items))
+            return len(w_obj.items)
         elif isinstance(w_obj, DictObject):
-            return self.wrap(len(w_obj.items))
+            return len(w_obj.items)
         raise NotImplementedError
 
     def getattr(self, w_obj, w_attr):
@@ -153,7 +156,7 @@ class FakeSpace(ObjSpace):
 
         raise NotImplementedError
 
-    def decode_index4(self, w_idx, size):
+    def decode_index4_unsafe(self, w_idx, size):
         if isinstance(w_idx, IntObject):
             return (self.int_w(w_idx), 0, 0, 1)
         else:
@@ -410,6 +413,9 @@ class FakeSpace(ObjSpace):
     def newtuple(self, list_w):
         return ListObject(list_w)
 
+    def newtuple2(self, w_a, w_b):
+        return ListObject([w_a, w_b])
+
     def newdict(self, module=True):
         return DictObject({})
 
