@@ -27,10 +27,11 @@ class AppTestMMap:
 
     def test_attributes(self):
         import mmap
-        import os
+        import os, sys
         assert isinstance(mmap.ACCESS_READ, int)
         assert isinstance(mmap.ACCESS_WRITE, int)
         assert isinstance(mmap.ACCESS_COPY, int)
+        assert isinstance(mmap.ACCESS_DEFAULT, int)
         if os.name == "posix":
             assert isinstance(mmap.MAP_ANON, int)
             assert isinstance(mmap.MAP_ANONYMOUS, int)
@@ -39,6 +40,8 @@ class AppTestMMap:
             assert isinstance(mmap.PROT_EXEC, int)
             assert isinstance(mmap.PROT_READ, int)
             assert isinstance(mmap.PROT_WRITE, int)
+            if sys.platform.startswith('linux'):
+                assert isinstance(mmap.MAP_POPULATE, int)
 
         assert mmap.error is OSError
 
@@ -80,7 +83,8 @@ class AppTestMMap:
         f = open(self.tmpname + "a", "wb+")
 
         f.write(b"c")
-        f.flush()
+        ret = f.flush()
+        assert ret is None
         raises(ValueError, mmap, f.fileno(), 123)
         f.close()
 
@@ -104,6 +108,7 @@ class AppTestMMap:
         m = mmap(f.fileno(), 1)
         m.close()
         raises(ValueError, m.read, 1)
+        raises(ValueError, len, m)
 
     def test_read_byte(self):
         from mmap import mmap
@@ -120,19 +125,13 @@ class AppTestMMap:
     def test_readline(self):
         from mmap import mmap
         import os
-        f = open(self.tmpname + "e", "wb+")
-
-        f.write(b"foo\n")
-        f.flush()
-        m = mmap(f.fileno(), 4)
-        if os.name == "nt":
-            # windows replaces \n with \r. it's time to change to \n only MS!
-            assert m.readline() == b"foo\r"
-        elif os.name == "posix":
-            assert m.readline() == b"foo\n"
-        assert m.readline() == b""
-        m.close()
-        f.close()
+        with open(self.tmpname + "e", "wb+") as f:
+            f.write(b"foo\n")
+            f.flush()
+            with mmap(f.fileno(), 4) as m:
+                result = m.readline()
+                assert result == b"foo\n"
+                assert m.readline() == b""
 
     def test_read(self):
         from mmap import mmap
@@ -431,6 +430,15 @@ class AppTestMMap:
         m.close()
         f.close()
 
+    def test_get_crash(self):
+        import sys
+        from mmap import mmap
+        s = b'hallo!!!'
+        m = mmap(-1, len(s))
+        m[:] = s
+        assert m[1:None:sys.maxsize] == b'a'
+        m.close()
+
     def test_set_item(self):
         import mmap
 
@@ -536,20 +544,22 @@ class AppTestMMap:
         f.close()
 
     def test_memoryview(self):
-        from mmap import mmap, PROT_READ
+        from mmap import mmap
         filename = self.tmpname + "y"
-        f = open(filename, "bw+")
-        f.write(b"foobar")
-        f.flush()
-        m = mmap(f.fileno(), 6)
-        b = memoryview(m)
-        assert len(b) == 6
-        assert b.readonly is False
-        assert b[3] == ord(b"b")
-        assert b[:] == b"foobar"
-        del b  # For CPython: "exported pointers exist"
-        m.close()
-        f.close()
+        with open(filename, "bw+") as f:
+            f.write(b"foobar")
+            f.flush()
+            with mmap(f.fileno(), 6) as m:
+                b = memoryview(m)
+                assert len(b) == 6
+                assert b.readonly is False
+                assert b[3] == ord(b"b")
+                assert b[:] == b"foobar"
+                del b  # For CPython: "exported pointers exist"
+        try:
+            from mmap import PROT_READ
+        except ImportError:
+            skip('no PROT_READ') 
         with open(filename, "rb") as f:
             m = mmap(f.fileno(), 6, prot=PROT_READ)
             b = memoryview(m)
@@ -561,16 +571,15 @@ class AppTestMMap:
     def test_offset(self):
         from mmap import mmap, ALLOCATIONGRANULARITY
         filename = self.tmpname + "y"
-        f = open(filename, "wb+")
-        f.write(b"foobar" * ALLOCATIONGRANULARITY)
-        f.flush()
-        size = ALLOCATIONGRANULARITY
-        offset = 2 * ALLOCATIONGRANULARITY
-        m = mmap(f.fileno(), size, offset=offset)
-        assert m[:] == (b"foobar" * ALLOCATIONGRANULARITY)[offset:offset+size]
-        assert len(m) == size
-        m.close()
-        f.close()
+        with open(filename, "wb+") as f:
+            f.write(b"foobar" * ALLOCATIONGRANULARITY)
+            f.flush()
+            size = ALLOCATIONGRANULARITY
+            offset = 2 * ALLOCATIONGRANULARITY
+            m = mmap(f.fileno(), size, offset=offset)
+            assert m[:] == (b"foobar" * ALLOCATIONGRANULARITY)[offset:offset+size]
+            assert len(m) == size
+            m.close()
 
     def test_offset_more(self):
         from mmap import mmap, ALLOCATIONGRANULARITY
@@ -874,3 +883,60 @@ class AppTestMMap:
         raises(ValueError, m.write, b'abc')
         assert m.tell() == 5000
         m.close()
+
+    def test_close_while_indexing(self):
+        import mmap
+        with open(self.tmpname + "_crash", 'w+b') as f:
+            f.write(b"foobar")
+            f.flush()
+
+            class X(object):
+                def __index__(self):
+                    m.close()
+                    return 1
+
+            m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_READ)
+            assert m[1] == ord(b"o")
+            with raises(ValueError):
+                m[X()]
+
+            m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_READ)
+            with raises(ValueError):
+                m[X()] = b"u"
+
+    def test_iter_yields_bytes(self):
+        # issue 3282: inconsistency in Python 3
+        from mmap import mmap
+        f = open(self.tmpname + "iter", "wb+")
+        f.write(b"AB")
+        f.flush()
+
+        m = mmap(f.fileno(), 2)
+        assert [m[0], m[1]] == [65, 66]
+        assert list(m) == [b"A", b"B"]
+        assert list(iter(m)) == [b"A", b"B"]
+        assert list(reversed(m)) == [b"B", b"A"]
+        assert list(enumerate(m)) == [(0, b"A"), (1, b"B")]
+
+    def test_madvise(self):
+        import mmap, sys
+        m = mmap.mmap(-1, 1024)
+        if not hasattr(m, "madvise"):
+            m.close()
+            skip("no madvise")
+
+        m.madvise(mmap.MADV_NORMAL)
+        m.close()
+
+    def test_repr(self):
+        import mmap
+        m = mmap.mmap(-1, 1024)
+        assert repr(m) == "<mmap.mmap closed=False, access=ACCESS_DEFAULT, length=1024, pos=0, offset=0>"
+        m.close()
+        assert repr(m) == "<mmap.mmap closed=True>"
+
+    def test_assign_buffer(self):
+        import mmap
+        m = mmap.mmap(-1, 1024)
+        m[0:5] = bytearray(b'12345')
+        assert m[0:5] == b'12345'
