@@ -1,4 +1,4 @@
-import py
+import pytest
 from pypy.module.pypyjit.test_pypy_c.test_00_model import BaseTestPyPyC
 from pypy.module.pypyjit.test_pypy_c.model import OpMatcher
 
@@ -82,7 +82,6 @@ class TestCall(BaseTestPyPyC):
         assert entry_bridge.match_by_id('call', """
             dummy_get_utf8?
             p38 = call_r(ConstClass(_ll_1_threadlocalref_get__Ptr_GcStruct_objectLlT_Signed), #, descr=<Callr . i EF=1 OS=5>)
-            p99 = getfield_gc_r(p38, descr=<FieldP pypy.interpreter.executioncontext.ExecutionContext.inst_sys_exc_operror .*>)
             p39 = getfield_gc_r(p38, descr=<FieldP pypy.interpreter.executioncontext.ExecutionContext.inst_topframeref .*>)
             i40 = force_token()
             p41 = getfield_gc_r(p38, descr=<FieldP pypy.interpreter.executioncontext.ExecutionContext.inst_w_tracefunc .*>)
@@ -116,6 +115,8 @@ class TestCall(BaseTestPyPyC):
                     self.a = a
                 def f(self, i):
                     return self.a + i
+            a = A("a") # stop field unboxing
+
             i = 0
             a = A(1)
             while i < n:
@@ -129,12 +130,12 @@ class TestCall(BaseTestPyPyC):
         # first, we test the entry bridge
         # -------------------------------
         entry_bridge, = log.loops_by_filename(self.filepath, is_entry_bridge=True)
-        ops = entry_bridge.ops_by_id('meth1', opcode='LOOKUP_METHOD')
+        ops = entry_bridge.ops_by_id('meth1', opcode='LOAD_METHOD')
         assert log.opnames(ops) == ['guard_value', 'getfield_gc_r',
                                     'guard_value',
                                     'guard_not_invalidated']
-        # the second LOOKUP_METHOD is folded away
-        assert list(entry_bridge.ops_by_id('meth2', opcode='LOOKUP_METHOD')) == []
+        # the second LOAD_METHOD is folded away
+        assert list(entry_bridge.ops_by_id('meth2', opcode='LOAD_METHOD')) == []
         #
         # then, the actual loop
         # ----------------------
@@ -388,10 +389,9 @@ class TestCall(BaseTestPyPyC):
             {{{
             setfield_gc(p0, p20, descr=<FieldP .*PyFrame.vable_token .*>)
             setfield_gc(p22, ConstPtr(null), descr=<FieldP pypy.interpreter.argument.Arguments.inst_keywords_w .*>)
-            setfield_gc(p22, ConstPtr(null), descr=<FieldP pypy.interpreter.argument.Arguments.inst_keywords .*>)
+            setfield_gc(p22, ConstPtr(null), descr=<FieldP pypy.interpreter.argument.Arguments.inst_keyword_names_w .*>)
             setfield_gc(p22, 1, descr=<FieldU pypy.interpreter.argument.Arguments.inst__jit_few_keywords .*>)
             setfield_gc(p22, 0, descr=<FieldU pypy.interpreter.argument.Arguments.inst_methodcall .*>)
-            setfield_gc(p22, ConstPtr(null), descr=<FieldP pypy.interpreter.argument.Arguments.inst_keyword_names_w .*>)
             setfield_gc(p26, ConstPtr(ptr22), descr=<FieldP pypy.objspace.std.listobject.W_ListObject.inst_strategy .*>)
             setfield_gc(p26, ConstPtr(null), descr=<FieldP pypy.objspace.std.listobject.W_ListObject.inst_lstorage .*>)
             setarrayitem_gc(p24, 0, p26, descr=<ArrayP .>)
@@ -444,7 +444,6 @@ class TestCall(BaseTestPyPyC):
             dummy_get_utf8?
             guard_not_invalidated(descr=...)
             p29 = call_r(ConstClass(_ll_1_threadlocalref_get__Ptr_GcStruct_objectLlT_Signed), #, descr=<Callr . i EF=1 OS=5>)
-            p99 = getfield_gc_r(p29, descr=<FieldP pypy.interpreter.executioncontext.ExecutionContext.inst_sys_exc_operror .*>)
             p30 = getfield_gc_r(p29, descr=<FieldP pypy.interpreter.executioncontext.ExecutionContext.inst_topframeref .*>)
             p31 = force_token()
             p32 = getfield_gc_r(p29, descr=<FieldP pypy.interpreter.executioncontext.ExecutionContext.inst_w_tracefunc .*>)
@@ -452,11 +451,7 @@ class TestCall(BaseTestPyPyC):
             i34 = getfield_gc_i(p29, descr=<FieldU pypy.interpreter.executioncontext.ExecutionContext.inst_profilefunc .*>)
             i35 = int_is_zero(i34)
             guard_true(i35, descr=...)
-            p37 = getfield_gc_r(ConstPtr(ptr36), descr=<FieldP pypy.interpreter.nestedscope.Cell.inst_w_value .*>)
-            guard_nonnull_class(p37, ConstClass(W_IntObject), descr=...)
-            i39 = getfield_gc_i(p37, descr=<FieldS pypy.objspace.std.intobject.W_IntObject.inst_intval .*>)
-            i40 = int_add_ovf(i22, i39)
-            guard_no_overflow(descr=...)
+            i40 = int_add(i22, 5)
             --TICK--
         """)
 
@@ -560,6 +555,26 @@ class TestCall(BaseTestPyPyC):
         assert len(calls) == 0
         assert len([op for op in allops if op.name.startswith('new')]) == 0
 
+    def test_kwargs_update_virtual1(self):
+        log = self.run("""
+        def f(**kwargs):
+            return len(kwargs)
+        def main(stop):
+            i = 0
+            res = 0
+            while i < stop:
+                d = {'a': 1, 'b': 2}
+                # used to force the dict!
+                res += f(c=2, **d) # ID: call
+                i += 1
+            return res
+        """, [1000])
+        loop, = log.loops_by_id('call')
+        ops = loop.ops_by_id('call')
+        assert log.opnames(ops) == ["guard_not_invalidated", "force_token",
+                "int_add_ovf", "guard_no_overflow"]
+
+
     def test_kwargs_non_virtual(self):
         log = self.run("""
         def f(a, b, c):
@@ -579,7 +594,7 @@ class TestCall(BaseTestPyPyC):
         calls = [op for op in allops if op.name.startswith('call')]
         assert OpMatcher(calls).match('''
         p93 = call_r(ConstClass(view_as_kwargs), p35, p12, descr=<.*>)
-        i103 = call_i(ConstClass(_match_keywords), ConstPtr(ptr52), 0, 0, p94, p98, 0, descr=<.*>)
+        i103 = call_i(ConstClass(_match_keywords), ConstPtr(ptr52), 0, 0, 0, p94, p98, 0, descr=<.*>)
         ''')
         assert len([op for op in allops if op.name.startswith('new')]) == 1
         # 1 alloc
@@ -634,3 +649,80 @@ class TestCall(BaseTestPyPyC):
                 i += 1
             return 13
         """, [1000])
+
+    def test_nonstd_jitdriver_distinguishes_map(self):
+        log = self.run("""
+        def f(a):
+            return a + 1
+        def g(a):
+            return a + 2
+        def main():
+            # test the "contains" jitdriver, but the others are the same
+            res = (9999 in map(f, range(100000)))
+            res += (9999 in map(g, range(200000)))
+            return res
+        """, [])
+        assert len([l for l in log.loops if len(l.chunks) > 1 and l.chunks[1].bytecode_name.startswith("DescrOperation.contains")]) == 2
+
+    def test_methodcall_kwargs_regression(self):
+        log = self.run("""
+        class A:
+            def f(self, x, y, z):
+                return x + y + z
+        def main():
+            a = A()
+            res = 0
+            for i in range(10000):
+                a.f(x=i, y=i+1, z=i*2) # ID: meth
+                res += i
+        """, [])
+        
+        loop, = log.loops_by_id('meth')
+
+        assert loop.match_by_id("meth", """
+            setfield_gc(p15, i65, descr=...)
+            guard_not_invalidated(descr=...)
+            i68 = int_mul_ovf(i62, 2)
+            guard_no_overflow(descr=...)
+            p69 = force_token()
+        """)
+
+    @pytest.mark.skip("this kind of call works really different on python3")
+    def test_starargs_list_copying(self):
+        log = self.run("""
+        def f(a, *args):
+            return 1 + len(args)
+        def main(stop):
+            i = 0
+            l = [1, 2, "abc", None]
+            while i < stop:
+                f(1, *l) # ID: call
+                i += 1
+            return 13
+        """, [1000])
+        loop, = log.loops_by_id('call')
+        assert loop.match_by_id('call', '''
+        ...
+        guard_not_invalidated(descr=...)
+        p79 = new_array_clear(i34, descr=...)
+        call_n(ConstClass(ll_arraycopy__arrayPtr_arrayPtr_Signed_Signed_Signed), p36, p79, 0, 0, i34, descr=...)
+        p83 = new_array_clear(i41, descr=...)
+        setarrayitem_gc(p83, 0, ConstPtr(ptr85), descr=...)
+        call_n(ConstClass(ll_arraycopy__arrayPtr_arrayPtr_Signed_Signed_Signed), p79, p83, 0, 1, i34, descr=...)
+        p85 = force_token()
+        ...
+        ''')
+
+    def test_function_escape_without_defaults(self):
+        def main():
+            global f
+            res = 0
+            for i in range(10000):
+                def f():
+                    return i + 1
+                res += f() # ID: call
+
+        log = self.run(main, [])
+        loop, = log.loops_by_id('call')
+        opnames = log.opnames(loop._allops())
+        assert opnames.count("new_array_clear") == 2 # for the closure, but not for defaults
