@@ -1,5 +1,7 @@
 """
 Logic to find sys.executable and the initial sys.path containing the stdlib
+In CPython much of this is done in Modules/getpath.py which is frozen into
+the executable as bytecode
 """
 
 import errno
@@ -62,6 +64,8 @@ def _readlink_maybe(filename):
 
 
 def resolvedirof(filename):
+    if isinstance(filename, unicode):
+        filename = filename.encode('utf8')
     filename = rpath.rabspath(filename)
     dirname = rpath.rabspath(os.path.join(filename, '..'))
     if os.path.islink(filename):
@@ -105,7 +109,7 @@ def find_pyvenv_cfg(dirname):
     return ''
 
 
-def find_stdlib(state, executable):
+def find_stdlib(state, platlibdir, executable):
     """
     Find and compute the stdlib path, starting from the directory where
     ``executable`` is and going one level up until we find it.  Return a
@@ -128,7 +132,7 @@ def find_stdlib(state, executable):
             if home:
                 dirname = home
                 search_pyvenv_cfg = 0
-        newpath = compute_stdlib_path_maybe(state, dirname)
+        newpath = compute_stdlib_path_maybe(state, platlibdir, dirname)
         if newpath is not None:
             return newpath, dirname
         search = dirname    # walk to the parent directory
@@ -139,13 +143,42 @@ def _checkdir(path):
     if not stat.S_ISDIR(st[0]):
         raise OSError(errno.ENOTDIR, path)
 
+def _checkfile(path, fname):
+    pth = os.path.join(path, fname)
+    if not os.path.isfile(pth):
+        raise OSError(errno.EEXIST, pth)
 
-def compute_stdlib_path(state, prefix):
+def compute_stdlib_path_packaged(state, platlibdir, prefix):
     """
     Compute the paths for the stdlib rooted at ``prefix``. ``prefix``
-    must at least contain a directory called ``lib-python/X.Y`` and
-    another one called ``lib_pypy``. If they cannot be found, it raises
-    OSError.
+    must at least contain a directory called ``lib/pypyX.Y`` (``Lib`` on
+    windows), and there must be a site.py in that directory.  If it cannot
+    be found, it raises OSError. This version is called first, and handles
+    packaged PyPy.
+    """
+    from pypy.module.sys.version import CPYTHON_VERSION
+    lib_pyzip = os.path.join(prefix, 'python%d%d.zip' % CPYTHON_VERSION[:2])
+    if os.path.isfile(lib_pyzip):
+        python_std_lib = lib_pyzip
+    else:
+        if _WIN32:
+            lib_python = os.path.join(prefix, 'Lib')
+        else:
+            dirname = 'pypy%d.%d' % CPYTHON_VERSION[:2]
+            lib_python = os.path.join(prefix, platlibdir)
+            lib_python = os.path.join(lib_python, dirname)
+        python_std_lib = os.path.join(prefix, lib_python)
+        # In a source checkout, the directory will exist but site.py will not
+        # yet exist in it
+        _checkfile(python_std_lib, 'site.py')
+    return compute_lib_pypy_path(state, python_std_lib, prefix, use_lib_pypy=False)
+
+def compute_stdlib_path_sourcetree(state, platlibdir, prefix):
+    """
+    Compute the paths for the stdlib rooted at ``prefix``. ``prefix``
+    must at least contain a directory called ``lib-python/X.Y``. If it
+    cannot be found, it raises OSError. This version is called if
+    compute_stdlib_path_packaged fails.
     """
     from pypy.module.sys.version import CPYTHON_VERSION
     lib_pyzip = os.path.join(prefix, 'python%d%d.zip' % CPYTHON_VERSION[:2])
@@ -156,46 +189,53 @@ def compute_stdlib_path(state, prefix):
         lib_python = os.path.join(prefix, 'lib-python')
         python_std_lib = os.path.join(lib_python, dirname)
         _checkdir(python_std_lib)
+    return compute_lib_pypy_path(state, python_std_lib, prefix)
 
-    lib_pypy = os.path.join(prefix, 'lib_pypy')
-    _checkdir(lib_pypy)
-
+def compute_lib_pypy_path(state, python_std_lib, prefix, use_lib_pypy=True):
     importlist = []
 
-    if state is not None:    # 'None' for testing only
-        lib_extensions = os.path.join(lib_pypy, '__extensions__')
-        state.w_lib_extensions = state.space.newfilename(lib_extensions)
-        importlist.append(lib_extensions)
-
-    importlist.append(lib_pypy)
+    if use_lib_pypy:
+        lib_pypy = os.path.join(prefix, 'lib_pypy')
+        _checkdir(lib_pypy)
+        importlist.append(lib_pypy)
     importlist.append(python_std_lib)
 
-    lib_tk = os.path.join(python_std_lib, 'lib-tk')
-    importlist.append(lib_tk)
-
     # List here the extra platform-specific paths.
-    if not _WIN32:
-        importlist.append(os.path.join(python_std_lib, 'plat-' + PLATFORM))
     if _MACOSX:
+        # Is this still desirable?
         platmac = os.path.join(python_std_lib, 'plat-mac')
         importlist.append(platmac)
         importlist.append(os.path.join(platmac, 'lib-scriptpackages'))
-
     return importlist
 
 
-def compute_stdlib_path_maybe(state, prefix):
+def compute_stdlib_path_maybe(state, platlibdir, prefix):
     """Return the stdlib path rooted at ``prefix``, or None if it cannot
     be found.
     """
     try:
-        return compute_stdlib_path(state, prefix)
+        return compute_stdlib_path_packaged(state, platlibdir, prefix)
     except OSError:
-        return None
+        try:
+            return compute_stdlib_path_sourcetree(state, platlibdir, prefix)
+        except OSError:
+            return None
 
 
 @unwrap_spec(executable='fsencode')
 def pypy_find_executable(space, executable):
+    if _WIN32:
+        module_filename = pypy_init_executable()
+        if module_filename:
+            if _WIN32:
+                module_path, lgt = rffi.wcharp2utf8(module_filename)
+                module_path = assert_str0(module_path)
+            else:
+                module_path = rffi.charp2str(module_filename)
+            pypy_init_free(module_filename)
+            module_path = rpath.rabspath(module_path)
+            if _exists_and_is_executable(module_path):
+                return space.newfilename(module_path)
     return space.newfilename(find_executable(executable))
 
 
@@ -208,14 +248,18 @@ def pypy_resolvedirof(space, filename):
 def pypy_find_stdlib(space, executable):
     path, prefix = None, None
     if executable != '*':
-        path, prefix = find_stdlib(get_state(space), executable)
+        path, prefix = find_stdlib(get_state(space), space.config.objspace.platlibdir, executable)
     if path is None:
         if space.config.translation.shared:
             dynamic_location = pypy_init_home()
             if dynamic_location:
-                dyn_path = rffi.charp2str(dynamic_location)
+                if _WIN32:
+                    dyn_path, lgt = rffi.wcharp2utf8(dynamic_location)
+                    dyn_path = assert_str0(dyn_path)
+                else:
+                    dyn_path = rffi.charp2str(dynamic_location)
                 pypy_init_free(dynamic_location)
-                path, prefix = find_stdlib(get_state(space), dyn_path)
+                path, prefix = find_stdlib(get_state(space), space.config.objspace.platlibdir, dyn_path)
         if path is None:
             return space.w_None
     w_prefix = space.newfilename(prefix)
@@ -246,20 +290,20 @@ char *_pypy_init_home(void)
 {
     HMODULE hModule = 0;
     DWORD res;
-    char *p;
+    WCHAR *p;
 
-    GetModuleHandleEx(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS |
+    GetModuleHandleExW(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS |
                        GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
-                       (LPCTSTR)&_pypy_init_home, &hModule);
+                       (LPCWSTR)&_pypy_init_home, &hModule);
 
     if (hModule == 0 ) {
-        fprintf(stderr, "PyPy initialization: GetModuleHandleEx() failed\n");
+        fprintf(stderr, "PyPy initialization: GetModuleHandleExW() failed\n");
         return NULL;
     }
-    p = malloc(_MAX_PATH);
+    p = malloc(_MAX_PATH * sizeof(WCHAR));
     if (p == NULL)
         return NULL;
-    res = GetModuleFileName(hModule, p, _MAX_PATH);
+    res = GetModuleFileNameW(hModule, p, _MAX_PATH);
     if (res >= _MAX_PATH || res <= 0) {
         free(p);
         fprintf(stderr, "PyPy initialization: GetModuleFileName() failed\n");
@@ -267,6 +311,28 @@ char *_pypy_init_home(void)
     }
     return p;
 }
+
+wchar_t *_pypy_init_executable(void)
+{
+    DWORD res;
+    WCHAR *p;
+
+    p = (WCHAR *)malloc(_MAX_PATH * sizeof(WCHAR));
+    if (p == NULL)
+        return NULL;
+    res = GetModuleFileNameW(NULL, p, _MAX_PATH);
+    if (res >= _MAX_PATH || res <= 0) {
+        free(p);
+        fprintf(stderr, "PyPy initialization: GetModuleFileName() failed\n");
+        return NULL;
+    }
+    return p;
+}
+inline
+void _pypy_init_free(WCHAR *p)
+{
+    free(p);
+}
 """
 
 else:
@@ -292,9 +358,7 @@ char *_pypy_init_home(void)
     }
     return p;
 }
-"""
 
-_source_code += """
 inline
 void _pypy_init_free(char *p)
 {
@@ -306,15 +370,29 @@ if we_are_translated():
    post_include_bits = []
 else:
     # for tests 
-    post_include_bits=['RPY_EXPORTED char *_pypy_init_home(void);',
-                       'RPY_EXPORTED void _pypy_init_free(char*);',
-                      ]
+    if _WIN32:
+        post_include_bits=['RPY_EXPORTED char *_pypy_init_home(void);',
+                           'RPY_EXPORTED void _pypy_init_free(char*);',
+                           'RPY_EXPORTED wchar_t *_pypy_init_executable(void);',
+                          ]
+    else:
+        post_include_bits=['RPY_EXPORTED char *_pypy_init_home(void);',
+                           'RPY_EXPORTED void _pypy_init_free(char*);',
+                          ]
 
 _eci = ExternalCompilationInfo(separate_module_sources=[_source_code],
                                post_include_bits=post_include_bits)
 _eci = _eci.merge(rdynload.eci)
 
-pypy_init_home = rffi.llexternal("_pypy_init_home", [], rffi.CCHARP,
+if _WIN32:
+    pypy_init_executable = rffi.llexternal("_pypy_init_executable", [], rffi.CWCHARP,
+                                           _nowrapper=True, compilation_info=_eci)
+    pypy_init_free = rffi.llexternal("_pypy_init_free", [rffi.CWCHARP], lltype.Void,
                                  _nowrapper=True, compilation_info=_eci)
-pypy_init_free = rffi.llexternal("_pypy_init_free", [rffi.CCHARP], lltype.Void,
+    pypy_init_home = rffi.llexternal("_pypy_init_home", [], rffi.CWCHARP,
+                                 _nowrapper=True, compilation_info=_eci)
+else:
+    pypy_init_free = rffi.llexternal("_pypy_init_free", [rffi.CCHARP], lltype.Void,
+                                 _nowrapper=True, compilation_info=_eci)
+    pypy_init_home = rffi.llexternal("_pypy_init_home", [], rffi.CCHARP,
                                  _nowrapper=True, compilation_info=_eci)
