@@ -3,16 +3,16 @@
 
 import _io
 
-stdin = _io.open(0, "r", encoding="ascii",
+stdin = _io.open(0, "r", encoding="utf-8",
                 closefd=False)
 stdin.buffer.raw.name = "<stdin>"
 
-stdout = _io.open(1, "w", encoding="ascii",
+stdout = _io.open(1, "w", encoding="utf-8",
                  buffering=1,
                  closefd=False)
 stdout.buffer.raw.name = "<stdout>"
 
-stderr = _io.open(2, "w", encoding="ascii",
+stderr = _io.open(2, "w", encoding="utf-8",
                  errors="backslashreplace",
                  buffering=1,
                  closefd=False)
