@@ -1,5 +1,6 @@
 # -*- coding: iso-8859-1 -*-
 import codecs
+import pytest
 import sys
 
 def test_stdin_exists(space):
@@ -56,8 +57,6 @@ class AppTestAppSysTests:
 
     def setup_class(cls):
         cls.w_appdirect = cls.space.wrap(cls.runappdirect)
-        filesystemenc = codecs.lookup(sys.getfilesystemencoding()).name
-        cls.w_filesystemenc = cls.space.wrap(filesystemenc)
 
     def test_sys_in_modules(self):
         import sys
@@ -136,12 +135,8 @@ class AppTestAppSysTests:
     def test_getfilesystemencoding(self):
         import sys
         enc = sys.getfilesystemencoding()
-        if self.appdirect:
-            assert enc == self.filesystemenc
-        else:
-            # see comment in 'setup_after_space_initialization'
-            untranslated_enc = {'win32': 'mbcs', 'darwin': 'utf-8'}.get(enc, 'ascii')
-            assert enc == untranslated_enc
+        # always utf-8
+        assert enc == "utf-8"
 
     def test_float_info(self):
         import sys
@@ -187,6 +182,20 @@ class AppTestAppSysTests:
         assert isinstance(li.nan, int)
         assert isinstance(li.imag, int)
         assert isinstance(li.algorithm, str)
+        assert str(li).startswith('sys.hash_info(')
+
+    def test_sys_flags(self):
+        import sys
+        # sanity check
+        assert sys.flags.optimize is not None
+        # make sure the flags are read-only
+        exc = raises(AttributeError, 'sys.flags.optimize = 3')
+        assert 'readonly' in str(exc.value)
+        raises(AttributeError, 'sys.flags.not_a_sys_flag = 2')
+
+    def test_sys_flags_warn_default_encoding(self):
+        import sys
+        sys.flags.warn_default_encoding # check it exists
 
     def test_sys_exit(self):
         import sys
@@ -213,6 +222,42 @@ class AppTestAppSysTests:
         import sys
         assert type(sys.flags.dev_mode) is bool
 
+    def test_multiarch(self):
+        import sys
+        multiarch = ''
+        try:
+            multiarch = sys.implementation._multiarch
+        except AttributeError:
+            assert sys.platform in ('win32',)
+        else:
+            assert 'linux' in multiarch or 'darwin' in multiarch
+
+    def test_audit(self):
+        import sys
+        sys.audit("os.chdir", "bla", 1, 2, 12) # does not crash
+
+    @pytest.mark.skipif(sys.platform=="win32", reason="platlibdir is posix only")
+    def test_platlibdir(self):
+        if self.appdirect:
+            skip("only relevant for untranslated tests")
+        import sys
+        assert sys.platlibdir == "lib" # default
+
+
+@pytest.mark.skipif(sys.platform=="win32", reason="platlibdir is posix only")
+class AppTestPlatlibdirNotDefault:
+    spaceconfig = {
+        "platlibdir": "lib64"
+    }
+    def setup_class(cls):
+        cls.w_appdirect = cls.space.wrap(cls.runappdirect)
+
+    def test_platlibdir(self):
+        if self.appdirect:
+            skip("only relevant for untranslated tests")
+        import sys
+        assert sys.platlibdir == "lib64"
+
 
 class AppTestSysModulePortedFromCPython:
     spaceconfig = {
@@ -316,7 +361,7 @@ class AppTestSysModulePortedFromCPython:
             traceback.print_exception = original_print_exception
             sys.stderr = savestderr
 
-        assert err.getvalue() == "ValueError: 42\n"
+        assert "ValueError: 42\n" in err.getvalue()
 
     def test_original_excepthook_pypy_encoding(self):
         import sys
@@ -547,7 +592,7 @@ class AppTestSysModulePortedFromCPython:
     def test_winver(self):
         import sys
         if hasattr(sys, "winver"):
-            assert sys.winver == sys.version[:3]
+            assert sys.winver == sys.version[:4]
 
     def test_dllhandle(self):
         import sys
@@ -744,27 +789,6 @@ class AppTestSysModulePortedFromCPython:
         import sys
         sys.__dict__   # check that we don't crash initializing any attribute
 
-    def test_subversion(self):
-        import sys
-        if '__pypy__' not in sys.builtin_module_names:
-            skip("only on PyPy")
-        assert sys.subversion == ('PyPy', '', '')
-
-    def test__mercurial(self):
-        import sys, re
-        if '__pypy__' not in sys.builtin_module_names:
-            skip("only on PyPy")
-        project, hgtag, hgid = sys._mercurial
-        assert project == 'PyPy'
-        # the tag or branch may be anything, including the empty string
-        assert isinstance(hgtag, str)
-        # the id is either nothing, or an id of 12 hash digits, with a possible
-        # suffix of '+' if there are local modifications
-        assert hgid == '' or re.match('[0-9a-f]{12}\+?', hgid)
-        # the id should also show up in sys.version
-        if hgid != '':
-            assert hgid in sys.version
-
     def test_float_repr_style(self):
         import sys
 
@@ -817,8 +841,6 @@ class AppTestSysModulePortedFromCPython:
         finally:
             sys.set_coroutine_origin_tracking_depth(0)
 
-
-
 class AppTestSysSettracePortedFromCpython(object):
     def test_sys_settrace(self):
         import sys
@@ -836,7 +858,10 @@ class AppTestSysSettracePortedFromCpython(object):
 
         def compare_events(line_offset, events, expected_events):
             events = [(l - line_offset, e) for (l, e) in events]
-            assert events == expected_events
+            length = len(expected_events)
+            assert len(events) >= length
+            # unexpected cruft can accumulate, check only the last audit events
+            assert events[-length:] == expected_events
 
         def run_test2(func):
             tracer = Tracer()
@@ -873,32 +898,13 @@ class AppTestSysSettracePortedFromCpython(object):
         run_test2(settrace_and_raise)
 
 
-class AppTestCurrentFrames:
-    def test_current_frames(self):
-        try:
-            import _thread
-        except ImportError:
-            pass
-        else:
-            skip('This test requires an intepreter without threads')
-        import sys
-
-        def f():
-            return sys._current_frames()
-        frames = f()
-        assert list(frames) == [0]
-        assert frames[0].f_code.co_name in ('f', '?')
-
-
-class AppTestCurrentFramesWithThread(AppTestCurrentFrames):
-    spaceconfig = {
-        "usemodules": ["time", "thread"],
-    }
+class AppTestCurrentFramesWithThread:
 
     def test_current_frames(self):
         import sys
-        import time
         import _thread
+        if sys.platform == 'darwin':
+            skip('test can hang on macos')
 
         # XXX workaround for now: to prevent deadlocks, call
         # sys._current_frames() once before starting threads.
@@ -929,6 +935,43 @@ class AppTestCurrentFramesWithThread(AppTestCurrentFrames):
         _, other_frame = frames.popitem()
         assert other_frame.f_code.co_name in ('other_thread', '?')
 
+    def test_current_exceptions(self):
+        import sys
+        import _thread
+
+        # XXX workaround for now: to prevent deadlocks, call
+        # sys._current_frames() once before starting threads.
+        # This is an issue in non-translated versions only.
+        sys._current_frames()
+
+        thread_id = _thread.get_ident()
+        def other_thread():
+            #print("thread started")
+            lock2.release()
+            try:
+                raise ValueError("oops")
+            except ValueError:
+                lock1.acquire()
+        lock1 = _thread.allocate_lock()
+        lock2 = _thread.allocate_lock()
+        lock1.acquire()
+        lock2.acquire()
+        _thread.start_new_thread(other_thread, ())
+
+        def f():
+            lock2.acquire()
+            return sys._current_exceptions()
+
+        exc = f()
+        lock1.release()
+        thisexc = exc.pop(thread_id)
+        assert thisexc == (None, None, None)
+
+        assert len(exc) == 1
+        key, values = exc.popitem()
+        exc_type, exc_value, exc_tb = values
+        assert str(exc_value) == "oops"
+
     def test_intern(self):
         from sys import intern
         raises(TypeError, intern)
