@@ -49,11 +49,12 @@ class AppTestUnicodeData:
         assert int(unicodedata.unidata_version.split(".")[0]) >= 8
         cases = [
             ('3400', '4DB5'),
-            ('4E00', '9FEF'),
-            ('20000', '2A6D6'),
-            ('2A700', '2B734'),
-            ('2B740', '2B81D'),
-            ('2B820', '2CEA1'),
+            ('4E00', '9FFD'),
+            ('20000', '2A6DF'),
+            ('2A700', '2B738'),
+            ('2B740', '2CEA1'),
+            ('2CEB0', '2EBE0'),
+            ('30000', '3134A'),
         ]
         for first, last in cases:
             first = int(first, 16)
@@ -78,6 +79,11 @@ class AppTestUnicodeData:
         import unicodedata
         assert unicodedata.lookup("GOTHIC LETTER FAIHU") == '\U00010346'
 
+    def test_yes_alias_python3(self):
+        import unicodedata
+        assert unicodedata.lookup("LATIN SMALL LETTER GHA") == u"\u01a3"
+        assert unicodedata.name(u"\u01a3") == "LATIN SMALL LETTER OI"
+
     def test_normalize_bad_argcount(self):
         import unicodedata
         raises(TypeError, unicodedata.normalize, 'x')
@@ -92,6 +98,11 @@ class AppTestUnicodeData:
         import unicodedata
         assert unicodedata.normalize('NFC', '\U000110a5\U000110ba') == '\U000110ab'
 
+    def test_is_normalized(self):
+        import unicodedata
+        assert unicodedata.is_normalized("NFC", '\U000110ab')
+        assert not unicodedata.is_normalized("NFC", '\U000110a5\U000110ba')
+
     def test_linebreaks(self):
         linebreaks = (0x0a, 0x0b, 0x0c, 0x0d, 0x85,
                       0x1c, 0x1d, 0x1e, 0x2028, 0x2029)
@@ -162,3 +173,15 @@ class AppTestUnicodeData:
     def test_11_change(self):
         import unicodedata
         assert unicodedata.name(chr(0x1f9b9)) == "SUPERVILLAIN"
+
+    def test_12_1_change(self):
+        import unicodedata
+        assert unicodedata.name(chr(0x32ff)) == 'SQUARE ERA NAME REIWA'
+
+    def test_13_0_change(self):
+        import unicodedata
+        assert unicodedata.lookup('CIRCLED CC')
+
+    def test_14_0_change(self):
+        import unicodedata
+        assert unicodedata.lookup('PLAYGROUND SLIDE')
