@@ -5,9 +5,27 @@ from pypy.interpreter.typedef import TypeDef, interp_attrproperty
 from pypy.interpreter.error import OperationError, oefmt
 from rpython.rlib.rarithmetic import intmask, r_uint, r_uint32
 from rpython.rlib.objectmodel import keepalive_until_here
+from rpython.rtyper.tool import rffi_platform
 
 from rpython.rlib import rzlib
 
+# Add constants for CPython3.11+ compatibility
+constantnames = '''
+    Z_NO_COMPRESSION Z_RLE Z_PARTIAL_FLUSH Z_FIXED Z_BLOCK Z_TREES
+    '''.split()
+class SimpleCConfig:
+    """
+    Definitions for additional constants in python3.11+.
+    """
+    _compilation_info_ = rzlib.eci
+
+for _name in constantnames:
+    setattr(SimpleCConfig, _name, rffi_platform.ConstantInteger(_name))
+
+config = rffi_platform.configure(SimpleCConfig)
+for _name in constantnames:
+    globals()[_name] = config[_name]
+
 
 @unwrap_spec(string='bufferstr', start='truncatedint_w')
 def crc32(space, string, start = rzlib.CRC32_DEFAULT_START):
@@ -44,16 +62,21 @@ def zlib_error(space, msg):
     return OperationError(w_error, space.newtext(msg))
 
 
-@unwrap_spec(data='bufferstr', level=int)
-def compress(space, data, level=rzlib.Z_DEFAULT_COMPRESSION):
+@unwrap_spec(data='bufferstr', level=int, wbits=int)
+def compress(space, data, __posonly__=None, level=rzlib.Z_DEFAULT_COMPRESSION, wbits=rzlib.MAX_WBITS):
     """
-    compress(data[, level]) -- Returned compressed string.
+    Returns a bytes object containing compressed data.
 
-    Optional arg level is the compression level, in 1-9.
+    data
+      Binary data to be compressed.
+    level
+      Compression level, in 0-9 or -1.
+    wbits
+      The window buffer size and container format.
     """
     try:
         try:
-            stream = rzlib.deflateInit(level)
+            stream = rzlib.deflateInit(level=level, wbits=wbits)
         except ValueError:
             raise zlib_error(space, "Bad compression level")
         try:
@@ -66,7 +89,7 @@ def compress(space, data, level=rzlib.Z_DEFAULT_COMPRESSION):
 
 
 @unwrap_spec(string='bufferstr', wbits="c_int", bufsize=int)
-def decompress(space, string, wbits=rzlib.MAX_WBITS, bufsize=0):
+def decompress(space, string, __posonly__=None, wbits=rzlib.MAX_WBITS, bufsize=0):
     """
     decompress(string[, wbits[, bufsize]]) -- Return decompressed string.
 
@@ -125,7 +148,7 @@ class Compress(ZLibObject):
             self.stream = rzlib.null_stream
 
     @unwrap_spec(data='bufferstr')
-    def compress(self, space, data):
+    def compress(self, space, data, __posonly__=None):
         """
         compress(data) -- Return a string containing data compressed.
 
@@ -166,6 +189,10 @@ class Compress(ZLibObject):
             raise zlib_error(space, e.msg)
         return Compress(space=space, stream=copied)
 
+    def descr_deepcopy(self, space, w_memo):
+        return self.copy(space)
+
+
     @unwrap_spec(mode="c_int")
     def flush(self, space, mode=rzlib.Z_FINISH):
         """
@@ -228,6 +255,8 @@ Compress.typedef = TypeDef(
     'Compress',
     __new__ = interp2app(Compress___new__),
     copy = interp2app(Compress.copy),
+    __copy__ = interp2app(Compress.copy),
+    __deepcopy__ = interp2app(Compress.descr_deepcopy),
     compress = interp2app(Compress.compress),
     flush = interp2app(Compress.flush),
     __doc__ = """compressobj([level]) -- Return a compressor object.
@@ -276,7 +305,7 @@ class Decompress(ZLibObject):
             self.unconsumed_tail = tail
 
     @unwrap_spec(data='bufferstr', max_length=int)
-    def decompress(self, space, data, max_length=0):
+    def decompress(self, space, data, __posonly__=None, max_length=0):
         """
         decompress(data[, max_length]) -- Return a string containing the
         decompressed version of the data.
@@ -332,6 +361,9 @@ class Decompress(ZLibObject):
             zdict=self.zdict,
         )
 
+    def descr_deepcopy(self, space, w_memo):
+        return self.copy(space)
+
     def flush(self, space, w_length=None):
         """
         flush( [length] ) -- This is kept for backward compatibility,
@@ -391,6 +423,8 @@ Decompress.typedef = TypeDef(
     'Decompress',
     __new__ = interp2app(Decompress___new__),
     copy = interp2app(Decompress.copy),
+    __copy__ = interp2app(Decompress.copy),
+    __deepcopy__ = interp2app(Decompress.descr_deepcopy),
     decompress = interp2app(Decompress.decompress),
     flush = interp2app(Decompress.flush),
     unused_data = interp_attrproperty('unused_data', Decompress, wrapfn="newbytes"),
