@@ -1,11 +1,12 @@
 import operator
-from pypy.interpreter.error import OperationError, oefmt
+from pypy.interpreter.error import OperationError, oefmt, oefmt_attribute_error
+from pypy.interpreter.error import enrich_attribute_error
 from pypy.interpreter.baseobjspace import ObjSpace
 from pypy.interpreter.function import Function, Method, FunctionWithFixedCode
 from pypy.interpreter.argument import Arguments
-from pypy.interpreter.typedef import default_identity_hash
+from pypy.interpreter.typedef import default_identity_hash, use_special_method_shortcut
 from rpython.tool.sourcetools import compile2, func_with_new_name
-from rpython.rlib.objectmodel import specialize
+from rpython.rlib.objectmodel import specialize, always_inline
 from rpython.rlib import jit
 
 @specialize.memo()
@@ -55,10 +56,12 @@ def dict_getitem(space):
 
 
 def raiseattrerror(space, w_obj, w_name, w_descr=None):
-    # space.repr always returns an encodable string.
     if w_descr is None:
-        raise oefmt(space.w_AttributeError,
-                    "'%T' object has no attribute %R", w_obj, w_name)
+        if space.isinstance_w(w_obj, space.w_type):
+            raise oefmt_attribute_error(
+                space, w_obj, w_name, "type object '%N' has no attribute %R")
+        raise oefmt_attribute_error(space,
+                    w_obj, w_name, "'%T' object has no attribute %R")
     else:
         raise oefmt(space.w_AttributeError,
                     "'%T' object attribute %R is read-only", w_obj, w_name)
@@ -71,8 +74,14 @@ def get_attribute_name(space, w_obj, w_name):
             raiseattrerror(space, w_obj, w_name)
         raise
 
+@always_inline
 def _same_class_w(space, w_obj1, w_obj2, w_typ1, w_typ2):
-    return space.is_w(w_typ1, w_typ2)
+    from pypy.objspace.std.typeobject import W_TypeObject
+    # micro-optimization, call on the instances instead of using space.is_w so
+    # we can inline it
+    assert isinstance(w_typ1, W_TypeObject)
+    assert isinstance(w_typ2, W_TypeObject)
+    return w_typ1.is_w(space, w_typ2)
 
 
 class Object(object):
@@ -106,9 +115,15 @@ class Object(object):
         name = get_attribute_name(space, w_obj, w_name)
         w_descr = space.lookup(w_obj, name)
         if w_descr is not None:
-            if space.is_data_descr(w_descr):
-                space.set(w_descr, w_obj, w_value)
-                return
+            # shortcut for:
+            # if space.is_data_descr(w_descr):
+            #     return space.set(w_descr, w_obj, w_value)
+            w_set = space.lookup(w_descr, '__set__')
+            if w_set is not None:
+                return space.get_and_call_function(w_set, w_descr, w_obj, w_value)
+            if space.lookup(w_descr, '__delete__') is not None:
+                raise oefmt(space.w_AttributeError,
+                            "'%T' object is not a descriptor with set", w_descr)
         if w_obj.setdictvalue(space, name, w_value):
             return
         raiseattrerror(space, w_obj, w_name, w_descr)
@@ -210,7 +225,11 @@ class DescrOperation(object):
     def getattr(space, w_obj, w_name):
         # may be overridden in StdObjSpace
         w_descr = space.lookup(w_obj, '__getattribute__')
-        return space._handle_getattribute(w_descr, w_obj, w_name)
+        try:
+            return space._handle_getattribute(w_descr, w_obj, w_name)
+        except OperationError as e:
+            enrich_attribute_error(space, e, w_obj, w_name)
+            raise e
 
     def _handle_getattribute(space, w_descr, w_obj, w_name):
         try:
@@ -226,6 +245,7 @@ class DescrOperation(object):
             return space.get_and_call_function(w_descr, w_obj, w_name)
 
     def setattr(space, w_obj, w_name, w_val):
+        # overridden in StdObjSpace
         w_descr = space.lookup(w_obj, '__setattr__')
         if w_descr is None:
             raise oefmt(space.w_AttributeError,
@@ -240,6 +260,8 @@ class DescrOperation(object):
                         w_obj)
         return space.get_and_call_function(w_descr, w_obj, w_name)
 
+    # would be cool to shortcut this too, but old-style classes return int from
+    # __nonzero__
     def is_true(space, w_obj):
         w_descr = space.lookup(w_obj, "__bool__")
         if w_descr is None:
@@ -248,7 +270,7 @@ class DescrOperation(object):
                 return True
             # call __len__
             w_res = space.get_and_call_function(w_descr, w_obj)
-            return space._check_len_result(w_res) != 0
+            return space._check_len_result(space.index(w_res)) != 0
         # call __bool__
         w_res = space.get_and_call_function(w_descr, w_obj)
         # more shortcuts for common cases
@@ -268,20 +290,29 @@ class DescrOperation(object):
         else:
             return space.w_False
 
-    def len(space, w_obj):
+    @use_special_method_shortcut('__len__')
+    def _len(space, w_obj):
         w_descr = space.lookup(w_obj, '__len__')
         if w_descr is None:
             raise oefmt(space.w_TypeError, "'%T' has no length", w_obj)
-        w_res = space.get_and_call_function(w_descr, w_obj)
-        return space.newint(space._check_len_result(w_res))
+        return space.get_and_call_function(w_descr, w_obj)
 
-    def _check_len_result(space, w_obj):
+    def len_w(space, w_obj):
+        w_res = space._len(w_obj)
+        return space._check_len_result(space.index(w_res))
+
+    def len(space, w_obj):
+        w_res = space.index(space._len(w_obj))
+        # check for error or overflow
+        space._check_len_result(w_res)
+        return w_res
+
+    def _check_len_result(space, w_int):
         # Will complain if result is too big.
-        w_result = space.index(w_obj)
-        assert space.isinstance_w(w_result, space.w_int)
-        if space.is_true(space.lt(w_result, space.newint(0))):
+        assert space.isinstance_w(w_int, space.w_int)
+        if space.is_true(space.lt(w_int, space.newint(0))):
             raise oefmt(space.w_ValueError, "__len__() should return >= 0")
-        result = space.getindex_w(w_result, space.w_OverflowError)
+        result = space.getindex_w(w_int, space.w_OverflowError)
         assert result >= 0
         return result
 
@@ -294,6 +325,8 @@ class DescrOperation(object):
                 return False
         return True
 
+    @use_special_method_shortcut('__iter__',
+            lambda space, w_iter: space.lookup(w_iter, '__next__') is not None)
     def iter(space, w_obj):
         w_descr = space.lookup(w_obj, '__iter__')
         if w_descr is None:
@@ -303,12 +336,16 @@ class DescrOperation(object):
                 raise oefmt(space.w_TypeError,
                             "'%T' object is not iterable", w_obj)
             return space.newseqiter(w_obj)
+        if space.is_w(w_descr, space.w_None):
+            raise oefmt(space.w_TypeError,
+                        "'%T' object is not iterable", w_obj)
         w_iter = space.get_and_call_function(w_descr, w_obj)
         w_next = space.lookup(w_iter, '__next__')
         if w_next is None:
             raise oefmt(space.w_TypeError, "iter() returned non-iterator")
         return w_iter
 
+    @use_special_method_shortcut('__next__')
     def next(space, w_obj):
         w_descr = space.lookup(w_obj, '__next__')
         if w_descr is None:
@@ -316,16 +353,34 @@ class DescrOperation(object):
                         "'%T' object is not an iterator", w_obj)
         return space.get_and_call_function(w_descr, w_obj)
 
+    @use_special_method_shortcut('__getitem__')
     def getitem(space, w_obj, w_key):
+        from pypy.objspace.std.typeobject import W_TypeObject
         w_descr = space.lookup(w_obj, '__getitem__')
         if w_descr is None and space.isinstance_w(w_obj, space.w_type):
-            w_descr = space.getattr(w_obj, space.newtext('__class_getitem__'))
+            # you've got to be kidding me :-( - cpython does the same
+            if space.is_w(w_obj, space.w_type):
+                from pypy.objspace.std.util import generic_alias_class_getitem
+                return generic_alias_class_getitem(space, w_obj, w_key)
+            try:
+                w_descr = space.getattr(w_obj, space.newtext('__class_getitem__'))
+            except OperationError as e:
+                if e.match(space, space.w_AttributeError):
+                    w_descr = None
+                else:
+                    raise e
+            if space.is_none(w_descr):
+                assert isinstance(w_obj, W_TypeObject)
+                raise oefmt(space.w_TypeError,
+                            "type '%8' is not subscriptable (key %R)",
+                            w_obj.name, w_key)
         if w_descr is None:
             raise oefmt(space.w_TypeError,
                         "'%T' object is not subscriptable (key %R)",
                         w_obj, w_key)
         return space.get_and_call_function(w_descr, w_obj, w_key)
 
+    @use_special_method_shortcut('__setitem__')
     def setitem(space, w_obj, w_key, w_val):
         w_descr = space.lookup(w_obj, '__setitem__')
         if w_descr is None:
@@ -333,6 +388,7 @@ class DescrOperation(object):
                         "'%T' object does not support item assignment", w_obj)
         return space.get_and_call_function(w_descr, w_obj, w_key, w_val)
 
+    @use_special_method_shortcut('__delitem__')
     def delitem(space, w_obj, w_key):
         w_descr = space.lookup(w_obj, '__delitem__')
         if w_descr is None:
@@ -352,7 +408,7 @@ class DescrOperation(object):
                         w_obj, w_res)
         return w_res
 
-    def pow(space, w_obj1, w_obj2, w_obj3):
+    def _pow_find_impls(space, w_obj1, w_obj2):
         w_typ1 = space.type(w_obj1)
         w_typ2 = space.type(w_obj2)
         w_left_src, w_left_impl = space.lookup_in_type_where(w_typ1, '__pow__')
@@ -368,23 +424,64 @@ class DescrOperation(object):
                     not space.abstract_issubclass_w(w_typ1, w_right_src)):
                     w_obj1, w_obj2 = w_obj2, w_obj1
                     w_left_impl, w_right_impl = w_right_impl, w_left_impl
+        return w_left_impl, w_right_impl, w_obj1, w_obj2
+
+    def pow_binary(space, w_obj1, w_obj2):
+        w_left_impl, w_right_impl, w_obj1, w_obj2 = space._pow_find_impls(w_obj1, w_obj2)
         if w_left_impl is not None:
-            if space.is_w(w_obj3, space.w_None):
-                w_res = space.get_and_call_function(w_left_impl, w_obj1, w_obj2)
-            else:
-                w_res = space.get_and_call_function(w_left_impl, w_obj1, w_obj2, w_obj3)
+            w_res = space.get_and_call_function(w_left_impl, w_obj1, w_obj2)
             if _check_notimplemented(space, w_res):
                 return w_res
         if w_right_impl is not None:
-            if space.is_w(w_obj3, space.w_None):
-                w_res = space.get_and_call_function(w_right_impl, w_obj2, w_obj1)
-            else:
-                w_res = space.get_and_call_function(w_right_impl, w_obj2, w_obj1,
-                                                   w_obj3)
+            w_res = space.get_and_call_function(w_right_impl, w_obj2, w_obj1)
             if _check_notimplemented(space, w_res):
                 return w_res
+        return None
 
-        raise oefmt(space.w_TypeError, "operands do not support **")
+    def pow(space, w_obj1, w_obj2, w_obj3):
+        if space.is_w(w_obj3, space.w_None):
+            w_res = space.pow_binary(w_obj1, w_obj2)
+            if w_res is None:
+                raise oefmt(space.w_TypeError,
+                    "unsupported operand type(s) for ** or pow(): '%T' and '%T'",
+                    w_obj1, w_obj2
+                )
+            return w_res
+        # Three-arg power does not use __rpow__, unless the second argument is
+        # a cpyext type, then we use the bug-to-bug compatibility version
+        w_typ1 = space.type(w_obj1)
+        w_typ2 = space.type(w_obj2)
+        if w_typ2.is_cpytype():
+            w_res = space.pow3_bug_compat_cpyext(w_obj1, w_obj2, w_obj3)
+            if w_res is not None:
+                return w_res
+        else:
+            w_left_src, w_left_impl = space.lookup_in_type_where(w_typ1, '__pow__')
+            if w_left_impl is not None:
+                w_res = space.get_and_call_function(w_left_impl, w_obj1, w_obj2, w_obj3)
+                if _check_notimplemented(space, w_res):
+                    return w_res
+        if space.type(w_obj3).is_cpytype() and space.lookup(w_obj3, '__pow__') is not None:
+            raise oefmt(space.w_TypeError,
+                "unsupported operand type(s) for pow(): '%T', '%T', '%T' (This case is unsupported in PyPy. Please comment here if you hit it: https://github.com/pypy/pypy/issues/5207 )",
+                w_obj1, w_obj2, w_obj3
+            )
+        raise oefmt(space.w_TypeError,
+            "unsupported operand type(s) for pow(): '%T', '%T', '%T'",
+            w_obj1, w_obj2, w_obj3
+        )
+
+    def pow3_bug_compat_cpyext(space, w_obj1, w_obj2, w_obj3):
+        w_left_impl, w_right_impl, w_obj1, w_obj2 = space._pow_find_impls(w_obj1, w_obj2)
+        if w_left_impl is not None:
+            w_res = space.get_and_call_function(w_left_impl, w_obj1, w_obj2, w_obj3)
+            if _check_notimplemented(space, w_res):
+                return w_res
+        if w_right_impl is not None:
+            w_res = space.get_and_call_function(w_right_impl, w_obj2, w_obj1, w_obj3)
+            if _check_notimplemented(space, w_res):
+                return w_res
+        return None
 
     def inplace_pow(space, w_lhs, w_rhs):
         w_impl = space.lookup(w_lhs, '__ipow__')
@@ -392,8 +489,17 @@ class DescrOperation(object):
             w_res = space.get_and_call_function(w_impl, w_lhs, w_rhs)
             if _check_notimplemented(space, w_res):
                 return w_res
-        return space.pow(w_lhs, w_rhs, space.w_None)
+        w_res = space.pow_binary(w_lhs, w_rhs)
+        if w_res is None:
+            raise oefmt(
+                space.w_TypeError,
+                "unsupported operand type(s) for **=: '%T' and '%T'",
+                w_lhs, w_rhs
+            )
+        return w_res
 
+    @use_special_method_shortcut('__contains__',
+            lambda space, w_res: space.is_w(space.type(w_res), space.w_bool))
     def contains(space, w_container, w_item):
         w_descr = space.lookup(w_container, '__contains__')
         if w_descr is not None:
@@ -413,7 +519,7 @@ class DescrOperation(object):
                 if not e.match(space, space.w_StopIteration):
                     raise
                 return space.w_False
-            if space.eq_w(w_item, w_next):
+            if space.eq_w(w_next, w_item):
                 return space.w_True
 
     def sequence_count(space, w_container, w_item):
@@ -444,21 +550,22 @@ class DescrOperation(object):
                 return space.newint(index)
             index += 1
 
-    def hash(space, w_obj):
+    def hash_w(space, w_obj):
+        """compute the unwrapped hash of w_obj"""
         w_hash = space.lookup(w_obj, '__hash__')
         if w_hash is None:
             # xxx there used to be logic about "do we have __eq__ or __cmp__"
             # here, but it does not really make sense, as 'object' has a
             # default __hash__.  This path should only be taken under very
             # obscure circumstances.
-            return default_identity_hash(space, w_obj)
+            return space.int_w(default_identity_hash(space, w_obj))
         if space.is_w(w_hash, space.w_None):
             raise oefmt(space.w_TypeError,
                         "unhashable type: '%T'", w_obj)
         w_result = space.get_and_call_function(w_hash, w_obj)
         if not space.isinstance_w(w_result, space.w_int):
             raise oefmt(space.w_TypeError,
-                        "__hash__ method should return an integer")
+                        "__hash__ method should return an integer not '%T'", w_result)
 
         from pypy.objspace.std.intobject import (
             W_AbstractIntObject, W_IntObject)
@@ -470,7 +577,10 @@ class DescrOperation(object):
         # turn -1 into -2 without using a condition, which would
         # create a potential bridge in the JIT
         result -= (result == -1)
-        return space.newint(result)
+        return result
+
+    def hash(space, w_obj):
+        return space.newint(space.hash_w(w_obj))
 
     def issubtype_w(space, w_sub, w_type):
         return space._type_issubtype(w_sub, w_type)
@@ -486,7 +596,7 @@ class DescrOperation(object):
     def isinstance(space, w_inst, w_type):
         return space.newbool(space.isinstance_w(w_inst, w_type))
 
-    def index(space, w_obj):
+    def _index(space, w_obj):
         if space.isinstance_w(w_obj, space.w_int):
             return w_obj
         w_impl = space.lookup(w_obj, '__index__')
@@ -498,17 +608,23 @@ class DescrOperation(object):
 
         if space.is_w(space.type(w_result), space.w_int):
             return w_result
-        if space.isinstance_w(w_result, space.w_int):
-            tp = space.type(w_result).name
-            space.warn(space.newtext(
-                "__index__ returned non-int (type %s).  "
-                "The ability to return an instance of a strict subclass of int "
-                "is deprecated, and may be removed in a future version of "
-                "Python." % (tp,)), space.w_DeprecationWarning)
-            return w_result
-        raise oefmt(space.w_TypeError,
+        if not space.isinstance_w(w_result, space.w_int):
+            raise oefmt(space.w_TypeError,
                     "__index__ returned non-int (type %T)", w_result)
+        tp = space.type(w_result).name
+        space.warn(space.newtext(
+            "__index__ returned non-int (type %s).  "
+            "The ability to return an instance of a strict subclass of int "
+            "is deprecated, and may be removed in a future version of "
+            "Python." % (tp,)), space.w_DeprecationWarning)
+        return w_result
 
+    def index(space, w_obj):
+        w_result = space._index(w_obj)
+        if not space.is_w(space.type(w_result), space.w_int):
+            result = space.bigint_w(w_result)
+            w_result = space.newlong_from_rbigint(result)
+        return w_result
 
 # helpers
 
@@ -526,12 +642,47 @@ class PrintCache(object):
     def __init__(self, space):
         self.w_print = space.getattr(space.builtin, space.newtext("print"))
 
+# specialize on the special method names to use the lookup_in_type_where fast
+# paths
+@specialize.arg(3, 4)
+def _call_binop_impl(space, w_obj1, w_obj2, left, right, seq_bug_compat):
+    w_typ1 = space.type(w_obj1)
+    w_typ2 = space.type(w_obj2)
+    w_left_src, w_left_impl = space.lookup_in_type_where(w_typ1, left)
+    if space.is_w(w_typ1, w_typ2):
+        w_right_impl = None
+    else:
+        w_right_src, w_right_impl = space.lookup_in_type_where(w_typ2, right)
+        # the logic to decide if the reverse operation should be tried
+        # before the direct one is very obscure.  For now, and for
+        # sanity reasons, we just compare the two places where the
+        # __xxx__ and __rxxx__ methods where found by identity.
+        # Note that space.is_w() is potentially not happy if one of them
+        # is None...
+        if w_right_src and (w_left_src is not w_right_src) and w_left_src:
+            # 'seq_bug_compat' is for cpython bug-to-bug compatibility:
+            # see objspace/std/test/test_unicodeobject.*concat_overrides
+            # and objspace/test/test_descrobject.*rmul_overrides.
+            # For cases like "unicode + string subclass".
+            if ((seq_bug_compat and w_typ1.flag_sequence_bug_compat
+                                and not w_typ2.flag_sequence_bug_compat)
+                    # the non-bug-compat part is the following check:
+                    or space.issubtype_w(w_typ2, w_typ1)):
+                if (not space.abstract_issubclass_w(w_left_src, w_right_src) and
+                    not space.abstract_issubclass_w(w_typ1, w_right_src)):
+                    w_obj1, w_obj2 = w_obj2, w_obj1
+                    w_left_impl, w_right_impl = w_right_impl, w_left_impl
+
+    w_res = _invoke_binop(space, w_left_impl, w_obj1, w_obj2)
+    if w_res is not None:
+        return w_res
+    return _invoke_binop(space, w_right_impl, w_obj2, w_obj1)
 
 # regular methods def helpers
 
 def _make_binop_impl(symbol, specialnames):
     left, right = specialnames
-    errormsg = "unsupported operand type(s) for %s: '%%N' and '%%N'" % (
+    errormsg = "unsupported operand type(s) for %s: '%%T' and '%%T'" % (
         symbol.replace('%', '%%'),)
     seq_bug_compat = (symbol == '+' or symbol == '*')
 
@@ -541,43 +692,29 @@ def _make_binop_impl(symbol, specialnames):
     if symbol == "-":
         printerrormsg = errormsg + '. Did you mean "print(<-number>)"?'
 
-    def binop_impl(space, w_obj1, w_obj2):
-        w_typ1 = space.type(w_obj1)
-        w_typ2 = space.type(w_obj2)
-        w_left_src, w_left_impl = space.lookup_in_type_where(w_typ1, left)
-        if space.is_w(w_typ1, w_typ2):
-            w_right_impl = None
-        else:
-            w_right_src, w_right_impl = space.lookup_in_type_where(w_typ2, right)
-            # the logic to decide if the reverse operation should be tried
-            # before the direct one is very obscure.  For now, and for
-            # sanity reasons, we just compare the two places where the
-            # __xxx__ and __rxxx__ methods where found by identity.
-            # Note that space.is_w() is potentially not happy if one of them
-            # is None...
-            if w_right_src and (w_left_src is not w_right_src) and w_left_src:
-                # 'seq_bug_compat' is for cpython bug-to-bug compatibility:
-                # see objspace/std/test/test_unicodeobject.*concat_overrides
-                # and objspace/test/test_descrobject.*rmul_overrides.
-                # For cases like "unicode + string subclass".
-                if ((seq_bug_compat and w_typ1.flag_sequence_bug_compat
-                                    and not w_typ2.flag_sequence_bug_compat)
-                        # the non-bug-compat part is the following check:
-                        or space.issubtype_w(w_typ2, w_typ1)):
-                    if (not space.abstract_issubclass_w(w_left_src, w_right_src) and
-                        not space.abstract_issubclass_w(w_typ1, w_right_src)):
-                        w_obj1, w_obj2 = w_obj2, w_obj1
-                        w_left_impl, w_right_impl = w_right_impl, w_left_impl
+    @use_special_method_shortcut(left)
+    def shortcut_binop(space, w_obj1, w_obj2):
+        # makes a few assumptions: if the two rpython types are the same, then
+        # the left and the right implementations are the same too
+        w_impl = space.lookup(w_obj1, left)
+        if w_impl is not None:
+            w_res = space.get_and_call_function(w_impl, w_obj1, w_obj2)
+            if _check_notimplemented(space, w_res):
+                return w_res
+        raise oefmt(space.w_TypeError, errormsg, w_obj1, w_obj2)
 
-        w_res = _invoke_binop(space, w_left_impl, w_obj1, w_obj2)
-        if w_res is not None:
-            return w_res
-        w_res = _invoke_binop(space, w_right_impl, w_obj2, w_obj1)
+    def binop_impl(space, w_obj1, w_obj2):
+        # shortcut: rpython classes are the same
+        if type(w_obj1) is type(w_obj2) and not w_obj1.user_overridden_class:
+            w_res = shortcut_binop(space, w_obj1, w_obj2)
+            if _check_notimplemented(space, w_res):
+                return w_res
+        w_res = _call_binop_impl(space, w_obj1, w_obj2, left, right, seq_bug_compat)
         if w_res is not None:
             return w_res
         if printerrormsg is not None and w_obj1 is space.fromcache(PrintCache).w_print:
-            raise oefmt(space.w_TypeError, printerrormsg, w_typ1, w_typ2)
-        raise oefmt(space.w_TypeError, errormsg, w_typ1, w_typ2)
+            raise oefmt(space.w_TypeError, printerrormsg, w_obj1, w_obj2)
+        raise oefmt(space.w_TypeError, errormsg, w_obj1, w_obj2)
 
     return func_with_new_name(binop_impl, "binop_%s_impl"%left.strip('_'))
 
@@ -608,9 +745,26 @@ def _invoke_comparison(space, w_descr, w_obj1, w_obj2):
 def _make_comparison_impl(symbol, specialnames):
     left, right = specialnames
     op = getattr(operator, left)
+
+    @use_special_method_shortcut(left)
+    def shortcut_comparison(space, w_obj1, w_obj2):
+        # see shortcut_binop
+        w_impl = space.lookup(w_obj1, left)
+        if w_impl is not None:
+            w_res = space.get_and_call_function(w_impl, w_obj1, w_obj2)
+            if _check_notimplemented(space, w_res):
+                return w_res
+        return _no_special_method_found_cmp(space, w_obj1, w_obj2, w_obj1, w_obj2)
+
+
     def comparison_impl(space, w_obj1, w_obj2):
         w_orig_obj1 = w_obj1
         w_orig_obj2 = w_obj2
+        # shortcut, rpython classes are the same
+        if left == right and type(w_obj1) is type(w_obj2) and not w_obj1.user_overridden_class:
+            w_res = shortcut_comparison(space, w_obj1, w_obj2)
+            if _check_notimplemented(space, w_res):
+                return w_res
         w_typ1 = space.type(w_obj1)
         w_typ2 = space.type(w_obj2)
         w_left_src, w_left_impl = space.lookup_in_type_where(w_typ1, left)
@@ -633,7 +787,9 @@ def _make_comparison_impl(symbol, specialnames):
         w_res = _invoke_comparison(space, w_right_impl, w_obj2, w_obj1)
         if w_res is not None:
             return w_res
-        #
+        return _no_special_method_found_cmp(space, w_obj1, w_obj2, w_orig_obj1, w_orig_obj2)
+
+    def _no_special_method_found_cmp(space, w_obj1, w_obj2, w_orig_obj1, w_orig_obj2):
         # we did not find any special method, let's do the default logic for
         # == and !=
         if left == '__eq__':
@@ -662,12 +818,15 @@ def _make_inplace_impl(symbol, specialnames):
         noninplacespacemethod += '_'     # not too clean
     seq_bug_compat = (symbol == '+=' or symbol == '*=')
     rhs_method = '__r' + specialname[3:]
+    lhs_method = '__' + specialname[3:]
+    errormsg = "unsupported operand type(s) for %s: '%%T' and '%%T'" % (
+        symbol.replace('%', '%%'),)
 
     def inplace_impl(space, w_lhs, w_rhs):
         w_impl = space.lookup(w_lhs, specialname)
         if w_impl is not None:
             # 'seq_bug_compat' is for cpython bug-to-bug compatibility:
-            # see objspace/test/test_descrobject.*rmul_overrides.
+            # see objspace/test/apptest_descroperation.*rmul_overrides.
             # For cases like "list += object-overriding-__radd__".
             if (seq_bug_compat and space.type(w_lhs).flag_sequence_bug_compat
                            and not space.type(w_rhs).flag_sequence_bug_compat):
@@ -683,14 +842,20 @@ def _make_inplace_impl(symbol, specialnames):
             w_res = space.get_and_call_function(w_impl, w_lhs, w_rhs)
             if _check_notimplemented(space, w_res):
                 return w_res
-        # XXX fix the error message we get here
-        return getattr(space, noninplacespacemethod)(w_lhs, w_rhs)
+
+        w_res = _call_binop_impl(space, w_lhs, w_rhs, lhs_method,
+                                 rhs_method, seq_bug_compat)
+        if w_res is not None:
+            return w_res
+
+        raise oefmt(space.w_TypeError, errormsg, w_lhs, w_rhs)
 
     return func_with_new_name(inplace_impl, 'inplace_%s_impl'%specialname.strip('_'))
 
 def _make_unaryop_impl(symbol, specialnames):
     specialname, = specialnames
     errormsg = "unsupported operand type for unary %s: '%%T'" % symbol
+    @use_special_method_shortcut(specialname)
     def unaryop_impl(space, w_obj):
         w_impl = space.lookup(w_obj, specialname)
         if w_impl is None:
@@ -708,9 +873,11 @@ for targetname, specialname, checkerspec in [
     l = ["space.isinstance_w(w_result, %s)" % x
                 for x in checkerspec]
     checker = " or ".join(l)
-    msg = "unsupported operand type for %(targetname)s(): '%%T'"
+    msg = "must be real number, not %%T"
     msg = msg % locals()
     source = """if 1:
+        @use_special_method_shortcut(%(specialname)r,
+            lambda space, w_result: %(checker)s)
         def %(targetname)s(space, w_obj):
             w_impl = space.lookup(w_obj, %(specialname)r)
             if w_impl is None:
@@ -735,6 +902,9 @@ for targetname, specialname in [
     ('repr', '__repr__')]:
 
     source = """if 1:
+        @use_special_method_shortcut(%(specialname)r,
+            lambda space, w_res: space.isinstance_w(w_res, space.w_text)
+        )
         def %(targetname)s(space, w_obj):
             w_impl = space.lookup(w_obj, %(specialname)r)
             if w_impl is None:
