@@ -5,13 +5,14 @@ from rpython.rlib.rarithmetic import intmask, r_ulonglong
 from rpython.rlib.rbigint import rbigint
 from rpython.rlib.rfloat import (
     DTSF_STR_PRECISION, formatd, string_to_float)
+from rpython.rlib.special_value import NAN
 from rpython.rlib.rstring import ParseStringError
 from rpython.tool.sourcetools import func_with_new_name
 
 from pypy.interpreter.baseobjspace import W_Root
 from pypy.interpreter.error import OperationError, oefmt
 from pypy.interpreter.gateway import WrappedDefault, interp2app, unwrap_spec
-from pypy.interpreter.typedef import GetSetProperty, TypeDef
+from pypy.interpreter.typedef import GetSetProperty, TypeDef, default_identity_hash
 from pypy.objspace.std import newformat
 from pypy.objspace.std.floatobject import _hash_float, _remove_underscores
 from pypy.objspace.std.unicodeobject import unicode_to_decimal_w
@@ -20,6 +21,9 @@ HASH_IMAG = 1000003
 
 
 def _split_complex(s):
+    def iswhitespace(ch):
+        return (ch == ' ' or ch == '\f' or ch == '\n' or ch == '\r' or
+            ch == '\t' or ch == '\v')
     slen = len(s)
     if slen == 0:
         raise ValueError
@@ -30,9 +34,9 @@ def _split_complex(s):
     imagsign = ' '
     i = 0
     # ignore whitespace at beginning and end
-    while i < slen and s[i] == ' ':
+    while i < slen and iswhitespace(s[i]):
         i += 1
-    while slen > 0 and s[slen-1] == ' ':
+    while slen > 0 and iswhitespace(s[slen-1]):
         slen -= 1
 
     if s[i] == '(' and s[slen-1] == ')':
@@ -55,7 +59,7 @@ def _split_complex(s):
 
     realstop = i
 
-    # return appropriate strings is only one number is there
+    # return appropriate strings if only one number is there
     if i >= slen:
         newstop = realstop - 1
         if newstop < 0:
@@ -128,19 +132,17 @@ def format_float(x, code, precision):
 def repr_format(x):
     return format_float(x, 'r', 0)
 
-def str_format(x):
-    return format_float(x, 'g', DTSF_STR_PRECISION)
 
-
-def unpackcomplex(space, w_complex, strict_typing=True, firstarg=True):
+def unpackcomplex(space, w_complex, allow_subclass=False, firstarg=True):
     """
     convert w_complex into a complex and return the unwrapped (real, imag)
-    tuple. If strict_typing==True, we also typecheck the value returned by
-    __complex__ to actually be a complex (and not e.g. a float).
-    See test___complex___returning_non_complex.
+    tuple. If allow_subclass (i.e. _PyComplex_AsCComplex) we fast path subclasses
+    of complex.
     """
     if type(w_complex) is W_ComplexObject:
         return (w_complex.realval, w_complex.imagval)
+    if allow_subclass and isinstance(w_complex, W_ComplexObject):
+        return (w_complex.realval, w_complex.imagval)
     #
     # test for a '__complex__' method, and call it if found.
     w_z = None
@@ -150,12 +152,21 @@ def unpackcomplex(space, w_complex, strict_typing=True, firstarg=True):
     #
     if w_z is not None:
         # __complex__() must return a complex
-        # (XXX should not use isinstance here)
         if isinstance(w_z, W_ComplexObject):
+            if type(w_z) is not W_ComplexObject:
+                space.warn(
+                    space.newtext("__complex__ returned non-complex (type %s).  The ability to return an instance of a strict subclass of complex is deprecated, and may be removed in a future version of Python." % (space.type(w_z).getname(space))),
+                    space.w_DeprecationWarning
+                )
             return (w_z.realval, w_z.imagval)
         raise oefmt(space.w_TypeError,
                     "__complex__() must return a complex number")
 
+    # try to see whether it has an __index__
+    if space.lookup(w_complex, '__index__') is not None:
+        result = space.float_w(space.float(space.index(w_complex)))
+        return (result, 0.0)
+
     #
     # no '__complex__' method, so we assume it is a float,
     # unless it is an instance of some subclass of complex.
@@ -184,6 +195,31 @@ def unpackcomplex(space, w_complex, strict_typing=True, firstarg=True):
                         "complex() second argument must be a number, not '%T'",
                          w_complex)
 
+# Python3 semantics of c_pow have changed slightly
+
+def c_pow(x, y):
+    (a_r, a_i), (b_r, b_i) = x, y
+    if b_r == 0 and b_i == 0:
+        rr, ir = (1., 0.)
+    elif (a_r == 0. or math.isinf(a_r)) and a_i == 0:
+        if b_i != 0. or b_r < 0:
+            raise ZeroDivisionError
+        rr, ir = (0.0, 0.0)
+    else:
+        vabs = math.hypot(a_r, a_i)
+        len = math.pow(vabs, b_r)
+        at = math.atan2(a_i, a_r)
+        phase = at * b_r 
+        if b_i != 0.0:
+            len /= math.exp(at * b_i)
+            phase += b_i * math.log(vabs)
+        try:
+            rr = len * math.cos(phase)
+            ir = len * math.sin(phase)
+        except ValueError:
+            rr = NAN
+            ir = NAN
+    return (rr, ir)
 
 
 class W_ComplexObject(W_Root):
@@ -217,7 +253,7 @@ class W_ComplexObject(W_Root):
         return W_ComplexObject(rr, ir)
 
     def pow(self, other):
-        rr, ir = rcomplex.c_pow(self.as_tuple(), other.as_tuple())
+        rr, ir = c_pow(self.as_tuple(), other.as_tuple())
         return W_ComplexObject(rr, ir)
 
     def pow_small_int(self, n):
@@ -268,9 +304,6 @@ class W_ComplexObject(W_Root):
         val = real_b.lshift(64).or_(imag_b).lshift(IDTAG_SHIFT).int_or_(tag)
         return space.newlong_from_rbigint(val)
 
-    def descr_int(self, space):
-        raise oefmt(space.w_TypeError, "can't convert complex to int")
-
     def _to_complex(self, space, w_obj):
         if isinstance(w_obj, W_ComplexObject):
             return w_obj
@@ -339,8 +372,8 @@ class W_ComplexObject(W_Root):
         return w_obj
 
     def descr___getnewargs__(self, space):
-        return space.newtuple([space.newfloat(self.realval),
-                               space.newfloat(self.imagval)])
+        return space.newtuple2(space.newfloat(self.realval),
+                               space.newfloat(self.imagval))
 
     def descr_repr(self, space):
         if self.realval == 0 and math.copysign(1., self.realval) == 1.:
@@ -350,17 +383,15 @@ class W_ComplexObject(W_Root):
         return space.newtext('(' + repr_format(self.realval)
                              + sign + repr_format(self.imagval) + 'j)')
 
-    def descr_str(self, space):
-        if self.realval == 0 and math.copysign(1., self.realval) == 1.:
-            return space.newtext(str_format(self.imagval) + 'j')
-        sign = (math.copysign(1., self.imagval) == 1. or
-                math.isnan(self.imagval)) and '+' or ''
-        return space.newtext('(' + str_format(self.realval)
-                             + sign + str_format(self.imagval) + 'j)')
-
     def descr_hash(self, space):
-        hashreal = _hash_float(space, self.realval)
-        hashimg = _hash_float(space, self.imagval)   # 0 if self.imagval == 0
+        if math.isnan(self.realval):
+            hashreal = space.int_w(default_identity_hash(space, self))
+        else:
+            hashreal = _hash_float(self.realval)
+        if math.isnan(self.imagval):
+            hashimg = space.int_w(default_identity_hash(space, self))
+        else:
+            hashimg = _hash_float(self.imagval)
         h = intmask(hashreal + HASH_IMAG * hashimg)
         h -= (h == -1)
         return space.newint(h)
@@ -369,7 +400,7 @@ class W_ComplexObject(W_Root):
         w_other = self._to_complex(space, w_other)
         if w_other is None:
             return space.w_NotImplemented
-        return space.newtuple([self, w_other])
+        return space.newtuple2(self, w_other)
 
     def descr_format(self, space, w_format_spec):
         return newformat.run_formatter(space, w_format_spec, "format_complex",
@@ -378,15 +409,16 @@ class W_ComplexObject(W_Root):
     def descr_bool(self, space):
         return space.newbool((self.realval != 0.0) or (self.imagval != 0.0))
 
-    def descr_float(self, space):
-        raise oefmt(space.w_TypeError, "can't convert complex to float")
-
     def descr_neg(self, space):
         return W_ComplexObject(-self.realval, -self.imagval)
 
     def descr_pos(self, space):
         return W_ComplexObject(self.realval, self.imagval)
 
+    def descr_complex(self, space):
+        """Convert this value to exact type complex."""
+        return W_ComplexObject(self.realval, self.imagval)
+
     def descr_abs(self, space):
         try:
             return space.newfloat(math.hypot(self.realval, self.imagval))
@@ -476,19 +508,6 @@ class W_ComplexObject(W_Root):
         except ZeroDivisionError as e:
             raise oefmt(space.w_ZeroDivisionError, "complex division by zero")
 
-    def descr_floordiv(self, space, w_rhs):
-        raise oefmt(space.w_TypeError, "can't take floor of complex number.")
-    descr_rfloordiv = func_with_new_name(descr_floordiv, 'descr_rfloordiv')
-
-    def descr_mod(self, space, w_rhs):
-        raise oefmt(space.w_TypeError, "can't mod complex numbers.")
-    descr_rmod = func_with_new_name(descr_mod, 'descr_rmod')
-
-    def descr_divmod(self, space, w_rhs):
-        raise oefmt(space.w_TypeError,
-                    "can't take floor or mod of complex number.")
-    descr_rdivmod = func_with_new_name(descr_divmod, 'descr_rdivmod')
-
     @unwrap_spec(w_third_arg=WrappedDefault(None))
     def descr_pow(self, space, w_exponent, w_third_arg):
         w_exponent = self._to_complex(space, w_exponent)
@@ -543,12 +562,9 @@ This is equivalent to (real + imag*1j) where imag defaults to 0.""",
     imag = complexwprop('imagval',
                         doc="the imaginary part of a complex number"),
     __repr__ = interp2app(W_ComplexObject.descr_repr),
-    __str__ = interp2app(W_ComplexObject.descr_str),
     __hash__ = interp2app(W_ComplexObject.descr_hash),
     __format__ = interp2app(W_ComplexObject.descr_format),
     __bool__ = interp2app(W_ComplexObject.descr_bool),
-    __int__ = interp2app(W_ComplexObject.descr_int),
-    __float__ = interp2app(W_ComplexObject.descr_float),
     __neg__ = interp2app(W_ComplexObject.descr_neg),
     __pos__ = interp2app(W_ComplexObject.descr_pos),
     __abs__ = interp2app(W_ComplexObject.descr_abs),
@@ -568,14 +584,9 @@ This is equivalent to (real + imag*1j) where imag defaults to 0.""",
     __rmul__ = interp2app(W_ComplexObject.descr_rmul),
     __truediv__ = interp2app(W_ComplexObject.descr_truediv),
     __rtruediv__ = interp2app(W_ComplexObject.descr_rtruediv),
-    __floordiv__ = interp2app(W_ComplexObject.descr_floordiv),
-    __rfloordiv__ = interp2app(W_ComplexObject.descr_rfloordiv),
-    __mod__ = interp2app(W_ComplexObject.descr_mod),
-    __rmod__ = interp2app(W_ComplexObject.descr_rmod),
-    __divmod__ = interp2app(W_ComplexObject.descr_divmod),
-    __rdivmod__ = interp2app(W_ComplexObject.descr_rdivmod),
     __pow__ = interp2app(W_ComplexObject.descr_pow),
     __rpow__ = interp2app(W_ComplexObject.descr_rpow),
 
     conjugate = interp2app(W_ComplexObject.descr_conjugate),
+    __complex__ = interp2app(W_ComplexObject.descr_complex),
 )
