@@ -53,11 +53,11 @@ FLAG_DONE      = '\x00'
 TYPE_INT64     = 'I'     # no longer generated
 
 # the following typecodes have been added in version 4.
-TYPE_ASCII                = 'a'   # never generated so far by pypy
-TYPE_ASCII_INTERNED       = 'A'   # never generated so far by pypy
+TYPE_ASCII                = 'a'
+TYPE_ASCII_INTERNED       = 'A'
 TYPE_SMALL_TUPLE          = ')'
-TYPE_SHORT_ASCII          = 'z'   # never generated so far by pypy
-TYPE_SHORT_ASCII_INTERNED = 'Z'   # never generated so far by pypy
+TYPE_SHORT_ASCII          = 'z'
+TYPE_SHORT_ASCII_INTERNED = 'Z'
 
 
 _marshallers = []
@@ -72,10 +72,7 @@ def marshaller(type):
 def unmarshaller(tc, save_ref=False):
     def _decorator(f):
         assert tc < '\x80'
-        _unmarshallers.append((tc, f))
-        if save_ref:
-            tcref = chr(ord(tc) + 0x80)
-            _unmarshallers.append((tcref, f))
+        _unmarshallers.append((tc, f, save_ref))
         return f
     return _decorator
 
@@ -118,9 +115,6 @@ def marshal(space, w_obj, m):
     if typecode != FLAG_DONE:
         m.atom_str(typecode, s.as_str())
 
-def get_unmarshallers():
-    return _unmarshallers
-
 
 @marshaller(W_NoneObject)
 def marshal_none(space, w_none, m):
@@ -201,7 +195,7 @@ def marshal_long(space, w_long, m):
     SHIFT = 15
     MASK = (1 << SHIFT) - 1
     num = space.bigint_w(w_long)
-    sign = num.sign
+    sign = num.get_sign()
     num = num.abs()
     total_length = (num.bit_length() + (SHIFT - 1)) / SHIFT
     m.put_int(total_length * sign)
@@ -384,6 +378,7 @@ def marshal_pycode(space, w_pycode, m):
     # see pypy.interpreter.pycode for the layout
     x = space.interp_w(PyCode, w_pycode)
     m.put_int(x.co_argcount)
+    m.put_int(x.co_posonlyargcount)
     m.put_int(x.co_kwonlyargcount)
     m.put_int(x.co_nlocals)
     m.put_int(x.co_stacksize)
@@ -397,10 +392,11 @@ def marshal_pycode(space, w_pycode, m):
     _marshal_tuple(space, co_varnames_w, m)  # more lists, now of w_unicodes
     _marshal_tuple(space, co_freevars_w, m)
     _marshal_tuple(space, co_cellvars_w, m)
-    _marshal_unicode(space, x.co_filename, m)
+    marshal(space, x.w_filename, m)
     _marshal_unicode(space, x.co_name, m)
+    _marshal_unicode(space, x.co_qualname, m)
     m.put_int(x.co_firstlineno)
-    m.atom_str(TYPE_STRING, x.co_lnotab)
+    m.atom_str(TYPE_STRING, x.co_linetable)
 
 # helper for unmarshalling "tuple of string" objects
 # into rpython-level lists of strings.  Only for code objects.
@@ -423,6 +419,7 @@ def unmarshal_pycode(space, u, tc):
     w_codeobj = objectmodel.instantiate(PyCode)
     u.save_ref(tc, w_codeobj)
     argcount    = u.get_int()
+    posonlyargcount = u.get_int()
     kwonlyargcount = u.get_int()
     nlocals     = u.get_int()
     stacksize   = u.get_int()
@@ -433,19 +430,50 @@ def unmarshal_pycode(space, u, tc):
     varnames    = _unmarshal_strlist(u)
     freevars    = _unmarshal_strlist(u)
     cellvars    = _unmarshal_strlist(u)
-    filename    = space.utf8_0_w(u.load_w_obj())
+    w_fn = u.load_w_obj()
+    filename    = space.bytes0_w(space.fsencode(w_fn))
+
     name        = space.utf8_w(u.load_w_obj())
+    qualname    = space.utf8_w(u.load_w_obj())
     firstlineno = u.get_int()
-    lnotab      = space.bytes_w(u.load_w_obj())
+    position_info = space.bytes_w(u.load_w_obj())
     filename = assert_str0(filename)
     PyCode.__init__(w_codeobj,
-                  space, argcount, kwonlyargcount, nlocals, stacksize, flags,
+                  space, argcount, posonlyargcount, kwonlyargcount, nlocals, stacksize, flags,
                   code, consts_w[:], names, varnames, filename,
-                  name, firstlineno, lnotab, freevars, cellvars)
+                  name, qualname, firstlineno, position_info, freevars, cellvars,
+                  hidden_applevel=u.hidden_applevel)
     return w_codeobj
 
+def _marshal_ascii_unicode(space, s, m, w_unicode, w_interned):
+    # determine typecode
+    if len(s) < 256:
+        is_short = True
+        if w_interned is not None:
+            typecode = TYPE_SHORT_ASCII_INTERNED
+        else:
+            typecode = TYPE_SHORT_ASCII
+    else:
+        is_short = False
+        if w_interned is not None:
+            typecode = TYPE_ASCII_INTERNED
+        else:
+            typecode = TYPE_ASCII
+    # use ref
+    if w_unicode is not None:
+        typecode = write_ref(typecode, w_unicode, m)
+        if typecode == FLAG_DONE:
+            return
+    # write
+    m.start(typecode)
+    if is_short:
+        m.put(chr(len(s)))
+    else:
+        m.put_int(len(s))
+    m.put(s)
 
 def _marshal_unicode(space, s, m, w_unicode=None):
+    from rpython.rlib import rutf8
     if m.version >= 3:
         w_interned = space.get_interned_str(s)
     else:
@@ -455,10 +483,21 @@ def _marshal_unicode(space, s, m, w_unicode=None):
         typecode = TYPE_INTERNED  #   as a key for u.all_refs
     else:
         typecode = TYPE_UNICODE
+    if m.version >= 4:
+        # check whether it's ascii
+        if w_unicode is not None:
+            is_ascii = w_unicode.is_ascii()
+        else:
+            is_ascii = len(s) == rutf8.check_utf8(s, True)
+        if is_ascii:
+            _marshal_ascii_unicode(space, s, m, w_unicode, w_interned)
+            return
+    # general case, write unicode
     if w_unicode is not None:
         typecode = write_ref(typecode, w_unicode, m)
-    if typecode != FLAG_DONE:
-        m.atom_str(typecode, s)
+        if typecode == FLAG_DONE:
+            return
+    m.atom_str(typecode, s)
 
 # surrogate-preserving variants
 _decode_utf8 = unicodehelper.decode_utf8sp
@@ -478,32 +517,32 @@ def unmarshal_interned(space, u, tc):
     w_ret = unmarshal_unicode(space, u, tc)
     return u.space.new_interned_w_str(w_ret)
 
-def _unmarshal_ascii(u, short_length, interned):
+def _unmarshal_ascii(space, u, short_length, interned):
     from rpython.rlib import rutf8
     if short_length:
         lng = ord(u.get1())
     else:
         lng = u.get_lng()
     s = u.get(lng)
-    # Treat each chr as a single codepoint
-    utf8 = ''.join([rutf8.unichr_as_utf8(ord(c), True) for c in s])
-    w_u = u.space.newtext(utf8)
-    if interned:
-        w_u = u.space.new_interned_w_str(w_u)
+    try:
+        rutf8.check_ascii(s)
+    except rutf8.CheckError:
+        raise oefmt(space.w_ValueError, "bad marshal data (string is not ascii)")
+    w_u = u.space.newtext(s, len(s)) # ascii is valid utf-8
     return w_u
 
-@unmarshaller(TYPE_ASCII)    # nb. never generated by pypy so far
+@unmarshaller(TYPE_ASCII)
 def unmarshal_ascii(space, u, tc):
-    return _unmarshal_ascii(u, False, False)
+    return _unmarshal_ascii(space, u, False, False)
 @unmarshaller(TYPE_ASCII_INTERNED)
 def unmarshal_ascii(space, u, tc):
-    return _unmarshal_ascii(u, False, True)
+    return _unmarshal_ascii(space, u, False, True)
 @unmarshaller(TYPE_SHORT_ASCII)
 def unmarshal_ascii(space, u, tc):
-    return _unmarshal_ascii(u, True, False)
+    return _unmarshal_ascii(space, u, True, False)
 @unmarshaller(TYPE_SHORT_ASCII_INTERNED)
 def unmarshal_ascii(space, u, tc):
-    return _unmarshal_ascii(u, True, True)
+    return _unmarshal_ascii(space, u, True, True)
 
 
 @marshaller(W_SetObject)
@@ -554,3 +593,4 @@ def unmarshal_ref(space, u, tc):
 
 
 _marshallers_unroll = unrolling_iterable(_marshallers)
+_unmarshallers_unroll = unrolling_iterable(_unmarshallers)
