@@ -12,10 +12,7 @@ from rpython.rlib.rarithmetic import r_uint, intmask
 from pypy.interpreter.signature import Signature
 
 @specialize.argtype(1)
-@jit.look_inside_iff(lambda space, s, start, end:
-       jit.isconstant(s) and
-       jit.isconstant(start) and
-       jit.isconstant(end))
+@jit.elidable
 def _parse_int(space, s, start, end):
     """Parse a number and check for overflows"""
     result = 0
@@ -91,12 +88,12 @@ def make_template_formatting_class(for_unicode):
                     markup_follows = True
                     if c == "}":
                         if at_end or s[i] != "}":
-                            raise oefmt(space.w_ValueError, "Single '}'")
+                            raise oefmt(space.w_ValueError, "Single '}' encountered in format string")
                         i += 1
                         markup_follows = False
                     if c == "{":
                         if at_end:
-                            raise oefmt(space.w_ValueError, "Single '{'")
+                            raise oefmt(space.w_ValueError, "Single '{' encountered in format string")
                         if s[i] == "{":
                             i += 1
                             markup_follows = False
@@ -117,6 +114,7 @@ def make_template_formatting_class(for_unicode):
                     nested = 1
                     field_start = i
                     recursive = False
+                    in_second_part = False
                     while i < end:
                         c = s[i]
                         if c == "{":
@@ -126,14 +124,16 @@ def make_template_formatting_class(for_unicode):
                             nested -= 1
                             if not nested:
                                 break
-                        elif c == "[":
+                        elif c == "[" and not in_second_part:
                             i += 1
                             while i < end and s[i] != "]":
                                 i += 1
                             continue
+                        elif c == ':' or c == '!':
+                            in_second_part = True
                         i += 1
                     if nested:
-                        raise oefmt(space.w_ValueError, "Unmatched '{'")
+                        raise oefmt(space.w_ValueError, "expected '}' before end of string")
                     rendered = self._render_field(field_start, i, recursive, level)
                     out.append(rendered)
                     i += 1
@@ -156,13 +156,14 @@ def make_template_formatting_class(for_unicode):
                         i += 1
                         if i == end:
                             raise oefmt(self.space.w_ValueError,
-                                        "expected conversion")
+                                        "end of string while looking for "
+                                        "conversion specifier")
                         conversion = s[i]
                         i += 1
                         if i < end:
                             if s[i] != ':':
                                 raise oefmt(self.space.w_ValueError,
-                                            "expected ':' after format "
+                                            "expected ':' after conversion "
                                             "specifier")
                             i += 1
                     else:
@@ -256,8 +257,8 @@ def make_template_formatting_class(for_unicode):
                     if w_obj is not None:
                         w_obj = space.getattr(w_obj, w_attr)
                     else:
-                        self.parser_list_w.append(space.newtuple([
-                            space.w_True, w_attr]))
+                        self.parser_list_w.append(space.newtuple2(
+                            space.w_True, w_attr))
                 elif c == "[":
                     got_bracket = False
                     i += 1
@@ -279,8 +280,8 @@ def make_template_formatting_class(for_unicode):
                     if w_obj is not None:
                         w_obj = space.getitem(w_obj, w_item)
                     else:
-                        self.parser_list_w.append(space.newtuple([
-                            space.w_False, w_item]))
+                        self.parser_list_w.append(space.newtuple2(
+                            space.w_False, w_item))
                 else:
                     raise oefmt(space.w_ValueError,
                                 "Only '[' and '.' may follow ']'")
@@ -310,8 +311,8 @@ def make_template_formatting_class(for_unicode):
             self.parser_list_w = []
             self._resolve_lookups(None, name, i, end)
             #
-            return space.newtuple([w_first,
-                                   space.iter(space.newlist(self.parser_list_w))])
+            return space.newtuple2(w_first,
+                                   space.iter(space.newlist(self.parser_list_w)))
 
         def _convert(self, w_obj, conversion):
             space = self.space
@@ -437,6 +438,7 @@ def make_formatting_class(for_unicode):
         def __init__(self, space, spec):
             self.space = space
             self.spec = spec
+            self.w_obj = space.w_None
 
         def _is_alignment(self, c):
             return (c == "<" or
@@ -455,6 +457,7 @@ def make_formatting_class(for_unicode):
             self._align = default_align
             self._alternate = False
             self._sign = "\0"
+            self._zeros = False # do we have the 3.11 z specifier
             self._thousands_sep = "\0"
             self._precision = -1
             the_type = default_type
@@ -483,12 +486,15 @@ def make_formatting_class(for_unicode):
             if length - i >= 1 and self._is_sign(spec[i]):
                 self._sign = spec[i]
                 i += 1
+            if length - i >= 1 and spec[i] == 'z':
+                self._zeros = True
+                i += 1
             if length - i >= 1 and spec[i] == "#":
                 self._alternate = True
                 i += 1
             if not got_fill_char and length - i >= 1 and spec[i] == "0":
                 self._fill_char = self._lit("0")[0]
-                if not got_align:
+                if not got_align and default_align == '>':
                     self._align = "="
                 i += 1
             self._width, i = _parse_int(self.space, spec, i, length)
@@ -510,7 +516,7 @@ def make_formatting_class(for_unicode):
                 if self._precision == -1:
                     raise oefmt(space.w_ValueError, "no precision given")
             if length - i > 1:
-                raise oefmt(space.w_ValueError, "invalid format spec")
+                raise oefmt(space.w_ValueError, "Invalid format specifier '%8' for object of type '%T'", spec, self.w_obj)
             if length - i == 1:
                 presentation_type = spec[i]
                 if self.is_unicode:
@@ -591,7 +597,7 @@ def make_formatting_class(for_unicode):
 
         def _unknown_presentation(self, w_val):
             raise oefmt(self.space.w_ValueError,
-                        "Unknown format code %s for object of type '%T'", self._type, w_val)
+                        "Unknown format code '%s' for object of type '%T'", self._type, w_val)
 
         def format_string(self, w_string):
             space = self.space
@@ -603,6 +609,9 @@ def make_formatting_class(for_unicode):
             if self._type != "s":
                 self._unknown_presentation(w_string)
             if self._sign != "\0":
+                if self._sign == " ":
+                    raise oefmt(space.w_ValueError,
+                                "Space not allowed in string format specifier")
                 raise oefmt(space.w_ValueError,
                             "Sign not allowed in string format specifier")
             if self._alternate:
@@ -613,12 +622,20 @@ def make_formatting_class(for_unicode):
                 raise oefmt(space.w_ValueError,
                             "'=' alignment not allowed in string format "
                             "specifier")
+            if self._zeros:
+                raise oefmt(space.w_ValueError, "Negative zero coercion (z) not allowed")
             length = space.len_w(w_string)
             precision = self._precision
             if precision != -1 and length >= precision:
                 assert precision >= 0
                 length = precision
-                string = string[:precision]
+                if for_unicode:
+                    w_slice = space.newslice(
+                        space.newint(0), space.newint(precision), space.w_None)
+                    w_string = space.getitem(w_string, w_slice)
+                    string = space.utf8_w(w_string)
+                else:
+                    string = string[:precision]
             self._calc_padding(string, length)
             return self.wrap(self._pad(string))
 
@@ -683,8 +700,12 @@ def make_formatting_class(for_unicode):
                             spec.n_remainder) # Not padding or digits
             if self._fill_char == "0" and self._align == "=":
                 spec.n_min_width = self._width - extra_length
-            if self._loc_thousands:
-                self._group_digits(spec, digits[to_number:])
+            if self._loc_thousands and spec.n_digits:
+                self._grouped_digits = self._group_digits(digits[to_number:],
+                                                          spec.n_min_width,
+                                                          spec.n_digits,
+                                                          self._loc_grouping,
+                                                          self._loc_thousands)
                 n_grouped_digits = len(self._grouped_digits)
             else:
                 n_grouped_digits = spec.n_digits
@@ -707,7 +728,9 @@ def make_formatting_class(for_unicode):
                            spec.n_decimal + spec.n_remainder + spec.n_rpadding
             return spec
 
-        def _fill_digits(self, buf, digits, d_state, n_chars, n_zeros,
+        @staticmethod
+        @jit.elidable
+        def _fill_digits(buf, digits, d_state, n_chars, n_zeros,
                          thousands_sep):
             if thousands_sep:
                 for c in thousands_sep:
@@ -717,13 +740,13 @@ def make_formatting_class(for_unicode):
             for i in range(n_zeros):
                 buf.append("0")
 
-        def _group_digits(self, spec, digits):
+        @staticmethod
+        @jit.elidable
+        def _group_digits(digits, min_width, left, loc_grouping, loc_thousands):
             buf = []
-            grouping = self._loc_grouping
-            min_width = spec.n_min_width
+            grouping = loc_grouping
             grouping_state = 0
-            left = spec.n_digits
-            n_ts = len(self._loc_thousands)
+            n_ts = len(loc_thousands)
             need_separator = False
             done = False
             previous = 0
@@ -741,8 +764,8 @@ def make_formatting_class(for_unicode):
                 final_grouping = min(group, max(left, max(min_width, 1)))
                 n_zeros = max(0, final_grouping - left)
                 n_chars = max(0, min(left, final_grouping))
-                ts = self._loc_thousands if need_separator else None
-                self._fill_digits(buf, digits, left, n_chars, n_zeros, ts)
+                ts = loc_thousands if need_separator else None
+                Formatter._fill_digits(buf, digits, left, n_chars, n_zeros, ts)
                 need_separator = True
                 left -= n_chars
                 min_width -= final_grouping
@@ -754,20 +777,13 @@ def make_formatting_class(for_unicode):
                 group = max(max(left, min_width), 1)
                 n_zeros = max(0, group - left)
                 n_chars = max(0, min(left, group))
-                ts = self._loc_thousands if need_separator else None
-                self._fill_digits(buf, digits, left, n_chars, n_zeros, ts)
+                ts = loc_thousands if need_separator else None
+                Formatter._fill_digits(buf, digits, left, n_chars, n_zeros, ts)
             buf.reverse()
-            self._grouped_digits = "".join(buf)
-
-        def _upcase_string(self, s):
-            buf = []
-            for c in s:
-                index = ord(c)
-                if ord("a") <= index <= ord("z"):
-                    c = chr(index - 32)
-                buf.append(c)
             return "".join(buf)
 
+        def _upcase_string(self, s):
+            return s.upper()
 
         def _fill_number(self, spec, num, to_digits, to_prefix, fill_char,
                          to_remainder, upper, grouped_digits=None):
@@ -799,7 +815,7 @@ def make_formatting_class(for_unicode):
                     digits = self._upcase_string(digits)
                 out.append(digits)
             if spec.n_decimal:
-                out.append(self._lit(".")[0])
+                out.append(self._lit(self._loc_dec)[0])
             if spec.n_remainder:
                 out.append(num[to_remainder:])
             if spec.n_rpadding:
@@ -811,13 +827,17 @@ def make_formatting_class(for_unicode):
             space = self.space
             if self._precision != -1:
                 raise oefmt(space.w_ValueError,
-                            "precision not allowed in integer type")
+                            "Precision not allowed in integer format specifier")
+            if self.spec == 'z':
+                raise oefmt(space.w_ValueError,
+                        "Negative zero coercion (z) not allowed in integer"
+                        " format specifier")
             sign_char = "\0"
             tp = self._type
             if tp == "c":
                 if self._sign != "\0":
                     raise oefmt(space.w_ValueError,
-                                "sign not allowed with 'c' presentation type")
+                                "Sign not allowed with integer format specifier 'c'")
                 if self._alternate:
                     raise oefmt(space.w_ValueError,
                                 "Alternate form (#) not allowed "
@@ -888,12 +908,11 @@ def make_formatting_class(for_unicode):
                 return rutf8.decode_latin_1(as_str)
             return as_str
 
-        def _int_to_base(self, base, value):
+        @staticmethod
+        @jit.elidable
+        def _int_to_base(base, value):
             if base == 10:
-                s = str(value)
-                if self.is_unicode:
-                    return rutf8.decode_latin_1(s)
-                return s
+                return str(value)
             # This part is slow.
             negative = value < 0
             base = r_uint(base)
@@ -949,6 +968,8 @@ def make_formatting_class(for_unicode):
                 tp == "x" or
                 tp == "X" or
                 tp == "n"):
+                if self._zeros:
+                    raise oefmt(space.w_ValueError, "Negative zero coercion (z) not allowed")
                 return self._format_int_or_long(w_num, kind)
             elif (tp == "e" or
                   tp == "E" or
@@ -982,6 +1003,8 @@ def make_formatting_class(for_unicode):
             default_precision = 6
             if self._alternate:
                flags |= rfloat.DTSF_ALT
+            if self._zeros:
+                flags |= rfloat.DTSF_NO_NEG_0
 
             tp = self._type
             self._get_locale(tp)
@@ -1063,13 +1086,15 @@ def make_formatting_class(for_unicode):
                             "specifier")
             if self._alternate:
                 flags |= rfloat.DTSF_ALT
+            if self._zeros:
+                flags |= rfloat.DTSF_NO_NEG_0
 
             skip_re = 0
             add_parens = 0
             if tp == "\0":
                 #should mirror str() output
-                tp = "g"
-                default_precision = 12
+                tp = "r"
+                default_precision = 0
                 #test if real part is non-zero
                 if (w_complex.realval == 0 and
                     math.copysign(1., w_complex.realval) == 1.):
@@ -1084,6 +1109,8 @@ def make_formatting_class(for_unicode):
             #check if precision not set
             if self._precision == -1:
                 self._precision = default_precision
+            elif tp == "r":
+                tp = "g"
 
             #in CPython it's named 're' - clashes with re module
             re_num, special = rfloat.double_to_string(w_complex.realval, tp, self._precision, flags)
@@ -1209,6 +1236,7 @@ unicode_formatter = make_formatting_class(for_unicode=True)
 
 
 @specialize.arg(2)
-def run_formatter(space, w_format_spec, meth, *args):
+def run_formatter(space, w_format_spec, meth, w_obj, *args):
     formatter = unicode_formatter(space, space.utf8_w(w_format_spec))
-    return getattr(formatter, meth)(*args)
+    formatter.w_obj = w_obj
+    return getattr(formatter, meth)(w_obj, *args)
