@@ -80,8 +80,7 @@ def test_div():
 
     # Just for fun.
     for i in range(100):
-        check_div(complex(random(), random()),
-                       complex(random(), random()))
+        check_div(complex(random(), random()), complex(random(), random()))
 
     raises(ZeroDivisionError, complex.__truediv__, 1+1j, 0+0j)
     # FIXME: The following currently crashes on Alpha
@@ -92,7 +91,8 @@ def test_truediv():
     raises(ZeroDivisionError, complex.__truediv__, 1+1j, 0+0j)
 
 def test_floordiv():
-    raises(TypeError, "3+0j // 0+0j")
+    with raises(TypeError):
+        3+0j // 0+0j
 
 def test_convert():
     exc = raises(TypeError, complex.__int__, 3j)
@@ -137,7 +137,8 @@ def test_richcompare_boundaries():
 
 def test_mod():
     a = 3.33+4.43j
-    raises(TypeError, "a % a")
+    with raises(TypeError):
+        a % a
 
 def test_divmod():
     raises(TypeError, divmod, 1+1j, 0+0j)
@@ -157,8 +158,10 @@ def test_pow():
     assert 3j ** 0j == 1
     assert 3j ** 0 == 1
 
-    raises(ZeroDivisionError, "0j ** a")
-    raises(ZeroDivisionError, "0j ** (3-2j)")
+    with raises(ZeroDivisionError):
+        0j ** a
+    with raises(ZeroDivisionError):
+        0j ** (3-2j)
 
     # The following is used to exercise certain code paths
     assert a ** 105 == a ** 105
@@ -171,11 +174,14 @@ def test_pow():
     b = 5.1+2.3j
     raises(ValueError, pow, a, b, 0)
 
-    b = complex(float('inf'), 0.0) ** complex(10., 3.)
-    assert repr(b) == "(nan+nanj)"
+    raises(ZeroDivisionError, pow, complex(float('inf'), 0.0), complex(10., 3.))
+
+    # issue 3944
+    b = pow(1, -1.j)
+    assert repr(b.imag) == "-0.0"
+
 
 def test_boolcontext():
-    from random import random
     for i in range(100):
         assert complex(random() + 1e-6, random() + 1e-6)
     assert not complex(0.0, 0.0)
@@ -253,7 +259,8 @@ def test_constructor():
 
     # SF bug 543840:  complex(string) accepts strings with \0
     # Fixed in 2.3.
-    raises(ValueError, complex, '1+1j\0j')
+    with raises(ValueError):
+        complex("1+1j\0j")
 
     raises(TypeError, int, 5+3j)
     raises(TypeError, float, 5+3j)
@@ -338,11 +345,12 @@ def test_constructor_bad_error_message():
     assert str(err) == "complex() second argument must be a number, not 'dict'"
 
 def test_error_messages():
-    err = raises(ZeroDivisionError, "1+1j / 0").value
-    assert str(err) == "complex division by zero"
-    err = raises(TypeError, "1+1j // 0").value
-    assert str(err) == "can't take floor of complex number."
-
+    with raises(ZeroDivisionError) as err:
+        1+1j / 0
+    assert str(err.value) == "complex division by zero"
+    with raises(TypeError) as err:
+        1+1j // 0
+    assert str(err.value) == "unsupported operand type(s) for //: 'complex' and 'int'"
 
 def test_hash():
     for x in range(-30, 30):
@@ -411,15 +419,14 @@ def test_repr_roundtrip():
             z = complex(x, y)
             roundtrip = eval(repr(z))
             # adding 0.0 has no effect beside changing -0.0 to 0.0
-            floats_identical(0.0 + z.real,
-                                  0.0 + roundtrip.real)
-            floats_identical(0.0 + z.imag,
-                                  0.0 + roundtrip.imag)
+            floats_identical(0.0 + z.real, 0.0 + roundtrip.real)
+            floats_identical(0.0 + z.imag, 0.0 + roundtrip.imag)
 
 def test_neg():
     assert -(1+6j) == -1-6j
 
 def test_file():
+    import os
     import tempfile
 
     a = 3.33+4.43j
@@ -443,6 +450,7 @@ def test_file():
             pass
 
 def test_convert():
+    import warnings
     raises(TypeError, int, 1+1j)
     raises(TypeError, float, 1+1j)
 
@@ -458,7 +466,32 @@ def test_convert():
             return complex.__new__(self, 2*value)
         def __complex__(self):
             return self
-    assert complex(complex1(1j)) == 2j
+    with warnings.catch_warnings(record=True) as log:
+        warnings.simplefilter("always", DeprecationWarning)
+        assert complex(complex1(1j)) == 2j
+        assert len(log) == 1
+        assert log[0].category == DeprecationWarning
+
+    class complex1b(complex):
+        """Test usage of a complex subclass without __complex__() method"""
+        def __new__(self, value=0j):
+            return complex.__new__(self, 2*value)
+    with warnings.catch_warnings(record=True) as log:
+        warnings.simplefilter("always", DeprecationWarning)
+        assert complex(complex1b(1j)) == 2j
+        assert len(log) == 0
+
+    class complex1_proxy:
+        """Test usage of __complex__() without subclassing complex"""
+        def __init__(self, value=0j):
+            self.value = value
+        def __complex__(self):
+            return complex1(self.value)
+    with warnings.catch_warnings(record=True) as log:
+        warnings.simplefilter("always", DeprecationWarning)
+        assert complex(complex1_proxy(1j)) == 2j
+        assert len(log) == 1
+        assert log[0].category == DeprecationWarning
 
     class complex2(complex):
         """Make sure that __complex__() calls fail if anything other than a
@@ -550,10 +583,12 @@ def test_format():
     assert format(1.5+0.5j, '#f') == '1.500000+0.500000j'
 
     # zero padding is invalid
-    raises(ValueError, (1.5+0.5j).__format__, '010f')
+    with raises(ValueError):
+        (1.5+0.5j).__format__("010f")
 
     # '=' alignment is invalid
-    raises(ValueError, (1.5+3j).__format__, '=20')
+    with raises(ValueError):
+        (1.5+3j).__format__("=20")
 
     # integer presentation types are an error
     for t in 'bcdoxX%':
@@ -584,6 +619,10 @@ def test_format():
     assert '{0:f}'.format(complex(NAN, NAN)) == 'nan+nanj'
     assert '{0:F}'.format(complex(NAN, NAN)) == 'NAN+NANj'
 
+def test_str_zeros_bug():
+    assert str(1000000000001j) == '1000000000001j'
+    assert complex.__str__ is object.__str__
+
 def test_complex_two_arguments():
     raises(TypeError, complex, 5, None)
 
@@ -614,3 +653,30 @@ def test_int_override():
 
     c = MyComplex(0.j)
     assert int(c) == 42
+
+def test_complex_constructor_calls_index():
+    class A:
+        def __init__(self, val):
+            self.val = val
+        def __index__(self):
+            return self.val
+    assert complex(A(1), A(2)) == (1.0+2.0j)
+
+def test_parse_complex_whitespace_bug():
+    c = complex('\t( -1.23+4.5J )\n\r\v ')
+    assert c == -1.23+4.5j
+
+def test_has_complex_special_method():
+    x = 1+1j
+    assert x.__complex__() is x
+    class C(complex): pass
+    x = C(1+1j)
+    assert x.__complex__() == x
+    assert x.__complex__().__class__ is complex
+
+    class C(complex):
+        def __complex__(self):
+            return -10-10j
+
+    assert complex(C(1 + 1j)) == -10-10j
+
