@@ -59,6 +59,33 @@ class TestTypeObject:
             space.warn = prev_warn
         assert len(warnings) == 2
 
+    def test_setattr_if_not_from_object(self):
+        space = self.space
+        w_A, w_B = space.unpackiterable(space.appexec([], """():
+            class A(object):
+                pass
+            class B(object):
+                def __setattr__(self, obj, name, value):
+                    pass
+            return A, B
+        """))
+        assert w_A.uses_object_setattr is False
+        w_res = w_A.setattr_if_not_from_object()
+        assert w_res is None
+        assert w_A.uses_object_setattr is True
+
+        assert w_B.uses_object_setattr is False
+        w_res = w_B.setattr_if_not_from_object()
+        assert w_res is not None
+        assert w_B.uses_object_setattr is False
+
+        # check invalidation
+        space.setattr(w_A, space.newtext("__setattr__"), space.newint(1))
+        assert w_A.uses_object_setattr is False
+        w_res = w_A.setattr_if_not_from_object()
+        assert w_res is not None
+        assert w_A.uses_object_setattr is False
+
 
 class AppTestTypeObject:
     def test_abstract_methods(self):
@@ -93,6 +120,16 @@ class AppTestTypeObject:
         x = X()
         exc = raises(AttributeError, "x.a")
         assert str(exc.value) == "'X' object has no attribute 'a'"
+        assert exc.value.name == "a"
+        exc = raises(AttributeError, "X.a")
+        assert str(exc.value) == "type object 'X' has no attribute 'a'"
+        assert exc.value.name == "a"
+        exc = raises(AttributeError, "del x.a")
+        assert str(exc.value) == "'X' object has no attribute 'a'"
+        assert exc.value.name == "a"
+        exc = raises(AttributeError, "del X.a")
+        assert str(exc.value) == "type object 'X' has no attribute 'a'"
+        assert exc.value.name == "a"
 
     def test_call_type(self):
         assert type(42) is int
@@ -517,6 +554,32 @@ class AppTestTypeObject:
         assert type.mro(B_mro) == [B_mro, A_mro, object]
         """
 
+    def test_mro_mutation_interaction_bug(self):
+        """
+        class Base(object):
+            value = 1
+
+        class Meta(type):
+            def mro(cls):
+                return (cls, Base, object)
+
+        class WeirdClass(metaclass=Meta):
+            pass
+
+        assert Base.value == 1
+        assert WeirdClass.value == 1
+
+        Base.value = 2
+        assert Base.value == 2
+        assert WeirdClass.value == 2
+
+        Base.value = 3
+        assert Base.value == 3
+        assert WeirdClass.value == 3
+
+        assert len(Base.__subclasses__()) == 0
+        """
+
     def test_abstract_mro(self):
         """
         class A1:    # in py3k is a new-style class
@@ -533,45 +596,6 @@ class AppTestTypeObject:
         assert E1.__mro__ == (E1, D1, B1, C1, A1, object)
         """
 
-    def test_nodoc(self):
-        class NoDoc(object):
-            pass
-
-        try:
-            assert NoDoc.__doc__ == None
-        except AttributeError:
-            raise AssertionError("__doc__ missing!")
-
-    def test_explicitdoc(self):
-        class ExplicitDoc(object):
-            __doc__ = 'foo'
-
-        assert ExplicitDoc.__doc__ == 'foo'
-
-    def test_implicitdoc(self):
-        class ImplicitDoc(object):
-            "foo"
-
-        assert ImplicitDoc.__doc__ == 'foo'
-
-    def test_set_doc(self):
-        class X:
-            "elephant"
-        X.__doc__ = "banana"
-        assert X.__doc__ == "banana"
-        raises(TypeError, lambda:
-               type(list).__dict__["__doc__"].__set__(list, "blah"))
-        raises((AttributeError, TypeError), lambda:
-               type(X).__dict__["__doc__"].__delete__(X))
-        assert X.__doc__ == "banana"
-
-    def test_text_signature(self):
-        assert object.__text_signature__ == '()'
-
-        class A:
-            pass
-        assert A.__text_signature__ is None
-
     def test_metaclass_conflict(self):
         """
         class T1(type):
@@ -964,8 +988,9 @@ class AppTestTypeObject:
         assert Abc.__name__ == 'Def'
         raises(TypeError, "Abc.__name__ = 42")
         raises(TypeError, "Abc.__name__ = b'A'")
+        raises(UnicodeEncodeError, "Abc.__name__ = 'A\udcdcB'")
         for v, err in [('G\x00hi', "type name must not contain null characters"),
-                       ('A\udcdcB', "surrogates not allowed")]:
+                       ]:
             try:
                 Abc.__name__ = v
             except ValueError as e:
@@ -1020,6 +1045,23 @@ class AppTestTypeObject:
         assert not A == B
         assert not A != A
 
+    def test_compare_overridden(self):
+        """
+        class AMeta(type):
+            def __eq__(self, other):
+                return True
+
+        class A(metaclass=AMeta):
+            pass
+
+        assert (A == 1) == True
+        assert (1 == A) == True
+
+        assert (A != 1) == False
+        assert (1 != A) == False
+        """
+
+
     def test_class_variations(self):
         class A(object):
             pass
@@ -1442,6 +1484,13 @@ class AppTestTypeObject:
         assert WithMetaclass[int] == "Metaclass[int]"
         """
 
+    def test_class_getitem_None(self):
+        class DontWantToBeGeneric(tuple):
+            __class_getitem__ = None
+        with raises(TypeError) as info:
+            DontWantToBeGeneric[int]
+        assert "DontWantToBeGeneric" in str(info.value)
+
     def test_mro_entries(self):
         """
         class BaseA: pass
@@ -1498,6 +1547,108 @@ class AppTestTypeObject:
         assert not hasattr(TestNoOrigBases, '__orig_bases__')
         """
 
+    def test_mro_entries_bug(self):
+        class A:
+            pass
+
+        def f(): pass
+        f.__mro_entries__ = lambda bases: (A, )
+        class BuggyMcBugFace(f):
+            pass
+        assert BuggyMcBugFace.__base__ is A
+
+    def test_mro_entries_type_ignored(self):
+        class A:
+            def __mro_entries__(self, bases):
+                assert 0
+        class B(A):
+            pass
+        assert B.__base__ is A
+
+    def test_classcell_missing(self):
+        """
+        # Some metaclasses may not pass the original namespace to type.__new__
+        # We test that case here by forcibly deleting __classcell__
+        class Meta(type):
+            def __new__(cls, name, bases, namespace):
+                namespace.pop('__classcell__', None)
+                return super().__new__(cls, name, bases, namespace)
+
+        # The default case should continue to work without any errors
+        class WithoutClassRef(metaclass=Meta):
+            pass
+
+        # With zero-arg super() or an explicit __class__ reference, we expect
+        # __build_class__ to raise a RuntimeError complaining that
+        # __class__ was not set, and asking if __classcell__ was propagated
+        # to type.__new__.
+        with raises(RuntimeError):
+            class WithClassRef(metaclass=Meta):
+                def f(self):
+                    return __class__
+        """
+
+    def test_class_getitem(self):
+        ga = type[int]
+        assert ga.__origin__ is type
+        assert ga.__args__ == (int, )
+
+    def test_or_types_operator(self):
+        """
+        class Example:
+            pass
+
+        assert (int | str)
+        assert (int | int | str) == int | str
+        assert (int | list) != int | str
+        assert (str | int) == int | str
+        assert (int | None)
+        assert (None | int) == int | None
+        assert (int | str | list)
+        assert (int | (str | list)) == (int | str) | list
+        assert (list[int] | str)
+        assert (str | list[int]) == list[int] | str
+        assert (str | float | int | complex | int) == ((int | str) | (float | complex))
+        assert (int | int) == int
+
+        with raises(TypeError):
+            int | 3
+
+        with raises(TypeError):
+            3 | int
+
+        with raises(TypeError):
+            Example() | int
+
+        with raises(TypeError):
+            (int | str) < (int | bool)
+
+        with raises(TypeError):
+            (int | str) <= (int | str)
+        """
+
+    def test_or_type_repr(self):
+        """
+        assert repr(int | None) == "int | None"
+        assert repr(None | int) == "None | int"
+        assert repr(int | list[int]) == "int | list[int]"
+        assert repr(int | str | list[int]) == "int | str | list[int]"
+        assert repr(int | str | int) == "int | str"
+        """
+
+    def test_metaclass_new_args_when_calling_type(self):
+        """
+        class T1(type):
+            def __new__(metacls, cls, bases, body, extraarg=False):
+                assert metacls is T1
+                assert extraarg
+                return super().__new__(metacls, cls, bases, body)
+
+        class D1(metaclass=T1, extraarg=True):
+            pass
+
+        type('D2', (D1, ), {}, extraarg=True)
+        """
 
 class AppTestWithMethodCacheCounter:
     spaceconfig = {"objspace.std.withmethodcachecounter": True}
@@ -1691,7 +1842,7 @@ class AppTestComparesByIdentity:
         assert type.__dict__['__doc__'].__name__ == '__doc__'
 
     def test_type_construct_unicode_surrogate_issue(self):
-        raises(ValueError, type, 'A\udcdcb', (), {})
+        raises(UnicodeEncodeError, type, 'A\udcdcb', (), {})
 
     def test_set_name(self):
         class Descriptor:
@@ -1715,6 +1866,19 @@ class AppTestComparesByIdentity:
         assert str(excinfo.value) == "Error calling __set_name__ on 'Descriptor' instance 'd' in 'A'"
         print(excinfo.value)
 
+    def test_set_name_self(self):
+        # issue 3326: modifying self.__dict__ in self.__set_name__
+        class Descriptor:
+            def __set_name__(self, owner, name):
+                setattr(owner, "attr", self)
+
+        class Foo:
+            desc = Descriptor()
+            desc2 = Descriptor() 
+
+
+        pass # does not crash
+
     def test_type_init_accepts_kwargs(self):
         type.__init__(type, "a", (object, ), {}, a=1)
 
@@ -1764,3 +1928,53 @@ class AppTestComparesByIdentity:
         assert "takes exactly 3 arguments (1 given)" in str(info.value)
         info = raises(TypeError, Meta, 5, 7)
         assert "takes exactly 3 arguments (1 given)" in str(info.value)
+
+    def test_hash_comparison_of_methods(self):
+        def check_ordering(a, b):
+            with raises(TypeError):
+                a < b
+            with raises(TypeError):
+                a > b
+            with raises(TypeError):
+                a <= b
+            with raises(TypeError):
+                a >= b
+
+        class A:
+            def __init__(self, x):
+                self.x = x
+            def f(self):
+                pass
+            def g(self):
+                pass
+            def __eq__(self, other):
+                return True
+            def __hash__(self):
+                raise TypeError
+
+        class B(A):
+            pass
+
+        a1 = A(1)
+        a2 = A(1)
+        assert a1.f == a1.f
+        assert not a1.f != a1.f
+        assert not a1.f == a2.f
+        assert a1.f != a2.f
+        assert not a1.f == a1.g
+        assert a1.f != a1.g
+        check_ordering(a1.f, a1.f)
+        assert hash(a1.f) == hash(a1.f)
+
+        assert not A.f == a1.f
+        assert A.f != a1.f
+        assert not A.f == A.g
+        assert A.f != A.g
+        assert B.f == A.f
+        assert not B.f != A.f
+        check_ordering(A.f, A.f)
+        assert hash(B.f) == hash(A.f)
+
+        # the following triggers a SystemError in 2.4
+        a = A(hash(A.f)^(-1))
+        hash(a.f)
