@@ -95,3 +95,20 @@ app = gateway.applevel(r'''
 
 _classdir = app.interphook('_classdir')
 _objectdir = app.interphook('_objectdir')
+
+def generic_alias_class_getitem(space, w_cls, w_item):
+    "See PEP 585"
+    w_builtins = space.getbuiltinmodule('builtins')
+    w_mod = space.call_method(w_builtins, '__import__',
+            space.newtext("_pypy_generic_alias"))
+    w_GenericAlias = space.getattr(w_mod, space.newtext("GenericAlias"))
+    return space.call_function(w_GenericAlias, w_cls, w_item)
+
+def builtinclass_new_args_check(space, name, w_basetyp, w_subtyp, __args__):
+    if space.is_w(w_basetyp, w_subtyp) or space.is_w(
+            space.getattr(w_basetyp, space.newtext('__init__')),
+            space.getattr(w_subtyp, space.newtext('__init__'))):
+        if __args__.arguments_w:
+            raise oefmt(space.w_TypeError, "%8() expected at most 1 argument, got %d", name, 1 + len(__args__.arguments_w))
+        if __args__.keyword_names_w:
+            raise oefmt(space.w_TypeError, "%8() takes no keyword arguments", name)
