@@ -1,122 +1,111 @@
+from pytest import raises
+import sys
+import operator
+
 def test_special_methods():
-    class OldStyle:
+    class A:
+        def __lt__(self, other):
+            return "lt"
+        def __imul__(self, other):
+            return "imul"
+        def __sub__(self, other):
+            return "sub"
+        def __rsub__(self, other):
+            return "rsub"
+        def __pow__(self, other):
+            return "pow"
+        def __rpow__(self, other):
+            return "rpow"
+        def __neg__(self):
+            return "neg"
+    a = A()
+    assert (a < 5) == "lt"
+    assert (object() > a) == "lt"
+    a1 = a
+    a1 *= 4
+    assert a1 == "imul"
+    assert a - 2 == "sub"
+    assert a - object() == "sub"
+    assert 2 - a == "rsub"
+    assert object() - a == "rsub"
+    assert a ** 2 == "pow"
+    assert a ** object() == "pow"
+    assert 2 ** a == "rpow"
+    assert object() ** a == "rpow"
+    assert -a == "neg"
+
+    class B(A):
+        def __lt__(self, other):
+            return "B's lt"
+        def __imul__(self, other):
+            return "B's imul"
+        def __sub__(self, other):
+            return "B's sub"
+        def __rsub__(self, other):
+            return "B's rsub"
+        def __pow__(self, other):
+            return "B's pow"
+        def __rpow__(self, other):
+            return "B's rpow"
+        def __neg__(self):
+            return "B's neg"
+
+    b = B()
+    assert (a < b) == "lt"
+    assert (b > a) == "lt"
+    b1 = b
+    b1 *= a
+    assert b1 == "B's imul"
+    a1 = a
+    a1 *= b
+    assert a1 == "imul"
+
+    assert a - b == "B's rsub"
+    assert b - a == "B's sub"
+    assert b - b == "B's sub"
+    assert a ** b == "B's rpow"
+    assert b ** a == "B's pow"
+    assert b ** b == "B's pow"
+    assert -b == "B's neg"
+
+    class C(B):
         pass
-    for base in (object, OldStyle,):
-        class A(base):
-            def __lt__(self, other):
-                return "lt"
-            def __imul__(self, other):
-                return "imul"
-            def __sub__(self, other):
-                return "sub"
-            def __rsub__(self, other):
-                return "rsub"
-            def __pow__(self, other):
-                return "pow"
-            def __rpow__(self, other):
-                return "rpow"
-            def __neg__(self):
-                return "neg"
-        a = A()
-        assert (a < 5) == "lt"
-        assert (object() > a) == "lt"
-        a1 = a
-        a1 *= 4
-        assert a1 == "imul"
-        assert a - 2 == "sub"
-        assert a - object() == "sub"
-        assert 2 - a == "rsub"
-        assert object() - a == "rsub"
-        assert a ** 2 == "pow"
-        assert a ** object() == "pow"
-        assert 2 ** a == "rpow"
-        assert object() ** a == "rpow"
-        assert -a == "neg"
+    c = C()
+    assert c - 1 == "B's sub"
+    assert 1 - c == "B's rsub"
+    assert c - b == "B's sub"
+    assert b - c == "B's sub"
 
-        class B(A):
-            def __lt__(self, other):
-                return "B's lt"
-            def __imul__(self, other):
-                return "B's imul"
-            def __sub__(self, other):
-                return "B's sub"
-            def __rsub__(self, other):
-                return "B's rsub"
-            def __pow__(self, other):
-                return "B's pow"
-            def __rpow__(self, other):
-                return "B's rpow"
-            def __neg__(self):
-                return "B's neg"
-
-        b = B()
-        assert (a < b) == "lt"
-        assert (b > a) == "lt"
-        b1 = b
-        b1 *= a
-        assert b1 == "B's imul"
-        a1 = a
-        a1 *= b
-        assert a1 == "imul"
-
-        if base is object:
-            assert a - b == "B's rsub"
-        else:
-            assert a - b == "sub"
-        assert b - a == "B's sub"
-        assert b - b == "B's sub"
-        if base is object:
-            assert a ** b == "B's rpow"
-        else:
-            assert a ** b == "pow"
-        assert b ** a == "B's pow"
-        assert b ** b == "B's pow"
-        assert -b == "B's neg"
-
-        class C(B):
-            pass
-        c = C()
-        assert c - 1 == "B's sub"
-        assert 1 - c == "B's rsub"
-        assert c - b == "B's sub"
-        assert b - c == "B's sub"
-
-        assert c ** 1 == "B's pow"
-        assert 1 ** c == "B's rpow"
-        assert c ** b == "B's pow"
-        assert b ** c == "B's pow"
+    assert c ** 1 == "B's pow"
+    assert 1 ** c == "B's rpow"
+    assert c ** b == "B's pow"
+    assert b ** c == "B's pow"
 
 def test_getslice():
     class Sq(object):
-        def __getslice__(self, start, stop):
-            return (start, stop)
         def __getitem__(self, key):
-            return "booh"
+            return key.start, key.stop
         def __len__(self):
             return 100
 
     sq = Sq()
 
     assert sq[1:3] == (1,3)
-    slice_min, slice_max = sq[:]
-    assert slice_min == 0
-    assert slice_max >= 2**31-1
-    assert sq[1:] == (1, slice_max)
-    assert sq[:3] == (0, 3)
-    assert sq[:] == (0, slice_max)
+    assert sq[:] == (None, None)
+    assert sq[1:] == (1, None)
+    assert sq[:3] == (None, 3)
+    assert sq[:] == (None, None)
     # negative indices
-    assert sq[-1:3] == (99, 3)
-    assert sq[1:-3] == (1, 97)
-    assert sq[-1:-3] == (99, 97)
-    # extended slice syntax always uses __getitem__()
-    assert sq[::] == "booh"
+    assert sq[-1:3] == (-1, 3)
+    assert sq[1:-3] == (1, -3)
+    assert sq[-1:-3] == (-1, -3)
+    # extended slice syntax also uses __getitem__()
+    assert sq[::] == (None, None)
 
 def test_setslice():
     class Sq(object):
-        def __setslice__(self, start, stop, sequence):
-            ops.append((start, stop, sequence))
         def __setitem__(self, key, value):
-            raise AssertionError(key)
+            ops.append((key.start, key.stop, value))
         def __len__(self):
             return 100
 
@@ -126,22 +115,18 @@ def test_setslice():
     sq[12:] = 'world'
     sq[:-1] = 'spam'
     sq[:] = 'egg'
-    slice_max = ops[-1][1]
-    assert slice_max >= 2**31-1
 
     assert ops == [
-        (95, 3,          'hello'),
-        (12, slice_max, 'world'),
-        (0,  99,         'spam'),
-        (0,  slice_max, 'egg'),
+        (-5,    3,   'hello'),
+        (12,   None, 'world'),
+        (None, -1,   'spam'),
+        (None, None, 'egg'),
         ]
 
 def test_delslice():
     class Sq(object):
-        def __delslice__(self, start, stop):
-            ops.append((start, stop))
         def __delitem__(self, key):
-            raise AssertionError(key)
+            ops.append((key.start, key.stop))
         def __len__(self):
             return 100
 
@@ -151,38 +136,32 @@ def test_delslice():
     del sq[-12:]
     del sq[:1]
     del sq[:]
-    slice_max = ops[-1][1]
-    assert slice_max >= 2**31-1
 
     assert ops == [
-        (5,   97),
-        (88,  slice_max),
-        (0,   1),
-        (0,   slice_max),
+        (5,   -3),
+        (-12, None),
+        (None,1),
+        (None,None),
         ]
 
 def test_getslice_nolength():
     class Sq(object):
-        def __getslice__(self, start, stop):
-            return (start, stop)
         def __getitem__(self, key):
-            return "booh"
+            return key.start, key.stop
 
     sq = Sq()
 
     assert sq[1:3] == (1,3)
-    slice_min, slice_max = sq[:]
-    assert slice_min == 0
-    assert slice_max >= 2**31-1
-    assert sq[1:] == (1, slice_max)
-    assert sq[:3] == (0, 3)
-    assert sq[:] == (0, slice_max)
+    assert sq[:] == (None, None)
+    assert sq[1:] == (1, None)
+    assert sq[:3] == (None, 3)
+    assert sq[:] == (None, None)
     # negative indices, but no __len__
     assert sq[-1:3] == (-1, 3)
     assert sq[1:-3] == (1, -3)
     assert sq[-1:-3] == (-1, -3)
-    # extended slice syntax always uses __getitem__()
-    assert sq[::] == "booh"
+    # extended slice syntax also uses __getitem__()
+    assert sq[::] == (None, None)
 
 def test_ipow():
     x = 2
@@ -193,9 +172,10 @@ def test_typechecks():
     class myint(int):
         pass
     class X(object):
-        def __nonzero__(self):
+        def __bool__(self):
             return myint(1)
-    raises(TypeError, "not X()")
+    with raises(TypeError):
+        not X()
 
 def test_string_subclass():
     class S(str):
@@ -211,13 +191,13 @@ def test_string_subclass():
 
 def test_notimplemented():
     #import types
-    import operator
 
     def specialmethod(self, other):
         return NotImplemented
 
     def check(expr, x, y, operator=operator):
-        raises(TypeError, expr)
+        with raises(TypeError):
+            exec(expr, locals={"x":x, "y":y})
 
     for metaclass in [type]:   # [type, types.ClassType]:
         for name, expr, iexpr in [
@@ -235,11 +215,8 @@ def test_notimplemented():
                 ('__and__',      'x & y',                   'x &= y'),
                 ('__or__',       'x | y',                   'x |= y'),
                 ('__xor__',      'x ^ y',                   'x ^= y'),
-                ('__coerce__',   'coerce(x, y)',            None)]:
-            if name == '__coerce__':
-                rname = name
-            else:
-                rname = '__r' + name[2:]
+                ]:
+            rname = '__r' + name[2:]
             A = metaclass('A', (), {name: specialmethod})
             B = metaclass('B', (), {rname: specialmethod})
             a = A()
@@ -269,68 +246,128 @@ def test_string_results():
             return answer * 2
         def __repr__(self):
             return answer * 3
-        def __hex__(self):
-            return answer * 4
-        def __oct__(self):
-            return answer * 5
 
-    for operate, n in [(str, 2), (repr, 3), (hex, 4), (oct, 5)]:
+    for operate, n in [(str, 2), (repr, 3)]:
         answer = "hello"
         assert operate(A()) == "hello" * n
-        if operate not in (hex, oct):
-            answer = u"world"
-            assert operate(A()) == "world" * n
         assert type(operate(A())) is str
         answer = 42
         excinfo = raises(TypeError, operate, A())
         assert "returned non-string (type 'int')" in str(excinfo.value)
 
+def test_string_results_unicode():
+    class A(object):
+        def __str__(self):
+            return 'àèì'
+        def __repr__(self):
+            return 'àèì'
+
+    for operate in (str, repr):
+        x = operate(A())
+        assert x == 'àèì'
+        assert type(x) is str
+
+
+def test_byte_results_unicode():
+    class A(object):
+        def __str__(self):
+            return b'foo'
+        def __repr__(self):
+            return b'bar'
+
+    for operate in (str, repr):
+        raises(TypeError, operate, A())
+
 def test_missing_getattribute():
+    """
     class X(object):
         pass
 
-    class Y(X):
-        class __metaclass__(type):
-            def mro(cls):
-                return [cls, X]
+    class metaclass(type):
+        def mro(cls):
+            return [cls, X]
+    class Y(X, metaclass=metaclass):
+        pass
 
     x = X()
     x.__class__ = Y
     raises(AttributeError, getattr, x, 'a')
+    """
 
-def test_silly_but_consistent_order():
-    # incomparable objects sort by type name :-/
-    class A(object):
-        pass
-    class zz(object):
-        pass
-    assert A() < zz()
-    assert zz() > A()
-    # if in doubt, CPython sorts numbers before non-numbers
-    assert 0 < ()
-    assert 0L < ()
-    assert 0.0 < ()
-    assert 0j < ()
-    assert 0 < []
-    assert 0L < []
-    assert 0.0 < []
-    assert 0j < []
-    assert 0 < A()
-    assert 0L < A()
-    assert 0.0 < A()
-    assert 0j < A()
-    assert 0 < zz()
-    assert 0L < zz()
-    assert 0.0 < zz()
-    assert 0j < zz()
-    # what if the type name is the same... whatever, but
-    # be consistent
-    a1 = A()
-    a2 = A()
+def test_unordeable_types():
     class A(object): pass
-    a3 = A()
-    a4 = A()
-    assert (a1 < a3) == (a1 < a4) == (a2 < a3) == (a2 < a4)
+    class zz(object): pass
+    with raises(TypeError):
+        A() < zz()
+    with raises(TypeError):
+        zz() > A()
+    with raises(TypeError):
+        A() < A()
+    with raises(TypeError):
+        A() < None
+    with raises(TypeError):
+        None < A()
+    with raises(TypeError):
+        0 < ()
+    with raises(TypeError):
+        0.0 < ()
+    with raises(TypeError):
+        0j < ()
+    with raises(TypeError):
+        0 < []
+    with raises(TypeError):
+        0.0 < []
+    with raises(TypeError):
+        0j < []
+    with raises(TypeError):
+        0 < A()
+    with raises(TypeError):
+        0.0 < A()
+    with raises(TypeError):
+        0j < A()
+    with raises(TypeError):
+        0 < zz()
+    with raises(TypeError):
+        0.0 < zz()
+    with raises(TypeError):
+        0j < zz()
+
+def test_correct_order_error_msg():
+    class A(object):
+        def __lt__(self, other):
+            return NotImplemented
+        __gt__ = __lt__
+
+    class B(A):
+        pass
+
+    with raises(TypeError) as e:
+        A() < B()
+    assert str(e.value) == "'<' not supported between instances of 'A' and 'B'"
+
+def test_equality_among_different_types():
+    class A(object): pass
+    class zz(object): pass
+    a = A()
+    assert a == a
+    for x, y in [(A(), A()),
+                 (A(), zz()),
+                 (A(), A()),
+                 (A(), None),
+                 (None, A()),
+                 (0, ()),
+                 (0.0, ()),
+                 (0j, ()),
+                 (0, []),
+                 (0.0, []),
+                 (0j, []),
+                 (0, A()),
+                 (0.0, A()),
+                 (0j, A()),
+                 ]:
+        assert not x == y
+        assert x != y
+
 
 def test_setattrweakref():
     skip("fails, works in cpython")
@@ -360,6 +397,16 @@ def test_subclass_addition():
     assert res1 == res2 == 123
     assert l == [A, B, B, A]
 
+def test__eq__called():
+    l = []
+    class A(object):
+        def __eq__(self, other):
+            l.append((self, other))
+            return True
+    a = A()
+    a == a
+    assert l == [(a, a)]
+
 def test_subclass_comparison():
     # the __eq__ *is* called with reversed arguments
     l = []
@@ -402,7 +449,6 @@ def test_subclass_comparison_more():
     assert l == [B, A, B, A]
 
 def test_rich_comparison():
-    # Old-style
     class A:
         def __init__(self, a):
             self.a = a
@@ -411,32 +457,51 @@ def test_rich_comparison():
     class B:
         def __init__(self, a):
             self.a = a
-    # New-style
-    class C(object):
-        def __init__(self, a):
-            self.a = a
-        def __eq__(self, other):
-            return self.a == other.a
-    class D(object):
-        def __init__(self, a):
-            self.a = a
 
     assert A(1) == B(1)
     assert B(1) == A(1)
-    assert A(1) == C(1)
-    assert C(1) == A(1)
-    assert A(1) == D(1)
-    assert D(1) == A(1)
-    assert C(1) == D(1)
-    assert D(1) == C(1)
     assert not(A(1) == B(2))
     assert not(B(1) == A(2))
-    assert not(A(1) == C(2))
-    assert not(C(1) == A(2))
-    assert not(A(1) == D(2))
-    assert not(D(1) == A(2))
-    assert not(C(1) == D(2))
-    assert not(D(1) == C(2))
+    assert A(1) != B(2)
+    assert B(1) != A(2)
+    assert not(A(1) != B(1))
+    assert not(B(1) != A(1))
+
+def test_ne_high_priority():
+    """object.__ne__() should allow reflected __ne__() to be tried"""
+    calls = []
+    class Left:
+        # Inherits object.__ne__()
+        def __eq__(*args):
+            calls.append('Left.__eq__')
+            return NotImplemented
+    class Right:
+        def __eq__(*args):
+            calls.append('Right.__eq__')
+            return NotImplemented
+        def __ne__(*args):
+            calls.append('Right.__ne__')
+            return NotImplemented
+    Left() != Right()
+    assert calls == ['Left.__eq__', 'Right.__ne__']
+
+def test_ne_low_priority():
+    """object.__ne__() should not invoke reflected __eq__()"""
+    calls = []
+    class Base:
+        # Inherits object.__ne__()
+        def __eq__(*args):
+            calls.append('Base.__eq__')
+            return NotImplemented
+    class Derived(Base):  # Subclassing forces higher priority
+        def __eq__(*args):
+            calls.append('Derived.__eq__')
+            return NotImplemented
+        def __ne__(*args):
+            calls.append('Derived.__ne__')
+            return NotImplemented
+    Base() != Derived()
+    assert calls == ['Derived.__ne__', 'Base.__eq__']
 
 def test_partial_ordering():
     class A(object):
@@ -516,8 +581,18 @@ def test_eq_order():
     assert (D() >  A()) == 'D:A.gt'
     assert (D() >= A()) == 'D:A.ge'
 
+def test_binop_rule():
+    called = []
+    class A:
+        def __eq__(self, other):
+            called.append(self)
+            return NotImplemented
+    a1 = A()
+    a2 = A()
+    a1 == a2
+    assert called == [a1, a2]
+
 def test_addition():
-    # Old-style
     class A:
         def __init__(self, a):
             self.a = a
@@ -527,25 +602,9 @@ def test_addition():
     class B:
         def __init__(self, a):
             self.a = a
-    # New-style
-    class C(object):
-        def __init__(self, a):
-            self.a = a
-        def __add__(self, other):
-            return self.a + other.a
-        __radd__ = __add__
-    class D(object):
-        def __init__(self, a):
-            self.a = a
 
     assert A(1) + B(2) == 3
     assert B(1) + A(2) == 3
-    assert A(1) + C(2) == 3
-    assert C(1) + A(2) == 3
-    assert A(1) + D(2) == 3
-    assert D(1) + A(2) == 3
-    assert C(1) + D(2) == 3
-    assert D(1) + C(2) == 3
 
 def test_mod_failure():
     try:
@@ -561,6 +620,14 @@ def test_invalid_iterator():
             return self
     raises(TypeError, iter, x())
 
+def test_iterator_is_None():
+    class x(object):
+        __iter__ = None
+    with raises(TypeError) as info:
+        iter(x())
+    assert str(info.value) == "'x' object is not iterable"
+
+
 def test_attribute_error():
     class classmethodonly(classmethod):
         def __get__(self, instance, type):
@@ -575,6 +642,18 @@ def test_attribute_error():
 
     raises(AttributeError, lambda: A().a)
 
+def test_attribute_error2():
+    class A(object):
+        def __eq__(self, other):
+            raise AttributeError('doh')
+    raises(AttributeError, operator.eq, A(), A())
+
+    class E(object):
+        @property
+        def __eq__(self):
+            raise AttributeError('doh')
+    assert not (E() == E())
+
 def test_delete_descriptor():
     class Prop(object):
         def __get__(self, obj, cls):
@@ -601,8 +680,8 @@ def test_isinstance_and_issubclass():
             if cls is B:
                 return True
             return False
-    class A:
-        __metaclass__ = Meta
+    A = Meta('A', (), {})  # like "class A(metaclass=Meta)", but
+                           # Python2 cannot parse this
     class B(A):
         pass
     a = A()
@@ -618,29 +697,16 @@ def test_isinstance_and_issubclass():
     assert issubclass(B, B)
     assert issubclass(23, B)
 
-def test_issubclass_and_method():
-    class Meta(type):
-        def __subclasscheck__(cls, sub):
-            if sub is Dict:
-                return True
+def test_rebind_method():
+    # No check is done on a method copied to another type
     class A:
-        __metaclass__ = Meta
         def method(self):
             return 42
     class Dict:
         method = A.method
     assert Dict().method() == 42
 
-def test_truth_of_long():
-    class X(object):
-        def __len__(self): return 1L
-        __nonzero__ = __len__
-    raises(TypeError, bool, X())  # must return bool or int, not long
-    del X.__nonzero__
-    assert X()
-
 def test_len_overflow():
-    import sys
     class X(object):
         def __len__(self):
             return sys.maxsize + 1
@@ -648,17 +714,11 @@ def test_len_overflow():
     raises(OverflowError, bool, X())
 
 def test_len_underflow():
-    import sys
     class X(object):
         def __len__(self):
             return -1
     raises(ValueError, len, X())
     raises(ValueError, bool, X())
-    class Y(object):
-        def __len__(self):
-            return -1L
-    raises(ValueError, len, Y())
-    raises(ValueError, bool, Y())
 
 def test_len_custom__int__():
     class X(object):
@@ -669,14 +729,54 @@ def test_len_custom__int__():
         def __int__(self):
             return self.x
 
-    l = len(X(3.0))
-    assert l == 3 and type(l) is int
-    assert X(3.0)
-    assert not X(0.0)
-    l = len(X(X(2)))
-    assert l == 2 and type(l) is int
-    assert X(X(2))
-    assert not X(X(0))
+    raises(TypeError, len, X(3.0))
+    raises(TypeError, len, X(X(2)))
+    raises(TypeError, bool, X(3.0))
+    raises(TypeError, bool, X(X(2)))
+    raises(OverflowError, len, X(sys.maxsize + 1))
+
+def test_len_index():
+    class Index(object):
+        def __index__(self):
+            return 42
+    class X(object):
+        def __len__(self):
+            return Index()
+    n = len(X())
+    assert type(n) is int
+    assert n == 42
+
+    class BadIndex(object):
+        def __index__(self):
+            return 'foo'
+    class Y(object):
+        def __len__(self):
+            return BadIndex()
+    excinfo = raises(TypeError, len, Y())
+    assert excinfo.value.args[0].startswith("__index__ returned non-")
+
+    class BadIndex2(object):
+        def __index__(self):
+            return 2**100
+    class Z(object):
+        def __len__(self):
+            return BadIndex2()
+    excinfo = raises(OverflowError, len, Z())
+
+def test_sane_len():
+    # this test just tests our assumptions about __len__
+    # this will start failing if __len__ changes assertions
+    for badval in ['illegal', -1, 1 << 32]:
+        class A:
+            def __len__(self):
+                return badval
+        try:
+            bool(A())
+        except (Exception) as e_bool:
+            try:
+                len(A())
+            except (Exception) as e_len:
+                assert str(e_bool) == str(e_len)
 
 def test_bool___contains__():
     class X(object):
@@ -694,25 +794,14 @@ def test_bool___contains__():
     class MyError(Exception):
         pass
     class CannotConvertToBool(object):
-        def __nonzero__(self):
+        def __bool__(self):
             raise MyError
     class X(object):
         def __contains__(self, item):
             return CannotConvertToBool()
     x = X()
-    raises(MyError, "'foo' in x")
-
-def test___cmp___fake_int():
-    class MyInt(object):
-        def __init__(self, x):
-            self.x = x
-        def __int__(self):
-            return self.x
-    class X(object):
-        def __cmp__(self, other):
-            return MyInt(0)
-
-    assert X() == 'hello'
+    with raises(MyError):
+        'foo' in x
 
 def test_sequence_rmul_overrides():
     class oops(object):
@@ -720,13 +809,13 @@ def test_sequence_rmul_overrides():
             return 42
         def __index__(self):
             return 3
-    assert '2' * oops() == 42
+    assert b'2' * oops() == 42
     assert [2] * oops() == 42
     assert (2,) * oops() == 42
     assert u'2' * oops() == 42
-    assert bytearray('2') * oops() == 42
+    assert bytearray(b'2') * oops() == 42
     assert 1000 * oops() == 42
-    assert '2'.__mul__(oops()) == '222'
+    assert b'2'.__mul__(oops()) == b'222'
     x = '2'
     x *= oops()
     assert x == 42
@@ -740,13 +829,13 @@ def test_sequence_rmul_overrides_oldstyle():
             return 42
         def __index__(self):
             return 3
-    assert '2' * oops() == 42
+    assert b'2' * oops() == 42
     assert [2] * oops() == 42
     assert (2,) * oops() == 42
     assert u'2' * oops() == 42
-    assert bytearray('2') * oops() == 42
+    assert bytearray(b'2') * oops() == 42
     assert 1000 * oops() == 42
-    assert '2'.__mul__(oops()) == '222'
+    assert b'2'.__mul__(oops()) == b'222'
 
 def test_sequence_radd_overrides():
     class A1(list):
@@ -770,7 +859,8 @@ def test_data_descriptor_without_delete():
             pass
     class A(object):
         d = D()
-    raises(AttributeError, "del A().d")
+    with raises(AttributeError):
+        del A().d
 
 def test_data_descriptor_without_set():
     class D(object):
@@ -778,9 +868,58 @@ def test_data_descriptor_without_set():
             pass
     class A(object):
         d = D()
-    raises(AttributeError, "A().d = 5")
+    with raises(AttributeError):
+        A().d = 5
 
 def test_not_subscriptable_error_gives_keys():
     d = {'key1': {'key2': {'key3': None}}}
-    excinfo = raises(TypeError, "d['key1']['key2']['key3']['key4']['key5']")
+    with raises(TypeError) as excinfo:
+        d['key1']['key2']['key3']['key4']['key5']
     assert "key4" in str(excinfo.value)
+
+def test_64bit_hash():
+    class BigHash(object):
+        def __hash__(self):
+            return sys.maxsize + 2
+        def __eq__(self, other):
+            return isinstance(other, BigHash)
+    # previously triggered an OverflowError
+    d = {BigHash(): None}
+    assert BigHash() in d
+
+def test_class_getitem():
+    with raises(TypeError) as excinfo:
+        int[int]
+    assert "type 'int' is not subscriptable" in str(excinfo.value)
+
+def test_error_ipow():
+    with raises(TypeError) as info:
+        x = None
+        x **= 2
+    assert "unsupported operand type(s) for **=: 'NoneType' and 'int'" in str(info.value)
+
+def test_error_pow():
+    with raises(TypeError) as info:
+        None ** 2
+    assert "unsupported operand type(s) for ** or pow(): 'NoneType' and 'int'" in str(info.value)
+    with raises(TypeError) as info:
+        pow(None, 2, 2.1)
+    assert "unsupported operand type(s) for pow(): 'NoneType', 'int', 'float'" in str(info.value)
+
+def test_type_subscription_message():
+    with raises(TypeError) as info:
+        Exception[int]
+    assert "type 'Exception' is not subscriptable" in str(info.value)
+
+def test_rpow_not_used_when_third_argument():
+    class RPOW(object):
+        def __pow__(self, other, mod=None):
+            print("__pow__")
+            return 123
+        def __rpow__(self, other, mod=None):
+            print("__rpow__")
+            return 321
+
+    a = RPOW()
+    with raises(TypeError):
+        pow(1, a, a)
