@@ -1,11 +1,7 @@
 """Rewrite assertion AST to produce nice error messages"""
-from __future__ import absolute_import, division, print_function
 from pypy.interpreter.astcompiler import ast
 import itertools
-import sys
 
-def ast_Call(a, b, c):
-    return ast.Call(a, b, c, None, None, -1, -1)
 
 def rewrite_asserts(space, source, filename):
     """Parse the source code and rewrite asserts statements
@@ -52,17 +48,10 @@ binop_map = {
     ast.Is: "is",
     ast.IsNot: "is not",
     ast.In: "in",
-    ast.NotIn: "not in"
+    ast.NotIn: "not in",
+    ast.MatMult: "@",
 }
-# Python 3.5+ compatibility
-try:
-    binop_map[ast.MatMult] = "@"
-except AttributeError:
-    pass
 
-# Python 3.4+ compatibility
-def _NameConstant(c):
-    return b(ast.Name, str(c), ast.Load)
 
 def set_location(node, lineno, col_offset):
     """Set node location information recursively."""
@@ -79,7 +68,7 @@ def set_location(node, lineno, col_offset):
     return node
 
 FIELDS = {name: (fields1 or []) + (fields2 or [])
-        for (name, base, fields1, fields2) in ast.State.AST_TYPES}
+        for (name, base, fields1, fields2, _, _) in ast.State.AST_TYPES}
 
 def iter_fields(node):
     fields = FIELDS[type(node).__name__]
@@ -87,7 +76,7 @@ def iter_fields(node):
         yield name, getattr(node, name)
 
 def b(cls, *args):
-    return cls(*args + (-1, -1))
+    return cls(*args + (-1, -1, -1, -1))
 
 class AssertionRewriter(ast.ASTVisitor):
     """Assertion rewriting implementation.
@@ -148,10 +137,10 @@ class AssertionRewriter(ast.ASTVisitor):
         return getattr(self, "visit_" + type(node).__name__, self.default_visitor)(node)
 
     def Str(self, s):
-        return b(ast.Str, self.space.newtext(s))
+        return b(ast.Constant, self.space.newtext(s), None)
 
     def Num(self, i):
-        return b(ast.Num, self.space.newint(i))
+        return b(ast.Constant, self.space.newint(i), None)
 
     def run(self, mod):
         """Find all assert statements in *mod* and rewrite them."""
@@ -160,12 +149,9 @@ class AssertionRewriter(ast.ASTVisitor):
             return
         # Insert some special imports at the top of the module but after any
         # docstrings and __future__ imports.
-        if sys.version_info[0] >= 3:
-            builtin_name = 'builtins'
-        else:
-            builtin_name = '__builtin__'
-        aliases = [ast.alias(builtin_name, "@py_builtins"),
-                   ast.alias("pytest", "@pytest")]
+        builtin_name = 'builtins'
+        aliases = [ast.alias(builtin_name, "@py_builtins", 0, 0, 0, 0),
+                   ast.alias("pytest", "@pytest", 0, 0, 0, 0)]
         doc = getattr(mod, "docstring", None)
         expect_docstring = doc is None
         if doc is not None and self.is_rewrite_disabled(doc):
@@ -174,8 +160,9 @@ class AssertionRewriter(ast.ASTVisitor):
         lineno = 1
         for item in mod.body:
             if (expect_docstring and isinstance(item, ast.Expr) and
-                    isinstance(item.value, ast.Str)):
-                doc = item.value.s
+                    isinstance(item.value, ast.Constant) and
+                    self.space.isinstance_w(item.value.value, self.space.w_unicode)):
+                doc = item.value.value
                 if self.is_rewrite_disabled(doc):
                     return
                 expect_docstring = False
@@ -186,7 +173,7 @@ class AssertionRewriter(ast.ASTVisitor):
             pos += 1
         else:
             lineno = item.lineno
-        imports = [ast.Import([alias], lineno=lineno, col_offset=0)
+        imports = [ast.Import([alias], lineno=lineno, col_offset=0, end_lineno=lineno, end_col_offset=0)
                    for alias in aliases]
         mod.body[pos:pos] = imports
         # Collect asserts.
@@ -224,7 +211,7 @@ class AssertionRewriter(ast.ASTVisitor):
     def assign(self, expr):
         """Give *expr* a name."""
         name = self.variable()
-        self.statements.append(b(ast.Assign, [b(ast.Name, name, ast.Store)], expr))
+        self.statements.append(b(ast.Assign, [b(ast.Name, name, ast.Store)], expr, None))
         return b(ast.Name, name, ast.Load)
 
     def display(self, expr):
@@ -235,7 +222,7 @@ class AssertionRewriter(ast.ASTVisitor):
         """Call a helper in this module."""
         py_name = b(ast.Name, "@pytest", ast.Load)
         attr = b(ast.Attribute, py_name, "ar_" + name, ast.Load)
-        return ast_Call(attr, list(args), [])
+        return b(ast.Call, attr, list(args), [])
 
     def builtin(self, name):
         """Return the builtin called *name*."""
@@ -285,7 +272,7 @@ class AssertionRewriter(ast.ASTVisitor):
         format_dict = b(ast.Dict, keys, list(current.values()))
         form = b(ast.BinOp, expl_expr, ast.Mod, format_dict)
         name = "@py_format" + str(next(self.variable_counter))
-        self.on_failure.append(b(ast.Assign, [b(ast.Name, name, ast.Store)], form))
+        self.on_failure.append(b(ast.Assign, [b(ast.Name, name, ast.Store)], form, None))
         return b(ast.Name, name, ast.Load)
 
     def default_visitor(self, node):
@@ -325,14 +312,14 @@ class AssertionRewriter(ast.ASTVisitor):
         msg = self.pop_format_context(template)
         fmt = self.helper("format_explanation", msg)
         err_name = b(ast.Name, "AssertionError", ast.Load)
-        exc = ast_Call(err_name, [fmt], [])
-        raise_ = b(ast.Raise, exc, None, None)
+        exc = b(ast.Call, err_name, [fmt], [])
+        raise_ = b(ast.Raise, exc, None)
         body.append(raise_)
         # Clear temporary variables by setting them to None.
         if self.variables:
             variables = [b(ast.Name, name, ast.Store)
                          for name in self.variables]
-            clear = b(ast.Assign, variables, _NameConstant(None))
+            clear = b(ast.Assign, variables, b(ast.Constant, self.space.w_None, None), None)
             self.statements.append(clear)
         # Fix line numbers.
         for stmt in self.statements:
@@ -342,7 +329,7 @@ class AssertionRewriter(ast.ASTVisitor):
     def visit_Name(self, name):
         # Display the repr of the name if it's a local variable or
         # _should_repr_global_name() thinks it's acceptable.
-        locs = ast_Call(self.builtin("locals"), [], [])
+        locs = b(ast.Call, self.builtin("locals"), [], [])
         inlocs = b(ast.Compare, self.Str(name.id), [ast.In], [locs])
         dorepr = self.helper("should_repr_global_name", name)
         test = b(ast.BoolOp, ast.Or, [inlocs, dorepr])
@@ -367,9 +354,9 @@ class AssertionRewriter(ast.ASTVisitor):
                 self.on_failure = fail_inner
             self.push_format_context()
             res, expl = self.visit(v)
-            body.append(b(ast.Assign, [b(ast.Name, res_var, ast.Store)], res))
+            body.append(b(ast.Assign, [b(ast.Name, res_var, ast.Store)], res, None))
             expl_format = self.pop_format_context(self.Str(expl))
-            call = ast_Call(app, [expl_format], [])
+            call = b(ast.Call, app, [expl_format], [])
             self.on_failure.append(b(ast.Expr, call))
             if i < levels:
                 cond = res
@@ -398,7 +385,7 @@ class AssertionRewriter(ast.ASTVisitor):
         res = self.assign(b(ast.BinOp, left_expr, binop.op, right_expr))
         return res, explanation
 
-    def visit_Call_35(self, call):
+    def visit_Call(self, call):
         """
         visit `ast.Call` nodes on Python3.5 and after
         """
@@ -406,20 +393,20 @@ class AssertionRewriter(ast.ASTVisitor):
         arg_expls = []
         new_args = []
         new_kwargs = []
-        for arg in call.args:
+        for arg in call.args or []:
             res, expl = self.visit(arg)
             arg_expls.append(expl)
             new_args.append(res)
-        for keyword in call.keywords:
+        for keyword in call.keywords or []:
             res, expl = self.visit(keyword.value)
-            new_kwargs.append(ast.keyword(keyword.arg, res))
+            new_kwargs.append(b(ast.keyword, keyword.arg, res))
             if keyword.arg:
                 arg_expls.append(keyword.arg + "=" + expl)
             else:  # **args have `arg` keywords with an .arg of None
                 arg_expls.append("**" + expl)
 
         expl = "%s(%s)" % (func_expl, ', '.join(arg_expls))
-        new_call = ast.Call(new_func, new_args, new_kwargs)
+        new_call = b(ast.Call, new_func, new_args, new_kwargs)
         res = self.assign(new_call)
         res_expl = self.explanation_param(self.display(res))
         outer_expl = "%s\n{%s = %s\n}" % (res_expl, res_expl, expl)
@@ -430,45 +417,6 @@ class AssertionRewriter(ast.ASTVisitor):
         res, expl = self.visit(starred.value)
         return starred, '*' + expl
 
-    def visit_Call_legacy(self, call):
-        """
-        visit `ast.Call nodes on 3.4 and below`
-        """
-        new_func, func_expl = self.visit(call.func)
-        arg_expls = []
-        new_args = []
-        new_kwargs = []
-        new_star = new_kwarg = None
-        for arg in call.args or []:
-            res, expl = self.visit(arg)
-            new_args.append(res)
-            arg_expls.append(expl)
-        for keyword in call.keywords or []:
-            res, expl = self.visit(keyword.value)
-            new_kwargs.append(ast.keyword(keyword.arg, res))
-            arg_expls.append(keyword.arg + "=" + expl)
-        if call.starargs:
-            new_star, expl = self.visit(call.starargs)
-            arg_expls.append("*" + expl)
-        if call.kwargs:
-            new_kwarg, expl = self.visit(call.kwargs)
-            arg_expls.append("**" + expl)
-        expl = "%s(%s)" % (func_expl, ', '.join(arg_expls))
-        new_call = b(ast.Call, new_func, new_args, new_kwargs,
-                            new_star, new_kwarg)
-        res = self.assign(new_call)
-        res_expl = self.explanation_param(self.display(res))
-        outer_expl = "%s\n{%s = %s\n}" % (res_expl, res_expl, expl)
-        return res, outer_expl
-
-    # ast.Call signature changed on 3.5,
-    # conditionally change  which methods is named
-    # visit_Call depending on Python version
-    if sys.version_info >= (3, 5):
-        visit_Call = visit_Call_35
-    else:
-        visit_Call = visit_Call_legacy
-
     def visit_Attribute(self, attr):
         if attr.ctx != ast.Load:
             return self.default_visitor(attr)
@@ -501,7 +449,7 @@ class AssertionRewriter(ast.ASTVisitor):
             expl = "%s %s %s" % (left_expl, sym, next_expl)
             expls.append(self.Str(expl))
             res_expr = b(ast.Compare, left_res, [op], [next_res])
-            self.statements.append(b(ast.Assign, [store_names[i]], res_expr))
+            self.statements.append(b(ast.Assign, [store_names[i]], res_expr, None))
             left_res, left_expl = next_res, next_expl
         # Use pytest.assertion.util._reprcompare if that's available.
         expl_call = self.helper("call_reprcompare",
