@@ -2,9 +2,12 @@ import py
 from rpython.jit.metainterp.heapcache import HeapCache
 from rpython.jit.metainterp.resoperation import rop, InputArgInt
 from rpython.jit.metainterp.history import ConstInt, ConstPtr, BasicFailDescr
-from rpython.jit.metainterp.history import IntFrontendOp, RefFrontendOp
+from rpython.jit.metainterp.history import IntFrontendOp, RefFrontendOp as OrigRefFrontendOp
 from rpython.rtyper.lltypesystem import  llmemory, rffi
 
+def RefFrontendOp(pos):
+    return OrigRefFrontendOp(pos, OrigRefFrontendOp._resref)
+
 descr1 = object()
 descr2 = object()
 descr3 = object()
@@ -103,25 +106,33 @@ class TestHeapCache(object):
         h = HeapCache()
         box1 = RefFrontendOp(1)
         box2 = RefFrontendOp(2)
-        assert not h.is_nonstandard_virtualizable(box1)
-        assert not h.is_nonstandard_virtualizable(box2)
+        assert not h.is_known_nonstandard_virtualizable(box1)
+        assert not h.is_known_nonstandard_virtualizable(box2)
         h.nonstandard_virtualizables_now_known(box1)
-        assert h.is_nonstandard_virtualizable(box1)
-        assert not h.is_nonstandard_virtualizable(box2)
+        assert h.is_known_nonstandard_virtualizable(box1)
+        assert not h.is_known_nonstandard_virtualizable(box2)
 
         h.reset()
-        assert not h.is_nonstandard_virtualizable(box1)
-        assert not h.is_nonstandard_virtualizable(box2)
+        assert not h.is_known_nonstandard_virtualizable(box1)
+        assert not h.is_known_nonstandard_virtualizable(box2)
+
+    def test_nonstandard_virtualizable_const(self):
+        h = HeapCache()
+        # rare but not impossible situation for some interpreters: we have a
+        # *constant* nonstandard virtualizable
+        c_box = ConstPtr(ConstPtr.value)
+        h.nonstandard_virtualizables_now_known(c_box) # should not crash
+        assert not h.is_known_nonstandard_virtualizable(c_box)
 
     def test_nonstandard_virtualizable_allocation(self):
         h = HeapCache()
         box1 = RefFrontendOp(1)
         h.new(box1)
         # we've seen the allocation, so it's not the virtualizable
-        assert h.is_nonstandard_virtualizable(box1)
+        assert h.is_known_nonstandard_virtualizable(box1)
 
         h.reset()
-        assert not h.is_nonstandard_virtualizable(box1)
+        assert not h.is_known_nonstandard_virtualizable(box1)
 
     def test_heapcache_fields(self):
         h = HeapCache()
@@ -331,8 +342,8 @@ class TestHeapCache(object):
         h = HeapCache()
         box1 = RefFrontendOp(1)
         box2 = RefFrontendOp(2)
-        lengthbox1 = IntFrontendOp(11)
-        lengthbox2 = IntFrontendOp(12)
+        lengthbox1 = IntFrontendOp(11, 11)
+        lengthbox2 = IntFrontendOp(12, 11)
         h.new_array(box1, lengthbox1)
         assert h.arraylen(box1) is lengthbox1
 
@@ -349,29 +360,32 @@ class TestHeapCache(object):
         h.setfield(box1, box2, descr1)
         h.setarrayitem(box1, index1, box2, descr1)
         h.setarrayitem(box1, index2, box4, descr1)
-        h.invalidate_caches(rop.INT_ADD, None, [])
-        h.invalidate_caches(rop.INT_ADD_OVF, None, [])
-        h.invalidate_caches(rop.SETFIELD_RAW, None, [])
-        h.invalidate_caches(rop.SETARRAYITEM_RAW, None, [])
+        h.invalidate_caches(rop.INT_ADD, None)
+        h.invalidate_caches(rop.INT_ADD_OVF, None)
+        h.invalidate_caches(rop.SETFIELD_RAW, None)
+        h.invalidate_caches(rop.SETARRAYITEM_RAW, None)
+        h.invalidate_caches(rop.RECORD_EXACT_CLASS, None)
+        h.invalidate_caches(rop.RECORD_EXACT_VALUE_I, None)
+        h.invalidate_caches(rop.RECORD_EXACT_VALUE_R, None)
         assert h.getfield(box1, descr1) is box2
         assert h.getarrayitem(box1, index1, descr1) is box2
         assert h.getarrayitem(box1, index2, descr1) is box4
 
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_N, FakeCallDescr(FakeEffectinfo.EF_ELIDABLE_CANNOT_RAISE), [])
         assert h.getfield(box1, descr1) is box2
         assert h.getarrayitem(box1, index1, descr1) is box2
         assert h.getarrayitem(box1, index2, descr1) is box4
 
-        h.invalidate_caches(rop.GUARD_TRUE, None, [])
+        h.invalidate_caches(rop.GUARD_TRUE, None)
         assert h.getfield(box1, descr1) is box2
         assert h.getarrayitem(box1, index1, descr1) is box2
         assert h.getarrayitem(box1, index2, descr1) is box4
 
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_LOOPINVARIANT_N, FakeCallDescr(FakeEffectinfo.EF_LOOPINVARIANT), [])
 
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_N, FakeCallDescr(FakeEffectinfo.EF_RANDOM_EFFECTS), [])
         assert h.getfield(box1, descr1) is None
         assert h.getarrayitem(box1, index1, descr1) is None
@@ -477,16 +491,14 @@ class TestHeapCache(object):
     def test_replace_box_with_const_in_array(self):
         h = HeapCache()
         box1 = RefFrontendOp(1)
-        lengthbox2 = IntFrontendOp(2)
-        lengthbox2.setint(10)
+        lengthbox2 = IntFrontendOp(2, 10)
         h.arraylen_now_known(box1, lengthbox2)
         assert h.arraylen(box1) is lengthbox2
         c10 = ConstInt(10)
         h.replace_box(lengthbox2, c10)
         assert c10.same_constant(h.arraylen(box1))
 
-        box2 = IntFrontendOp(2)
-        box2.setint(12)
+        box2 = IntFrontendOp(2, 12)
         h.setarrayitem(box1, index2, box2, descr1)
         assert h.getarrayitem(box1, index2, descr1) is box2
         c12 = ConstInt(12)
@@ -500,19 +512,19 @@ class TestHeapCache(object):
         box3 = RefFrontendOp(3)
         box4 = RefFrontendOp(4)
         box5 = RefFrontendOp(5)
-        lengthbox1 = IntFrontendOp(11)
-        lengthbox2 = IntFrontendOp(12)
+        lengthbox1 = IntFrontendOp(11, 12)
+        lengthbox2 = IntFrontendOp(12, 178)
         h.new_array(box1, lengthbox1)
         h.setarrayitem(box1, index1, box2, descr1)
         h.new_array(box2, lengthbox1)
         # Just need the destination box for this call
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_N,
             arraycopydescr1,
             [None, box5, box2, index1, index1, index1]
         )
         assert h.getarrayitem(box1, index1, descr1) is box2
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_N,
             arraycopydescr1,
             [None, box5, box3, index1, index1, index1]
@@ -521,7 +533,7 @@ class TestHeapCache(object):
 
         h.setarrayitem(box4, index1, box2, descr1)
         assert h.getarrayitem(box4, index1, descr1) is box2
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_N,
             arraycopydescr1,
             [None, box3, box5, index1, index1, index2]
@@ -533,11 +545,11 @@ class TestHeapCache(object):
         box1 = RefFrontendOp(1)
         box2 = RefFrontendOp(2)
         box3 = RefFrontendOp(3)
-        lengthbox2 = IntFrontendOp(12)
+        lengthbox2 = IntFrontendOp(12, 12)
         h.setarrayitem(box1, index1, box2, descr2)
         assert h.getarrayitem(box1, index1, descr2) is box2
         h.new_array(box2, lengthbox2)
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_N,
             arraycopydescr1,
             [None, box3, box2, index1, index1, index2]
@@ -551,7 +563,7 @@ class TestHeapCache(object):
         box3 = RefFrontendOp(3)
         h.setarrayitem(box1, index1, box2, descr2)
         assert h.getarrayitem(box1, index1, descr2) is box2
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_N,
             arraycopydescr1,
             [None, box3, box2, index1, index1, InputArgInt()]
@@ -564,7 +576,7 @@ class TestHeapCache(object):
         box2 = RefFrontendOp(2)
         box3 = RefFrontendOp(3)
         h.setarrayitem(box1, index1, box2, descr1)
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_N,
             arraycopydescr1,
             [None, box1, box3, index1, index1, index2]
@@ -577,10 +589,10 @@ class TestHeapCache(object):
         box2 = RefFrontendOp(2)
         box3 = RefFrontendOp(3)
         box4 = RefFrontendOp(4)
-        lengthbox1 = IntFrontendOp(11)
+        lengthbox1 = IntFrontendOp(11, 435)
         h.new_array(box1, lengthbox1)
         h.setarrayitem(box3, index1, box4, descr1)
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_N,
             arraycopydescr1,
             [None, box2, box1, index1, index1, index2]
@@ -590,18 +602,20 @@ class TestHeapCache(object):
         h = HeapCache()
         box1 = RefFrontendOp(1)
         box2 = RefFrontendOp(2)
-        lengthbox1 = IntFrontendOp(11)
-        lengthbox2 = IntFrontendOp(12)
-        h.new_array(box1, lengthbox1)
-        h.new_array(box2, lengthbox2)
-        h.invalidate_caches(
+        lengthbox1 = IntFrontendOp(11, 1)
+        lengthbox2 = IntFrontendOp(12, 3)
+        h.new_array(box1, index2)
+        h.new_array(box2, index2)
+        assert h.is_unescaped(box1)
+        assert h.is_unescaped(box2)
+        h.invalidate_caches_varargs(
             rop.CALL_N,
             arraycopydescr1,
             [None, box2, box1, index1, index1, index2]
         )
         assert h.is_unescaped(box1)
         assert h.is_unescaped(box2)
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_N,
             arraycopydescr1,
             [None, box2, box1, index1, index1, InputArgInt()]
@@ -616,9 +630,9 @@ class TestHeapCache(object):
         assert not h.is_unescaped(box1)
         h.new(box2)
         assert h.is_unescaped(box2)
-        h.invalidate_caches(rop.SETFIELD_GC, None, [box2, box1])
+        h.invalidate_caches(rop.SETFIELD_GC, None, box2, box1)
         assert h.is_unescaped(box2)
-        h.invalidate_caches(rop.SETFIELD_GC, None, [box1, box2])
+        h.invalidate_caches(rop.SETFIELD_GC, None, box1, box2)
         assert not h.is_unescaped(box2)
 
     def test_unescaped_testing(self):
@@ -631,14 +645,14 @@ class TestHeapCache(object):
         assert h.is_unescaped(box1)
         assert h.is_unescaped(box2)
         # Putting a virtual inside of another virtual doesn't escape it.
-        h.invalidate_caches(rop.SETFIELD_GC, None, [box1, box2])
+        h.invalidate_caches(rop.SETFIELD_GC, None, box1, box2)
         assert h.is_unescaped(box2)
         # Reading a field from a virtual doesn't escape it.
-        h.invalidate_caches(rop.GETFIELD_GC_I, None, [box1])
+        h.invalidate_caches(rop.GETFIELD_GC_I, None, box1)
         assert h.is_unescaped(box1)
         # Escaping a virtual transitively escapes anything inside of it.
         assert not h.is_unescaped(box3)
-        h.invalidate_caches(rop.SETFIELD_GC, None, [box3, box1])
+        h.invalidate_caches(rop.SETFIELD_GC, None, box3, box1)
         assert not h.is_unescaped(box1)
         assert not h.is_unescaped(box2)
 
@@ -650,10 +664,10 @@ class TestHeapCache(object):
         h.new(box2)
         assert h.is_unescaped(box1)
         assert h.is_unescaped(box2)
-        h.invalidate_caches(rop.INSTANCE_PTR_EQ, None, [box1, box2])
+        h.invalidate_caches(rop.INSTANCE_PTR_EQ, None, box1, box2)
         assert h.is_unescaped(box1)
         assert h.is_unescaped(box2)
-        h.invalidate_caches(rop.INSTANCE_PTR_NE, None, [box1, box2])
+        h.invalidate_caches(rop.INSTANCE_PTR_NE, None, box1, box2)
         assert h.is_unescaped(box1)
         assert h.is_unescaped(box2)
 
@@ -664,32 +678,32 @@ class TestHeapCache(object):
         box3 = RefFrontendOp(3)
         h.new(box1)
         h.new(box2)
-        h.invalidate_caches(rop.SETFIELD_GC, None, [box1, box2])
-        h.invalidate_caches(rop.SETFIELD_GC, None, [box2, box1])
-        h.invalidate_caches(rop.SETFIELD_GC, None, [box3, box1]) # does not crash
+        h.invalidate_caches(rop.SETFIELD_GC, None, box1, box2)
+        h.invalidate_caches(rop.SETFIELD_GC, None, box2, box1)
+        h.invalidate_caches(rop.SETFIELD_GC, None, box3, box1) # does not crash
 
     def test_unescaped_array(self):
         h = HeapCache()
         box1 = RefFrontendOp(1)
         box2 = RefFrontendOp(2)
-        lengthbox1 = IntFrontendOp(11)
-        lengthbox2 = IntFrontendOp(12)
-        h.new_array(box1, lengthbox1)
+        lengthbox1 = IntFrontendOp(11, 3)
+        lengthbox2 = IntFrontendOp(12, 3)
+        h.new_array(box1, index2)
         assert h.is_unescaped(box1)
-        h.invalidate_caches(rop.SETARRAYITEM_GC, None, [box1, index1, box2])
+        h.invalidate_caches(rop.SETARRAYITEM_GC, None, box1, index1, box2)
         assert h.is_unescaped(box1)
-        h.invalidate_caches(rop.SETARRAYITEM_GC, None, [box2, index1, box1])
+        h.invalidate_caches(rop.SETARRAYITEM_GC, None, box2, index1, box1)
         assert not h.is_unescaped(box1)
 
         h = HeapCache()
-        h.new_array(box1, lengthbox1)
+        h.new_array(box1, index2)
         h.new(box2)
         assert h.is_unescaped(box1)
         assert h.is_unescaped(box2)
-        h.invalidate_caches(rop.SETARRAYITEM_GC, None, [box1, lengthbox2, box2])
+        h.invalidate_caches(rop.SETARRAYITEM_GC, None, box1, lengthbox2, box2)
         assert h.is_unescaped(box1)
         assert h.is_unescaped(box2)
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_N, FakeCallDescr(FakeEffectinfo.EF_RANDOM_EFFECTS), [box1]
         )
         assert not h.is_unescaped(box1)
@@ -702,7 +716,7 @@ class TestHeapCache(object):
         h.new(box1)
         assert h.is_unescaped(box1)
         h.setfield(box1, box2, descr1)
-        h.invalidate_caches(rop.CALL_N,
+        h.invalidate_caches_varargs(rop.CALL_N,
             FakeCallDescr(FakeEffectinfo.EF_CAN_RAISE),
             []
         )
@@ -712,11 +726,11 @@ class TestHeapCache(object):
         h = HeapCache()
         box1 = RefFrontendOp(1)
         box3 = RefFrontendOp(3)
-        lengthbox1 = IntFrontendOp(11)
-        h.new_array(box1, lengthbox1)
+        lengthbox1 = IntFrontendOp(11, 3)
+        h.new_array(box1, index2)
         assert h.is_unescaped(box1)
         h.setarrayitem(box1, index1, box3, descr1)
-        h.invalidate_caches(rop.CALL_N,
+        h.invalidate_caches_varargs(rop.CALL_N,
             FakeCallDescr(FakeEffectinfo.EF_CAN_RAISE),
             []
         )
@@ -730,12 +744,12 @@ class TestHeapCache(object):
         h.new(box2)
         h.setfield(box1, box2, descr1)
         assert h.getfield(box1, descr1) is box2
-        h.invalidate_caches(rop.STRSETITEM, None, [])
-        h.invalidate_caches(rop.UNICODESETITEM, None, [])
-        h.invalidate_caches(rop.SETFIELD_RAW, None, [])
-        h.invalidate_caches(rop.SETARRAYITEM_RAW, None, [])
-        h.invalidate_caches(rop.SETINTERIORFIELD_RAW, None, [])
-        h.invalidate_caches(rop.RAW_STORE, None, [])
+        h.invalidate_caches(rop.STRSETITEM, None)
+        h.invalidate_caches(rop.UNICODESETITEM, None)
+        h.invalidate_caches(rop.SETFIELD_RAW, None)
+        h.invalidate_caches(rop.SETARRAYITEM_RAW, None)
+        h.invalidate_caches(rop.SETINTERIORFIELD_RAW, None)
+        h.invalidate_caches(rop.RAW_STORE, None)
         assert h.is_unescaped(box1)
         assert h.is_unescaped(box2)
         assert h.getfield(box1, descr1) is box2
@@ -753,9 +767,9 @@ class TestHeapCache(object):
         h.new(box1)
         h.new(box2)
         h.setfield(box1, box2, descr1)
-        h.invalidate_caches(rop.SETFIELD_GC, None, [box1, box2])
+        h.invalidate_caches(rop.SETFIELD_GC, None, box1, box2)
         assert h.getfield(box1, descr1) is box2
-        h.invalidate_caches(rop.CALL_MAY_FORCE_N, FakeCallDescr(FakeEffectinfo.EF_RANDOM_EFFECTS), [])
+        h.invalidate_caches_varargs(rop.CALL_MAY_FORCE_N, FakeCallDescr(FakeEffectinfo.EF_RANDOM_EFFECTS), [])
         assert not h.is_unescaped(box1)
         assert not h.is_unescaped(box2)
         assert h.getfield(box1, descr1) is None
@@ -767,7 +781,7 @@ class TestHeapCache(object):
         h.new(box1)
         h.new(box2)
         h.setfield(box1, box2, descr1)
-        h.invalidate_caches(rop.SETFIELD_GC, None, [box1, box2])
+        h.invalidate_caches(rop.SETFIELD_GC, None, box1, box2)
         assert h.getfield(box1, descr1) is box2
         descr = BasicFailDescr()
         class XTra:
@@ -780,7 +794,7 @@ class TestHeapCache(object):
             EF_ELIDABLE_OR_MEMORYERROR = 3
             EF_ELIDABLE_CAN_RAISE = 4
         descr.get_extra_info = XTra
-        h.invalidate_caches(rop.CALL_N, descr, [])
+        h.invalidate_caches_varargs(rop.CALL_N, descr, [])
         assert h.is_unescaped(box1)
         assert h.is_unescaped(box2)
         assert h.getfield(box1, descr1) is box2
@@ -824,6 +838,17 @@ class TestHeapCache(object):
         assert not h.is_unescaped(box1)
         assert h.is_likely_virtual(box1)
 
+    def test_is_likely_virtual_array(self):
+        h = HeapCache()
+        box1 = RefFrontendOp(1)
+        h.new_array(box1, index2)
+        assert h.is_likely_virtual(box1)
+        box2 = RefFrontendOp(2)
+        lengthbox = IntFrontendOp(11, 3)
+        # arrays are only virtual if the length is constant
+        h.new_array(box2, lengthbox)
+        assert not h.is_likely_virtual(box2)
+
     def test_quasiimmut_seen(self):
         h = HeapCache()
         box1 = RefFrontendOp(1)
@@ -862,7 +887,7 @@ class TestHeapCache(object):
         assert not h.is_quasi_immut_known(descr1, box2)
 
         # a call invalidates everything
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_N, FakeCallDescr(FakeEffectinfo.EF_CAN_RAISE), [])
         assert not h.is_quasi_immut_known(descr2, box3)
         assert not h.is_quasi_immut_known(descr2, box4)
@@ -898,7 +923,29 @@ class TestHeapCache(object):
         assert not h.is_quasi_immut_known(descr1, box2)
 
         # a call invalidates everything
-        h.invalidate_caches(
+        h.invalidate_caches_varargs(
             rop.CALL_N, FakeCallDescr(FakeEffectinfo.EF_CAN_RAISE), [])
         assert not h.is_quasi_immut_known(descr2, box3)
         assert not h.is_quasi_immut_known(descr2, box4)
+
+    def test_heapcache_on_const(self):
+        h = HeapCache()
+        # two different boxes of the same value
+        box1 = ConstPtr(ConstPtr.value)
+        box2 = ConstPtr(ConstPtr.value)
+        box3 = RefFrontendOp(3)
+        box4 = RefFrontendOp(4)
+
+        # the cache unifies a bit
+        assert h.getfield(box1, descr1) is None
+        assert h.getfield(box2, descr1) is None
+        h.getfield_now_known(box1, descr1, box3)
+        assert h.heap_cache[descr1].last_const_box is box1
+        assert h.heap_cache[descr1]._unique_const_heuristic(box2) is box1
+        assert h.getfield(box1, descr1) is box3
+        assert h.getfield(box2, descr1) is box3
+
+        h.reset()
+        h.setfield(box2, box4, descr1)
+        assert h.getfield(box1, descr1) is box4
+        assert h.getfield(box2, descr1) is box4
