@@ -1,13 +1,14 @@
 from rpython.rtyper.lltypesystem import rffi
 from rpython.rtyper.lltypesystem import lltype
 from rpython.rtyper.tool import rffi_platform as platform
-from rpython.rtyper.lltypesystem.rffi import CCHARP
+from rpython.rtyper.lltypesystem.rffi import CCHARP, CONST_CCHARP
 from rpython.rlib import jit
 from rpython.translator.tool.cbuild import ExternalCompilationInfo
 from rpython.translator.platform import platform as target_platform
 
-from rpython.rlib.rarithmetic import intmask, r_uint
+from rpython.rlib.rarithmetic import intmask, r_uint, widen
 import os,sys
+from textwrap import dedent
 
 _POSIX = os.name == "posix"
 _WIN32 = sys.platform == "win32"
@@ -34,11 +35,12 @@ if _POSIX:
                 'stdint.h',
                 'errno.h',
                 'limits.h',
+                'net/if.h',
                 )
     if _HAS_AF_PACKET:
         includes += ('netpacket/packet.h',
                      'sys/ioctl.h',
-                     'net/if.h')
+                    )
 
     cond_includes = [('AF_NETLINK', 'linux/netlink.h')]
 
@@ -53,7 +55,7 @@ if _SOLARIS:
 
 if _WIN32:
     includes = ()
-    libraries = ('ws2_32',)
+    libraries = ('ws2_32', 'Iphlpapi')
     calling_conv = 'win'
     header_lines = [
         '#include <WinSock2.h>',
@@ -65,7 +67,13 @@ if _WIN32:
         header_lines.extend([
             '#include <Mstcpip.h>',
             # these types do not exist on microsoft compilers
+            '#ifdef _WIN64',
+            'typedef long long ssize_t;',
+            'typedef unsigned long long size_t;',
+            '#else',
             'typedef int ssize_t;',
+            'typedef unsigned int size_t;',
+            '#endif',
             'typedef unsigned __int16 uint16_t;',
             'typedef unsigned __int32 uint32_t;',
             ])
@@ -257,23 +265,24 @@ for name, default in constants_w_defaults:
 
 # types
 if _MSVC:
-    socketfd_type = rffi.UINT
+    socketfd_type = lltype.Unsigned
 else:
-    socketfd_type = rffi.INT
+    socketfd_type = rffi.INT_real
 
 CConfig.uint16_t = platform.SimpleType('uint16_t', rffi.USHORT)
 CConfig.uint32_t = platform.SimpleType('uint32_t', rffi.UINT)
 CConfig.size_t = platform.SimpleType('size_t', rffi.INT)
 CConfig.ssize_t = platform.SimpleType('ssize_t', rffi.INT)
-CConfig.socklen_t = platform.SimpleType('socklen_t', rffi.INT)
+CConfig.socklen_t = platform.SimpleType('socklen_t', rffi.UINT_real)
 sockaddr_ptr = lltype.Ptr(lltype.ForwardReference())
 addrinfo_ptr = lltype.Ptr(lltype.ForwardReference())
 
 
 # struct types
 CConfig.sockaddr = platform.Struct('struct sockaddr',
-                                             [('sa_family', rffi.INT),
-                                   ('sa_data', rffi.CFixedArray(rffi.CHAR, 1))])
+                                      [('sa_family', rffi.INT),
+                                       ('sa_data', rffi.CFixedArray(rffi.CHAR, 1)),
+                                      ])
 CConfig.in_addr = platform.Struct('struct in_addr',
                                          [('s_addr', rffi.UINT)])
 CConfig.in6_addr = platform.Struct('struct in6_addr',
@@ -338,6 +347,9 @@ if _POSIX:
                                             [('fd', socketfd_type),
                                              ('events', rffi.SHORT),
                                              ('revents', rffi.SHORT)])
+    CConfig.if_nameindex_s = platform.Struct('struct if_nameindex',
+                                        [('if_index', rffi.UINT),
+                                         ('if_name', rffi.CCHARP)])
 
     if _HAS_AF_PACKET:
         CConfig.sockaddr_ll = platform.Struct('struct sockaddr_ll',
@@ -352,11 +364,18 @@ if _POSIX:
         CConfig.ifreq = platform.Struct('struct ifreq',
                                 [('ifr_ifindex', rffi.INT),
                                  ('ifr_name', rffi.CFixedArray(rffi.CHAR, 8))])
+else:
+    CConfig.if_nameindex_s = rffi.CStruct('if_nameindex',
+                                        ('if_index', rffi.UINT),
+                                         ('if_name', rffi.CCHARP))
 
+includes = []
+separate_module_sources = []
+post_include_bits = []
 # insert handler for sendmsg / recvmsg here
 HAVE_SENDMSG = bool(_POSIX)
 if HAVE_SENDMSG:
-    includes = ['stddef.h',
+    includes += ['stddef.h',
                 'sys/socket.h',
                 'unistd.h',
                 'string.h',
@@ -367,7 +386,7 @@ if HAVE_SENDMSG:
                 'sys/types.h',
                 'netinet/in.h',
                 'arpa/inet.h']
-    separate_module_sources = ['''
+    separate_module_sources += ['''
 
         // special defines for returning from recvmsg
         #define BAD_MSG_SIZE_GIVEN -10000
@@ -385,7 +404,7 @@ if HAVE_SENDMSG:
         #define SOCKLEN_T_LIMIT INT_MAX
         #endif
 
-        // ################################################################################################
+        // #########################################################################
         // Recvmsg implementation and associated functions
 
         // Taken from CPython. Determines the minimum memory space required for the ancillary data.
@@ -719,7 +738,7 @@ if HAVE_SENDMSG:
         }
 
 
-        // ################################################################################################
+        // ##########################################################################
         // Sendmsg implementation and associated functions
 
         #ifdef CMSG_LEN
@@ -878,7 +897,7 @@ if HAVE_SENDMSG:
             return retval;
         }
 
-        // ################################################################################################
+        // ############################################################################
         // Wrappers for CMSG_SPACE and CMSG_LEN
 
         /*
@@ -907,7 +926,7 @@ if HAVE_SENDMSG:
         }
         #endif
 
-        // ################################################################################################
+        // ###########################################################################
         // Extra functions that I needed
 
         /*
@@ -924,7 +943,7 @@ if HAVE_SENDMSG:
             These functions free memory that was allocated in C (sendmsg or recvmsg) was used in rsocket and now needs cleanup
         */
         RPY_EXTERN
-        int free_pointer_to_signedp(int** ptrtofree){
+        int free_pointer_to_signedp(long** ptrtofree){
             free(*ptrtofree);
             return 0;
         }
@@ -937,7 +956,7 @@ if HAVE_SENDMSG:
 
     ''',]
 
-    post_include_bits =[ "RPY_EXTERN "
+    post_include_bits +=[ "RPY_EXTERN "
                          "int sendmsg_implementation(int socket, struct sockaddr* address, socklen_t addrlen, long* length_of_messages, char** messages, int no_of_messages, long* levels, long* types, char** file_descriptors, long* no_of_fds, int control_length, int flag );\n"
                          "RPY_EXTERN "
                          "int recvmsg_implementation(int socket_fd, int ancillary_size, int flags, struct sockaddr* address, socklen_t* addrlen, int* message_lengths, char** messages, int no_of_messages, long* size_of_ancillary, long** levels, long** types, char** file_descr, long** descr_per_ancillary, long* flag);\n"
@@ -958,21 +977,11 @@ if HAVE_SENDMSG:
                          "RPY_EXTERN "
                          "int memcpy_from_CCHARP_at_offset_and_size(char* stringfrom, char** stringto, int offset, int size);\n"
                          "RPY_EXTERN "
-                         "int free_pointer_to_signedp(int** ptrtofree);\n"
+                         "int free_pointer_to_signedp(long** ptrtofree);\n"
                          "RPY_EXTERN "
                          "int free_ptr_to_charp(char** ptrtofree);\n"
                          ]
 
-
-    compilation_info = eci.merge(ExternalCompilationInfo(
-                                    includes=includes,
-                                    separate_module_sources=separate_module_sources,
-                                    post_include_bits=post_include_bits,
-                               ))
-else:
-    compilation_info = eci
-
-
 if _WIN32:
     CConfig.WSAEVENT = platform.SimpleType('WSAEVENT', rffi.VOIDP)
     CConfig.WSANETWORKEVENTS = platform.Struct(
@@ -987,6 +996,73 @@ if _WIN32:
 
     CConfig.FROM_PROTOCOL_INFO = platform.DefinedConstantInteger(
         'FROM_PROTOCOL_INFO')
+    post_include_bits +=[dedent("""\
+        struct if_nameindex {
+            unsigned int if_index;
+            char * if_name;
+        };
+        RPY_EXTERN struct if_nameindex * if_nameindex(void);
+        RPY_EXTERN void if_freenameindex(struct if_nameindex *ptr);
+        """)]
+    separate_module_sources += [dedent("""\
+        #include <Iphlpapi.h>
+        #include <stdlib.h>
+        RPY_EXTERN
+        struct if_nameindex * if_nameindex(void) {
+            PMIB_IF_TABLE2 tbl;
+            int ret;
+            if (GetIfTable2Ex(MibIfTableRaw, &tbl) != NO_ERROR) {
+                return NULL;
+            }
+            size_t memsize = sizeof(struct if_nameindex) * (tbl->NumEntries + 1);
+            struct if_nameindex *out = malloc(memsize);
+            if (out == NULL) {
+                return NULL;
+            }
+            ULONG i = 0;
+            for (i; i < tbl->NumEntries; i++) {
+                MIB_IF_ROW2 r = tbl->Table[i];
+                WCHAR buf[NDIS_IF_MAX_STRING_SIZE + 1];
+                if ((ret = ConvertInterfaceLuidToNameW(&r.InterfaceLuid, buf,
+                                                       NDIS_IF_MAX_STRING_SIZE))) {
+                    FreeMibTable(tbl);
+                    for (ULONG j=0; j<i; j++) {
+                        free(out[j].if_name);
+                    }
+                    free(out);
+                    return NULL;
+                }
+                /* convert from wchar_t to char */
+                size_t origsize = wcslen(buf) + 1;
+                size_t newsize = origsize * 2 + 2;
+                size_t convertedChars = 0; 
+                out[i].if_name = malloc(newsize);
+                out[i].if_name[0] = '\\0';
+                wcstombs_s(&convertedChars, out[i].if_name, newsize, buf, newsize - 1);
+                out[i].if_index = r.InterfaceIndex;
+            }
+            out[i].if_name = NULL;
+            out[i].if_index = 0;
+            return out;
+        }
+ 
+        RPY_EXTERN
+        void if_freenameindex(struct if_nameindex *ptr) {
+            if (ptr == NULL) return;
+            int index = 0;
+            while (1) {
+                if (ptr[index].if_name == NULL) break;
+                free(ptr[index].if_name);
+                index += 1;
+            }
+            free(ptr);
+        }""")]
+
+compilation_info = eci.merge(ExternalCompilationInfo(
+                                includes=includes,
+                                separate_module_sources=separate_module_sources,
+                                post_include_bits=post_include_bits,
+                           ))
 
 CConfig.timeval = platform.Struct('struct timeval',
                                          [('tv_sec', rffi.LONG),
@@ -1110,11 +1186,11 @@ assert WIN32 == _WIN32
 
 if _MSVC:
     def invalid_socket(fd):
-        return fd == INVALID_SOCKET
+        return widen(fd) == INVALID_SOCKET
     INVALID_SOCKET = r_uint(cConfig.INVALID_SOCKET)
 else:
     def invalid_socket(fd):
-        return fd < 0
+        return widen(fd) < 0
     INVALID_SOCKET = -1
 
 uint16_t = cConfig.uint16_t
@@ -1139,10 +1215,12 @@ if _POSIX:
     if _HAS_AF_PACKET:
         sockaddr_ll = cConfig.sockaddr_ll
         ifreq = cConfig.ifreq
+    if_nameindex_s = cConfig.if_nameindex_s
 if WIN32:
     WSAEVENT = cConfig.WSAEVENT
     WSANETWORKEVENTS = cConfig.WSANETWORKEVENTS
     SAVE_ERR = rffi.RFFI_SAVE_WSALASTERROR
+    if_nameindex_s = CConfig.if_nameindex_s
 else:
     SAVE_ERR = rffi.RFFI_SAVE_ERRNO
 timeval = cConfig.timeval
@@ -1158,7 +1236,7 @@ def external_c(name, args, result, **kwargs):
 
 if _POSIX:
     dup = external('dup', [socketfd_type], socketfd_type, save_err=SAVE_ERR)
-    gai_strerror = external('gai_strerror', [rffi.INT], CCHARP)
+    gai_strerror = external('gai_strerror', [rffi.INT], CONST_CCHARP)
 
 #h_errno = c_int.in_dll(socketdll, 'h_errno')
 #
@@ -1212,7 +1290,7 @@ inet_pton = external('inet_pton', [rffi.INT, rffi.CCHARP,
                      save_err=SAVE_ERR)
 
 inet_ntop = external('inet_ntop', [rffi.INT, rffi.VOIDP, CCHARP,
-                                   socklen_t], CCHARP,
+                                   socklen_t], CONST_CCHARP,
                      save_err=SAVE_ERR)
 
 inet_addr = external('inet_addr', [rffi.CCHARP], rffi.UINT)
@@ -1242,15 +1320,19 @@ socketgetsockopt = external('getsockopt', [socketfd_type, rffi.INT,
 socketsetsockopt = external('setsockopt', [socketfd_type, rffi.INT,
                                    rffi.INT, rffi.VOIDP, socklen_t], rffi.INT,
                             save_err=SAVE_ERR)
-socketrecv = external('recv', [socketfd_type, rffi.VOIDP, rffi.INT,
-                               rffi.INT], ssize_t, save_err=SAVE_ERR)
+if WIN32:
+    socketrecv = external('recv', [socketfd_type, rffi.VOIDP, rffi.INT,
+                                   rffi.INT], rffi.INT, save_err=SAVE_ERR)
+else:
+    socketrecv = external('recv', [socketfd_type, rffi.VOIDP, rffi.INT,
+                                   rffi.INT], ssize_t, save_err=SAVE_ERR)
 recvfrom = external('recvfrom', [socketfd_type, rffi.VOIDP, size_t,
                            rffi.INT, sockaddr_ptr, socklen_t_ptr], rffi.INT,
                     save_err=SAVE_ERR)
 recvmsg = jit.dont_look_inside(rffi.llexternal(
     "recvmsg_implementation",
     [rffi.INT, rffi.INT, rffi.INT, sockaddr_ptr, socklen_t_ptr,
-        rffi.INTP, rffi.CCHARPP, rffi.INT, rffi.SIGNEDP, rffi.SIGNEDPP, rffi.SIGNEDPP,
+        rffi.INT_realP, rffi.CCHARPP, rffi.INT, rffi.SIGNEDP, rffi.SIGNEDPP, rffi.SIGNEDPP,
         rffi.CCHARPP, rffi.SIGNEDPP, rffi.SIGNEDP], rffi.INT,
     save_err=SAVE_ERR, compilation_info=compilation_info))
 
@@ -1287,7 +1369,7 @@ getprotobyname = external('getprotobyname', [rffi.CCHARP], lltype.Ptr(cConfig.pr
 
 if _POSIX:
     fcntl = external('fcntl', [socketfd_type, rffi.INT, rffi.INT], rffi.INT,
-                     save_err=SAVE_ERR)
+                     save_err=SAVE_ERR, natural_arity=2)
     socketpair_t = rffi.CArray(socketfd_type)
     socketpair = external('socketpair', [rffi.INT, rffi.INT, rffi.INT,
                           lltype.Ptr(socketpair_t)], rffi.INT,
@@ -1309,6 +1391,14 @@ select = external('select',
                   rffi.INT,
                   save_err=SAVE_ERR)
 
+if_nameindex = rffi.llexternal('if_nameindex', [], lltype.Ptr(if_nameindex_s),
+                        save_err=SAVE_ERR, compilation_info=compilation_info,
+                        calling_conv=calling_conv)
+
+if_freenameindex = rffi.llexternal('if_freenameindex', [lltype.Ptr(if_nameindex_s)],
+                        lltype.Void, save_err=SAVE_ERR, compilation_info=compilation_info,
+                        calling_conv=calling_conv)
+
 FD_CLR = external_c('FD_CLR', [rffi.INT, fd_set], lltype.Void, macro=True)
 FD_ISSET = external_c('FD_ISSET', [rffi.INT, fd_set], rffi.INT, macro=True)
 FD_SET = external_c('FD_SET', [rffi.INT, fd_set], lltype.Void, macro=True)
@@ -1422,7 +1512,7 @@ else:
 
     socket_strerror_str = os.strerror
     def gai_strerror_str(errno):
-        return rffi.charp2str(gai_strerror(errno))
+        return rffi.constcharp2str(gai_strerror(errno))
 
     def socket_strerror_unicode(errno):
         return socket_strerror_str(errno).decode('latin-1')
