@@ -10,19 +10,29 @@ from rpython.rtyper.tool import rffi_platform
 from rpython.tool.udir import udir
 from rpython.translator.tool.cbuild import ExternalCompilationInfo
 from rpython.translator.platform import CompilationError
+from rpython.translator import cdir
 from rpython.rtyper.lltypesystem import lltype, rffi
-from rpython.rlib.rarithmetic import intmask, r_longlong
+from rpython.rlib.rutf8 import codepoints_in_utf8
+from rpython.rlib.rarithmetic import intmask, r_longlong, widen
 from rpython.rlib import jit
 
 # This module can be imported on any platform,
 # but most symbols are not usable...
 WIN32 = os.name == "nt"
+srcdir = os.path.join(os.path.dirname(__file__), 'src')
 
 if WIN32:
     eci = ExternalCompilationInfo(
-        includes = ['windows.h', 'stdio.h', 'stdlib.h', 'io.h'],
-        libraries = ['kernel32'],
+        includes = ['windows.h', 'stdio.h', 'stdlib.h', 'io.h', 'winhelpers.h'],
+        include_dirs = [srcdir, cdir],
+        libraries = ['kernel32', 'Advapi32'],
+        separate_module_files = [os.path.join(srcdir, "winhelpers.c")],
         )
+
+    def external(name, args, result, compilation_info=eci, **kwds):
+        return rffi.llexternal(name, args, result,
+                               compilation_info=compilation_info, **kwds)
+
 else:
     eci = ExternalCompilationInfo()
 
@@ -30,7 +40,7 @@ class CConfig:
     _compilation_info_ = eci
 
     if WIN32:
-        DWORD_PTR = rffi_platform.SimpleType("DWORD_PTR", rffi.LONG)
+        DWORD_PTR = rffi_platform.SimpleType("DWORD_PTR", rffi.UNSIGNED)
         WORD = rffi_platform.SimpleType("WORD", rffi.USHORT)
         DWORD = rffi_platform.SimpleType("DWORD", rffi.UINT)
         BOOL = rffi_platform.SimpleType("BOOL", rffi.LONG)
@@ -50,7 +60,7 @@ class CConfig:
         LPBOOL = rffi_platform.SimpleType("LPBOOL", rffi.LONGP)
         LPBYTE = rffi_platform.SimpleType("LPBYTE", rffi.UCHARP)
         SIZE_T = rffi_platform.SimpleType("SIZE_T", rffi.SIZE_T)
-        ULONG_PTR = rffi_platform.SimpleType("ULONG_PTR", rffi.ULONG)
+        ULONG_PTR = rffi_platform.SimpleType("ULONG_PTR", rffi.UNSIGNED)
 
         HRESULT = rffi_platform.SimpleType("HRESULT", rffi.LONG)
         HLOCAL = rffi_platform.SimpleType("HLOCAL", rffi.VOIDP)
@@ -101,7 +111,7 @@ class CConfig:
 
         defines = """FORMAT_MESSAGE_ALLOCATE_BUFFER FORMAT_MESSAGE_FROM_SYSTEM
                        MAX_PATH _MAX_ENV FORMAT_MESSAGE_IGNORE_INSERTS
-                       WAIT_OBJECT_0 WAIT_TIMEOUT INFINITE
+                       WAIT_OBJECT_0 WAIT_TIMEOUT INFINITE WAIT_FAILED
                        ERROR_INVALID_HANDLE
                        DELETE READ_CONTROL SYNCHRONIZE WRITE_DAC
                        WRITE_OWNER PROCESS_ALL_ACCESS
@@ -117,6 +127,9 @@ class CConfig:
                        STD_ERROR_HANDLE HANDLE_FLAG_INHERIT FILE_TYPE_CHAR
                        LOAD_WITH_ALTERED_SEARCH_PATH CT_CTYPE3 C3_HIGHSURROGATE
                        CP_ACP CP_UTF8 CP_UTF7 CP_OEMCP MB_ERR_INVALID_CHARS
+                       LOAD_LIBRARY_SEARCH_DEFAULT_DIRS SEM_FAILCRITICALERRORS
+                       LOAD_LIBRARY_SEARCH_DLL_LOAD_DIR ENABLE_VIRTUAL_TERMINAL_PROCESSING
+                       _Py_MAXIMUM_REPARSE_DATA_BUFFER_SIZE
                     """
         from rpython.translator.platform import host_factory
         static_platform = host_factory()
@@ -146,6 +159,8 @@ if WIN32:
     GENERIC_ALL      = rffi.cast(DWORD, r_longlong(0x10000000))
     FILE_SHARE_READ  = rffi.cast(DWORD, r_longlong(0x00000001))
     FILE_SHARE_WRITE = rffi.cast(DWORD, r_longlong(0x00000002))
+    ALL_READ_WRITE   = rffi.cast(DWORD, r_longlong(0xC0000003))
+    SHARE_READ_WRITE = rffi.cast(DWORD, r_longlong(0x00000003))
 
     PFILETIME = rffi.CArrayPtr(FILETIME)
 
@@ -165,7 +180,9 @@ if WIN32:
         (into the same "saved LastError" variable).
         """
         from rpython.rlib import rthread
-        return rffi.cast(lltype.Signed, rthread.tlfield_rpy_lasterror.getraw())
+        # extra cast to LONG to match CPython behaviour
+        lasterror = rffi.cast(rffi.LONG, rthread.tlfield_rpy_lasterror.getraw())
+        return rffi.cast(lltype.Signed, lasterror)
 
     def SetLastError_saved(err):
         """Set the value of the saved LastError.  This value will be used in
@@ -188,7 +205,9 @@ if WIN32:
         (into the same "saved alt LastError" variable).
         """
         from rpython.rlib import rthread
-        return rffi.cast(lltype.Signed, rthread.tlfield_alt_lasterror.getraw())
+        # extra cast to LONG to match CPython behaviour
+        lasterror = rffi.cast(rffi.LONG, rthread.tlfield_alt_lasterror.getraw())
+        return rffi.cast(lltype.Signed, lasterror)
 
     def SetLastError_alt_saved(err):
         """Set the value of the saved alt LastError.  This value will be used in
@@ -207,7 +226,7 @@ if WIN32:
     LoadLibrary = winexternal('LoadLibraryA', [rffi.CCHARP], HMODULE,
                               save_err=rffi.RFFI_SAVE_LASTERROR)
     def wrap_loadlibraryex(func):
-        def loadlibrary(name, flags=LOAD_WITH_ALTERED_SEARCH_PATH):
+        def loadlibrary(name, flags):
             # Requires a full path name with '/' -> '\\'
             return func(name, NULL_HANDLE, flags)
         return loadlibrary
@@ -245,8 +264,8 @@ if WIN32:
     _get_osfhandle = rffi.llexternal('_get_osfhandle', [rffi.INT], rffi.INTP)
 
     def get_osfhandle(fd):
-        from rpython.rlib.rposix import FdValidator
-        with FdValidator(fd):
+        from rpython.rlib.rposix import SuppressIPH
+        with SuppressIPH():
             handle = rffi.cast(HANDLE, _get_osfhandle(fd))
         if handle == INVALID_HANDLE_VALUE:
             raise WindowsError(ERROR_INVALID_HANDLE, "Invalid file handle")
@@ -255,12 +274,12 @@ if WIN32:
     _open_osfhandle = rffi.llexternal('_open_osfhandle', [rffi.INTP, rffi.INT], rffi.INT)
 
     def open_osfhandle(handle, flags):
-        from rpython.rlib.rposix import FdValidator
+        from rpython.rlib.rposix import SuppressIPH
         fd = _open_osfhandle(handle, flags)
-        with FdValidator(fd):
+        with SuppressIPH():
             return fd
-    
-    wcsncpy_s = rffi.llexternal('wcsncpy_s', 
+
+    wcsncpy_s = rffi.llexternal('wcsncpy_s',
                     [rffi.CWCHARP, rffi.SIZE_T, rffi.CWCHARP, rffi.SIZE_T], rffi.INT)
 
     def build_winerror_to_errno():
@@ -363,7 +382,10 @@ if WIN32:
         return WindowsError(code, context)
 
     def FAILED(hr):
-        return rffi.cast(HRESULT, hr) < 0
+        # XXX convert to int before checking result
+        #     because 32-bit arithmetic is unimplemented on win64
+        #     this is fine since HRESULT is defined as (signed) LONG
+        return int(rffi.cast(HRESULT, hr)) < 0
 
     _GetModuleFileName = winexternal('GetModuleFileNameA',
                                      [HMODULE, rffi.CCHARP, DWORD],
@@ -542,7 +564,7 @@ if WIN32:
         try:
             if not _GetHandleInformation(handle, pflags):
                 raise lastSavedWindowsError("GetHandleInformation")
-            flags = pflags[0]
+            flags = widen(pflags[0])
         finally:
             lltype.free(pflags, flavor='raw')
         return (flags & HANDLE_FLAG_INHERIT) != 0
@@ -574,34 +596,107 @@ if WIN32:
                 os.close(fd2)
                 raise
         return res
-    
+
     GetConsoleMode = winexternal(
         'GetConsoleMode', [HANDLE, LPDWORD], BOOL)
-        
+
     GetNumberOfConsoleInputEvents = winexternal(
         'GetNumberOfConsoleInputEvents', [HANDLE, LPDWORD], BOOL)
 
     ERROR_INSUFFICIENT_BUFFER = 122
     ERROR_OPERATION_ABORTED   = 995
-    CP_UTF8 = 65001 
-    
+    CP_UTF8 = 65001
+
     ReadConsoleW = winexternal(
         'ReadConsoleW', [HANDLE, LPWSTR, DWORD, LPDWORD, LPVOID], BOOL,
         save_err=rffi.RFFI_SAVE_LASTERROR)
-        
+
     WriteConsoleW = winexternal(
         'WriteConsoleW', [HANDLE, LPVOID, DWORD, LPDWORD, LPVOID], BOOL,
         save_err=rffi.RFFI_SAVE_LASTERROR)
 
     GetStringTypeW = winexternal(
-        'GetStringTypeW', [DWORD, rffi.CWCHARP, rffi.INT, LPWORD], BOOL, 
+        'GetStringTypeW', [DWORD, rffi.CWCHARP, rffi.INT, LPWORD], BOOL,
         save_err=rffi.RFFI_SAVE_LASTERROR)
 
     _SetEnvironmentVariableW = winexternal(
         'SetEnvironmentVariableW', [LPWSTR, LPWSTR], BOOL,
         save_err=rffi.RFFI_SAVE_LASTERROR)
 
-    def SetEnvironmentVariableW(name, value):
-        with rffi.scoped_unicode2wcharp(name) as nameWbuf:
-            with rffi.scoped_unicode2wcharp(value) as valueWbuf:
-                return _SetEnvironmentVariableW(nameWbuf, valueWbuf) 
+    def DelEnvironmentVariableW(name):
+        with rffi.scoped_utf82wcharp(name) as nameWbuf:
+            valueWbuf = lltype.nullptr(rffi.CWCHARP.TO)
+            return _SetEnvironmentVariableW(nameWbuf, valueWbuf)
+
+    _AddDllDirectory = winexternal('AddDllDirectory', [LPWSTR], rffi.VOIDP,
+        save_err=rffi.RFFI_SAVE_LASTERROR)
+
+    def AddDllDirectory(path, length):
+        with rffi.scoped_utf82wcharp(path, length) as pathW:
+            return _AddDllDirectory(pathW)
+
+    RemoveDllDirectory = winexternal('RemoveDllDirectory', [rffi.VOIDP], BOOL,
+        save_err=rffi.RFFI_SAVE_LASTERROR)
+
+    # Don't save the err since this is called before checking err in rdynload
+    SetErrorMode = winexternal('SetErrorMode', [rffi.UINT], rffi.UINT) 
+
+    # int os_readlink_impl(wchar_t *path_to_check, char * reparse_data_buffer, 
+    #                  wchar_t **result);
+    # returns the number of wchar_t chars in result, -1 if error, for a given
+    # path_to_check. The result pointer will be somewhere inside the pre-allocated
+    # reparse_data_buffer, which should be of length _Py_MAXIMUM_REPARSE_DATA_BUFFER_SIZE
+    os_readlink_impl = winexternal("os_readlink_impl",
+        [rffi.CWCHARP, rffi.VOIDP, rffi.CWCHARPP], rffi.INT,
+        save_err=rffi.RFFI_SAVE_LASTERROR)
+
+    os_symlink_impl = winexternal("os_symlink_impl",
+        [rffi.CWCHARP, rffi.CWCHARP, rffi.INT], rffi.INT,
+        save_err=rffi.RFFI_SAVE_LASTERROR)
+
+    os_unlink_impl = winexternal("os_unlink_impl", [rffi.CWCHARP], rffi.INT,
+        save_err=rffi.RFFI_SAVE_LASTERROR)
+
+    GetUserNameW = winexternal('GetUserNameW', [LPWSTR, LPDWORD], rffi.INT,
+        save_err=rffi.RFFI_SAVE_LASTERROR)
+
+    def getlogin():
+        UNLEN = 256  # from lmcons.h
+        with lltype.scoped_alloc(LPWSTR.TO, UNLEN) as user_name:
+            num_chars = lltype.malloc(LPDWORD.TO, 1, flavor='raw')
+            try:
+                num_chars[0] = rffi.cast(DWORD, UNLEN)
+                n = GetUserNameW(user_name, num_chars)
+                if n == -1:
+                    error = GetLastError_saved()
+                    raise WindowsError(error, "getlogin failed")
+                utf8, codepoints = rffi.wcharp2utf8n(user_name,
+                                                     widen(num_chars[0]))
+                return utf8
+            finally:
+                lltype.free(num_chars, flavor="raw")
+
+    BOOLP = lltype.Ptr(lltype.Array(BOOL, hints={'nolength': True}))
+
+    MultiByteToWideChar = rffi.llexternal('MultiByteToWideChar',
+                                          [rffi.UINT, DWORD,
+                                           LPCSTR, rffi.INT,
+                                           rffi.CWCHARP, rffi.INT],
+                                          rffi.INT,
+                                          calling_conv='win',
+                                          save_err=rffi.RFFI_SAVE_LASTERROR)
+
+    WideCharToMultiByte = rffi.llexternal('WideCharToMultiByte',
+                                          [rffi.UINT, DWORD,
+                                           rffi.CWCHARP, rffi.INT,
+                                           LPCSTR, rffi.INT,
+                                           LPCSTR, BOOLP],
+                                          rffi.INT,
+                                          calling_conv='win',
+                                          save_err=rffi.RFFI_SAVE_LASTERROR)
+
+
+    os_createdirectory_impl = winexternal("os_createdirectory_impl",
+                                          [rffi.CWCHARP, rffi.INT_real],
+                                          rffi.INT,
+                                          save_err=rffi.RFFI_SAVE_LASTERROR)
