@@ -62,6 +62,7 @@ class TestPosixFunction:
         compile(f, (str, float))(str(fname), t1)
         assert t1 == os.stat(str(fname)).st_mtime
 
+    @py.test.mark.skipif("sys.platform == 'darwin'")
     def test_utime_negative_fraction(self):
         def f(fname, t1):
             os.utime(fname, (t1, t1))
@@ -85,12 +86,15 @@ class TestPosixFunction:
 
     @win_only
     def test__getfullpathname_long(self):
-        stuff = "C:" + "\\abcd" * 100
-        py.test.raises(WindowsError, rposix.getfullpathname, stuff)
         ustuff = u"C:" + u"\\abcd" * 100
         res = rposix.getfullpathname(ustuff)
         assert res == ustuff
 
+    @win_only
+    def test__getfullpathname_unicode(self):
+        res = rposix.getfullpathname(u"foo\xf2\xf2")
+        assert res
+
     def test_getcwd(self):
         assert rposix.getcwd() == os.getcwd()
 
@@ -149,7 +153,8 @@ class TestPosixFunction:
         arg = '%s -c "print 1+1" > %s' % (sys.executable, filename)
         data = rposix.system(arg)
         assert data == 0
-        assert file(filename).read().strip() == '2'
+        with file(filename) as f:
+            assert f.read().strip() == '2'
         os.unlink(filename)
 
 
@@ -298,6 +303,14 @@ class TestPosixFunction:
         assert rposix.major(dev) == 24
         assert rposix.minor(dev) == 7
 
+    @py.test.mark.skipif("not hasattr(rposix, 'memfd_create')")
+    def test_memfd_create(self):
+        fd = rposix.memfd_create("abc", rposix.MFD_CLOEXEC)
+        try:
+            s = "defghi?"
+            os.write(fd, s)
+        finally:
+            os.close(fd)
 
 @py.test.mark.skipif("not hasattr(os, 'ttyname')")
 class TestOsExpect(ExpectTest):
@@ -330,21 +343,16 @@ class UnicodeWithEncoding:
     def __init__(self, unistr):
         self.unistr = unistr
 
-    if sys.platform == 'win32':
-        def as_bytes(self):
-            from rpython.rlib.runicode import unicode_encode_mbcs
-            res = unicode_encode_mbcs(self.unistr, len(self.unistr),
-                                      "strict")
-            return rstring.assert_str0(res)
-    else:
-        def as_bytes(self):
-            from rpython.rlib.runicode import unicode_encode_utf_8
-            res = unicode_encode_utf_8(self.unistr, len(self.unistr),
-                                       "strict")
-            return rstring.assert_str0(res)
+    def as_utf8(self):
+        from rpython.rlib.runicode import unicode_encode_utf_8
+        res = unicode_encode_utf_8(self.unistr, len(self.unistr), "strict")
+        return rstring.assert_str0(res)
 
-    def as_unicode(self):
-        return self.unistr
+    def as_bytes(self):
+        from rpython.rlib.runicode import unicode_encode_utf_8
+        res = unicode_encode_utf_8(self.unistr, len(self.unistr),
+                                   "strict")
+        return rstring.assert_str0(res)
 
 class BasePosixUnicodeOrAscii:
     def setup_method(self, method):
@@ -405,7 +413,7 @@ class BasePosixUnicodeOrAscii:
     def test_chmod(self):
         def f():
             return rposix.chmod(self.path, 0777)
-
+        f()
         interpret(f, []) # does not crash
 
     def test_unlink(self):
@@ -436,13 +444,9 @@ class BasePosixUnicodeOrAscii:
 
         if sys.platform == 'win32':
             def f():
-                if isinstance(udir.as_unicode(), str):
-                    _udir = udir.as_unicode()
-                    _res = ', '
-                else:
-                    _udir = udir
-                    _res = u', '
-                return _res.join(rposix.listdir(_udir))
+                contents = ', '.join(rposix.listdir(udir))
+                return contents.decode('utf8')
+            f()  # sanity check
             result = interpret(f, [])
             assert os.path.basename(self.ufilename) in ll_to_string(result)
         else:
@@ -726,6 +730,33 @@ if sys.platform.startswith('linux'):
         s2.close()
         s1.close()
 
+if sys.platform == "darwin":
+   def test_sendfile_partial(tmpdir):
+        # issue 3964
+        from rpython.rlib.rsocket import socketpair
+        wsock, rsock = socketpair()
+        wsock.setblocking(False)
+        rsock.setblocking(False)
+
+        recvd = 0
+        with open(str(tmpdir / "sendfile.txt"), "wb+") as file:
+            file.write(b"x" * 131072)
+            file.flush()
+            file.seek(0)
+
+            sent = rposix.sendfile(wsock.fd, file.fileno(),
+                               offset=0, count=131072)
+            assert sent > 0
+            try:
+                chunk = rsock.recv(1024)
+                while chunk:
+                    recvd += len(chunk)
+                    chunk = rsock.recv(1024)
+            except Exception as e:
+                pass
+        assert recvd  == sent
+
+
 @rposix_requires('pread')
 def test_pread():
     fname = str(udir.join('os_test.txt'))
@@ -795,14 +826,16 @@ def test_sched_get_priority_max():
     assert rposix.sched_get_priority_max(rposix.SCHED_FIFO) != -1
     assert rposix.sched_get_priority_max(rposix.SCHED_RR) != -1
     assert rposix.sched_get_priority_max(rposix.SCHED_OTHER) != -1
-    assert rposix.sched_get_priority_max(rposix.SCHED_BATCH) != -1
+    if sys.platform != 'darwin':
+        assert rposix.sched_get_priority_max(rposix.SCHED_BATCH) != -1
 
 @rposix_requires('sched_get_priority_min')
 def test_sched_get_priority_min():
     assert rposix.sched_get_priority_min(rposix.SCHED_FIFO) != -1
     assert rposix.sched_get_priority_min(rposix.SCHED_RR) != -1
     assert rposix.sched_get_priority_min(rposix.SCHED_OTHER) != -1
-    assert rposix.sched_get_priority_min(rposix.SCHED_BATCH) != -1
+    if sys.platform != 'darwin':
+        assert rposix.sched_get_priority_min(rposix.SCHED_BATCH) != -1
 
 @rposix_requires('sched_get_priority_min')
 def test_os_sched_priority_max_greater_than_min():
